/* -*- C++ -*- */

/* COPYRIGHT
 *
 * This file is part of DIVA base library.
 *
 * Author(s): Jarmo Hiipakka and Tommi Ilmonen, 2000-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * DIVA - Digital Interactive Virtual Acoustics group 
 * Telecommunications Software and Multimedia Laboratory
 * Helsinki University of Technology
 * Finland
 *
 * http://www.tml.hut.fi/Research/DIVA/
 *
 * DIVA base library is licensed under GNU Lesser General Public
 * License (LGPL). The LGPL conditions can be found in file "LGPL"
 * that is distributed with this source package or obtained from the
 * GNU organization (www.gnu.org).
 * 
 */

#ifndef DI_MIDI_EVENT_H
#define DI_MIDI_EVENT_H

#include <di_types.h>

#include <stdlib.h> // free

class DI_DataBuffer2;

/** MIDI timer resolution. Indicates the time stamp accuracy in
    Hertz. Currently 10k. This exceeds the accuracy of MIDI by far,
    but is not too big for most applications using only 32-bit
    calculations.  */
#define DI_MIDI_TIME_RESOLUTION 10000
#define DI_MIDI_CHANNELS 16

/** MIDI event. Very direct approach.

    The three generic bytes are stored within an integer (that takes
    four bytes). This makes parts of the code a bit ugly, but some
    copy operations a bit faster - I hope.

    <B>About sysex (system exlusive) messages:</B> All sysex data is
    stored to a variable length buffer (exluding the initial 0xf0
    byte). The final 0xf7 byte is included in the buffer.

    <UL>

    <LI> Introduction to MIDI:
    http://crystal.apana.org.au/ghansper/midi_introduction/contents.html

    <LI> MIDI information http://www.borg.com/~jglatt/tech/midispec.htm

    <LI> MIDI MMC http://www.borg.com/~jglatt/tech/mmc.htm

    <LI> MIDI file format: http://www.its.caltech.edu/~legere/midifile.htm 
    
    </UL>

*/
class DI_MidiEvent
{
public:
  static const char *m_label;
  
  /// MIDI Machine Control (MMC) commands
  enum MmcCommand {
    MmcStop = 0x01,
    MmcPlay = 0x02,
    MmcDeferredPlay = 0x03,
    MmcFastForward = 0x04,
    MmcRewind = 0x05,
    /// Punch in
    MmcRecordStrobe = 0x06,
    /// Punch out
    MmcRecordExit = 0x07,
    MmcPause = 0x09,
    MmcGoto = 0x44
  };

  /** MIDI Real-time message codes. 

  http://crystal.apana.org.au/ghansper/midi_introduction/midi_system_real_time.html
*/

  enum RealTimeMessage {
    RtmClock = 0xf8,
    RtmStart = 0xfa,
    RtmContinue = 0xfb,
    RtmStop = 0xfc,
    RtmActive = 0xfe,
    RtmSystemReset = 0xff
  };

  enum Type {
    Controller = 0xB0,
    KeyPressure = 0xA0,
    NoteOff = 0x80,
    NoteOn = 0x90,
    PitchBend = 0xE0,
    ProgramChange = 0xC0
  };


  DI_MidiEvent() : m_msg(0), m_ticks(0), m_time(0), 
    m_sysexData(0), m_sysexSize(0) {}
  DI_MidiEvent(const DI_MidiEvent &);
 
  ~DI_MidiEvent() { if(m_sysexData) free(m_sysexData); }  

  unsigned ticks() const { return m_ticks; }
  void setTicks(unsigned t) { m_ticks = t; }

  const unsigned long &time() const { return m_time; }
  void setTime(const unsigned long &t) { m_time = t; }

  /// First message byte
  inline unsigned char a() const { return ((const unsigned char*)&m_msg)[0]; }
  /// Second message byte
  inline unsigned char b() const { return ((const unsigned char*)&m_msg)[1]; }
  /// Third message byte
  inline unsigned char c() const { return ((const unsigned char*)&m_msg)[2]; }
  /// Obsolete (would be fourth message byte).
  inline unsigned char d() const { return ((const unsigned char*)&m_msg)[3]; }

  /// Get a given byte.
  inline unsigned char get(int i) const 
  { return ((const unsigned char*) &m_msg)[i]; }

  /// Set given byte.
  inline void set(int i, unsigned char value) 
  { ((unsigned char*) &m_msg)[i] = value; }
  /// Set all data.
  inline void set(uchar a, uchar b, uchar c, uint ticks)
  { set(0, a); set(1, b); set(2, c); m_ticks = ticks; }
  
  /** The four bytes of the event. */
  inline unsigned char *message() { return ((unsigned char *) &m_msg); }
  inline int messageInt() const { return m_msg; }
  inline void setMessage(int m) { m_msg = m; }
  inline void clearMessage() { m_msg = 0; freeSysex(); }

  /** @name Sysex data. */
  //@{

  /** Set system exlusive message. If the "data" is non-NULL then that
      data is written to current buffer. Otherwise the local buffer is
      simply resized. */

  bool setSysex(const char *data, unsigned l);
  void resizeSysex(unsigned n) 
  { m_sysexData = (unsigned char *) realloc(m_sysexData, n); m_sysexSize = n; }
  /// Remove all sysex-information.
  void freeSysex() { free(m_sysexData); m_sysexData = 0; m_sysexSize = 0; }

  unsigned char *sysexData() { return m_sysexData; }
  unsigned const char *sysexDataConst() const { return m_sysexData; }
  unsigned sysexSize() const { return m_sysexSize; }

  //@}
  
  void clear();
  /** Makes the semi-standard note-off real. */
  void realNoteOff()
  {
    if(isNoteOn() && !noteVelocity()) {
      set(0, 0x80 + channel());
      set(2, 64); // Most neutral velocity.
    }
  }

  DI_MidiEvent& operator= (const DI_MidiEvent &);

  /** @name MIDI types. */
  //@{

  bool isController() const {return (a() & 0xf0) == 0xB0; }
  bool isKeyPressure() const {return (a() & 0xf0) == 0xA0; }
  bool isNoteOff() const {return (a() & 0xf0) == 0x80; }
  bool isNoteOn()  const {return (a() & 0xf0) == 0x90; }
  bool isPitchBend() const {return (a() & 0xf0) == 0xE0; }
  bool isProgramChange() const {return (a() & 0xf0) == 0xC0; }
  bool isSysex() const { return m_sysexData != 0; }
  bool isRealTimeMessage() const { return a() >= 0xf8; }

  unsigned char channel() const { return a() & 0xf; }

  unsigned char program() const { return b(); }

  unsigned char keyNumber() const { return b(); }
  unsigned char noteVelocity() const { return c(); }
  
  unsigned char keyPressure() const { return c(); }

  unsigned char controllerNumber() const { return b(); }
  unsigned char controllerValue() const { return c(); }

  int pitchBendValue() const 
  { return (((int) c()) << 7) + (int) b(); }

  void noteOn(unsigned char key, unsigned char velocity, 
	      unsigned char channel, unsigned ticks);

  void noteOff(unsigned char key, unsigned char velocity, 
	       unsigned char channel, unsigned ticks);

  void controller(uchar channel, uchar parameter, uchar value, unsigned ticks)
  { set(Controller | channel, parameter, value, ticks); }
  void pitchBend(uchar channel, uchar key, uchar value, unsigned ticks)
  { set(PitchBend | channel, key, value, ticks); }

  //@}

  /** @name MIDI Machine Control (MMC). */
  //@{
  
  /// Returns true if this an MMC message.
  bool isMmc() const 
  { return m_sysexData == 0 ? false : m_sysexData[0] == 0x7F; }
  unsigned char mmcDeviceId() const 
  { return m_sysexData == 0 ? false : m_sysexData[1]; }

  /// returns the type of the MMC command.
  MmcCommand mmcCommand() const
  { return (MmcCommand) m_sysexData[3]; }

  /** Returns the SMPTE time related to MMC goto -command. 
      
      There are be five fields that you can read:

      <OL>
      
      <LI> Hour
      <LI> Minute
      <LI> Second
      <LI> SMPTE frame number
      <LI> SMPTE sub-frame number

      </OL>
  */

  unsigned const char * mmcGotoTimeSMPTE() const 
  { return m_sysexData + 6; }

  //@}

  /** @name Utilities. */
  //@{
  /// Convert MIDI keynumber to frequency (Hertz).
  static double keyNumberToHz(int key);
  int messageBytes();
  bool write(DI_DataBuffer2 *b);
  bool read(DI_DataBuffer2 *b);
  //@}

protected:
  int            m_msg;
  /// Time stamp in MIDI file ticks.
  unsigned       m_ticks;
  /// Time stamp in 1/10000 seconds reso.
  unsigned long  m_time;

  unsigned char *m_sysexData;
  unsigned       m_sysexSize;
};


#endif
