/* -*- C++ -*- */

/* COPYRIGHT
 *
 * This file is part of the DIVA IO library.
 *
 * Author(s): Jarmo Hiipakka and Tommi Ilmonen, 2000-2001.
 * Jarmo.Hiipakka@hut.fi Tommi.Ilmonen@hut.fi
 *
 * DIVA - Digital Interactive Virtual Acoustics group 
 * Telecommunications Software and Multimedia Laboratory
 * Helsinki University of Technology
 * Finland
 *
 * http://www.tml.hut.fi/Research/DIVA/
 *
 * DIVA IO library is licensed under GNU Lesser General Public
 * License (LGPL). The LGPL conditions can be found in file "LGPL"
 * that is distributed with this source package or obtained from the
 * GNU organization (www.gnu.org).
 * 
 */

#ifndef DI_VIDEO_INPUT_H
#define DI_VIDEO_INPUT_H

#define DI_DONT_CARE 0

#include <di_image.h>

#include <vector>

enum DI_FrameRate {
  DI_FPS_IGNORE,
  DI_FPS_5,
  DI_FPS_10,
  DI_FPS_15,
  DI_FPS_30,
  DI_FPS_60
};

// #include <di_image_conversion.h>

/** Base class for video input classes. Potential child classes are:
    FireWire video input, USB video input and movie video input. 

    @authors Tommi Ilmonen and Juha Laitinen
*/

class DI_VideoInput
{
public:
  
  virtual ~DI_VideoInput();

  // ----------- Image capture and format query ---------------

  /** 
   * This method captures an image from the video source. The image
   * is returned in the native format of the device.
   * @see #imageFormat() for the current format of the device.
   * @see #DI_ImageConversion for conversions into application video formats 
   * @note The device has to be initialized, and the stream transmission has
   * to be started before this method can be called.
   * @see #initialize()
   * @see #start()
   */
  virtual const DI_Image * captureImage() = 0;
  virtual void doneImage();

  /// Returns the current width of a frame in the video stream images.
  virtual int width() const = 0;
  /// Returns the current height of a frame in the video stream images.
  virtual int height() const = 0;
  /** Returns the current frame rate (int frames per second) of the
      video stream. */
  virtual float fps() const = 0;
  /// Returns the current image format of the stream.
  virtual DI_ImageFormat imageFormat() const = 0;
  /// Returns the total size of one captured image frame in bytes.
  virtual unsigned int size() const = 0;

  /** Sets the shutter time. Larger values lead to longer shutter
      times. Negative values tell the system to use automatic shutter
      timing. Manual range [0-1]. */
  virtual void setShutter(float value);
  /** Sets the camera gain (if possible). Negative values tell the
      system to use automatic gain control. Manual range [0-1]. */
  virtual void setGain(float value);

  virtual void setExposure(float value);

  virtual void setBrightness(float value);

  // ----------- Source control ---------------

  // NOTE: for at least image size & format query could be combined into
  // a single list to avoid too may querying functions. 

  /**
   * Returns a list of supported formats for this device.
   * @see #getSupportedFpsForFormat()
   */
  std::vector<DI_ImageFormat> getSupportedFormats() const;
  
  /**
   * Returns the best-supported format for this device.
   */
  DI_ImageFormat getBestFormat() const;
  
  /**
   * Returns the list of supported framerates for this device.
   */
  std::vector<DI_FrameRate> getSupportedFrameRates() const;
  
  /**
   * Returns the best supported framerate (fastest framerate) for
   * given video format and frame size.
   */
  DI_FrameRate getBestFrameRate(DI_ImageFormat fmt, unsigned int width, unsigned int height) const;

  /**
   * Returns the list of supported image sizes. 
   */
  std::vector< std::pair<unsigned int, unsigned int> > getSupportedImageSizes();
  
  /**
   * Returns true if given format at given framerate and image size is supported.
   */  
  bool isSupported(DI_ImageFormat fmt, 
		   int width = DI_DONT_CARE, 
		   int height = DI_DONT_CARE, 
                   DI_FrameRate framerate = DI_FPS_IGNORE);
   
  /**
   * Initialize this video input.
   * @exception IO / etc type exception could be thrown if this fails.
   */
  virtual void initialize(DI_ImageFormat fmt = DI_IMAGE_UNKNOWN, 
			  int width = DI_DONT_CARE, 
			  int height = DI_DONT_CARE, 
			  DI_FrameRate framerate = DI_FPS_IGNORE) = 0;

  /**
   * Returns true if video input is initialized and ready.
   */
  virtual bool isInitialized() const = 0;

  /** 
   * Starts the data transmission.
   */
  virtual void start() = 0;

  /**
  * Returns true if data transmission is running.
  */
  virtual bool isStarted() const = 0;

  /** 
   * Stops the camera data transmission.
   */
  virtual void stop() = 0;
  
  /** 
   * Clean up.
   */
  virtual void finalize() = 0;
  
protected:
  DI_VideoInput() {}

};

#endif


