/* -*- C++ -*- */

/* COPYRIGHT
 *
 * This file is part of Mustajuuri LGPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the Lesser General Public License
 * (LGPL) - version 2.1. Extra clauses can be found from the file
 * MUSTAJUURI_LGPL and further explanations from the file
 * MUSTAJUURI_LICENSE. The license text can be found in file
 * "LGPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *    
 */

#ifndef MJ_APP_ACTION_H
#define MJ_APP_ACTION_H

#include <di_mutex.h>

#include <qptrlist.h>

class MJ_ModuleDsp;

/** Action that should be taken by the application.  

    The actions can be anything. Usually the commands are something in
    the range of play control (start/stop commands etc).
    
    All known actions are activated frequently.

    @author Tommi Ilmonen
*/

class MJ_AppAction
{
public:
  MJ_AppAction() : m_done(false) {}
  virtual ~MJ_AppAction() {}
  /** Do whatever the action requires. All child classes must inherit
      this method. This method gets called frequently by the
      system. */
  virtual void operate() = 0;
  /** Indicates that this action has done its work. This method should
      return true if the action has been carried out (successfully or
      not) and false otherwise. */
  bool isDone() { return m_done; }
  /** Indicates if the argument action is identical to this one.  It
      may happen that several identical actions are requested. To
      prevent the same commands from being repeated over and over
      again this method is used to check that the actions are truly
      different. */
  virtual bool isIdentical(MJ_AppAction *);
protected:
  /** This boolean is used to indicate that this action is over. Child
      classes should use it to inform the system that they can be
      safely deleted.*/
  bool m_done;
};

/** List of application action commands. Only one of these objects
    should exist at any time.

    @author Tommi Ilmonen
*/

class MJ_AppActions
{
public:
  MJ_AppActions();
  ~MJ_AppActions();

  void addAction(MJ_AppAction *);
  
  void operate();

  static void stopDsp();
  static void startDsp(MJ_ModuleDsp *);
  static void moduleModified(MJ_ModuleDsp *, const char *);
  static void moduleIOModified(MJ_ModuleDsp *, uint);
  static void reportError(const char *);
  static void newAction(MJ_AppAction *);
protected:
  QList<MJ_AppAction>  m_actionBuffers[2];
  QList<MJ_AppAction>  m_actions;
  DI_MutexAuto         m_mutex;

  uint                 m_currentBuffer;

  static MJ_AppActions *m_act;
};

#endif
