/* -*- C++ -*- */

/* COPYRIGHT
 *
 * This file is part of Mustajuuri LGPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the Lesser General Public License
 * (LGPL) - version 2.1. Extra clauses can be found from the file
 * MUSTAJUURI_LGPL and further explanations from the file
 * MUSTAJUURI_LICENSE. The license text can be found in file
 * "LGPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *    
 */

#ifndef MJ_ARRAY_H
#define MJ_ARRAY_H

#include <di_types.h>

#include <mj_allocator.h>
#include <mj_config.h>

/** Minimalistic dynamic array template. 

   It is hoped that inlined methods do not bloat final binary.  In the
   contrary they should not exist if they are not needed.  

   The basic methods are like in any array template.

   More exotic speedups may feel a bit strange (array adoption etc).

   @author Tommi Ilmonen
*/
template <class TElem> class MJ_Array
{
public:
  /// Create an empty array.
  MJ_Array() : m_array(0), m_size(0) {}

  /// Create an array of given size.
  MJ_Array(uint nSize) 
    : m_array(0), m_size(0) { resize(nSize); }

  /// Copy constructor.
  MJ_Array(const MJ_Array &array)
    : m_array(0), m_size(0) { *this = array; }

  virtual ~MJ_Array();
  
  TElem &get(uint nIndex);

  /// Get reference to a constant element.
  const TElem &getConst(uint nIndex) const
  { return ((MJ_Array<TElem> *) this)->get(nIndex); }
  
  /// Get pointer to an element fast, without safety checks.
  TElem *getPtrFast(uint nIndex) { return & m_array[nIndex]; }

  /// Get pointer to an e constant element fast, without safety checks.
  const TElem *getPtrFastConst(uint nIndex) const 
  { return & m_array[nIndex]; }

  /** Returns the last element. No safety checks. */
  TElem &last() { return m_array[m_size - 1]; }
  
  /** Array get operator. Calls "get". */
  inline TElem &operator [] (uint nIndex) { return get(nIndex); }

  /** Array constant get operator. Calls "getConst". */
  inline const TElem &operator () (uint nIndex) const 
  { return getConst(nIndex); }

  MJ_Array<TElem> &operator = (const MJ_Array &array);

  /** Removes an element from the array. This method DOES NOT
      reallocate array buffer. Instead, it decrements the array size
      by one. */ 
  bool removeFast(uint ind)
  { 
    if(ind >= m_size) return false;
    uint max = --m_size;
    for(uint i = ind; i < max; i++) m_array[i] = m_array[i+1];
    return true;
  }

  virtual bool resize(uint nSize, bool fatalOnFailure = true);

  /** Makes the array larger or smaller than what it was. */
  inline bool addSize(int add, bool fatalOnFailure = true) 
  { return resize((uint) ((int) m_size + add), fatalOnFailure); }

  /** Set the value of all array elements. */
  void setAll(const TElem &xrElem)
  {
    TElem   *xpPtr = m_array;
    TElem   *xpMax = &m_array[m_size];
    
    while(xpPtr < xpMax)
      *xpPtr++ = xrElem;
  }

  /** Set the value of all array elements in given range. 

      No safety checks !
   */
  void setRange(const TElem &xrElem, uint start, uint end)
  { 
    if(end > m_size) end = m_size; 
    for(uint i=start; i < end; i++) m_array[i] = xrElem;
  }

  /** Shift array elements up in array.*/
  void shiftUp(int first, uint steps)
  {
    for(int i = (int) m_size - steps - 1; i >= first; i--)
      m_array[i + steps] = m_array[i];
  }

  /** Array length. */
  inline uint size() const { return m_size; }

  /// Get a pointer to the data buffer.
  inline TElem   *data() { return m_array; }

  /// Get a const pointer to the data buffer.
  inline const TElem   *data() const { return m_array; }

  /// Get a const pointer to the data buffer.
  inline const TElem   *dataConst() const { return m_array; }

  /// Get element fast, without safety checks.
  TElem &getFast(uint nIndex) { return m_array[nIndex]; }

  /// Get constant element fast, without safety checks.
  const TElem &getFastConst(uint nIndex) const { return m_array[nIndex]; }

  /** Resize the array. Old values are lost. Does not warn upon failure. */
  inline bool resizeFast(uint nSize)
  { 
    if(m_array) delete [] m_array; 
    if(nSize) {
      m_array = new TElem[nSize];
      if(m_array) m_size = nSize;
      else {
	m_size = 0;
	return false;
      }
    }
    else {
      m_array = 0;
      m_size = 0;
    }
    return true;
  }
  
  /** Deletes the data buffer. For non-inlined way to do the same, 
      use {\em Resize(0)}.*/
  inline void empty() 
  { if(m_array) delete [] m_array; m_array = 0; m_size = 0; }

  /** Return the size of the data buffer in bytes. */
  inline uint dataSize() const { return m_size * sizeof(TElem); }

  /** Returns the size of a single element in bytes. */
  static inline uint elemSize() { return sizeof(TElem); }

  /** @name DANGEROUS methods. */
  //@{

  /** Share data with another array. This method DOES NOT delete old
      array values. */
  inline void adopt(const MJ_Array &array)
  { m_array = array.m_array; m_size = array.m_size; }

  /** Give an array to this object. Once this method is called the
      object will consider the data its own. This method DOES delete
      old array values. */
  inline void adoptData(TElem *array, uint nSize)
  { if(m_array) delete [] m_array; m_array = array; m_size = nSize; }
  
  /** Adopts a pointer without deleting the old pointer or adjusting
      the array size information. */
  inline void adoptPointer(TElem *ptr) { m_array = ptr; }
  
  /** Drop pointers to the array WITHOUT deleting the data. */
  inline void dropData() { m_array = 0; m_size = 0; }

  /** Sets the perceived length. The length value is modified without
      doing any actual memory management. */
  inline void setLength(uint l) { m_size = l; }

  /** Resize the array using an allocator. */
  inline void allocatorResize(MJ_Allocator *a, uint size)
  { 
    m_array = (TElem *) a->allocate((ulong) size * sizeof(TElem));
    m_size = m_array ? size : 0;
  }

  /** Use allocation object to allocate memory. If the allocator is
      NULL then normal memory management will take place. 
      
      If you use this method to allocate memory then you should use
      either this method, or customEmpty, to free the memory.  */
  inline void customResize(MJ_Allocator *allocator, uint size)
  {
    if(allocator) {
      allocatorResize(allocator, size);
    }
    else {
      resize(size);
    }
  }
  
  /** Empty the array. If the allocator is NULL, then the array will
      be deleted with the normal memory management approach.  */

  inline void customEmpty(MJ_Allocator *allocator)
  {
    if(allocator) dropData();
    else resize(0);
  }

  //@}

  /** Copy data from an array. */
  inline void copyData(const TElem *array, uint size)
    { resize(size); for(uint i=0; i < size; i++) m_array[i] = array[i]; }

  /** @name STL compatibility: */
  
  //@{
  TElem *begin() { return m_array; }
  const TElem *begin() const { return m_array; }
  TElem *end() { return m_array + m_size; }
  const TElem *end() const { return m_array + m_size; }
  //@}

protected:
  TElem *m_array;
  uint m_size;
};

#endif
