/* -*- C++ -*- */

/* COPYRIGHT
 *
 * This file is part of Mustajuuri LGPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the Lesser General Public License
 * (LGPL) - version 2.1. Extra clauses can be found from the file
 * MUSTAJUURI_LGPL and further explanations from the file
 * MUSTAJUURI_LICENSE. The license text can be found in file
 * "LGPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *    
 */

#ifndef MJ_AUDIO_BUF_PARAMS_H
#define MJ_AUDIO_BUF_PARAMS_H

/** 
    Audio buffering parameters.  

    This class is used to hold audio buffer parameters.  These relate
    to real-time audio IO. It contains desired limits for buffer
    length and detected buffer sizes.

    \b NOTE: This class has little to do with fragments etc.  The
    buffer size refers to number of samples that have been transferred
    to the audio hardware and are waiting to come out.  This class
    does not deal with the sample count directly.  Instead the sample
    count is transformed to absolute time value in milliseconds.

*/
class MJ_AudioBufParams
{
public:
  MJ_AudioBufParams();
  
  ~MJ_AudioBufParams() {}
  
  /** Set desired buffering parameters. */
  void setRangesMs(float minBufMs, float maxBufMs)
  { m_minBufferMs = minBufMs; m_maxBufferMs = maxBufMs; }


  /** Returns the maximum length of the buffer (ms). */
  inline float maxBuffer() const { return m_maxBufferMs; }

  /** Returns the minimum length of the buffer (ms). */
  inline float minBuffer() const { return m_minBufferMs; }

  /** Initializes the memory for real-time operation. */
  void initRun();

  /** Adds a buffer fill value during run-time. If the fill value is
      less than one millisecond an underflowq error is added. */
  inline void addRunValue(float bufMs)
  { 
    if(bufMs < m_lowestBufferMs)
      m_lowestBufferMs = bufMs;
    else if(bufMs > m_highestBufferMs)
      m_highestBufferMs = bufMs;
    
    if(bufMs <= 1.0)
      m_underFlows++;
  }

  /** Returns the lowest buffer during the run (ms). */
  inline float lowestBufMs() const 
  { return m_lowestBufferMs; }

  /** Returns the highest buffer during the run (ms). */
  inline float highestBufMs() const 
  { return m_highestBufferMs; }
  
  /** Returns the number of underflows in the system. */
  unsigned long underFlows() { return m_underFlows; }

protected:
  
  /// Minimum desired buffer length.
  float   m_minBufferMs;
  /// Maximum desired buffer length
  float   m_maxBufferMs;

  /// Lowest detected buffer length
  float   m_lowestBufferMs;
  /// Longest detected buffer length
  float   m_highestBufferMs;

  /// Number of underflows during operation. 
  unsigned long m_underFlows;

};


#endif
