/* -*- C++ -*- */

/* COPYRIGHT
 *
 * This file is part of Mustajuuri LGPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the Lesser General Public License
 * (LGPL) - version 2.1. Extra clauses can be found from the file
 * MUSTAJUURI_LGPL and further explanations from the file
 * MUSTAJUURI_LICENSE. The license text can be found in file
 * "LGPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *    
 */

#ifndef MJ_AUDIO_CONFIG_H
#define MJ_AUDIO_CONFIG_H

#include <qdom.h>
#include <qstring.h>
#include <qptrlist.h>

#include <mj_multi_int.h>

class QTextStream;

/** Audio device types. 

    @see MJ_AudioConf MJ_AudioDeviceTypeKeys MJ_AudioEnvVarName
*/
enum MJ_AudioDeviceType {
  /// Input device
  MJ_ADT_INPUT,
  /// Output device
  MJ_ADT_OUTPUT,
  MJ_ADT_SIZEOF
};

/** Name strings of the audio device types. */
extern const char * MJ_AudioDeviceTypeKeys[MJ_ADT_SIZEOF];
extern const char * MJ_AudioDeviceTypeKeys2[MJ_ADT_SIZEOF];

/** The string of the environment variable that tells the audio IO
    config. 

    An ALSA 0.9 axample of a device with 26 input and output channels.

    export MJ_AUDIO_CONF="Input=2=hw:0,0|Output=2=hw:0,0|Input=16=hw:1,0|Output=16=hw:1,0"

*/
extern const char * MJ_AudioEnvVarName;

/** Cross-platform audio configuration management.

    Currently user must specify the devices by setting a proper
    environment variable (see MJ_AudioEnvVarName). Class MJ_AudioConf
    knows how to parse this environment variable and extract the name
    of available audio devices. In future we may query the system
    directly for devices. Such advanced feature is potentially
    possible on Linux/Alsa and IRIX (not on Linux/OSS).

    Audio configuration is expected to consist of <I>devices</I> with
    certain characteristics. The known device types are enumarated
    with MJ_AudioDeviceType. The system may have more exotic devices
    than these. This class cannot handle them.

    Static class members are used to qeury the audio setup.

    Nonstatic class members are relative to the object and may specify
    different audio IO configurations (for example for mixers etc.)

    Audio devices can be described in multiple ways. The first is a
    list of audio devices. The second is list of lists with each
    device type presented in separate list. The third is a list with
    input and output devices in separate lists. It is assumed that the
    second and third type are used by the application and they are
    directly supported by this class.  
    
    @author Tommi Ilmonen

    @see MJ_AudioEnvVarName */

class MJ_AudioConf 
{
public:
  MJ_AudioConf();
  ~MJ_AudioConf();

  bool addDevice(MJ_AudioDeviceType type, int number);
  bool removeDevice(MJ_AudioDeviceType type, int number);
  int  findDevice(MJ_AudioDeviceType type, int number);
  int  inputChannels();
  int  outputChannels();

  uint numDevices() { return m_devices.count(); }
  QList<MJ_2Int> & devices() { return m_devices; }
  MJ_2Int *device(uint i) { return m_devices.at(i); }
  
  QString deviceName(uint i);

  bool defaultStereoIo();
  bool defaultMultiChannelIo();
  
  bool write(QTextStream *s);
  bool read(QTextStream *s);

  QDomElement writeDom(QDomDocument &);
  bool readDom(QDomElement &elem);

  static bool isInput(MJ_AudioDeviceType type)
  { return type == MJ_ADT_INPUT; } 

  static QString audioDeviceDescription(MJ_AudioDeviceType type,
					int number);
  static QString audioDevice(const QString &);
  static QString audioDevice(MJ_AudioDeviceType type,
			     int number);
  static int audioDevices(MJ_AudioDeviceType type);
  static int deviceChannels(MJ_AudioDeviceType type, uint index);

protected:

  /** List of devices. First integer implies the device type while the
      second tells the index. */
  QList<MJ_2Int> m_devices;
  
  /// String that is used to tell this is audio config.
  static const char *m_identifier;

private:
  // Disabled
  MJ_AudioConf(const MJ_AudioConf &) {}
  MJ_AudioConf &operator = (const MJ_AudioConf &) { return *this; }
};

#endif
