/* -*- C++ -*- */

/* COPYRIGHT
 *
 * This file is part of Mustajuuri LGPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the Lesser General Public License
 * (LGPL) - version 2.1. Extra clauses can be found from the file
 * MUSTAJUURI_LGPL and further explanations from the file
 * MUSTAJUURI_LICENSE. The license text can be found in file
 * "LGPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *    
 */

#ifndef MJ_DSP_PLUGIN_H
#define MJ_DSP_PLUGIN_H

#include <mj_dsp_options.h>
#include <mj_module_dsp.h>
#include <mj_plugin_base.h>

#include <qfileinfo.h> 

// #include <mj_plugin_loader.h>

class QApplication;
class QTranslator;
class MJ_LadspaDatas;
class MJ_ModuleDspGui;
class MJ_MultiInfoGui;
class MJ_PluginDbeDspLoader;

enum MJ_DspPluginFunctionality {
  MJ_DPF_AUDIO_IN,
  MJ_DPF_AUDIO_OUT,
  MJ_DPF_FILE_IN,
  MJ_DPF_FILE_OUT,
  MJ_DPF_MIDI_IN,
  MJ_DPF_MIDI_OUT,
  MJ_DPF_SIZEOF
};

enum MJ_DspPluginMetaFunctionality {
  MJ_DPMF_STAND_ALONE,
  MJ_DPMF_CHAINABLE,
  MJ_DPMF_SIZEOF
};

extern const char * MJ_DspPluginFunctionalityKeys[MJ_DPF_SIZEOF];
extern const char * MJ_DspPluginMetaFunctionalityKeys[MJ_DPMF_SIZEOF];

/** DSP plugin data base entry. 

    A list (or array) of these would make up the plugin data-base. 
  
    Deleting the DspDataBaseEntry also results in the shared library 
    being closed. Thus the module database should be cleared only after 
    all modules have been deleted.
    
    @see ModuleDSP ModuleDSPOptions 
*/
class MJ_PluginDbeDsp : public MJ_PluginDbeBase
{
public:
  MJ_PluginDbeDsp();
  virtual ~MJ_PluginDbeDsp();

  /// Returns true if the plugin has not been well-defined.
  bool isEmpty() const;
  
  
  virtual MJ_ModuleDsp *newPlugin();

  /** Create a new gui for the object. */ 
  virtual MJ_ModuleDspGui *newGui(MJ_RtControl *rtparent,
				  MJ_ModuleDsp *module,
				  MJ_GuiSize s,
				  QWidget *parent, 
				  const char *name);

  /// Set the GUI factory function for
  void setGuiFactory
  (MJ_GuiSize s, MJ_ModuleDspGui *(*f)
   (MJ_RtControl *, MJ_ModuleDsp *, QWidget *, const char *));
  
  void setWorksAsTopLevel(bool b);
  bool worksAsTopLevel();

  bool mayConflict() const;
  void setMayConflict(bool b);

  /** Tells if a plugin conflicts with another. 
      
      Default implementation returns false.
  */
  virtual bool conflicts(MJ_PluginDbeDsp *);

  MJ_LadspaDatas *ladspaDatas();
  void setLadspaDatas(MJ_LadspaDatas *ld);

  /// Return the options for this module type.
  MJ_DspOptions  *options() { return &m_options; }
  /// Return the options for this module type.
  const MJ_DspOptions *optionsConst() const { return &m_options; }

  /** Sets the creator function (factory function) for this entry. */
  void setCreator(MJ_ModuleDsp *(*c)())
    { m_pluginCreator = c; }

  /** Plugin factory function. This function pointer can be used to
      create new instances of the class. Compulsory. */
  MJ_ModuleDsp *(*m_pluginCreator)();  

  QTranslator    *translator();
  void setTranslator(QTranslator *);

  static void test(QApplication *xpQA, int iArgc, char **cppArgv);

  static void setDebug(bool Debug) { m_debug = Debug; }

  QTextStream &write(QTextStream &s);

  /** Read necessary module data from a shared lib. 
    
      Return new instance of DspDataBaseEntry id succesfull, NULL
      otherwise.

      The file should have a function which has the name of the file
      with _DSP-extension. If the filename has a path in it the path
      name is not included in the function name:
      /home/people/tilmonen/foo/poppana.so contains the same data as
      poppana.so.

\code
   
MJ_PluginDbeDsp *DspDataNew = MJ_PluginDbeDsp::Read("./foo/poppana.so");

\endcode
    
   */
  
  static bool read(const QFileInfo &FileInfo,
		   const QString   &funcName,
		   MJ_PluginDbeDspLoader *loader,
		   bool             WarnOnErrors = true,
		   QString         *errorStr = 0);

protected:

  void exportCreator(MJ_ModuleDsp *p);

  MJ_DspOptions m_options;
  
  /// Pointers to the GUI generators
  MJ_ModuleDspGui *(*m_pluginGuiCreator[MJ_GUI_SIZEOF])
    (MJ_RtControl *, MJ_ModuleDsp *, QWidget *, const char *);


private:
  class __Internal;
  
  __Internal *internalData() const;

  __Internal *m_internal;

  MJ_PluginDbeDsp(const MJ_PluginDbeDsp &) {}
  MJ_PluginDbeDsp & operator = (const MJ_PluginDbeDsp &) {return *this;}
};

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

extern "C" {
  /** The plugin resigtering function. This function must be defined
      in all plugin libraries.  */
  MJ_PluginDbeDsp *MJ_createPlugin(int);
}

#endif
