/* -*- C++ -*- */

/* COPYRIGHT
 *
 * This file is part of Mustajuuri LGPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the Lesser General Public License
 * (LGPL) - version 2.1. Extra clauses can be found from the file
 * MUSTAJUURI_LGPL and further explanations from the file
 * MUSTAJUURI_LICENSE. The license text can be found in file
 * "LGPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *    
 */

#ifndef MJ_MESSAGE_CARRIER_H
#define MJ_MESSAGE_CARRIER_H

#include <mj_array.h>
#include <mj_config.h>

#include <qdom.h> 
#include <qptrlist.h> 
#include <qstring.h> 

class MJ_Message;
class MJ_ModuleDsp;
class MJ_Parameters;

class QTextStream;

/** Base class for message carriers. 
    
    There are two ways to label a message. 1) Integer - a single
    integer carries some meaning to the recipient. 2) Character string
    - the message is identified based the string.  

    @author Tommi Ilmonen

*/
class MJ_MessageCarrier
{
public:
  static const char *m_label;

  MJ_MessageCarrier();
  MJ_MessageCarrier(const char *);
  /// The message is copied (not linked).
  MJ_MessageCarrier(const char *, MJ_Message *);
  /// The message is copied.
  MJ_MessageCarrier(const MJ_MessageCarrier &i);
  
  virtual ~MJ_MessageCarrier();

  inline bool idDefined() { return m_id != 0; }
    
  inline const char *id() const { return m_id; }
  void  setId(const char *s);
  
  /// Returns true if there is a message
  inline bool hasMessage() const { return m_msg != 0; }
  /// Returns the message
  inline MJ_Message *message() { return m_msg; }
  /// Returns the message with constant declaration
  inline const MJ_Message *messageConst() const { return m_msg; }

  /** Sets the actual message.  Does not duplicate the message. */
  void setMessage(MJ_Message *msg);

  /// Drop pointers to the data.
  void dropData();
  
  /// Create an independent copy of this message.
  MJ_MessageCarrier *duplicate();

  /// Write to a steam
  virtual bool write(QTextStream *s);
  /// Read from stream
  virtual bool read(QTextStream *s);

  /// Write to a steam
  virtual QDomElement write(QDomDocument &);
  virtual bool read(QDomElement &);

  /// Write the object to "cout".
  void toCout();
  
  /** The number of message ids in the system. */
  static int msgIdCount() { return m_counter; }

  MJ_MessageCarrier & operator = (const MJ_MessageCarrier &);

protected:

  static void incrCount();
  static void decrCount();

  char*       m_id;

  MJ_Message *m_msg;

  static int m_counter;
};

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

/** List of messages. Provides sorting by message time stamps. 

    @author Tommi Ilmonen

*/
class MJ_MessageList : public QList<MJ_MessageCarrier>
{
public:
  static const char *m_label;

  MJ_MessageList() {}
  ~MJ_MessageList() {}

  bool copyValues(MJ_Parameters *);

  bool writeMessages(QTextStream *);
  bool readMessages(QTextStream *);

  bool writeMessages(QDomElement &, QDomDocument &);
  bool readMessages(QDomElement &);

  bool deliverMessages(MJ_ModuleDsp *);

  static void setDebug(bool b) { m_debug = b; }

protected:
  int compareItems (void *, void *);

  static bool m_debug;
};

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

/** Array of messages

    @author Tommi Ilmonen
*/
class MJ_MessageArray : public MJ_Array<MJ_MessageCarrier *>
{
public:
  static const char *m_label;

  MJ_MessageArray() {}
  virtual ~MJ_MessageArray();

  virtual bool resize(uint nSize, bool fatalOnFailure = true);
  
  bool copyValues(MJ_Parameters *);

  void setMessage(uint index, MJ_MessageCarrier *msg);

  void empty();

  static void setDebug(bool b) { m_debug = b; }

protected:
  static bool m_debug;

private:
  // Disabled:
  MJ_MessageArray(const MJ_MessageArray &) {}
  MJ_MessageArray & operator = (const MJ_MessageArray &) 
  { return *this; }
};


#endif
