/* COPYRIGHT
 *
 * This file is part of Mustajuuri LGPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the Lesser General Public License
 * (LGPL) - version 2.1. Extra clauses can be found from the file
 * MUSTAJUURI_LGPL and further explanations from the file
 * MUSTAJUURI_LICENSE. The license text can be found in file
 * "LGPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *    
 */

#ifndef MJ_MIDI_CONFIG_H
#define MJ_MIDI_CONFIG_H

#include <qstring.h>
#include <qptrlist.h>

#include <mj_multi_int.h>

class QTextStream;

/** Midi device types. 

    @see MJ_MidiConf MJ_MidiDeviceTypeKeys MJ_MidiEnvVarName
*/
enum MJ_MidiDeviceType {
  /// Input device
  MJ_MIDID_INPUT,
  /// Output device
  MJ_MIDID_OUTPUT,
  MJ_MIDID_SIZEOF
};

/** Name strings of the midi device types. */
extern const char * MJ_MidiDeviceTypeKeys[MJ_MIDID_SIZEOF];
extern const char * MJ_MidiDeviceTypeKeys2[MJ_MIDID_SIZEOF];

/** The string of the environment variable that tells the midi IO
    config. 

    An ALSA 0.9 axample of a device with 26 input and output channels.

    export MJ_MIDI_CONF="Input=2=hw:0,0|Output=2=hw:0,0|Input=16=hw:1,0|Output=16=hw:1,0"

*/
extern const char * MJ_MidiEnvVarName;

/** Cross-platform midi configuration management.

    Currently user must specify the devices by setting a proper
    environment variable (see MJ_MidiEnvVarName). Class MJ_MidiConf
    knows how to parse this environment variable and extract the name
    of available midi devices. In future we may query the system
    directly for devices. Such advanced feature is potentially
    possible on Linux/Alsa and IRIX (not on Linux/OSS).

    Midi configuration is expected to consist of <I>devices</I> with
    certain characteristics. The known device types are enumarated
    with MJ_MidiDeviceType. The system may have more exotic devices
    than these. This class cannot handle them.

    Static class members are used to query the midi setup.

    Nonstatic class members are relative to the object and may specify
    different midi IO configurations (for example for mixers etc.)

    Midi devices can be described in multiple ways. The first is a
    list of midi devices. The second is list of lists with each
    device type presented in separate list. The third is a list with
    input and output devices in separate lists. It is assumed that the
    second and third type are used by the application and they are
    directly supported by this class.  
    
    @author Tommi Ilmonen

    @see MJ_MidiEnvVarName */

class MJ_MidiConf 
{
public:
  MJ_MidiConf();
  ~MJ_MidiConf();

  bool addDevice(MJ_MidiDeviceType type, int number);
  bool removeDevice(MJ_MidiDeviceType type, int number);
  int  findDevice(MJ_MidiDeviceType type, int number);

  uint numDevices() { return m_devices.count(); }
  uint numDevices(MJ_MidiDeviceType type);
  QList<MJ_2Int> & devices() { return m_devices; }
  MJ_2Int *device(uint i) { return m_devices.at(i); }
  
  QString deviceName(uint i);
  
  bool write(QTextStream *s);
  bool read(QTextStream *s);

  static bool isInput(MJ_MidiDeviceType type)
  { return type == MJ_MIDID_INPUT; } 

  static QString midiDeviceDescription(MJ_MidiDeviceType type,
					int number);
  static QString midiDevice(const QString &);
  static QString midiDevice(MJ_MidiDeviceType type,
			     int number);
  static int midiDevices(MJ_MidiDeviceType type);
  static int deviceChannels(MJ_MidiDeviceType type, uint index);

protected:

  /** List of devices. First integer implies the device type while the
      second tells the index. */
  QList<MJ_2Int> m_devices;
  
  /// String that is used to tell this is midi config.
  static const char *m_identifier;

private:
  // Disabled
  MJ_MidiConf(const MJ_MidiConf &) {}
  MJ_MidiConf &operator = (const MJ_MidiConf &) { return *this; }
};

#endif
