/* COPYRIGHT
 *
 * This file is part of Mustajuuri LGPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the Lesser General Public License
 * (LGPL) - version 2.1. Extra clauses can be found from the file
 * MUSTAJUURI_LGPL and further explanations from the file
 * MUSTAJUURI_LICENSE. The license text can be found in file
 * "LGPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *    
 */

#ifndef MJ_MIDI_SEQUENCE_H
#define MJ_MIDI_SEQUENCE_H

#include <mj_array.h>
#include <mj_midi.h>

#include <mj_list.h>

#include <qstring.h>

class QDataStream;

bool MJ_skipBytes(QDataStream *, unsigned);


/// List of MJ_MidiEvent
class MJ_MidiEventList : public MJ_List<MJ_MidiEvent>
{
public:
  MJ_MidiEventList() {}
  virtual ~MJ_MidiEventList() {}

protected:
  int compareItems(void *, void *);
};

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

/** Track of MIDI events. 

    This class has autodeletion on by default. 

    <B>File read is incomplete:</B> Tempo information is lost.  */
class MJ_MidiTrack : public MJ_MidiEventList
{
public:
  static const char *m_label;

  MJ_MidiTrack();
  MJ_MidiTrack(const MJ_MidiTrack &);

  virtual ~MJ_MidiTrack();

  void setTicksPerQuarter(double d) { m_ticksPerQuarter = d; }

  bool readTrack(QDataStream *, int);

  bool writeTrack(QTextStream *);

  /** The tick number of the last event. */
  uint endTicks();

  /** Find a note off-event that corresponds to given note on event. 
      
      The search is started from current list position.
  */
  MJ_MidiEvent *findNoteOff(MJ_MidiEvent *noteOn);
  
  /** Get events with "tick" value less than given value.
   
      The event pointers are copied, the events are NOT duplicated.
   */
  void getEvents(QList<MJ_MidiEvent> *events, ulong time);

  void prepare();

  /// Makes a copy of each list element.
  MJ_MidiTrack &operator = (const MJ_MidiTrack &);

protected:

  bool readEvent(QDataStream *);
  bool readSysex(QDataStream *, MJ_MidiEvent *);
  bool readNormal(QDataStream *, MJ_MidiEvent *);

  int  readDeltaTime(QDataStream *);

  ulong ticksToTimeStamp(uint ticks);

  static bool m_debug;

  /// Index of the "current" event
  int    m_current;
  
  /// Time stamp of the tempo start, 10000 stamps per second.
  ulong  m_tempoStartTime;
  ulong  m_tempoStartTicks;

  /// Tempo in quarters per second.
  double m_tempo;
  double m_ticksPerQuarter;
};

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

/// An array of MIDI tracks.
class MJ_MidiSequence : public MJ_Array<MJ_MidiTrack>
{
public:
  static const char *m_label;

  MJ_MidiSequence();
  ~MJ_MidiSequence();

  bool readSequence(const QString &);
  bool writeSequence(QTextStream *);

  const QString & fileName() const { return m_fileName; }

  uint ticksPerQuarter() { return m_ticksPerQuarter; }
  uint endTicks();

  /** Get tempo info etc. */
  void prepare();
  void getEvents(QList<MJ_MidiEvent> *events, ulong timeUs);
  
  /** Copy all events to a single track. */
  void copyTo(MJ_MidiTrack *);

  static void test(const char *fname = "bumblebee.mid");
  
protected:
  bool readHeader(QDataStream *);
  
  uint  m_ticksPerQuarter;
  /// microseconds per quarter note
  ulong m_tempo;
  /// The microsecond timereso of the beginning time of current tempo
  ulong m_tempoBeginUs;
  ulong m_tempoBeginTicks;

  QString m_fileName;

  static bool m_debug;
};

#endif
