/* -*- C++ -*- */

/* COPYRIGHT
 *
 * This file is part of Mustajuuri LGPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the Lesser General Public License
 * (LGPL) - version 2.1. Extra clauses can be found from the file
 * MUSTAJUURI_LGPL and further explanations from the file
 * MUSTAJUURI_LICENSE. The license text can be found in file
 * "LGPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *    
 */

#ifndef MJ_PARAMETER_H
#define MJ_PARAMETER_H

#include <mj_message_enums.h>
#include <mj_misc.h>

#include <qglobal.h>
#include <qptrlist.h>
#include <qstring.h>
#include <qstringlist.h> 

class QTextStream;
class MJ_MessageCarrier;

enum MJ_ParamFlags {
  MJ_PFLAG_AFFECTS = 0x1,
  MJ_PFLAG_REAL_TIME_ADJUST = 0x2
};

#if QT_VERSION < 300
//  typedef QString (*)(const char*) translatorPtr;
#define MJ_TRANSLATOR(x) QString (*x)(const char*) 
#else
//   typedef QString (*)(const char*, const char *) translatorPtr;
#define MJ_TRANSLATOR(x) QString (*x)(const char*, const char *) 
#endif

/** Parameter. Parameter describes the state of some internal variable
    in module. 

    @see MJ_Message
*/
class MJ_Parameter
{
public:
  

  /** Default constructor. */
  MJ_Parameter();

  MJ_Parameter(const QString &name,
	       const char *id,
	       MJ_Unit unit = MJ_UNIT_BLANK);

  MJ_Parameter(const char *name,
	       const char *id,
	       MJ_Unit unit = MJ_UNIT_BLANK);
  MJ_Parameter(const char *name,
	       const char *nameShort,
	       const char *id,
	       MJ_Unit unit = MJ_UNIT_BLANK);

  /** Default copy constructor. */
  MJ_Parameter(const MJ_Parameter &);

  virtual ~MJ_Parameter();
  /** Message (or parameter) type */
  virtual MJ_MessageType type() const = 0;
  
  /// Identifier
  const char * id() const { return m_id; }

  /// Identifier
  void setId(const char * id);

  void setName(const char *s);
  void setNameShort(const char *s);

  void setName(const QString &s) { m_name = s.copy(); }
  void setNameShort(const QString &s) { m_nameShort = s.copy(); }

  const QString &name();
  const QString &nameShort();

  /** Figure up a name for this parameter. */
  QString name(int guiSize);

  /** Measurement unit. */
  MJ_Unit unit() const { return m_unit; }
  /** Set measurement unit. */
  void setUnit(MJ_Unit unit) { m_unit=unit; }
  
  /** Write the parameter to a stream */
  virtual bool write(QTextStream *);

  /** Write contents of the parameter so humans can understand it. */
  virtual QStringList writeClean();

  /** Read the parameter from a stream. */
  virtual bool read(QTextStream *);

  /** Transform the parameter to a corresponding value message. */
  virtual MJ_MessageCarrier *valueMessage();
  
  /// Returns true if this parameter affects other parameters.
  bool affectsOthers() const { return m_flags & MJ_PFLAG_AFFECTS;}

  /// Returns true if the parameter can be adjusted in real-time.
  bool realTimeAdjustable() const 
  {return m_flags & MJ_PFLAG_REAL_TIME_ADJUST;}

  void setFlags(int flags) {m_flags = flags; }

  /** The number of parameters in the system. */
  static int paramCount() { return m_counter; }

  /// Set the translator function
    // void setTranslator(QString (*translator)(const char*))
    void setTranslator(MJ_TRANSLATOR(translator))
      { m_translator = translator; }

  static int defaultFlags() { return MJ_PFLAG_REAL_TIME_ADJUST; }

protected:

  /** Increment reference counter. Used for debugging. Thread safe. */
  static void incrCount();
  /** Decrement reference counter. Used for debugging. Thread safe. */
  static void decrCount();

  char   *m_nameC;
  char   *m_nameShortC;
  
  // QString (*m_translator)(const char*);
  // translatorPtr m_translator;

  MJ_TRANSLATOR(m_translator);
  
  QString m_name;
  QString m_nameShort;

  char *  m_id;
  MJ_Unit m_unit;

  int     m_flags;
  
  static int     m_counter;
};

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

/// List of MJ_Parameter
class MJ_Parameters : public QList<MJ_Parameter>
{
public:
  static const char *m_label;

  MJ_Parameters() {}
  virtual ~MJ_Parameters() {}
  
  int findParameter(const char *id);

  bool writeParams(QTextStream *);
  bool writeParamsClean(QValueList<QStringList> *);
  bool readParams(QTextStream *);

protected:
};

#endif
