/* -*- C++ -*- */

/* COPYRIGHT
 *
 * This file is part of Mustajuuri LGPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the Lesser General Public License
 * (LGPL) - version 2.1. Extra clauses can be found from the file
 * MUSTAJUURI_LGPL and further explanations from the file
 * MUSTAJUURI_LICENSE. The license text can be found in file
 * "LGPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *    
 */

#ifndef MJ_PARAMETER_NUMERIC_H
#define MJ_PARAMETER_NUMERIC_H

#include <mj_parameter.h>

class MJ_Mapper;

/** Base class for numeric parameters. 

    Features: 1) Limit values (minimum and maximum) 2) List of allowed
    values (by default the list is empty). */
template <class TElem>
class MJ_ParameterNumeric : public MJ_Parameter
{
public:

  MJ_ParameterNumeric();

  MJ_ParameterNumeric(TElem value, TElem minV, TElem maxV)
    : m_min(minV), m_max(maxV), m_value(value), m_mapper(0) 
  {  m_allowedValues.setAutoDelete(true); }

  MJ_ParameterNumeric(const QString &name,
		      const char *id,
		      TElem value, 
		      TElem min, 
		      TElem max, 
		      MJ_Unit unit = MJ_UNIT_BLANK)
    : MJ_Parameter(name, id, unit),
      m_min(min),
      m_max(max),
      m_value(value), 
      m_mapper(0)
  { m_allowedValues.setAutoDelete(true); }

  MJ_ParameterNumeric(const char *name,
		      const char *id,
		      TElem value, 
		      TElem min, 
		      TElem max, 
		      MJ_Unit unit = MJ_UNIT_BLANK)
    : MJ_Parameter(name, 0, id, unit),
      m_min(min),
      m_max(max),
      m_value(value), 
      m_mapper(0)
  { m_allowedValues.setAutoDelete(true); }

  virtual ~MJ_ParameterNumeric();

  virtual QStringList writeClean();

  virtual MJ_MessageType type() const;
  virtual MJ_MessageCarrier *valueMessage();
  
  /// Sets a large number of variables at one go.
  void set(const QString &name,
	   const char * id, 
	   TElem value, TElem min, TElem max, MJ_Unit unit = MJ_UNIT_BLANK)
  {
    setName(name); 
    setId(id); 
    m_value = value; 
    setRange(min, max); 
    m_unit = unit; 
  }

  /// Sets a large number of variables at one go.
  void set(const char *name,
	   const char *  id, 
	   TElem value, TElem min, TElem max, MJ_Unit unit = MJ_UNIT_BLANK)
  {
    setName(name); 
    setId(id); 
    m_value = value; 
    setRange(min, max); 
    m_unit = unit; 
  }

  /// Returns current value.
  TElem value() const { return m_value; }

  /** Sets the value. If the value does not directly fit to allowed
      range its value is modified to meet the limits and false is
      returned. */
  bool setValue(TElem value);

  /** Clamp the value to the limited range. Internal value is not
      changed. */

  inline TElem clamp(TElem value) const 
  {
    if(value > m_max) return m_max;
    if(value < m_min) return m_min;

    return value;
  }

  /// The minimum value
  TElem min() const { return m_min; }
  /// Sets the minimum value
  void setMin(TElem i) { m_min = i; }

  /// Maximum value
  TElem max() const { return m_max; }
  /// Sets the maximum value
  void setMax(TElem i) { m_max = i; }

  /// Sets the range limits.
  void setRange(TElem xMin, TElem xMax);
  /// Sets the range and value.
  void set(TElem xMin, TElem xMax, TElem value)
  { setRange(xMin, xMax); setValue(value); }

  /// Adds on value to the list of allowed values
  void addAllowed(const TElem &value) 
  { m_allowedValues.append(new TElem(value)); }

  /// List of allowed values.
  QList<TElem> &allowedValues() { return m_allowedValues; }

  /// Returns the mapper to be used in UI (or NULL).
  MJ_Mapper *mapper() { return m_mapper; }

  void setMapper(MJ_Mapper *);

protected:
  QList<TElem>  m_allowedValues;

  TElem m_min;
  TElem m_max;

  TElem m_value;

  MJ_Mapper *m_mapper;
};

typedef MJ_ParameterNumeric<double> MJ_ParameterDouble;
typedef MJ_ParameterNumeric<int> MJ_ParameterInt;

#endif
