/* -*- C++ -*- */

/* COPYRIGHT
 *
 * This file is part of Mustajuuri LGPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the Lesser General Public License
 * (LGPL) - version 2.1. Extra clauses can be found from the file
 * MUSTAJUURI_LGPL and further explanations from the file
 * MUSTAJUURI_LICENSE. The license text can be found in file
 * "LGPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *    
 */

#ifndef MJ_PLUGIN_BASE_H
#define MJ_PLUGIN_BASE_H

#include <qstring.h> 
#include <qptrlist.h>
#include <qmap.h>

class MJ_SoRefCount;
class MJ_Version;

class QTextStream;

/// Information strings related to any plugin.
enum MJ_PLuginParameter {
  /** A short text about what this module really does. */
  MJ_PP_ABOUT,
  /** Comma-separated list of authors. Use you native alphabet. */
  MJ_PP_AUTHORS,
  /** Latin version of the authors' names */
  MJ_PP_AUTHORS_LATIN1,
  /** Class name of the plugin. */
  MJ_PP_CLASSNAME,
  /** Copyright text, dates, etc.*/ 
  MJ_PP_COPYRIGHT,
  /** Functionality this plugin implements. */
  MJ_PP_FUNCTIONALITY,
  /** Some help text*/ 
  MJ_PP_HELP,
  /** Unique UNCHANGING identifier of the plugin. User never sees this
        and it can be fairly arbitrary LATIN-1 string. NO BLANKS in
        the string ! */
  MJ_PP_IDENTIFIER,
  /** Menu position. Looks like file hierarchy.
	Something like "/reverbs/ambient reverb".*/
  MJ_PP_MENU_POSITION,
  /** Metafunctionality this plugin implements. */
  MJ_PP_METAFUNCTIONALITY,
  /** Name of the plugin, if known. */
  MJ_PP_NAME,
  /** Name of the translation file. The system fill append the
      language suffix and "qm" to this string. With language "fi"
      (Finnish) name mj_urttu will be transformed to mj_urttu_fi.qm
      before the system tries to find the file. */
  MJ_PP_TRANSLATION_FILE,
  /** Some help page in the web. */
  MJ_PP_URL_HELP,
  /** Home page of the plugin. */
  MJ_PP_URL_HOMEPAGE,
  /** Compilation time of the plugin. */
  MJ_PP_COMPILE_TIME,
  MJ_PP_SIZEOF
};

extern const char *MJ_PLuginParameterKeys[MJ_PP_SIZEOF];

/** Base class for plugin database entriesy. Objects of this kind
    contain information about the type of the plugin and its
    capabilities. */
class MJ_PluginDbeBase
{
public:

  MJ_PluginDbeBase();

  virtual ~MJ_PluginDbeBase();
  
  const QString &parameter(MJ_PLuginParameter p) const;

  const QString &menuPosition() const
  { return parameter(MJ_PP_MENU_POSITION); }

  QString menuPosition2();

  void setParameter(MJ_PLuginParameter p, const QString &s);

  /** Sets one parameter string. An utility method that simply
      replicates the effect of method that uses the QString. */
  void setParameter(MJ_PLuginParameter p, const char *);

  /// Utility method that sets the authors string
  void setAuthors(const QString &a) { setParameter(MJ_PP_AUTHORS, a); }

  /// Utility method that sets the menu position string
  void setMenuPosition(const QString &a) 
  { setParameter(MJ_PP_MENU_POSITION, a); }

  void setMenuPosition(const char *a) 
  { setParameter(MJ_PP_MENU_POSITION, a); }

  const QString & about() const 
  { return parameter(MJ_PP_ABOUT); }

  void setAbout(const QString &a) 
  { setParameter(MJ_PP_ABOUT, a); }

  void setAbout(const char *a) 
  { setParameter(MJ_PP_ABOUT, a); }
  
  /// Utility method to set the identifier string
  void setIdentifier(const QString &a) { setParameter(MJ_PP_IDENTIFIER, a); }
  /// Utility method to set the identifier string
  void setIdentifier(const char *a) { setParameter(MJ_PP_IDENTIFIER, a); }
  /// Utility method that returns the identifier string
  const QString &identifier() const { return parameter(MJ_PP_IDENTIFIER); }

  void addMetaFunctionality(const QString &);
  
  bool containsMetaFunctionality(const QString &);

  bool containsFeature(MJ_PLuginParameter p, const QString &str);

  /** The name of objects created by this plugind dbe. Returns the
      last field of the the menu path string. */
  QString objectName();

  /// Returns true if the plugin is valid.
  bool isValid() const;
  
  /// Get a possible debugging variable.
  bool *pluginDebug() { return m_pluginDebug; }
  void setPluginDebug(bool *b) { m_pluginDebug = b; }

  void incrementReferenceCounter();
  void decrementReferenceCounter();

  MJ_Version *mjVersion() const { return m_mjVersion; }

  MJ_Version *pluginVersion() const { return m_pluginVersion; }

  /// Switch debugging state
  static void setDebug(bool b) { m_debug = b; }

  void setSoHandle(void *soHandle);

  bool setRefCounter(MJ_SoRefCount *RefCount);
  MJ_SoRefCount *soRefCounter() { return m_refCount; }

  static int pluginObjects();

protected:

  QMap<int, QString>  m_params;
  bool               *m_pluginDebug;
  /// Number of plugin instances created.
  int                 m_pluginRefCount;
  /// Holds information on the shared object file handle.
  MJ_SoRefCount      *m_refCount;
  MJ_Version         *m_mjVersion;
  MJ_Version         *m_pluginVersion;
  /// True if this entry owns the counter

  static bool         m_debug;

private:
  MJ_PluginDbeBase(const MJ_PluginDbeBase &) {}
  MJ_PluginDbeBase & operator = (const MJ_PluginDbeBase &) { return *this; }
};

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

/// Sortable Plugin DataBase Entry list.
class MJ_PluginDbeList
  : public QList<MJ_PluginDbeBase>
{
public:
  MJ_PluginDbeList() {}
  ~MJ_PluginDbeList() {}
  
  /** Find a data base entry. The search can be based on any of the
      string parameters. */
  MJ_PluginDbeBase *findDbe
  (const QString &matchStr, 
   MJ_PLuginParameter matchParameter = MJ_PP_IDENTIFIER);

  /** Find a data base entry. The search can be based on any of the
      string parameters. */
  MJ_PluginDbeBase *findFindDbe
  (const QString &regexp, 
   MJ_PLuginParameter matchParameter = MJ_PP_IDENTIFIER);

  /** Write plugins DBE menu position or some other parameter to a
      stream. */
  void writeParams(QTextStream *s, 
		   MJ_PLuginParameter matchParameter = MJ_PP_MENU_POSITION, 
		   bool writeIndexes = true);

  void sort();
protected:
  int compareItems (void *, void *);
};

#endif
