/* -*- C++ -*- */

/* COPYRIGHT
 *
 * This file is part of Mustajuuri LGPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the Lesser General Public License
 * (LGPL) - version 2.1. Extra clauses can be found from the file
 * MUSTAJUURI_LGPL and further explanations from the file
 * MUSTAJUURI_LICENSE. The license text can be found in file
 * "LGPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *    
 */

#ifndef RT_CONTROL_H
#define RT_CONTROL_H

#include <qptrlist.h>

#include <mj_mutex.h>

/** Real time status. 

    @see MJ_RtControl
*/
enum MJ_RtStatus {
  /// Not active 
  MJ_RT_NOT,
  /** Real-time and threaded. Requires a mutex. */
  MJ_RT_THREADED_RT,
  /// Real-time, but not threaded
  MJ_RT_UNTHREADED_RT,
  /// Off-line pseudo-real-time calculation
  MJ_RT_UNTHREADED_PSEUDO_RT,
  ///
  MJ_RT_SIZEOF
};

/** Base class for real-time-aware classes.

    It provides services to lock data (mutex lock).  It also does lock
    counting so the user knows when locks have leaked. You can also
    make tree of these. This is done to help spread the mutexes all
    over the application.
    
    Once the real-time mode is finished the mutex should be dropped. 
    
    The virtual method you may want to override is {\em rtStatusChange}.

    \b Linux: Linux seems to lack mutexes that inherit priority. As a
    result we are forced to do priority adjustments manually. As a
    mutex is locked the in "lock" the method tries to automatically
    increase priority of this thread so it will not be descheduled
    while hoding the precious mutex. In "unlock" the priority is
    dropped to normal level.

    @see MJ_RtStatus, MJ_ModuleDspUi */
class MJ_RtControl
{
public:
  /// Create an object with no mutex.
  MJ_RtControl();

  /// Create an object and link it instantly
  MJ_RtControl(MJ_RtControl *parent);

  virtual ~MJ_RtControl();
  
  /// Return current mutex or zero, if none.
  inline MJ_Mutex *getMutex() { return m_rtMutex; }

  /// Return true if the system is real-time.
  inline bool isRealTime() const { return (bool) m_rtStatus; }
  
  /// Return true if the object has a mutex (same as "IsRealTime").
  inline bool hasMutex() const { return m_rtMutex != 0; }
  
  
  /// Returns true if the critical resource is locked.
  inline bool isLocked() const 
  { return m_rtLocks > 0; }


  /** Set the mutex that controls the DSP-process. 
      
      Putting a mutex to place also makes the GUI object "real-time". 
      In real-time mode some operations may be impossible to perfoms.

      The mutex is automatically distributed to the 
      children of this object.
   */
  virtual void rtStatusChange(MJ_RtStatus status, MJ_Mutex *mutex);

  /// Set the parent of the control object.
  void setRtParent(MJ_RtControl *parent);

  /** Add a child to the object. The MJ_RtControl objects
      automatically distribute mutex changes to their children. */
  void addRtChild(MJ_RtControl *child);
  
  /** Remove a child from the list of objects.  Does not directly
      affect the child. */
  void removeRtChild(MJ_RtControl *child);
  
protected:

  /** Lock the target object for quick operation. 
      
      This method locks the mutex if it exists. It also increments the
      lock counter by one. 

      You could get higher performance by using: 

      \begin{verbatim}
      if(!m_rtMutex) m_rtMutex->lock();
      \end{verbatim}

      Usually the performance penalty is
      insignificant and the aid of lock counting helps debug things.

  */
  bool lock();

  /** Unlock the target object after quick operation. */
  void unlock();

  /** For you to override. This method is called as the system enters
      or exits real-time mode. It can be used to enable/disable GUI
      elements that should not be used during real-time operation. */

  virtual void realTimeChange(bool nowRt);

  MJ_RtStatus m_rtStatus;
  MJ_Mutex   *m_rtMutex;
  int         m_rtLocks;
  
  MJ_RtControl        *m_rtParent;
  /** List of children. Most modules do not have children. To conserve
      memory we only use a pointer here and allocate the list space as
      needed. */
  QList<MJ_RtControl> *m_rtChildren;
  
private:
  // Disable copy constructor
  MJ_RtControl(const MJ_RtControl &) { }  
};


#endif
