/* COPYRIGHT
 *
 * This file is part of Mustajuuri LGPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the Lesser General Public License
 * (LGPL) - version 2.1. Extra clauses can be found from the file
 * MUSTAJUURI_LGPL and further explanations from the file
 * MUSTAJUURI_LICENSE. The license text can be found in file
 * "LGPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *    
 */


#ifndef MJ_SLEEP_H
#define MJ_SLEEP_H

#include <mj_time_utils.h>

#include <time.h> 

class DI_Mutex;
class DI_TimeStamp;

enum MJ_Timing {
  MJ_TIMING_GOOD,
  MJ_TIMING_LATE,
  MJ_TIMING_EARLY
};



/// Sleeping services.
class MJ_Sleep
{
public:

  MJ_Sleep() { m_tspec.tv_sec = 0; m_tspec.tv_nsec = 0;}
  virtual ~MJ_Sleep() {}
#ifndef WIN32

  /// Sleep for n seconds.
  static bool sleepS(uint secs) { return ! sleep(secs); }

  /// Sleep for n milliseconds.
  bool sleepMs(uint msecs) 
  { 
    m_tspec.tv_nsec = (long) msecs * MJ_MILLION;
    return nanosleep(&m_tspec, 0) != -1;
  }

  /// Sleep for n microseconds.
  bool sleepUs(uint usecs)
  { 
    m_tspec.tv_nsec = (long) usecs * 1000;
    return nanosleep(&m_tspec, 0) != -1;
  }
#else

  /// Sleep for n seconds.
  static bool sleepS(uint secs) { 
	  Sleep(secs*1000); 
	  return true;
  }

  /** Sleep for n milliseconds. You cannot sleep more than one second
      with this function. */
  static bool sleepMs(uint msecs) 
  { 
	Sleep(msecs);
	return true;
  }

  /** Sleep for n microseconds. You cannot sleep more than one second
      with this function. */
  static bool sleepUs(uint usecs)
  { 
	  Sleep(usecs/1000);
	  return true;
  }
#endif

protected:
  struct timespec m_tspec;
};

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

/// Synchronized sleeping.
class MJ_SleepSync : public MJ_Sleep
{
public:
#ifndef WIN32
  MJ_SleepSync() { resetTiming(); }

  void resetTiming() 
  { gettimeofday(& m_timing, 0); m_initial = m_timing; }
  
  /** Sleep in synchronous mode. The argument value is added to
      current time value. The return value tells how many microseconds
      early or late the function call returns. Zero is optimal,
      negative values indicate the call returned early and positive
      values indicate the call returned too late. */
  long sleepSynchroUs(long us);

  /** Sleep until time indicated in the stamp object is passed.  The
      initialization time of this object is added to the stamp, before
      calculating sleeping period.  
      
      The second argument should contain a locked mutex. The mutex
      will be freed while the (potential) sleeping is taking place and
      locked again.  */
  long sleepTo(const DI_TimeStamp *stamp, DI_Mutex *mutex = 0);
#endif

protected:  
  struct timeval m_initial, m_timing, m_tmp;
};

#endif
