/* COPYRIGHT
 *
 * This file is part of Mustajuuri LGPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the Lesser General Public License
 * (LGPL) - version 2.1. Extra clauses can be found from the file
 * MUSTAJUURI_LGPL and further explanations from the file
 * MUSTAJUURI_LICENSE. The license text can be found in file
 * "LGPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *    
 */

#ifndef MJ_TO_LADSPA_PLUGIN_H
#define MJ_TO_LADSPA_PLUGIN_H

#include <mj_message.h>
#include <mj_message_carrier.h>
#include <mj_module_dsp.h>

/** Base class for module-adjustment operations. */

class MJ_ModuleAdjust
{
public:
  MJ_ModuleAdjust() {}
  virtual ~MJ_ModuleAdjust() {}

  /// Apply the adjustment
  virtual bool apply(MJ_ModuleDsp *) = 0;
};

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

/** Adjust module with a message. Most modules can be heavily modified
    with messages. Thus this class provides an efficient way modify a
    module. */
class MJ_ModuleAdjustMessage : public MJ_ModuleAdjust
{
public:
  MJ_ModuleAdjustMessage(const char *id, MJ_Message *msg);
  virtual ~MJ_ModuleAdjustMessage();
  
  virtual bool apply(MJ_ModuleDsp *);

protected:
  MJ_MessageCarrier m_msg;
};

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

/// Adjust the number of channels in the module.
class MJ_ModuleAdjustChannels : public MJ_ModuleAdjust
{
public:
  MJ_ModuleAdjustChannels(uint in, uint out) : m_in(in), m_out(out) {}
  virtual ~MJ_ModuleAdjustChannels();
  
  virtual bool apply(MJ_ModuleDsp *);

protected:
  uint m_in, m_out;
};

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

/** Collection of adjustments to a module and the LADSPA-specific
    id/label information. 

    This class contains a list of module adjusting objects that will
    modify the the module as it is created. This class is used since a
    particular Mustajuuri plugin may span several LADSPA plugins. This
    class helps in this multiplication task.

    @see MJ_LadspaDatas
    @author Tommi Ilmonen
*/
class MJ_LadspaData
{
public:
  MJ_LadspaData(ulong id, const char *idStr, const char * name = 0);
  ~MJ_LadspaData();
  
  bool apply(MJ_ModuleDsp *);

  void addMsgDouble(const char *id, double v);
  void addMsgInt(const char *id, int v);

  void setChannels(uint in, uint out);

  ulong uniqueId() const { return m_id; }
  const char *label() const { return m_idStr; }
  const char *name()  const { return m_name; }

protected:
  /// Disabled
  MJ_LadspaData(const MJ_LadspaData&) {}

  QList<MJ_ModuleAdjust> m_adjustments;
  ulong m_id;
  char *m_idStr;
  char *m_name;
};

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

/** Collection of data packets to make a LADSPA plugin. 
    
    @see MJ_LadspaData
    @author Tommi Ilmonen
*/

class MJ_LadspaDatas
{
public:
  MJ_LadspaDatas() { m_datas.setAutoDelete(true); }
  ~MJ_LadspaDatas();
  
  MJ_LadspaData *newElem(ulong id, const char *idStr, const char * name = 0);
  MJ_LadspaData *newElem(ulong id, const char *idStr, const char * name,
			 uint inchans, uint outchans);
  
  QList<MJ_LadspaData> & datas() { return m_datas;}

protected:
  /// Disabled
  MJ_LadspaDatas(const MJ_LadspaData&) {}

  QList<MJ_LadspaData> m_datas;
};

#endif
