/* COPYRIGHT
 *
 * This file is part of Mustajuuri LGPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the Lesser General Public License
 * (LGPL) - version 2.1. Extra clauses can be found from the file
 * MUSTAJUURI_LGPL and further explanations from the file
 * MUSTAJUURI_LICENSE. The license text can be found in file
 * "LGPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *    
 */

#ifndef MJ_UNDO_INFO_H
#define MJ_UNDO_INFO_H

#include <mj_message_carrier.h>
#include <mj_misc.h>

#include <qstring.h>
#include <qptrlist.h>

class MJ_MessageCarrier;
class QTextStream;
class QObject;

/** Undo information packet. This packet holds the message that caused
    new values to be active and the counter messages that would remove
    the effect. 

    Quite often one cn get away with just one redo/undo message. In
    other cases it is necessary to use multiple messages. To allow the
    usage of multiple message this class saves messages to a list. To
    minimize memory consumption a single list holds information about
    both undo and redo-messages. To the end user this internal
    optimization is transparent.

    @see MJ_UndoList MJ_UndoCenter */
class MJ_UndoInfo
{
public:
  /// Info label
  static const char *m_label;

  /// Constructor
  MJ_UndoInfo();

  /// Desctructor
  virtual ~MJ_UndoInfo();

  /** Sets the message and the counter-message of this
      undo-package. */
  void set(MJ_MessageCarrier *msg, MJ_MessageCarrier *counterMsg = 0);

  void addMsg(MJ_MessageCarrier *msg, bool copy);

  void addCounterMsg(MJ_MessageCarrier *msg, bool copy);

  void clear();

  uint nofUndoMsgs() const { return m_msgs.count() - m_firstCounterMsg; }

  uint nofRedoMsgs() const { return m_firstCounterMsg; }

  MJ_MessageCarrier *msg();

  MJ_MessageCarrier *msgIndex(uint);

  MJ_MessageCarrier *counterMsg();

  MJ_MessageCarrier *counterMsgIndex(uint);

  /** Returns object count. The system keeps track of how many objects
      have been created nd assigns each unique count. Potentially
      useless. */
  unsigned long count() const { return m_count; }

  /// The host this undo info refers to.
  QObject *host() { return m_host; }

  /// Sets the host to which this object refers to.
  void setHost(QObject *h) { m_host = h; }

  /// Parameter name
  const QString &paramName() const { return m_paramName; }

  /// Sets the parameter name
  void setParamName(const QString &name) { m_paramName = name.copy(); }

  /// Measurement unit.
  MJ_Unit unit() const { return m_unit; }

  /// Set the measurement unit.
  void setUnit(MJ_Unit unit) { m_unit=unit; }

  /// Describes the change that can be undone with this object.
  virtual QString describe();

  /** Performs the undo operation defined by this object to the
      host. */
  bool undo();
  /** Performs the redo operation defined by this object to the
      host. */
  bool redo();

  /// Writes the info to a stream
  bool write(QTextStream *s);
  /// Reads the info frmo a stream.
  bool read(QTextStream *s);

protected:
  QString describeSimple();
  QString describeMulti();

  /// Disabled
  MJ_UndoInfo(const MJ_UndoInfo &) {}
  
  /// List of undo/redo messages.
  QList<MJ_MessageCarrier>  m_msgs;
  /// Index of the first countermessage (undomessage)
  uint                 m_firstCounterMsg;

  QObject      *m_host;
  
  MJ_Unit m_unit;
  
  QString       m_paramName;

  unsigned long         m_count;

  /** Monotonically increasing count. */ 
  static unsigned long  m_units;
};

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

/// List of undo events.
class MJ_UndoList : public QList<MJ_UndoInfo>
{
public:

protected:
  int compareItems (void *, void *);
};



#endif
