/* -*- C++ -*- */

/* COPYRIGHT
 *
 * This file is part of Mustajuuri LGPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the Lesser General Public License
 * (LGPL) - version 2.1. Extra clauses can be found from the file
 * MUSTAJUURI_LGPL and further explanations from the file
 * MUSTAJUURI_LICENSE. The license text can be found in file
 * "LGPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *    
 */

#ifndef MJ_VECTOR_H
#define MJ_VECTOR_H

#include <string.h>  // memcpy for inlines

#include <mj_array.h>
#include <mj_vml.h>

/** Mathematical vector template class. The element type should be
    some sort of floating point number (float, double, long
    double). If it is not, then you probably need to override some
    methods to get MJ_Vector<funky_type> working.

    MJ_Vector supports some common vector operation like multiplying, summing
    and subtracting vectors. 

    It is a near-independent class and can be probably exported from
    Mustajuuri without taking the full library.

    <B>VML support:</B> MJ_Vector supports <A
    HREF="http://www.tml.hut.fi/~tilmonen/vml/">Vector Mathematics
    Library - VML</A>. The support is done by making special versions
    of MJ_Vector methods that can utilize VML. The VML-methods are
    marked by their "Vml" postfix (for example <I>macVml</I>). When
    VML is not present for such methods the system silently reverts to
    the normal methods. This wrapping is as transparent as possible.

    <B>Using VML:</B> You need to see the VML documentation (in VML
    WWW pages) to understand VML and its limitations. MJ_Vector
    provides its own methods that override vml_begin() and
    vml_end(). These methods are MJ_Vector<type>::vmlBegin() and
    MJ_Vector<type>::vmlEnd(). All calls to VML methods should be
    between these two calls. Multiple calls can reside in such
    location. The following example shows the usage pattern. 

\code

MJ_Vector<float> v1, v2, v3;
int vlen = 100;
v1.resize(vlen);
v2.resize(vlen);
v3.resize(vlen);

// You *should* put values to the vectors before any arithmetic

MJ_Vector<float>::vmlBegin();

v2.accVml(v1);       // Add v1 to v2.
v3.macVml(v1, v2);   // Multiply v1 and v2 and sum the product to v3.

MJ_Vector<float>::vmlEnd();

\endcode

    If VML is not present on your system the VML-methods are
    silently replaced with conventional methods. There is no
    performance penalty.

    <B>VML performance:</B> On x86/3dnow platform (the only current
    VML platform) VML methods tend to be faster than equivalent
    conventional methods - even in cases where where one cannot group
    many VML methods to one batch. The switch between normal mode and
    vector mode does not seem to take too long. VML WWW pages have
    information on the performance in a few cases. 

    @author Tommi Ilmonen
*/
#undef max
#undef min

template <class TElem> 
class MJ_Vector : public MJ_Array<TElem>
{
public:

  MJ_Vector() {}
  MJ_Vector(uint s) : MJ_Array<TElem>(s) {}
  MJ_Vector(const MJ_Vector &v) : MJ_Array<TElem>(v) {}

  virtual ~MJ_Vector();
  
  /** Gets an interpolated value. NO SAFETY CHECKS !!!*/
  TElem getInterpolated1(float index) const
  {
    uint ui = (uint) index;
    float frac = index - (TElem) ui;
    return this->m_array[ui] * (1.0f - frac) + this->m_array[ui + 1] * frac;
  }

  /** Gets an interpolated value. NO SAFETY CHECKS !!!*/
  TElem getInterpolated1(double index) const
  {
    uint ui = (uint) index;
    double frac = index - (TElem) ui;
    return this->m_array[ui] * (1.0 - frac) + this->m_array[ui + 1] * frac;
  }

  /// Copies the argument vector
  MJ_Vector<TElem> &operator = (const MJ_Vector &xrVect)
  { *((MJ_Array<TElem> *) this) = xrVect; return *this;}

  /** Euclidian length. */
  TElem norm() const;
  
  /** Cumulative sum of all elements. */
  TElem cumSum() const;

  TElem max() const;
  TElem max2(uint *location) const;

  TElem min() const;
  TElem min2(uint *location) const;

  /** Returns the absolute maximum value. If the vector is empty
      returns zero. */
  TElem maxAbs() const;

  bool allElemsFinite() const;
  
  /** @name Vector arithmetic */
  //@{

  /** Multiply and accumulate. */
  void mac(const MJ_Vector &v1, const MJ_Vector &v2);

  /** Multiplies and adds the argument to this vector. */
  void mac(const MJ_Vector &xrV, const TElem &xrMul);

  /** Vector subtract. Subtracts the contents of the argument vector
      and store the result in the second argument.*/
  void sub(const MJ_Vector &xrSub, MJ_Vector *xpTarget) const;

  /** Vector add. Adds the contents of the argument vector
      and store the result in the second argument.*/
  void add(const MJ_Vector &xrSub, MJ_Vector *xpTarget) const;

  /** Accumulates (adds) the argument vector to this vector. */
  void acc(const MJ_Vector &xrV);

  /** Multiples all vector elements with a scalar. */
  void mulThis(const TElem &);
  /** Elementwise multiplication with another vector. */
  void mulThis(const MJ_Vector &);

  /** Multiplies this vector with scalar and saves the result to
      another vector. */
  void mul(TElem mul, MJ_Vector *res) const;

  /** Multiplies this vector with a vector and saves the result to
      another vector. */
  void mul(const MJ_Vector &v, MJ_Vector *res) const;

  void mulRamp(TElem start, TElem delta, MJ_Vector *res) const;

  void abs();

  void abs(MJ_Vector *res) const;

  void square();

  void square(MJ_Vector *res) const;

  //@}

  /** @name Vector construction */
  //@{
  
  /** Creates a linear ramp. */
  void ramp(uint first, uint len, TElem firstVal, TElem delta);

  /** Creates a linear ramp. */
  void ramp(TElem firstVal, TElem delta);
  
  //@}

  ///
  inline void modulo1(const TElem &xrLimit)
  { 
    TElem xrNegLimit = -xrLimit;
    for(ulong nI=0; nI < this->m_size; nI++) {
      if(this->m_array[nI] > xrLimit) this->m_array[nI] -= xrLimit;
      if(this->m_array[nI] < xrNegLimit) this->m_array[nI] += xrLimit;
    }
  }

  /** Floating point partial modulo arithmetic. See the source. */
  inline void modulo2(const TElem &xrLimit, const TElem &xrMul = 2.0)
  { 
    TElem xOffset = xrMul * xrLimit;
    TElem xNegLimit = -xrLimit;
    for(ulong nI=0; nI < this->m_size; nI++) {
      if(this->m_array[nI] > xrLimit) this->m_array[nI] -= xOffset;
      else if(this->m_array[nI] < xNegLimit) this->m_array[nI] += xOffset;
    }
  }

  /** @name Copy methods */
  //@{

  /** Calls memcpy. Checks the arrays are of equal size. Does not resize. */
  inline void copyFast(const MJ_Vector<TElem> &v)
  { dimCheck1(v); memcpy(this->m_array, v.m_array, this->m_size * sizeof(TElem)); }

  /** Direct call to memcpy. No resize, no safety checks. */
  inline void copyFastest(const MJ_Vector<TElem> &v)
  { memcpy(this->m_array, v.m_array, this->m_size * sizeof(TElem)); }

  /** Copy values from an array and skip x values between each new
      value. Useful for decoding audio hardware or file input. */
  void copySkipping(const TElem *v, 
		    uint startIndex, 
		    uint samples, 
		    uint skip);
  
  void exportSkipping(TElem *v, 
		      uint startIndex, 
		      uint samples, 
		      uint skip) const;

  //@}

  /** Checks that sizes of two vectors are equal. If they are not,
      fatal error is generated. */
  void dimCheck1(const MJ_Vector &) const;

  /** Checks that two argument vectors have the same length as this. */
  void dimCheck2(const MJ_Vector &,
		 const MJ_Vector &) const;

  /** @name VML methods */
  //@{

  /** Enter VML mode. If this vector type does not support VML on this
      platform this function does nothing.*/
  static inline void vmlBegin() {}

  /** Exit VML mode. If this vector type does not support VML on this
      platform this function does nothing.*/
  static inline void vmlEnd() {}
  
  /** Multiply and accumulate.  */
  inline void macVml(const MJ_Vector &v1, const MJ_Vector &v2)
  { mac(v1, v2); }

  /** Multiplies and adds the argument to this vector. */
  inline void macVml(const MJ_Vector &v1, const TElem &v2)
    { mac(v1, v2); } 
    
  /** Accumulates (adds) the argument vector to this vector. */
  inline void accVml(const MJ_Vector &v)
    { acc(v); }

  /** Multiplies this vector with scalar and saves the result to
      another vector. */
  inline void mulVml(TElem m, MJ_Vector *res) const
  { mul(m, res); }

  /** Multiplies this vector with the dirst argument vector element by
      element and saves the result to the second argument vector. */
  void mulVml(const MJ_Vector &v, MJ_Vector *res) const
  { mul(v, res); }

  //@}
};

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////


/** A vector template that does not do memory management. This class
    is useful when you want to link multiple vectors a single
    array. In Mustajuuri it is necessary as the input array of a DSP
    module links to output array of another module. 

    All memory management methods have been overriden to result in
    zero effect (or intentional fatal error). */
template <class TElem> 
class MJ_Vector2 : public MJ_Vector<TElem>
{
public:
  ///
  MJ_Vector2() {}

  /// Links this vector to the argument
  MJ_Vector2(const MJ_Vector2 &xrV) { adopt(xrV); }

  /** Simply forgets the pointer to the data. */
  virtual ~MJ_Vector2();

  /** Never call !!! Generates a fatal error. */
  virtual bool resize(uint nSize, bool bFatalOnFailure = true);
  
  /// Adopts (ie. shares) data with given vector.
  MJ_Vector2<TElem> &operator = (const MJ_Vector<TElem> &xrV) 
  { adopt(xrV); return *this;}

  /// Adopts (ie. shares) data with given vector.
  MJ_Vector2<TElem> &operator = (const MJ_Vector2<TElem> &xrV) 
  { adopt(xrV); return *this; }

};


/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

// VML acceleration is possible

#if defined(MJ_HAVE_VML) && MJ_HAVE_VML

inline void MJ_Vector<float>::vmlBegin()  { vml_begin(); }
inline void MJ_Vector<float>::vmlEnd() { vml_end(); }

inline void MJ_Vector<float>::macVml(const MJ_Vector &v1, const MJ_Vector &v2)
{ 
  vml_mac_fvvv(this->m_size, v1.dataConst(), v2.dataConst(), this->m_array);
} 

inline void MJ_Vector<float>::macVml(const MJ_Vector &v1, const float &v2) 
{
  vml_mac_fsvv(this->m_size, v2, v1.dataConst(), this->m_array);
}

inline void MJ_Vector<float>::accVml(const MJ_Vector &v)
{
  vml_acc_fvv(this->m_size, v.this->m_array, this->m_array);
}

inline void MJ_Vector<float>::mulVml(float m, MJ_Vector *res) const
{ 
  vml_mul_fsvv(this->m_size, m, this->m_array, res->this->m_array);
} 

void MJ_Vector<float>::mulVml(const MJ_Vector &v, MJ_Vector *res) const
{
  vml_mul_fvvv(this->m_size, v.this->m_array, 
	       this->m_array, res->this->m_array);
}

#endif  // defined(MJ_HAVE_VML) && MJ_HAVE_VML

typedef MJ_Vector<MJ_Sample> MJ_SampleVector;

#endif  // MJ_VECTOR_H
