/* COPYRIGHT
 *
 * This file is part of Mustajuuri LGPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the Lesser General Public License
 * (LGPL) - version 2.1. Extra clauses can be found from the file
 * MUSTAJUURI_LGPL and further explanations from the file
 * MUSTAJUURI_LICENSE. The license text can be found in file
 * "LGPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *    
 */

#include <mj_vector.h>
#include <mj_error.h>

#include <math.h>

#ifdef WIN32
#include <float.h>
#define finite _finite
#endif

template <class TElem> 
MJ_Vector<TElem>::~MJ_Vector()
{}

template <class TElem> 
TElem MJ_Vector<TElem>::norm() const
{
  const TElem *data = this->m_array;
  const TElem *sentinel = &data[this->m_size];
  
  TElem xSum = 0.0;

  while(data < sentinel) {
    TElem xSquare = *data * *data;
    xSum += xSquare;
    data++;
  }
  return (TElem) sqrt((double) xSum);
}

template <class TElem> 
TElem MJ_Vector<TElem>::cumSum() const
{
  const TElem *data = this->m_array;
  const TElem *sentinel = &data[this->m_size];
  
  TElem xSum = 0.0;

  if(this->m_size & 1) {
    xSum += *data++;
  }
  while(data < sentinel) {
    xSum += *data++;
    xSum += *data++;
  }
  return xSum;
}

/** Returns the maximum value. If the vector is empty returns zero. */
template <class TElem> 
TElem MJ_Vector<TElem>::max() const
{
  if(!this->m_size) return 0.0;

  const TElem *data = this->m_array;
  const TElem *sentinel = &data[this->m_size];

  TElem r = *data++;

  while(data < sentinel) {
    if(r < *data) r = *data;
    data++;
  }
  return r;
}

/** Returns the maximum value and the location of the value. If the
    vector is empty returns zero. */
template <class TElem> 
TElem MJ_Vector<TElem>::max2(uint *location) const
{
  if(!this->m_size) return 0.0;

  const TElem *data = this->m_array;
  const uint size = this->m_size;

  TElem r = *data++;
  *location = 0;
  
  for(uint i = 1; i < size; i++) {
    if(r < *data) {
      r = *data;
      *location = i;
    }
    data++;
  }
  return r;
}

/** Returns the minimum value. If the vector is empty returns zero. */
template <class TElem> 
TElem MJ_Vector<TElem>::min() const
{
  if(!this->m_size) return 0.0;

  const TElem *data = this->m_array;
  const TElem *sentinel = &data[this->m_size];

  TElem r = *data++;

  while(data < sentinel) {
    if(r > *data) r = *data;
    data++;
  }
  return r;
}

/** Returns the minimum value and the location of the value. If the
    vector is empty returns zero. */
template <class TElem> 
TElem MJ_Vector<TElem>::min2(uint *location) const
{
  if(!this->m_size) return 0.0;

  const TElem *data = this->m_array;
  const uint size = this->m_size;

  TElem r = *data++;
  *location = 0;
  
  for(uint i = 1; i < size; i++) {
    if(r > *data) {
      r = *data;
      *location = i;
    }
    data++;
  }
  return r;
}

template <class TElem> 
TElem MJ_Vector<TElem>::maxAbs() const
{
  if(!this->m_size) return 0.0;

  const TElem *data = this->m_array;
  const TElem *sentinel = &data[this->m_size];

  TElem r = MJ_abs(*data++);

  while(data < sentinel) {
    TElem tmp = MJ_abs(*data++);
    if(r < tmp) r = tmp;
  }
  return r;
}

/** Checks that all vector elements are finite. 
    
    If all vector elements are finite (ei. neither nan nor +/-inf)
    this method returns true. In the other case this method returns
    false.  */

template <class TElem> 
bool MJ_Vector<TElem>::allElemsFinite() const
{
  const TElem *data = this->m_array;
  const TElem *dataMax = & this->m_array[this->m_size];

  while(data < dataMax) {
    if(!finite(*data++))
      return false;
  }
  return true;
}

template <class TElem> 
void MJ_Vector<TElem>::mac
(const MJ_Vector &v1, const MJ_Vector &v2)
{
  dimCheck2(v1, v2);

  const TElem *in1 = v1.m_array;
  const TElem *in2 = v2.m_array;
  TElem *tptr = this->m_array;
  
  uint size = this->m_size;

  const TElem *sentinel = &in1[size];
  
  if(size & 1) 
    *tptr++ += *in1++ * *in2++;

  if(size & 2) {
    *tptr++ += *in1++ * *in2++;
    *tptr++ += *in1++ * *in2++;
  }

  while(in1 < sentinel) {
    *tptr++ += *in1++ * *in2++;
    *tptr++ += *in1++ * *in2++;    
    *tptr++ += *in1++ * *in2++;
    *tptr++ += *in1++ * *in2++;    
  }

}

template <class TElem> 
void MJ_Vector<TElem>::sub
(const MJ_Vector &xrSub, MJ_Vector *xpTarget) const 
{
  dimCheck2(xrSub, *xpTarget);

  const TElem *data = this->dataConst();
  const TElem *data2 = xrSub.dataConst();
  TElem *xpTargetD = xpTarget->data();

  for(ulong nJ = 0; nJ < this->m_size; nJ++) {
    *xpTargetD++ = *data++ - *data2++;
  }
}

template <class TElem>
void MJ_Vector<TElem>::add
(const MJ_Vector &xrSub, MJ_Vector *xpTarget) const 
{
  dimCheck2(xrSub, *xpTarget);

  const TElem *data = this->dataConst();
  const TElem *data2 = xrSub.dataConst();
  TElem *xpTargetD = xpTarget->data();

  for(ulong nJ = 0; nJ < this->m_size; nJ++) {
    *xpTargetD++ = *data++ + *data2++;
  }  
}

template <class TElem>
void MJ_Vector<TElem>::mac(const MJ_Vector &xrV, const TElem &xrMul)
{
  dimCheck1(xrV);

  TElem *data = this->m_array;
  const TElem *data2 = xrV.m_array;
  const TElem *sentinel = &data[this->m_size];

  while(data < sentinel) {
    *data++ += *data2++ * xrMul;
  }
}

template <class TElem>
void MJ_Vector<TElem>::acc(const MJ_Vector &xrV)
{
  dimCheck1(xrV);

  TElem *data = this->m_array;
  const TElem *data2 = xrV.dataConst();

  const TElem *sentinel = &data[this->m_size];

  if(this->m_size & 1)
    *data++ += *data2++;

  while(data < sentinel) {
    *data++ += *data2++;
    *data++ += *data2++;
  }
}

template <class TElem>
void MJ_Vector<TElem>::mulThis(const TElem &m)
{
  TElem * v = this->m_array;
  const TElem *sentinel = &v[this->m_size];
  
  if(this->m_size & 1)
    *v++ *= m;

  while(v < sentinel) {
    *v++ *= m;
    *v++ *= m;
  }
}

template <class TElem>
void MJ_Vector<TElem>::mulThis(const MJ_Vector &vm)
{
  dimCheck1(vm);

  TElem * v = this->m_array;
  const TElem *sentinel = &v[this->m_size];
  TElem * m = vm.m_array;
  
  if(this->m_size & 1)
    *v++ *= *m++;

  while(v < sentinel) {
    *v++ *= *m++;
    *v++ *= *m++;
  }
}

template <class TElem>
void MJ_Vector<TElem>::mul(TElem m, MJ_Vector *res) const
{
  dimCheck1(*res);

  TElem *data = res->m_array;
  const TElem *data2 = this->m_array;
  const TElem *sentinel = &data[this->m_size];

  if(this->m_size & 1)
    *data++ = *data2++ * m;

  while(data < sentinel) {
    *data++ = *data2++ * m;
    *data++ = *data2++ * m;
  }
}

template <class TElem>
void MJ_Vector<TElem>::mul(const MJ_Vector &v, MJ_Vector *res) const
{
  dimCheck2(v, *res);

  TElem *target = res->m_array;
  const TElem *in1 = this->m_array;
  const TElem *in2 = v.m_array;
  const TElem *sentinel = &in1[this->m_size];

  if(this->m_size & 1)
    *target++ = *in1++ * *in2++;

  while(in1 < sentinel) {
    *target++ = *in1++ * *in2++;
    *target++ = *in1++ * *in2++;
  }
}


template <class TElem>
void MJ_Vector<TElem>::mulRamp(TElem start, TElem delta, MJ_Vector *res) const
{
  TElem *target = res->m_array;
  
  const TElem *in1 = this->m_array;
  const TElem *sentinel = & this->m_array[this->m_size];
  
  if(this->m_size & 1) {
    *target++ = *in1++ * start;
    start += delta;
  }

  while(in1 < sentinel) {
    *target++ = *in1++ * start;
    start += delta;

    *target++ = *in1++ * start;
    start += delta;
  }
}

/** Take elementwise in-place absolute value. */

template <class TElem>
void MJ_Vector<TElem>::abs()
{
  TElem *target =  this->m_array;
  
  const TElem * sentinel = target + this->m_size;

  while(target < sentinel) {
    *target = MJ_abs(*target);
    target++;
  }
}

/** Take elementwise absolute value and save the result to another vector. */

template <class TElem>
void MJ_Vector<TElem>::abs(MJ_Vector *res) const
{
  TElem *target = res->m_array;
  
  const TElem * source = this->m_array;
  const TElem * sentinel = source + this->m_size;

  while(source < sentinel) {
    *target++ = MJ_abs(*source++);
  }
}

/** Square all vector elements in-place. */

template <class TElem>
void MJ_Vector<TElem>::square()
{
  TElem *target =  this->m_array;
  const TElem * sentinel = target + this->m_size;

  while(target < sentinel) {
    TElem tmp = *target;
    *target++ = tmp * tmp;
  }
}

/** Square all vector elements and copy the result to another
    vector. The target vector \b MUST be resized *properly */

template <class TElem>
void MJ_Vector<TElem>::square(MJ_Vector *res) const
{
  TElem *target = res->m_array;

  const TElem * source = this->m_array;
  const TElem * sentinel = source + this->m_size;

  while(source < sentinel) {
    TElem tmp = *source++;
    *target++ = tmp * tmp;
  }
}

template <class TElem>
void MJ_Vector<TElem>::ramp(uint first, uint len, TElem firstVal, TElem delta)
  
{
  if((first + len) > this->m_size)
    MJ_fatal(MJ_ERR_ARGUMENT, "MJ_Vector::ramp");
  
  TElem *target =  &this->m_array[first];

  const TElem * sentinel = & target[len];
  
  if(len & 1) {
    *target++ = firstVal;
    firstVal += delta;
  }

  while(target < sentinel) {
    *target++ = firstVal;
    firstVal += delta;
    *target++ = firstVal;
    firstVal += delta;
  }  
}
template <class TElem>
void MJ_Vector<TElem>::ramp(TElem firstVal, TElem delta)
{  
  TElem *target =  this->m_array;

  const TElem * sentinel = & target[this->m_size];
  
  if(this->m_size & 1) {
    *target++ = firstVal;
    firstVal += delta;
  }

  while(target < sentinel) {
    *target++ = firstVal;
    firstVal += delta;
    *target++ = firstVal;
    firstVal += delta;
  }  
}

template <class TElem>
void MJ_Vector<TElem>::dimCheck1(const MJ_Vector &xrA) const
{
  if(xrA.size() < this->size())
    MJ_fatal(MJ_ERR_ARGUMENT, "MJ_Vector::dimCheck1");
}

template <class TElem>
void MJ_Vector<TElem>::dimCheck2(const MJ_Vector &xrA,
				 const MJ_Vector &xrB) const
{
  if(xrA.size() < this->size() || xrB.size() < this->size())
    MJ_fatal(MJ_ERR_ARGUMENT, "MJ_Vector::dimCheck2");
}



template <class TElem>
void MJ_Vector<TElem>::copySkipping(const TElem *v, 
				    uint startIndex, 
				    uint samples, 
				    uint skip)
{

  const TElem *s = & v[startIndex];
  
  if(samples > this->m_size) 
    MJ_fatal(MJ_ERR_ARGUMENT, "MJ_Vector::copySkipping");
  
  for(uint i=0; i < samples; i++) {
    this->m_array[i] = *s;
    s += skip;
  }
}

template <class TElem>
void MJ_Vector<TElem>::exportSkipping(TElem *v, 
				      uint startIndex, 
				      uint samples, 
				      uint skip) const
{
  TElem *s = & v[startIndex];
  
  if(samples > this->m_size) 
    MJ_fatal(MJ_ERR_ARGUMENT, "MJ_Vector::exportSkipping");
  
  for(uint i=0; i < samples; i++) {
    *s = this->m_array[i];
    s += skip;
  }
}

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////


template <class TElem> 
MJ_Vector2<TElem>::~MJ_Vector2()
{
  this->dropData();  
}

template <class TElem> 
bool MJ_Vector2<TElem>::resize(uint nSize, bool bFatalOnFailure )
{
  (void) bFatalOnFailure;

  if(nSize == this->m_size) return true;

  MJ_fatal(MJ_ERR_DO_NOT_CALL, "MJ_Vector2::resize");
  return true;
}
