/* COPYRIGHT
 *
 * This file is part of Mustajuuri LGPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the Lesser General Public License
 * (LGPL) - version 2.1. Extra clauses can be found from the file
 * MUSTAJUURI_LGPL and further explanations from the file
 * MUSTAJUURI_LICENSE. The license text can be found in file
 * "LGPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *    
 */

#include <mj_error.h>

// #include <mj_app_action.h>
#include <mj_error_streams.h>
// #include <mj_iostream.h>
#include <mj_mutex.h>

#include <stdarg.h>
#include <stdlib.h>
#include <stdio.h>

void (*MJ_errorHandle) (MJ_ErrorLevel, const char *) = 0;

const char *MJ_ErrorLevels[MJ_ERRL_SIZEOF] = {
  "Trace",
  "Information",
  "Warning",
  "Error",
  "Fatal error # INTENTIONAL SEGMENTATION FAULT"
};

const char *MJ_ErrorCodes[MJ_ERR_SIZEOF] = {
  "No error",
  "Unkown error",
  "Out of memory",
  "Data corrupted",
  "Internal error",
  "Array size exceeded",
  "List size exceeded",
  "Bad argument",
  "NULL argument",
  "Resource denied",
  "Virtual function called",
  "Do not call this function",
  "Not implemented yet",
  "Not initialized yet",
  "Real-time performance failure",
  "Floating point error",
  "CPU load too high",
  "Numeric overflow",
  "Input problem",
  "System resource unavailable",
  "Buffer underflow",
  "Buffer overflow",
  "Bad environment",
  "System error",
  "Bad state",
  ""
};

static MJ_MutexAuto MJ_FailureDiagnose_Mutex(true, false);

FILE *MJ_traceFile = 0;

bool MJ_errorsToFile = false;
bool MJ_silenceErrors = false;

const char *MJ_errorLevelStr(MJ_ErrorLevel level)
{
  if(level < MJ_ERRL_SIZEOF) return MJ_ErrorLevels[level];
  return "Bad error level";
}

const char *MJ_errorStr(MJ_ErrorCode code)
{
  if(code < MJ_ERR_SIZEOF) return MJ_ErrorCodes[code];
  return "Bad error type";
}

FILE * MJ_getErrorStream(MJ_ErrorLevel level)
{
  if(!MJ_traceFile) {
    MJ_traceFile = stdout;
  }
  
  FILE *stream = MJ_traceFile;

  if(MJ_errorsToFile || level == MJ_ERRL_TRACE) return stream;
  if((level == MJ_ERRL_ERROR) || (level == MJ_ERRL_FATAL))
    return stderr;

  return stdout;
}

void MJ_failure(MJ_ErrorLevel level,
		MJ_ErrorCode  code, 
		const char *c1, const char *c2, const char *c3)
{
  if(MJ_silenceErrors) return;

  char buf[4096];
  int index = 0;

  index += sprintf(buf, "MJ:%s", MJ_errorLevelStr(level));

  if(level != MJ_ERRL_TRACE && code != MJ_ERR_NO_STRING)
    index += sprintf(buf + index, " # %s", MJ_errorStr(code));
  
  if(c1) index += sprintf(buf + index, " # %s", c1);
  if(c2) index += sprintf(buf + index, " # %s", c2);
  if(c3) index += sprintf(buf + index, " # %s", c3);

  sprintf(buf + index,"\n");

  MJ_FailureDiagnose_Mutex.lock();

  FILE *s = MJ_getErrorStream(level);
  fprintf(s, buf);
  fflush(s);

  MJ_FailureDiagnose_Mutex.unlock();

  if(MJ_errorHandle)
    (*MJ_errorHandle) (level, buf);

  /* if(level != MJ_ERRL_TRACE)
     MJ_AppActions::reportError(tmpStr);*/
}

void MJ_warning(MJ_ErrorCode  code, const char *c1, const char *c2)
{
  MJ_failure(MJ_ERRL_WARNING, code, c1, c2);
}

void MJ_warning2(MJ_ErrorCode  code, const char *msg, ...)
{
  char buf[8196];
  va_list args;
  va_start(args, msg);
  vsprintf(buf, msg, args);
  MJ_warning(code, buf);
  va_end(args);
}

void MJ_error(MJ_ErrorCode  code, const char *c1, int i)
{
  char buf[32];
  sprintf(buf, "%d", i);
  
  MJ_error(code, c1, buf);
}

void MJ_error(MJ_ErrorCode  code, const char *c1)
{
  MJ_failure(MJ_ERRL_ERROR, code, c1, 0);
}

void MJ_error(MJ_ErrorCode  code, const char *c1, const char *c2)
{
  MJ_failure(MJ_ERRL_ERROR, code, c1, c2);
}

void MJ_error(MJ_ErrorCode  code, 
	      const char *c1, const char *c2, const char *c3)
{
  MJ_failure(MJ_ERRL_ERROR, code, c1, c2, c3);
}

void MJ_error2(MJ_ErrorCode  code, const char *msg, ...)
{
  char buf[8196];
  va_list args;
  va_start(args, msg);
  vsprintf(buf, msg, args);
  MJ_error(code, buf);
  va_end(args);
}

/// Conditionally print error message to the user.
void MJ_error3(bool show, MJ_ErrorCode  code, const char *msg, ...)
{
  if(!show) return;

  char buf[8196];
  va_list args;
  va_start(args, msg);
  vsprintf(buf, msg, args);
  MJ_error(code, buf);
  va_end(args);
}

void MJ_fatal(MJ_ErrorCode  code, const char *c1, const char *c2)
{
  MJ_failure(MJ_ERRL_FATAL, code, c1, c2);
  int *badAddr = 0;
  *badAddr = 1;
  exit(0); // We should never get this far...
}

void MJ_fatal2(MJ_ErrorCode  code, const char *msg, ...)
{
  char buf[8196];
  va_list args;
  va_start(args, msg);
  vsprintf(buf, msg, args);
  MJ_fatal(code, buf);
  va_end(args);
}

void MJ_trace(MJ_ErrorCode  code, const char *c1, const char *c2)
{
  MJ_failure(MJ_ERRL_TRACE, code, c1, c2);
}

void MJ_trace(const char *c1, const char *c2)
{
  MJ_trace(MJ_ERR_NONE, c1, c2);
}

void MJ_trace(const char *c1, int i)
{
  char buf[32];
  
  sprintf(buf, "%d", i);
  MJ_trace(c1, buf);
}

void MJ_trace(const char *c1, const char *c2, int i)
{
  char buf[32];
  
  sprintf(buf, "%d", i);

  MJ_failure(MJ_ERRL_TRACE, MJ_ERR_NONE, c1, c2, buf);
}

void MJ_trace(const char *c1, const void *v)
{
  char buf[32];
  
  sprintf(buf, "%p", v);
    
  MJ_trace(c1, buf);
}

void MJ_trace2(const char *msg, ...)
{
  char buf[8196];
  va_list args;
  va_start(args, msg);
  vsprintf(buf, msg, args);
  MJ_trace(buf);
  va_end(args);
}

/// Conditionally print information to the user.
void MJ_trace3(bool show, const char *msg, ...)
{
  if(!show) return;

  char buf[8196];
  va_list args;
  va_start(args, msg);
  vsprintf(buf, msg, args);
  MJ_trace(buf);
  va_end(args);
}

