/* -*- C++ -*- */

/* COPYRIGHT
 *
 * This file is part of Mustajuuri LGPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the Lesser General Public License
 * (LGPL) - version 2.1. Extra clauses can be found from the file
 * MUSTAJUURI_LGPL and further explanations from the file
 * MUSTAJUURI_LICENSE. The license text can be found in file
 * "LGPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *    
 */

#ifndef MJ_MESSAGE_FACTORY_H
#define MJ_MESSAGE_FACTORY_H

#include <string>

class MJ_Message;

class QTextStream;

/** Factory (or singleton) interface to create messages. You can add
    new factory functions to facilitate new message types on the fly.

    HOWTO: As a plugin library is opened the library must somehow (via
    static object for example) add a new custom message factory to the
    factory list. As the library is closed the plugin should remove
    the custom factory function (since it is no more functional).

    An example (in some plugin library):

    \code

class Foo
{
  Foo() { MJ_MessageFactory::addFactoryFunction("MSG_MICKEY_MOUSE", 0x123, funcPtr); }
  ~Foo() { MJ_MessageFactory::removeFactoryFunction("MSG_MICKEY_MOUSE", 0x123); }
};

static Foo;

    \endcode

    Thread safety: One must not add or remove message factories 
    while some other thread is trying to use the factory in any 
    other way. The "create"-methods are fully thread safe and can 
    be used by multiple threads in parallel.

*/

class MJ_MessageFactory
{
public:
  
  static void init();

  static MJ_Message * create(int);
  static MJ_Message * create(const char *);
  static MJ_Message * create(const std::string & str);

  static int determineId(const char *);
  static const char * determineName(int);

  static bool read(QTextStream *,MJ_Message*, int);
  static bool write(QTextStream *,MJ_Message*);

  static void addFactoryFunction
  (const char *,
   int,
   MJ_Message * (*)(),
   bool (*reader)(QTextStream *,MJ_Message*, int) = 0,
   bool (*writer)(QTextStream *,MJ_Message*) = 0);

  static void removeFactoryFunction(const std::string &,
				    int);

protected:
  MJ_MessageFactory() {}
  ~MJ_MessageFactory() {}
};

#endif
