/* -*- C++ -*- */

/* COPYRIGHT
 *
 * This file is part of Mustajuuri LGPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the Lesser General Public License
 * (LGPL) - version 2.1. Extra clauses can be found from the file
 * MUSTAJUURI_LGPL and further explanations from the file
 * MUSTAJUURI_LICENSE. The license text can be found in file
 * "LGPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *    
 */

#ifndef MJ_FILTER_COEFFS_H
#define MJ_FILTER_COEFFS_H

#include <di_types.h>

/** IIR filter coefficients. 

    MJ_FilterCoeffs stores the coefficients of a digital filter as
    coefficients of the standard difference equation:

\verbatim
    y(n) = b(0)*x(n) + b(1)*x(n-1) + ... + b(order)*x(n-order)
                     - a(1)*y(n-1) - ... - a(order)*y(n-order)

    or

            B(z)
    Y(z) = ------ .
            A(z)
\endverbatim

    <B>Pointer sharing (linking):</B> MJ_FilterCoeffs objects may
    share data pointers. Pointers are shared by using the
    "link"-method. This is mildly dangerous how-ever. The user must be
    careful to not delete the original buffers while other filters
    still point to them. The next example displays how the
    pointer sharing is used:

\code

MJ_FilterCoeffs<float> coeffs1, coeffs2;

coeffs1.setOrder(17);  // Sets the order of the coefficients
coeffs2.link(coeffs1); // Now the pointers are shared

coeffs2.setNumerator(2, 0.21); // Affects both coeffs1 and coeffs2

coeffs1.setOrder(19);  // Resize, old coefficients are lost
coeffs2.link(coeffs1); // Now the link must be re-created 

\endcode

    Pointer sharing is used to improve performance. It allows users to
    1) conserve memory when multiple identical filters are in use and
    2) avoid copying of changed (but identical) filter coefficients
    between filters.

    @see MJ_FilterDFII 
    @author Tommi Ilmonen
*/

template <class TElem> class MJ_FilterCoeffsIIR
{
public:
  MJ_FilterCoeffsIIR();
  MJ_FilterCoeffsIIR(uint order);
  MJ_FilterCoeffsIIR(const MJ_FilterCoeffsIIR &);
  MJ_FilterCoeffsIIR(MJ_FilterCoeffsIIR *);
  ~MJ_FilterCoeffsIIR();

  void empty();

  uint getOrder() const { return m_order; }
  bool setOrder(uint order, bool reset = true);
  
  TElem getNumerator(uint index)   const { return m_b[index]; }
  TElem getDenominator(uint index) const { return m_a[index]; }

  void setNumerator(uint index, TElem coeff);
  void setDenominator(uint index, TElem coeff);

  /// Pointer to numerator -- \b "b" -- buffer
  TElem *b() { return m_b; }
  /// Pointer to denominator --  \b "a" -- buffer
  TElem *a() { return m_a; }
  /// Const pointer to numerator -- \b "b" -- buffer
  const TElem *bConst() const { return m_b; }
  /// Const pointer to denominator -- \b "b" -- buffer
  const TElem *aConst() const { return m_a; }

  void copy(const MJ_FilterCoeffsIIR &);
  void link(MJ_FilterCoeffsIIR &);

  /** Copy operator. This operator uses copy() to do the real work. */
  inline MJ_FilterCoeffsIIR<TElem> &operator = (const MJ_FilterCoeffsIIR &c)
  { copy(c); return *this; }

protected:
  TElem *m_a;
  TElem *m_b;
  uint   m_order;
  bool   m_delete;
};

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

enum MJ_WindowFunction {
  MJ_WINDOW_RECTANGULAR,
  MJ_WINDOW_HANNING,
  MJ_WINDOW_HAMMING,
  MJ_WINDOW_SIZEOF
};

/** FIR filter coefficients.

    The API of this class is as close to the MJ_FilterCoeffs API to
    make application filter changes easier. For this reason the filter
    coefficients are called "numerators" even when denominators do not
    exist.

    @see MJ_FilterCoeffsIIR
    @author Tommi Ilmonen */
template <class TElem> class MJ_FilterCoeffsFIR
{
public:
  MJ_FilterCoeffsFIR();
  MJ_FilterCoeffsFIR(uint order);
  MJ_FilterCoeffsFIR(const MJ_FilterCoeffsFIR &);
  MJ_FilterCoeffsFIR(MJ_FilterCoeffsFIR *);
  ~MJ_FilterCoeffsFIR();

  void empty();

  uint getOrder() const { return m_order; }
  bool setOrder(uint order, bool reset = true);
  
  /** Get a filter coefficient. 

      @param index Coefficient index. Valid range: from zero to filter
      order (inclusive)
    
      @warning There are no safety checks.
  */ 

  TElem getNumerator(uint index) const { return m_b[index]; }

  void setNumerator(uint index, TElem coeff);

  /// Pointer to the filter coefficients
  TElem *b() { return m_b; }
  /// Const pointer to the filter coefficients
  const TElem *bConst() const { return m_b; }

  void copy(const MJ_FilterCoeffsFIR &);
  void link(MJ_FilterCoeffsFIR &);

  /** Copy operator. This operator uses copy() to do the real work. */
  inline MJ_FilterCoeffsFIR<TElem> &operator = (const MJ_FilterCoeffsFIR &c)
  { copy(c); return *this; }

  /** @name Filter design. */

  //@{

  void designLowPass(TElem frequency);

  void applyWindow(MJ_WindowFunction window = MJ_WINDOW_HANNING);

  void normalizeLowPass();

  void designLowPassNormalized(TElem frequency,
			       MJ_WindowFunction window)
  {
    designLowPass(frequency);
    applyWindow(window);
    normalizeLowPass();
  }

  //@}

protected:
  TElem *m_b;
  uint   m_order;
  bool   m_delete;
};


#endif
