/* -*- C++ -*- */

/* COPYRIGHT
 *
 * This file is part of Mustajuuri LGPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the Lesser General Public License
 * (LGPL) - version 2.1. Extra clauses can be found from the file
 * MUSTAJUURI_LGPL and further explanations from the file
 * MUSTAJUURI_LICENSE. The license text can be found in file
 * "LGPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *    
 */

#ifndef MJ_FILTER_FIR_H
#define MJ_FILTER_FIR_H

#include <mj_filter_coeffs.h>
#include <mj_ringbuffer.h>

/** FIR filter (nth order).

    This class implements a fairly optimized linear digital FIR
    (Finite Impulse Response) filter.

    @see MJ_FilterCoeffsFIR 
    @author Tommi Ilmonen */
template <class TElem> class MJ_FilterFIR
{
public:
  MJ_FilterFIR();
  MJ_FilterFIR(uint order);
  ~MJ_FilterFIR();

  /** Copy coefficients. Automatically resizes memory buffer. */
  void update(MJ_FilterCoeffsFIR<TElem> &coeffs) 
  { m_coeffs.copy(coeffs); m_buffer.resize(coeffs.getOrder()); }
  /** Link coefficients. Automatically resizes memory buffer. */
  void update(MJ_FilterCoeffsFIR<TElem> *coeffs) 
  { m_coeffs.link(*coeffs); m_buffer.resize(coeffs->getOrder()); }

  /** Sets the order of the filter. */
  inline void setOrder(uint order)
  { m_coeffs.setOrder(order); m_buffer.resize(order); }

  inline void put(TElem s);

  /// Get the current output value.
  inline const TElem & get() const { return m_val; }

  /// Resets filter buffer memory to zero.
  inline void resetMemory() 
  { m_buffer.reset(); m_buffer.setAll(0.0); m_val = 0.0; }

  /// Access the coefficients directly (discouraged)
  MJ_FilterCoeffsFIR<TElem> & coeffs() { return m_coeffs; }
  /// Access the feedback buffer directly (discouraged)
  MJ_RingBufferDelay<TElem> & buffer() { return m_buffer; }

protected:
  TElem m_val;
  MJ_FilterCoeffsFIR<TElem> m_coeffs;
  MJ_RingBufferDelay<TElem> m_buffer;
};

/** Put a new input sample to the filter. */
template <class TElem>
inline void MJ_FilterFIR<TElem>::put(TElem s)
{
  /* This implementation is fairly optimized and may look a bit
     strange...*/
  
  const TElem *bptr = m_coeffs.bConst();
  
  uint i, order = m_coeffs.getOrder();
  TElem bsum = s * *bptr++;
  
  /* These calculations are done prior to adding the new sample the
     the buffer. Thus the apparent mismatch between filter
     coefficients and delayed values. */
  for(i=0; i < order; i++) {
    bsum += *bptr++ * m_buffer.getNewest(i);
  }
  
  m_buffer.put(s);
  
  m_val = bsum;
}

#endif
