/* COPYRIGHT
 *
 * This file is part of Mustajuuri LGPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the Lesser General Public License
 * (LGPL) - version 2.1. Extra clauses can be found from the file
 * MUSTAJUURI_LGPL and further explanations from the file
 * MUSTAJUURI_LICENSE. The license text can be found in file
 * "LGPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *    
 */

#include <mj_filter_iir2.h>

#include <stdio.h> // sprintf


/** Put many samples to the filter without calculating the
    output. There can be any number of samples. */

template <class TElem>
void MJ_FilterIIR2<TElem>::putSamples(const TElem *fa, unsigned count)
{
  const TElem *famax = &fa[count];

  while(fa < famax) 
    putSample(*fa++);
}


/** Put many samples to the filter and also get them out. */

template <class TElem>
void MJ_FilterIIR2<TElem>::putSamples
(TElem * output, const TElem * input, unsigned count)
{
  const TElem *inputMax = &input[count];

  // Cache variables to registers.
  TElem m0 = m_m0;
  TElem m1 = m_m1;
  TElem m2 = m_m2;

  TElem b0 = m_b0;
  TElem b1 = m_b1;
  TElem b2 = m_b2;
  
  TElem a1 = m_a1;
  TElem a2 = m_a2;

  while(input < inputMax) {
    m2 = m1;
    m1 = m0;
    m0 = (*input++) - a1 * m1 - a2 * m2;
    *output++ = b0 * m0 + b1 * m1 +  b2 * m2;
  }

  m_m0 = m0;
  m_m1 = m1;
  m_m2 = m2;
}

/** Put many samples to the filter and also get them out. 

    The result will we stored in place of the input.
*/

template <class TElem>
void MJ_FilterIIR2<TElem>::putGetSamplesInplace(TElem *inputOutput, 
					 unsigned count)
{
  const TElem *inputMax = & inputOutput[count];

  // Cache variables to registers.
  TElem m0 = m_m0;
  TElem m1 = m_m1;
  TElem m2 = m_m2;

  TElem b0 = m_b0;
  TElem b1 = m_b1;
  TElem b2 = m_b2;
  
  TElem a1 = m_a1;
  TElem a2 = m_a2;

  while(inputOutput < inputMax) {
    m2 = m1;
    m1 = m0;
    m0 = (*inputOutput) - a1 * m1 - a2 * m2;
    *inputOutput++ = b0 * m0 + b1 * m1 +  b2 * m2;
  }

  m_m0 = m0;
  m_m1 = m1;
  m_m2 = m2;  
}

/** Design a parameteric EQ. This method does not check the arguments
    for validity. You may well corrupt the filter with improper
    parameters (negative bandwidth etc). 

    See MJ_dspParametricEq
*/


template <class TElem>
void MJ_FilterIIR2<TElem>::designParametric1(
				      /// Absolute gain
				      TElem boostAbs, 
				      /// Center frequency in radians
				      TElem centerRad,
				      /// Bandwidth in radians
				      TElem bandWidthRad)
{
  if(boostAbs == 1.0) {
    flatResponse();
    return;
  }

  TElem gb = (boostAbs + 1.0) * 0.5;
  
  iir2ParametricEq(1.0, boostAbs, gb, centerRad, bandWidthRad,
		   bptr(), aptr());
}

/** Design a parameteric EQ. This method differs from
    designParametric1 by the parameters it has. This method performs
    few checks on the arguments. It does - however - check that the
    center frequency does not exceed frequency Nyquist * 0.8. This
    check is done to ensure that the filter is stable.

    See MJ_dspParametricEq */

template <class TElem>
void MJ_FilterIIR2<TElem>::designParametric2(
				      /// Gain in Decibels
				      TElem boostDb, 
				      /// Center frequency in Hertz
				      TElem centerHz,
				      /// Bandwidth in octaves
				      TElem bandWidthOctaves,
				      /// Sampling rate
				      TElem samplingRate)
{

  TElem boostAbs = pow(10.0, boostDb / 20.0);

  TElem limitFreq = samplingRate * 0.4; // Way below Nyquist

  if(centerHz > limitFreq) centerHz = limitFreq;

  /* Scale between zero and pi. */
  TElem centerRad = centerHz / samplingRate * MJ_2pi;
  
  TElem bandWidthRad = centerRad * bandWidthOctaves;
  designParametric1(boostAbs, centerRad, bandWidthRad);
}

/** Write the coeficients to a buffer. The data is formatted to match
    Matlabs filter representation. You can just drop the resulting
    text into Matlab and enjoy yourself.
    
    The caller must provide a buffer large enough to contain the
    data. 256 bytes should be enough at all cases.  */

template <class TElem>
void MJ_FilterIIR2<TElem>::writeTo(char *buf)
{
  sprintf(buf, "b = [ %9.9f %9.9f %9.9f ]; a =  [ 1.0 %9.9f %9.9f ];",
	  m_b0, m_b1, m_b2, m_a1, m_a2);
}

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

// FILTER DESIGN

/** Design a second order linear parameteric filter. 

    No safety checks.

    See matlab-file parmeq.m 
*/
template <class TElem>
void MJ_FilterIIR2<TElem>::iir2ParametricEq
(TElem gr, TElem ga, TElem gb, TElem w0, TElem dw,
 TElem *b, TElem *a)
{
  TElem tmp = ga * ga - gb * gb;
  if(tmp < 0.0) tmp = -tmp;

  TElem beta = MJ_tan(dw * 0.5) * 
    MJ_sqrt(MJ_abs(gb * gb - gr * gr))  / MJ_sqrt(tmp);
  
  /* printf("gr = %f ga = %f gb = %f w0 = %f dw = %f BETA = %f", 
     gr, ga, gb, w0, dw, beta);
     fflush(0); */

  TElem scale = 1.0 / (1.0 + beta);
  tmp = ga * beta;
  TElem cosw = MJ_cos(w0);

  b[0] = (gr + tmp) * scale;
  b[1] = (-2.0 * gr * cosw) * scale;
  b[2] = (gr - tmp) * scale;

  a[0] = -2.0 * cosw * scale;
  a[1] = (1.0 - beta) * scale;
}


template <class TElem>
void MJ_FilterIIR2<TElem>::iir2ParametricEqB(
			 /// Linear center gain
			 TElem K, 
			 /// Linear edge freq gain
			 TElem F, 
			 /// Center frequency in radians
			 TElem omega, 
			 /// Band width in octaves
			 TElem bw, 
			 /// Pointer to "b" coefficients
			 TElem *b, 
			 /// Pointer to "a" coefficients		
			 TElem *a)
{
  TElem os = MJ_sin(omega);
  
  TElem gamma = MJ_sqrt(K*(F*F-1)/(K*K-F*F)) * 
    sinh(MJ_ln22 * bw * (omega/ os)) * os;

  TElem sk = sqrt(K);
  TElem gmsk = gamma*sk;
  TElem gdsk = gamma/sk;

  b[0] = (1.0 + gmsk) / (1.0 + gdsk);
  b[1] = (-2*MJ_cos(omega))/(1.0 + gamma/MJ_sqrt(K));
  b[2] = (1.0 - gmsk) / (1.0 + gdsk);
  a[0] = b[1];
  a[1] = (1.0 - gdsk)/(1.0 + gdsk);
}

/** Parametric EQ design. See matlab file "peq.m". 

    We assume G0 = 1.0;
*/

template <class TElem>
void MJ_FilterIIR2<TElem>::iir2ParametricEqC(
			 /// Linear center gain
			 TElem G, 
			 /// Linear edge freq gain
			 TElem GB, 
			 /// Center frequency in radians
			 TElem w0, 
			 /// Band width in octaves
			 TElem bw, 
			 /// Pointer to "b" coefficients
			 TElem *b, 
			 /// Pointer to "a" coefficients		
			 TElem *a)
{
  TElem Dw = bw * w0;
  TElem G2 = G * G;
  TElem GB2 = GB * GB;
  TElem w02 = w0 * w0;
  TElem Dw2 = Dw * Dw;

  TElem F = MJ_abs(G2 - GB2);
  TElem G00 = MJ_abs(G2 - 1.0);
  TElem F00 = MJ_abs(GB2 - 1.0);

  TElem tmp = w02 - MJ_piSquared;
  tmp *= tmp;
  TElem tmp2 = F00 * MJ_piSquared * Dw2 / F;
  TElem num = tmp + G2 * tmp2;
  TElem den = tmp + tmp2;

  TElem G1 = MJ_sqrt(num/den);
  TElem G12 = G1 * G1;

  
  TElem G01 = MJ_abs(G2  - G1);
  TElem G11 = MJ_abs(G2  - G12);
  TElem F01 = MJ_abs(GB2 - G1);
  TElem F11 = MJ_abs(GB2 - G12);
 
  tmp = MJ_tan(w0 / 2.0);
  TElem W2 = MJ_sqrt(G11 / G00) * tmp * tmp;
  TElem DW = (1.0 + MJ_sqrt(F00 / F11) * W2) * MJ_tan(Dw/2);
  
  TElem C = F11 * DW * DW - 2 * W2 * (F01 - MJ_sqrt(F00 * F11));
  TElem D = 2 * W2 * (G01 - MJ_sqrt(G00 * G11));
  
  TElem A = MJ_sqrt((C + D) / F);
  TElem B = MJ_sqrt((G2 * C + GB2 * D) / F);
  
  TElem scale = 1.0 / (1.0 + W2 + A);

  b[0] = (G1 + W2 + B) * scale;
  b[1] = -2*(G1 - W2) * scale;
  b[2] = (G1 - B + W2) * scale;

  a[0] = -2*(1 - W2)  * scale;
  a[1] = (1 + W2 - A)  * scale;
}

/** Design a second order high-shelving filter.
    
    See matlab-file high_shelf.m and url 
    http://shoko.calarts.edu/~majordom/music-dsp/1998c/msg00555.html

*/

template <class TElem>
void MJ_FilterIIR2<TElem>::iir2HighShelveEq
(TElem omega, TElem gainDb, TElem slope,
 TElem *b, TElem *a)
{
  TElem A  = MJ_pow(10, gainDb/40);
  TElem sn = MJ_sin(omega);
  TElem cs = MJ_cos(omega);

  TElem tmp = A-1;

  TElem beta  = MJ_sqrt( (A * A + 1)/slope - tmp * tmp );

  TElem scale = 1.0 / ((A+1) - (A-1)*cs + beta*sn); // 1.0 / a0
  
  a[0] =  2*( (A-1) - (A+1)*cs)          * scale;
  a[1] =     ((A+1) - (A-1)*cs - beta*sn) * scale;
  
  b[0] =    A* ( (A+1) + (A-1)*cs + beta*sn ) * scale;
  b[1] = -2*A* ( (A-1) + (A+1)*cs)            * scale;
  b[2] =    A* ( (A+1) + (A-1)*cs - beta*sn)  * scale;
}


/** Design a second order low-shelving filter.
    
    See matlab-file low_shelf.m and url 
    http://shoko.calarts.edu/~majordom/music-dsp/1998c/msg00555.html

 */

template <class TElem>
void MJ_FilterIIR2<TElem>::iir2LowShelveEq
(TElem omega, TElem gainDb, TElem slope,
 TElem *b, TElem *a)
{
  TElem A  = MJ_pow(10, gainDb/40);
  omega = MJ_pi - omega;
  TElem sn = MJ_sin(omega);
  TElem cs = MJ_cos(omega);

  TElem tmp = A-1;

  TElem beta  = MJ_sqrt( (A * A + 1)/slope - tmp * tmp );

  TElem scale = 1.0 / ((A+1) - (A-1)*cs + beta*sn); // 1.0 / a0
  
  a[0] = -2*( (A-1) - (A+1)*cs)          * scale;
  a[1] =    ((A+1) - (A-1)*cs - beta*sn) * scale;
  
  b[0] =    A* ( (A+1) + (A-1)*cs + beta*sn ) * scale;
  b[1] =  2*A* ( (A-1) + (A+1)*cs)            * scale;
  b[2] =    A* ( (A+1) + (A-1)*cs - beta*sn)  * scale;
}

template <class TElem>
void MJ_FilterIIR2<TElem>::iir2LowPassEq (TElem omega, TElem Q,
				   TElem *b, TElem *a)
{
  TElem cs = MJ_cos(omega);

  TElem tmp = 1.0 - cs;
  TElem tmp2 = tmp * 0.5;
  b[0] = tmp2;
  b[1] = tmp;
  b[2] = tmp2;
  
  tmp2 = MJ_sin(omega) / (2.0 * Q); // tmp2 = alpha

  tmp = 1.0 / (1 + tmp2) ; // 1.0 / a0 = scale
  
  a[0] = (-2.0 * cs)  * tmp;
  a[1] = (1.0 - tmp2) * tmp;
}

template <class TElem>
void MJ_FilterIIR2<TElem>::iir2HighPassEq (TElem omega, TElem Q,
				    TElem *b, TElem *a)
{
  TElem cs = MJ_cos(omega);

  TElem tmp = 1.0 + cs;
  TElem tmp2 = tmp * 0.5;
  b[0] = tmp2;
  b[1] = -tmp;
  b[2] = tmp2;
  
  tmp2 = MJ_sin(omega) / (2.0 * Q); // tmp2 = alpha

  tmp = 1.0 / (1 + tmp2) ; // 1.0 / a0 = scale
  
  a[0] = (-2.0 * cs) * tmp;
  a[1] = (1.0 - tmp2) * tmp;
}

/** Design a band-pass filter based on center frequency and Q-value.

    @arg omega Center frequency in radians
    @arg bw Bandwidth
*/


template <class TElem>
void MJ_FilterIIR2<TElem>::iir2BandPassEq (TElem omega, TElem Q,
					   TElem *b, TElem *a)
{
  const TElem ncs2 = -2.0 * MJ_cos(omega);
  const TElem alpha = MJ_sin(omega) / (2.0 * Q);
  const TElem scale = 1.0 / (1.0 + alpha);

  b[0] = alpha;
  b[1] = 0.0;
  b[2] = -alpha;
  

  a[0] = ncs2 * scale;
  a[1] = (1.0 - alpha) * scale;
}


/** Design a band-pass filter based on center frequency and bandwidth.

    @arg omega Center frequency in radians
    @arg bw Bandwidth
*/

template <class TElem>
void MJ_FilterIIR2<TElem>::iir2BandPassEqB (TElem omega, 
					    TElem bw,
					    TElem *b, TElem *a)
{
  const TElem ncs2 = -2.0 * MJ_cos(omega);
  const TElem sn   = MJ_sin(omega);
  const TElem alpha = sn * MJ_sinh(MJ_ln22 * bw * omega / sn);
  const TElem scale = 1.0 / (1.0 + alpha);

  b[0] = alpha;
  b[1] = 0.0;
  b[2] = -alpha;
  

  a[0] = ncs2 * scale;
  a[1] = (1.0 - alpha) * scale;
}

