/* -*- C++ -*- */

/* COPYRIGHT
 *
 * This file is part of Mustajuuri LGPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the Lesser General Public License
 * (LGPL) - version 2.1. Extra clauses can be found from the file
 * MUSTAJUURI_LGPL and further explanations from the file
 * MUSTAJUURI_LICENSE. The license text can be found in file
 * "LGPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *    
 */

#ifndef MJ_LOOK_UP_FUNCTION_H
#define MJ_LOOK_UP_FUNCTION_H

#include <di_types.h>

/** Fast table look-up function interpolation. 
 *
 * @author Tommi Ilmonen.
 */

class MJ_LookUpFunction
{
public:
  enum { MASK = 63, POINTS = 64 };
  
  /** Default functions. The order of these enums should not changed
   * due to backwards compatibility. */

  enum Function {
    LINEAR,
    SQRT,
    POW2,
    SIN,
    PARABOLIC,
    FUNCTIONS
  };

  MJ_LookUpFunction();
  ~MJ_LookUpFunction();

  void calculateLinear();
  void calculateSqrt();
  void calculatePow2();
  void calculateSin01();
  void calculateParabolic();

  /** 
   * Calculates the value of the function in given point.
   * 
   * @param in Input value in range [0,1]
   * 
   * @return Interpolated value in range [0,1]
   */

  inline float value(float in)
  {
    in *= (float) MASK;

    if(in <= 0.0f)
      return 0.0f;
    else if(in >= MASK)
      return 1.0f;

    uint base = (uint) in;
    float w = in - base;

    return m_values[base] * (1.0f - w) + m_values[base + 1] * w;
  }
  
  static MJ_LookUpFunction & getFunc(Function c) { return m_funcs[c]; }
  static MJ_LookUpFunction & getFunc(uint c) { return m_funcs[c]; }

  static const char * getFuncName(Function c);
  static const char * getFuncName(uint c) { return getFuncName((Function) c); }

protected:

  float m_values[POINTS];

  static MJ_LookUpFunction m_funcs[FUNCTIONS];

public:

  class Init;
};

#endif
