/* -*- C++ -*- */

/* COPYRIGHT
 *
 * This file is part of Mustajuuri LGPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the Lesser General Public License
 * (LGPL) - version 2.1. Extra clauses can be found from the file
 * MUSTAJUURI_LGPL and further explanations from the file
 * MUSTAJUURI_LICENSE. The license text can be found in file
 * "LGPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *    
 */

#ifndef MJ_MATH_H
#define MJ_MATH_H

#include <math.h>

#include <mj_config.h>

/** \file mj_math.h

    Overloaded mathematical functions.
    
    <B>WHY ?</B>

    Standard libm defines functions with different names for different
    data types. This is a problem, if you typedef variables and want
    to change the type afterwards - by default the functions remain
    what they were, resulting in either loss of precicion (float to
    double change) or excessive precision (double to float). The same
    problem arises with template systems.
   
    To make the transition between data types transparent, Mustajuuri
    defines an overloaded series of common mathematics functions with
    the same name and both float and double arguments.

    Besides the inline wrappers, mj_math also introduces some useful
    functions usually available as macros - minimum and maximum value
    for example. Function is better than macro.

    There are also some useful constants (pi, ln(2) etc.)
   
 */

/// The value of pi.
#define MJ_pi  3.1415926535897931
/// The value of 2 * pi.
#define MJ_2pi 6.2831853071795862
/// The value of pi^2.
#define MJ_piSquared 9.869604401089358
/// The value of pi / 2.
#define MJ_pi2 1.57079632679489660000
/// The value of pi / 4.
#define MJ_pi4 0.78539816339744828

/// The value of ln(2)
#define MJ_ln2 0.693147180559945290
/// The value of ln(2) / 2
#define MJ_ln22 0.346573590279972640


/* Float wrappers. */

inline float MJ_cos(float v)
{
  return cosf(v);
}

inline float MJ_log(float v)
{
  return logf(v);
}

inline float MJ_exp(float v)
{
  return expf(v);
}

inline float MJ_pow(float v, float power)
{
  return powf(v, power);
}

inline float MJ_square(float v)
{
  return v * v;
}

inline float MJ_sin(float v)
{
  return sinf(v);
}

inline float MJ_sinh(float v)
{
  return sinhf(v);
}

inline float MJ_sqrt(float v)
{
  return sqrtf(v);
}

inline float MJ_tan(float v)
{
  return tanf(v);
}

inline float MJ_atan2(float v, float u)
{
  return atan2f(v,u);
}

/* Double wrappers. */

inline double MJ_cos(double v)
{
  return cos(v);
}

inline double MJ_log(double v)
{
  return log(v);
}

inline double MJ_exp(double v)
{
  return exp(v);
}

inline double MJ_pow(double v, double power)
{
  return pow(v, power);
}

inline double MJ_square(double v)
{
  return v * v;
}

inline double MJ_sin(double v)
{
  return sin(v);
}

inline double MJ_sinh(double v)
{
  return sinh(v);
}

inline double MJ_sqrt(double v)
{
  return sqrt(v);
}

inline double MJ_tan(double v)
{
  return tan(v);
}

inline double MJ_atan2(double v, double u)
{
  return atan2(v,u);
}

// Clamp (limit) to the given range
template <class T>
inline T MJ_clamp(T v, T min, T max)
{
  if(v < min) return min;
  if(v > max) return max;
  return v;
}

inline double MJ_clampD(double v, double min, double max)
{
  if(v < min) return min;
  if(v > max) return max;
  return v;
}

inline float MJ_clampF(float v, float min, float max)
{
  if(v < min) return min;
  if(v > max) return max;
  return v;
}

inline MJ_Sample MJ_clampS(MJ_Sample v, MJ_Sample min, MJ_Sample max)
{
  if(v < min) return min;
  if(v > max) return max;
  return v;
}

// Minmimum value
template <class T>
inline T MJ_min(T x1, T x2)
{
  if(x1 > x2) return x2;
  return x1;
}

// Maximum value
template <class T>
inline T MJ_max(T x1, T x2)
{
  if(x1 > x2) return x1;
  return x2;
}

/// Absolute value

template <class T>
inline T MJ_abs(T x1)
{
  if(x1 > 0.0) return x1;
  else return -x1;
}

//@name Decibel conversions
//@{
/// Convert absolute amplitude to Decibels.
inline double MJ_absToDb(double d)
{ return (log10(d)) * 20.0; }
inline double MJ_squareToDb(double d)
{ return (log10(d)) * 10.0; }
/// Convert absolute amplitude to Decibels with clipping point.
inline double MJ_absToDb(double d, double clip)
{
  double tmp = (log10(d)) * 20.0;
  return tmp > clip ? tmp : clip;
}
/// Convert Decibels to absolute amplitude.
inline double MJ_dbToAbs(double d)
{  return pow(10.0, d * 0.05); }
inline double MJ_dbToSquare(double d)
{ return pow(10.0, d * 0.1); }

/// Convert absolute amplitude to Decibels.
inline float MJ_absToDb(float d)
{ return (log10f(d)) * 20.0; }
inline float MJ_squareToDb(float d)
{ return (log10f(d)) * 10.0; }
/// Convert Decibels to absolute amplitude.
inline float MJ_dbToAbs(float d)
{ return powf(10.0, d * 0.05); }
inline float MJ_dbToSquare(float d)
{ return powf(10.0, d * 0.1); }
//@}

/// Convert pitch relative cents to multiplier
inline float MJ_centToMul(float cents)
{
  return powf(2.0, cents / 1200.0);
}
/// Convert multiplier to relative pitch
inline float MJ_mulToCent(float mul)
{ return logf(mul) / MJ_ln2 * 1200.0; }

/// Convert pitch relative cents to multiplier
inline double MJ_centToMul(double cents)
{ return pow(2.0, cents / 1200.0); }
/// Convert multiplier to relative pitch
inline double MJ_mulToCent(double mul)
{ return log(mul) / MJ_ln2 * 1200.0; }

unsigned MJ_findDivider(unsigned);
/** Check if number is prime. @see MJ_findDivider. */
inline bool MJ_isPrime(unsigned l) { return !MJ_findDivider(l); }
unsigned MJ_nextPrime(unsigned l);
unsigned MJ_previousPrime(unsigned l);

inline float MJ_metersToSeconds(float meters) { return meters / 334.0f; }
inline double MJ_metersToSeconds(double meters) { return meters / 334.0; }

template <class T>
inline int MJ_compare(const T & v1, const T & v2)
{
  if(v1 > v2)
    return 1;
  else if(v1 == v2)
    return 0;
  return -1;
}

#endif
