/* -*- C++ -*- */

/* COPYRIGHT
 *
 * This file is part of Mustajuuri LGPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the Lesser General Public License
 * (LGPL) - version 2.1. Extra clauses can be found from the file
 * MUSTAJUURI_LGPL and further explanations from the file
 * MUSTAJUURI_LICENSE. The license text can be found in file
 * "LGPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *    
 */

#ifndef MJ_RAMP_H
#define MJ_RAMP_H

#include <mj_math.h>

enum MJ_RampType {
  MJ_RAMP_LINEAR,
  MJ_RAMP_EXPONENTIAL
};

/** Linear Interpolator.

    @author Tommi Ilmonen
 */
template <class TElem> class MJ_RampLinear
{
public:
  MJ_RampLinear() { clear(); }
  MJ_RampLinear(TElem value) : m_current(value), 
			       m_step(0.0), m_target(value), m_counter(0) { }

  /// Get the current sample.
  TElem get() const { return m_current; }

  /// Reset the ramp to zero
  void clear() { m_current = m_step = m_target = 0.0; m_counter = 0; }

  void decrementCounter()
  {if(m_counter)m_counter--;}

  /// Advance the interpolation 
  void advance() 
  { 
    if(m_counter) { 
      m_current += m_step; 
      m_counter --; 
    } 
    else m_current = m_target;
  }

  void set(unsigned interpolateSamples, 
	   TElem current, 
	   TElem target)
  {
    m_target = target;
    m_current = current;
    m_counter = interpolateSamples;
    m_step = (target - current) / (TElem) interpolateSamples;
  }

  /// Set the target value.
  void setTarget(unsigned interpolateSamples, TElem target)
  {
    m_target = target;
    m_counter = interpolateSamples;
    m_step = (target - m_current) / (TElem) interpolateSamples;
  }

  /** Recaluculate step size. This method is useful if the
      interpolation period is so long that there is a risk of
      incorrect behaviour due to inadequate numeric resolution. It is
      always ok to call this method. */
  void recalculateStep()
  { if(m_counter) m_step = (m_target - m_current) / (TElem) m_counter; }

  /// Force the ramp to reach the target value.
  inline void reachTarget() { m_current = m_target; m_counter = 0; }

  /// Reset the ramp to the given (constant) value.
  inline void reset(TElem value) 
  { m_current = m_target = value; m_counter = 0; }

  /// Returns true if the interpolation is still going on.
  inline bool isActive() const { return m_counter; }

protected:

  /// Current value.
  TElem       m_current;
  /// The increment that should be done at each step.
  TElem       m_step;
  TElem       m_target;
  unsigned    m_counter;

};

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

/** Linear/exponential interpolator. This class should not be used to
    do really long interpolations.

    @author Tommi Ilmonen
*/

template <class TElem> class MJ_Ramp
{
public:
  MJ_Ramp() { clear(); }
  MJ_Ramp(TElem value) : m_type(MJ_RAMP_LINEAR), m_current(value), 
			 m_step(0.0), m_target(value), m_counter(0) { }

  /// Reset the ramp to zero
  void clear() { m_current = m_step = m_target = 0.0; m_counter = 0; }

  void decrementCounter()
  {if(m_counter)m_counter--;}

  /// Advance the interpolation lienarly by one sample.
  void advanceLinear() 
  { 
    if(m_counter) { 
      m_current += m_step; 
      m_counter --; 
    } 
    else m_current = m_target;
  }

  void advanceLinearMac(TElem *target, const TElem *source, unsigned n);
  void advanceLinearMul(TElem *target, const TElem *source, unsigned n);

  /// Advance the interpolation exponentially by one sample.
  void advanceExponential() 
  { 
    if(m_counter) { 
      m_current *= m_step; 
      m_counter --; 
    } 
    else m_current = m_target;
  }

  void advance() 
  { 
    if(m_counter) { 
      if(m_type == MJ_RAMP_LINEAR) m_current += m_step; 
      else m_current *= m_step; 
      m_counter --; 
    } 
    else m_current = m_target;
  }

  void advanceN(unsigned n) 
  { 
    if(m_counter > n) { 
      if(m_type == MJ_RAMP_LINEAR) m_current += (n * m_step); 
      else
	for(unsigned i=0; i < n; i++) m_current *= m_step;
      m_counter -= n; 
    } 
    else m_current = m_target;
  }

  /// Get the current sample.
  TElem get() const { return m_current; }
  /// Get the current target value.
  TElem target() const { return m_target; }
  /// Get the current step value.
  TElem step() const { return m_step; }
  /// Returns the number of samples until the next target is reached.
  inline unsigned samplesToTarget() const { return m_counter; }


  void set(unsigned interpolateSamples, 
	   TElem current, 
	   TElem target, 
	   MJ_RampType type)
  {
    m_target = target;
    m_current = current;
    m_counter = interpolateSamples;
    m_type = type;
    if(type == MJ_RAMP_LINEAR)
      m_step = (target - current) / (TElem) interpolateSamples;
    else 
      m_step = MJ_pow(target / current, 
		      (TElem) 1.0 / (TElem) interpolateSamples);
  }
  
  /// Set the target value.
  void setTarget(unsigned interpolateSamples, TElem target, MJ_RampType type)
  {
    if(!interpolateSamples) {
      reset(target);
      return;
    }

    m_target = target;
    m_counter = interpolateSamples;
    m_type = type;
    if(type == MJ_RAMP_LINEAR)
      m_step = (target - m_current) / (TElem) interpolateSamples;
    else 
      m_step = MJ_pow(target / m_current, 
		      (TElem) 1.0 / (TElem) interpolateSamples);
  }
  
  /// Force the ramp to reach the target value.
  inline void reachTarget() { m_current = m_target; m_counter = 0; }

  /// Reset the ramp to the given (constant) value.
  inline void reset(TElem value) 
  { m_type = MJ_RAMP_LINEAR; m_current = m_target = value; m_step = 0; m_counter = 0; }

  /// Returns true if the interpolation is still going on.
  inline bool isActive() const { return m_counter; }

  MJ_Ramp & operator = (TElem v) { reset(v); return * this; }

protected:

  MJ_RampType m_type;
  /// Current value.
  TElem       m_current;
  /// The increment that should be done at each step.
  TElem       m_step;
  TElem       m_target;
  unsigned    m_counter;
};

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////


/// Helper for MJ_MultiRamp
template <class TElem> class MJ_RampPoint 
{
public:

  MJ_RampType m_type;
  TElem       m_value;
  long        m_time;
};

/** Special-purpose multi-segment interpolator. This class is most
    certainly not easy to use.

\code

MJ_MultiRamp<float> ramp;

// Maximum 20 points for interpolation.
ramp.resize(20);

// Then prepare the ramp for operation
ramp.clear();

// Now initialize the 10 target points with addTarget

ramp.addTarget(10, 0.5, MJ_RAMP_LINEAR);
ramp.addTarget(100, 0.1, MJ_RAMP_LINEAR);
ramp.addTarget(250, 0.100, MJ_RAMP_EXPONENTIAL);

ramp.prepare();

// Now the ramp can be used to both interpolate new values and new
// interpolation points can be added on the fly.

ramp.get();
ramp.get();
ramp.addTarget(300, 0.2, MJ_RAMP_LINEAR);
ramp.get();
ramp.get();

\endcode


*/

template <class TElem> class MJ_MultiRamp
{
public:
  MJ_MultiRamp(unsigned size = 10) 
    : m_points(0), m_counter(0), m_nofPoints(0) { resize(size);  }

  MJ_MultiRamp(const MJ_MultiRamp &r)
    : m_points(0), m_counter(0), m_nofPoints(0) { *this = r; }

  ~MJ_MultiRamp();

  bool resize(unsigned points);

  void advance();

  TElem get() const { return m_ramp.get(); }

  /// Use discouraged.
  MJ_Ramp<TElem> &ramp() { return m_ramp; }

  void clear()
  { m_pointsLeft = 0; m_counter = 0; }

  void prepare()
  { m_ramp.setTarget(m_points->m_time, m_points->m_value,m_points->m_type); }
  
  bool addTarget(long time, TElem value, MJ_RampType type);

  MJ_MultiRamp<TElem> &operator = (const MJ_MultiRamp &r);

protected:

  void nextPoint();

  MJ_Ramp<TElem>       m_ramp;
  MJ_RampPoint<TElem> *m_points;
  long                 m_counter;
  unsigned             m_nofPoints;
  unsigned             m_pointsLeft;
};

#endif
