/* -*- C++ -*- */

/* COPYRIGHT
 *
 * This file is part of Mustajuuri LGPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the Lesser General Public License
 * (LGPL) - version 2.1. Extra clauses can be found from the file
 * MUSTAJUURI_LGPL and further explanations from the file
 * MUSTAJUURI_LICENSE. The license text can be found in file
 * "LGPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *    
 */

#ifndef MJ_RESAMPLER_H
#define MJ_RESAMPLER_H

#include <mj_filter_fir.h>
#include <mj_ringbuffer.h>

/** Sample rate conversion utility. This class can use arbitrary and
    dynamic resampling ratios. Thus it is well suited to high-quality
    sample-playback sound synthesis and sound file resampling.

    First we upsample the signal and then interpolate new values from
    the upsampled signal.

    \code
    MJ_Resampler<float> resampler;
    double rate = myConversionRate; // Any value greater than zero.

    for(uint i=0; i < max; i++) {
      float newVal = resample.getResampled(myConversionRate);
      while(resample.needNewSample()) {
        resample.put(getNewSampleFromSource());
      }
    } 

    \endcode
    
*/

template <class TElem> class MJ_Resampler
{
public:
  MJ_Resampler() : m_lastInput(0), m_multiply(1), m_currPos(0.0) {}

  ~MJ_Resampler() {}

  void setParameters(TElem frequencyLimit,
		     uint upsample = 4, 
		     uint filterOrder = 16);

  void redesignFilter(TElem frequencyLimit)
  { m_filter.coeffs().designLowPassNormalized(frequencyLimit / m_multiply,
					      MJ_WINDOW_HAMMING); }
  /** Put a new sample to the resampler. */

  void put(TElem v);

  /** Check if the converter needs a new sample to convert
      anything. */

  bool needNewSample() const
  { return m_currPos <= 1.0; }

  /** Get a new resampled value from the converter.
      
      \b Note: You should never call this method without first
      checking that new samples can be obtained. Use method
      needNewSample to do that.

      @arg rate The conversion rate ie. the relationship between the
      original sampling rate and resampled sampling rate.

  */

  inline TElem getResampled(double rate)
  {
    // TElem ret = m_buffer.getInterpolatedNewest3DFast(m_currPos);
    TElem ret = m_buffer.getInterpolatedNewest1D(m_currPos);
    m_currPos -= (rate * m_multiply);
    return ret;
  }

  /** Returns the low-pass filter. Use of this method is
      discouraged. */
  MJ_FilterFIR<TElem> & filter() { return m_filter; }

protected:

  TElem  m_lastInput;
  uint   m_multiply;
  /** Current read-head position in upsampled delay buffer. */
  double m_currPos;

  MJ_FilterFIR<TElem>   m_filter;
  MJ_RingBuffer2<TElem> m_buffer;
};


#endif
