/* COPYRIGHT
 *
 * This file is part of FLUID
 *
 * Author: Tommi Ilmonen, Janne Kontkanen 1999-2002.
 * Tommi.Ilmonen@hut.fi, Janne.Kontkanen@hut.fi
 *
 * This file is licensed under the Lesser General Public License
 * (LGPL) - version 2.1. 
 *    
 */
#ifndef FLUID_VECTOR2T_HPP
#define FLUID_VECTOR2T_HPP

#include "di_types.h"
#include <iostream>
#include <math.h>

namespace Fluid {

  template <class T>
  class Vector2T
  {
  public:
    T		x;										// x-component of the vector
    T		y;										// y-component of the vector
  
    Vector2T	()						                {}
    Vector2T	(T cx, T cy)				                { x = (T)cx;	y = (T)cy; }
    template <class S>		Vector2T(const S * v)	                        { x = v[0]; y = v[1]; }
    // Vector2T	(double cx, double cy)				                { x = (T)(cx);	y = (T)(cy); }
    template <class S>		Vector2T	(const Vector2T<S>& v)          { x = (T)v.x; y = (T)v.y; }

    template <class S>		Vector2T& operator=  (const Vector2T<S>& v)	{ x = (T)v.x; y = (T)v.y; return *this; }
    template <class S>		Vector2T& operator=  (const S * v)	        { x = v[0]; y = v[1]; return *this; }

    Vector2T&	clear		(void)						{ x = (T)(0); y = (T)(0); return *this; }
    Vector2T&	make		(T cx, T cy)			                { x = cx; y = cy; return *this; }
    Vector2T&	make		(T xy)					{ x = (xy); y = (xy); return *this; }
    T *           data() { return &x; }
    const T *     data() const { return &x; }
    bool		operator==  (const Vector2T& src) const		                { return (x == src.x && y == src.y); }
    bool		operator!=  (const Vector2T& src) const		                { return !(x == src.x && y == src.y); }
    Vector2T&	operator+=	(const Vector2T& v)				{ x += v.x, y += v.y; return *this; }		
    Vector2T&	operator-=	(const Vector2T& v)				{ x -= v.x, y -= v.y; return *this; }
    Vector2T&	operator+=	(T s)					        { x += s, y += s; return *this; }
    Vector2T&	operator-=	(T s)					        { x -= s, y -= s; return *this; }
    Vector2T&	operator*=	(T s)					        { x = (x*s), y = (T)(y*s); return *this; }
    Vector2T&	operator/=	(T s)					        { s = 1.0/s; x = (x*s), y = (y*s); return *this; }
    bool		isOne		(void) const					{ return (x == 1.0f && y == 1.0f); }
    bool		isZero		(void) const					{ return (x == 0.0f && y == 0.0f); }
    double	length		(void) const				        { return sqrt(x*x+y*y); }
    double	lengthSqr	(void) const				        { return x*x+y*y; }
    Vector2T&	negate		(void)						{ x=-x; y=-y; return *this; }
    Vector2T&	normalize	(double len = 1.0)				{ double l = length(); if (l!=0.0) *this *= (len/l); return *this; }
    Vector2T&	scale		(const Vector2T& v)				{ x *= v.x, y *= v.y; return *this; }
    Vector2T&	descale		(const Vector2T& v)				{ x /= v.x, y /= v.y; return *this; }
    Vector2T&	rotate		(double s, double c)			        { T t = x; x = (T)(x*c+y*-s); y = (T)(t*s+y*c); return *this; }
    Vector2T&	rotate		(double angle)					{ return rotate(sin(angle), cos(angle)); }
    double	angle		(void) const				        { return atan2(y, x); }
    Vector2T&	clampUnit	(void)						{ if(x <= (T)0.0) x = (T)0.0; else if(x >= (T)1.0) x = (T)1.0; if(y <= (T)0.0) y = (T)0.0; else if(y >= (T)1.0) y = (T)1.0; return *this; }
    const	T&			operator[]	(int i) const			{ return ((T*)this)[i]; }
    T&			        operator[]	(int i)				{ return ((T*)this)[i]; }
  
    template <class S>
    void copy(const S * data) { x = data[0]; y = data[1]; }
  };

  template <class T> inline	Vector2T<T>	operator+	(const Vector2T<T>& v1, const Vector2T<T>& v2)	{ return Vector2T<T>(v1.x+v2.x, v1.y+v2.y); }
  template <class T> inline	Vector2T<T>	operator+	(const Vector2T<T>& v1, T v2)	{ return Vector2T<T>(v1.x+v2, v1.y+v2); }
  template <class T> inline	Vector2T<T>	operator-	(const Vector2T<T>& v1, const Vector2T<T>& v2)	{ return Vector2T<T>(v1.x-v2.x, v1.y-v2.y); }
  template <class T> inline	Vector2T<T>	operator-	(const Vector2T<T>& v1, T v2)	{ return Vector2T<T>(v1.x-v2, v1.y-v2); }
  template <class T> inline	Vector2T<T>	operator*	(const Vector2T<T>& v, const double s)		{ return Vector2T<T>(v.x*s, v.y*s); }
  template <class T> inline	Vector2T<T>	operator*	(const double s, const Vector2T<T>& v)		{ return v*s; }
  template <class T> inline	Vector2T<T>	operator/	(const Vector2T<T>& v, const double s)		{ double r = 1.0/s; return v*r; }
  template <class T> inline	Vector2T<T>	operator-	(const Vector2T<T>& v)						{ return Vector2T<T>(-v.x, -v.y); }

  template <class T>
  inline float abs(Vector2T<T> t)
  {
    return t.length();
  }

  template <class T>
  inline float dot(const Vector2T<T> &t1, const Vector2T<T> &t2)
  {
    return t1.x * t2.x + t1.y * t2.y;
  }

  template <class T>
  inline std::ostream &operator<<(std::ostream &os, const Fluid::Vector2T<T> &t)
  {
    os << t.x << ' ' << t.y;
    return os;
  }

  template <class T>
  inline std::istream &operator>>(std::istream &is, Fluid::Vector2T<T> &t)
  {
    is >> t.x;
    is >> t.y;
    return is;
  }

  typedef Vector2T<float> Vector2; 

} // namespace

#endif
