/* COPYRIGHT
 *
 * This file is part of the Gesture Recognition Toolkit - GRT
 *
 * Author:    Tommi Ilmonen, 2001.
 * Copyright: Tommi Ilmonen, 2001.
 *
 * Tommi.Ilmonen@hut.fi
 *    
 */

#ifndef FLUID_TIME_STAMP_H
#define FLUID_TIME_STAMP_H

#include <di_types.h> // For SGI

#include <stdio.h>
#include <sys/time.h>

namespace Fluid {

  /** Time stamp a la NTP. Time stamp is expressed as a 64-bit integer
      (or fixed point number). The first 32 bits indicate number of
      seconds while the next 32 bits indicate the fractional part of the
      second. */

  class TimeStamp
  {
  public:
    /** The type that actually holds the data. In practice it is a
	64-bit integer (long long). */
    typedef llong stamp_t;
  
    /// The maximum value that the timestamp variable can achieve.
    // static const stamp_t max_v = (stamp_t) -1;

    TimeStamp(stamp_t t = 0) : m_stamp(t) {}
    ~TimeStamp() {}

    static inline stamp_t getTimeOfDayStamp();
    inline void getTimeOfDay() { m_stamp = getTimeOfDayStamp(); }

    void zero() { m_stamp = 0; }
    TimeStamp &  setSeconds(uint t) 
    { ullong tmp = t; m_stamp = (tmp << 32); return * this; }
    TimeStamp &  setSeconds(float t) 
    { m_stamp = (stamp_t) (t * (float) ((unsigned) -1)); return * this; }
    TimeStamp & setSeconds(double t) 
    { m_stamp = (stamp_t) (t * (double) ((unsigned) -1)); return * this; }

    inline void write(char *) const;
    inline void read(const char *);
  
    /// Full seconds.
    unsigned secondsU() const { return (unsigned) (m_stamp >> 32); }
    int secondsI() const { return (int) (m_stamp >> 32); }
    unsigned subSecondsU() const { return (unsigned) m_stamp; }
    /// Seconds as floating point value
    float secondsF() const { return ((float) m_stamp) / (float) ((stamp_t) 0xffffffff); }
    double secondsD() const { return ((double) m_stamp) / (double) ((stamp_t) 0xffffffff); }
    float subSeconds() const 
    { return ((float) ((unsigned) m_stamp)) / (float) 0xffffffff; }

    stamp_t stamp() const { return m_stamp; }

    inline bool operator <= (const TimeStamp &ts) const
    { return m_stamp <= ts.m_stamp; }

    inline bool operator < (const TimeStamp &ts) const
    { return m_stamp < ts.m_stamp; }

    inline bool operator > (const TimeStamp &ts) const
    { return m_stamp > ts.m_stamp; }

    inline bool operator >= (const TimeStamp &ts) const
    { return m_stamp >= ts.m_stamp; }

    inline bool operator != (const TimeStamp &ts) const
    { return m_stamp != ts.m_stamp; }

    inline bool operator == (const TimeStamp &ts) const
    { return m_stamp == ts.m_stamp; }

    inline TimeStamp operator + (const TimeStamp &ts) const
    { return TimeStamp(m_stamp + ts.m_stamp); }

    inline TimeStamp & operator += (const TimeStamp &ts)
    { m_stamp += ts.m_stamp; return *this; }

    inline TimeStamp & operator *= (double v)
    { m_stamp = (stamp_t) (m_stamp * v); return *this; }

    inline TimeStamp operator * (double v) const
    { return TimeStamp((stamp_t) (m_stamp * v)); }

    inline TimeStamp operator * (int v) const
    { return TimeStamp((m_stamp * (stamp_t) v)); }

    inline TimeStamp & operator /= (double v)
    { m_stamp = (stamp_t) (m_stamp / v); return *this; }

    inline TimeStamp operator / (double v) const
    { return TimeStamp((stamp_t) (m_stamp / v)); }

    inline TimeStamp operator / (int v) const
    { return TimeStamp((m_stamp / (stamp_t) v)); }

    inline TimeStamp operator - (const TimeStamp &ts) const
    { return TimeStamp(m_stamp - ts.m_stamp); }

    inline TimeStamp & operator -= (const TimeStamp &ts)
    { m_stamp -= ts.m_stamp; return *this; }

    inline TimeStamp & operator = (const stamp_t &t)
    { m_stamp = t; return *this; } 

  protected:

    stamp_t m_stamp;
  };

  // Inined functions:

  TimeStamp::stamp_t TimeStamp::getTimeOfDayStamp()
  {
    struct timeval tp;
    
    gettimeofday(&tp,0);
    
    double fraction = 0.000001 * (double) tp.tv_usec;
    
    stamp_t tmp = tp.tv_sec;
    tmp = tmp << 32;
    
    unsigned lfraction = 
      (unsigned) (fraction * (double) ((unsigned) 0xffffffff));
    
    return tmp + (stamp_t) lfraction;
  }
  
  
  void TimeStamp::write(char *str) const
  {
    sprintf(str, "%u %u", secondsU(), subSecondsU());
  }

  void TimeStamp::read(const char *str)
  {
    unsigned sec = 0, sub = 0;

    sscanf(str, "%u %u", & sec, & sub);

    unsigned long long tmp = sec;

    tmp = tmp << 32;
    tmp += sub;    

    m_stamp = tmp;
  }
  

} // namespace

#endif
