
/* COPYRIGHT
 *
 * This file is part of the Visual Effects Engine - VEE
 *
 * Read the "VEE-LICENSE" file for the license.
 *
 * Authors & Copyright:   
 *
 * Tommi Ilmonen, Tuukka Heikura, Marko Myllymaa and 
 * Janne Kontkanen 2001-2004
 *
 * Additional copyrights: Tekes 2003-2004
 *
 * firstname.lastname@hut.fi
 *
 */

#include <vee_blow.h>

#include <vee_error.h>
#include <vee_math.h>
#include <vee_object.h>
#include <vee_particle.h>
#include <vee_scene.h>
#include <vee_strs.h>
#include <vee_value.h>


VEE_Blow::VEE_Blow()
  : m_range(1.0), m_magnitude(1.0), m_blown(0.0)
{
  m_rotation.identity();

  m_rotationInverse.identity();
  m_forceType = VEE_FORCE_DRAG;
}

VEE_Blow::~VEE_Blow()
{}


/** Receive and process a value. 

    Parameters:

    <UL>
    
    <LI> "agelimit", float, range: 0 to inf

    <LI> "direction", Vector3

    <LI> "magnitude", float, range: -1 to 1

    <LI> "range", float, range: 0 to inf

    <LI> "rotation", Matrix3

    </UL>

*/
bool VEE_Blow::setValue(const char *id, VEE_Value *msg)
{
  // VEE_trace("VEE_Blow::setValue # %s", id);

  if(VEE_Strs::equals(id, VEE_Strs::rotation)) {
    VEE_Matrix3 v;

    bool ok = msg->getMatrix3(&v);

    if(ok)
      setRotation(v);
    return ok;
  } 
  else if(VEE_Strs::equals(id, VEE_Strs::direction)) {
    VEE_Vector3 v;

    bool ok = msg->getVector3(&v);

    if(ok)
      setDirection(v);
    return ok;
  }
  else if(VEE_Strs::equals(id, VEE_Strs::magnitude)) {
    float v;

    bool ok = msg->getFloat(&v);

    if(ok)
      m_magnitude = v;

    return ok;
  }
  else if(VEE_Strs::equals(id, VEE_Strs::range)) {
    float v;

    bool ok = msg->getFloat(&v);

    if(ok)
      m_range = v;

    return ok;
  }
  else 
    return VEE_ForceObject::setValue(id, msg);
}

VEE_Object * VEE_Blow::clone() const
{
  return new VEE_Blow(*this);
}

/** Alternative way to set the direction of the blow field. */
void VEE_Blow::setDirection(const VEE_Vector3 &v)
{
  VEE_Matrix3 ry, rz;

  float ay = atan2f(-v[2], v[0]);
  float az = atan2f(v[1], sqrtf(v[0] * v[0] + v[2] * v[2]));

  ry.rotateY(ay);
  rz.rotateZ(az);

  setRotation(ry * rz);
}

void VEE_Blow::getForce(VEE_Vector3 *force,
			const VEE_Vector3 &,
			const VEE_Vector3 &location)
{
  float f=0, l, t1;

  VEE_Vector3 c(0.5 * m_range, 0.0, 0.0);
  VEE_Vector3 t(0.5 * m_range, 0.2 * m_range, 0.2 * m_range);
  VEE_Vector3 x;
  VEE_Vector3 p = m_rotationInverse*(location-m_location)-c;
  
  for(uint i=0; i<3; i++) {
    l = p[i] / t[i];
    f += l * l;
  }
  
  if(f<1) {
    t = m_rotation * (location - m_location);

    t1 = sqrt(t[1] * t[1] + t[2] * t[2]) / (0.2 * m_range);

    x[0] = m_magnitude * t1 * sqrt((m_range-t[0]) / m_range);
    x[1] = 0;
    x[2] = 0;
    *force = m_rotation * x;
  }
  else {
    force->clear();
  }
}

void VEE_Blow::update(float dt)
{
  VEE_ForceObject::update(dt);

  m_blown += dt * m_magnitude;
}

void VEE_Blow::boundingBoxOfForce(VEE_BoundingBox *box)
{
  VEE_Vector3 r(m_range, m_range, m_range);
  box->low() = m_location-r;
  box->high() = m_location+r;
}

VEE_Object * VEE_Blow::factory()
{
  return new VEE_Blow;
}
