/* -*- C++ -*- */

/* COPYRIGHT
 *
 * This file is part of the Visual Effects Engine - VEE
 *
 * Read the "VEE-LICENSE" file for the license.
 *
 * Authors & Copyright:   
 *
 * Tommi Ilmonen, Tuukka Heikura, Marko Myllymaa and 
 * Janne Kontkanen 2001-2004
 *
 * Additional copyrights: Tekes 2003-2004
 *
 * firstname.lastname@hut.fi
 *
 */

#ifndef VEE_CLONABLE_POINTER_H
#define VEE_CLONABLE_POINTER_H

/** This class implements pointer-carrier/cloning functionality. When
    you need to store and \b copy objects of abstract type then this
    class is handy. The base class that is the type "T" must have a
    method "T * clone() const" that is used to copy (or clone)
    objects.

    @author Tommi Ilmonen
*/

template <typename T>
class VEE_ClonablePointer
{
public:
  /// Construct pointer carrier with NULL pointer
  VEE_ClonablePointer() : m_ptr(0) {}
  /// Takes hold of a pointer (does not clone)
  VEE_ClonablePointer(T *ptr) : m_ptr(ptr) {}
  /// Clone a pointer from another carrier
  VEE_ClonablePointer(const VEE_ClonablePointer &that) 
    : m_ptr(that.m_ptr ? that.m_ptr->clone() : 0) {}
  
  /// Deletes the hidden pointer
  ~VEE_ClonablePointer() { delete m_ptr; }
    
  /// Clone a pointer from another carrier
  VEE_ClonablePointer & operator = (const VEE_ClonablePointer &that)
  { 
    if(this != &that) {
      delete m_ptr; 
      m_ptr = that.m_ptr ? that.m_ptr->clone() : 0; 
    }
    return *this; 
  }

  /// Takes hold of a pointer (does not clone)
  VEE_ClonablePointer & operator = (T * that)
  { delete m_ptr; m_ptr = that ; return *this; }

  /// Returns the pointer (potentially NULL).
  T * ptr() { return m_ptr; }

  /// Returns the pointer (potentially NULL).
  const T * ptr() const { return m_ptr; }

  /// Returns clone of the pointer (or NULL).
  T * clone() const { return m_ptr ? m_ptr->clone() : 0; }

  /// Forgets the pointer without deleting it.
  void forget() { m_ptr = 0; }

protected:
  T * m_ptr;
};

#endif
