/* COPYRIGHT
 *
 * This file is part of the Visual Effects Engine - VEE
 *
 * Read the "VEE-LICENSE" file for the license.
 *
 * Authors & Copyright:   
 *
 * Tommi Ilmonen, Tuukka Heikura, Marko Myllymaa and 
 * Janne Kontkanen 2001-2004
 *
 * Additional copyrights: Tekes 2003-2004
 *
 * firstname.lastname@hut.fi
 *
 */

#include <vee_curve_particle.h>

#include <vee_algorithm.h>
#include <vee_scene.h>
#include <vee_value.h>

VEE_CurveParticle::VEE_CurveParticle()
  : m_knotInerval(0.5),
    m_activeKnots(0)
{
  m_objectType = VEE_OBJECT_CURVE_PARTICLE;
  resizeCurve(20);
}

VEE_CurveParticle::~VEE_CurveParticle()
{}

bool VEE_CurveParticle::setValue(const char *id, VEE_Value *v)
{
  if(VEE_Particle::setValue(id, v))
    return true;

  bool ok = false;

  if(strcmp(id, "interval") == 0) {
    float tmp;
    ok = v->getFloat(&tmp);
    if(ok)
      m_knotInerval = tmp;
  }
  else if(strcmp(id, "knots") == 0) {
    int tmp;
    ok = v->getInt(&tmp);
    if(ok)
      resizeCurve(tmp);
  }
    

  return ok;
}

VEE_Variable VEE_CurveParticle::getVariable(const char *id) const
{
  VEE_Variable v = VEE_Particle::getVariable(id);
  if(v.isValid())
    return v;

  if(strcmp(id, "interval") == 0) {
    v = m_knotInerval;
  }
  else if(strcmp(id, "knots") == 0) {
    v =  (int) m_locations.size();
  }

  return v;
}

void VEE_CurveParticle::updateVelocity(float dt)
{
  VEE_Particle::updateVelocity(dt);

  uint i, useKnots = (uint) (m_age / m_knotInerval);

  if(!m_activeKnots) {
    VEE_Algorithm::setAll(m_locations.begin(), m_locations.end(), m_location);
    VEE_Algorithm::setAll(m_velocities.begin(), 
			  m_velocities.end(), m_velocity);
    m_activeKnots = 1;
  }
  
  if(useKnots > m_activeKnots) {

    uint old = m_activeKnots;

    m_activeKnots = VEE_min<unsigned int>(m_activeKnots + 1, m_locations.size());

    if(old != m_activeKnots) {
      uint older = old - 1;
      VEE_Vector3 diff = m_locations[older] - m_locations[old];
      diff.normalize(m_velocities[older].length());
      m_velocities[old] = diff;
    }
  }

  VEE_Vector3 v;
  v.clear();

  // external forces
  if(VEE_abs(m_mass) > 1.0e-5) {

    VEE_Algorithm::setAll(m_forces.begin(), m_forces.end(), v);
    
    sceneBox()->getForces(VEE_FORCE_GRAVITY, this,
			  &* m_forces.begin(),m_activeKnots ,
			  &* m_velocities.begin(), 
			  &* m_locations.begin());
    float tmp = m_mass * dt;

    for(i=0; i < m_activeKnots; i++)
      m_velocities[i] += m_forces[i] * tmp;
  }
  if(VEE_abs(m_area) > 1.0e-5) {

    VEE_Algorithm::setAll(m_forces.begin(), m_forces.end(), v);

    sceneBox()->getForces(VEE_FORCE_DRAG, this,
			  &* m_forces.begin(), m_activeKnots ,
			  &* m_velocities.begin(), 
			  &* m_locations.begin());
    float tmp = m_area * dt;

    for(i=0; i < m_activeKnots; i++)
      m_velocities[i] += m_forces[i] * tmp;
  }
}

void VEE_CurveParticle::update(float dt)
{
  VEE_Particle::update(dt);

  uint i;

  for(i=0; i < m_activeKnots; i++)
    m_locations[i] += m_velocities[i] * dt;
}
  
VEE_Object *VEE_CurveParticle::clone() const
{
  return new VEE_CurveParticle(*this);
}

VEE_Object * VEE_CurveParticle::factory()
{
  return new VEE_CurveParticle;
}

void VEE_CurveParticle::resizeCurve(uint n)
{
  VEE_Vector3 v;
  v.clear();

  m_locations.resize(n);
  VEE_Algorithm::setAll(m_locations.begin(), m_locations.end(), v);

  m_velocities = m_locations;
  m_forces     = m_locations;
}
