/* COPYRIGHT
 *
 * This file is part of the Visual Effects Engine - VEE
 *
 * Read the "VEE-LICENSE" file for the license.
 *
 * Authors & Copyright:   
 *
 * Tommi Ilmonen, Tuukka Heikura, Marko Myllymaa and 
 * Janne Kontkanen 2001-2004
 *
 * Additional copyrights: Tekes 2003-2004
 *
 * firstname.lastname@hut.fi
 *
 */

#include <vee_explosion.h>

#include <vee_bounding_box.h>
#include <vee_strs.h>
#include <vee_value.h>

VEE_Explosion::VEE_Explosion()
  : m_range(1.0),
    m_magnitude(1.0),
    m_currentShockRange(0.0)
{
  m_ageLimit = 1.0;
}

VEE_Explosion::VEE_Explosion(const VEE_Vector3 & location, 
			     float range,
			     float magnitude, 
			     float ageLimit)
  : VEE_ForceObject(location, ageLimit),
    m_range(range),
    m_magnitude(magnitude),
    m_currentShockRange(0.0)
{}

VEE_Explosion::~VEE_Explosion() 
{}

/** Receive and process a value. 

    Parameters:

    <UL>
    
    <LI> "agelimit", float, range: 0 to inf

    <LI> "magnitude", float, range: -1 to 1

    <LI> "range", float, range: -1 to 1

    <LI> "currentshockrange", float, range: -1 to 1

    </UL>

*/
bool VEE_Explosion::setValue(const char *id, VEE_Value *msg)
{
  if(VEE_ForceObject::setValue(id, msg))
    return true;

  bool ok = false;

  if(VEE_Strs::equals(id, "currentshockrange")) {
    float v;

    ok = msg->getFloat(&v);

    if(ok)
      m_currentShockRange = v;
  }
  else if(VEE_Strs::equals(id, VEE_Strs::magnitude)) {
    float v;

    ok = msg->getFloat(&v);

    if(ok)
      m_magnitude = v;
  }
  else if(VEE_Strs::equals(id, VEE_Strs::range)) {
    float v;

    ok = msg->getFloat(&v);

    if(ok)
      m_range = v;
  }

  return ok;
}

VEE_Variable VEE_Explosion::getVariable(const char *id) const
{
  VEE_Variable v;

  if(VEE_Strs::equals(id, "currentshockrange")) {
    v = m_currentShockRange ;
  }
  else if(VEE_Strs::equals(id, VEE_Strs::magnitude)) {
    v = m_magnitude;
  }
  else if(VEE_Strs::equals(id, VEE_Strs::range)) {
    v = m_range;
  }
  else 
    v = VEE_ForceObject::getVariable(id);
  
  return v;
}

VEE_Object * VEE_Explosion::clone() const
{
  return new VEE_Explosion(*this);
}

void VEE_Explosion::update(float dt)
{
  m_age += dt;

  m_currentShockRange = m_range * (m_age / m_ageLimit);

  m_done = m_age > m_ageLimit;
}

void VEE_Explosion::getForce(VEE_Vector3 *force,
			     const VEE_Vector3 &velocity,
			     const VEE_Vector3 &location)
{
  VEE_Vector3 diff = location - m_location;

  float distance = diff.length();

  if(distance > m_currentShockRange || distance < 1e-5)
    return;
  
  /* We could do the following via "Unitize" and then use m_magnitude
     to get the desired speed, this method is slightly faster. */
  VEE_Vector3 targetVel = diff * (m_magnitude / distance);
  
  *force = targetVel - velocity;
}

void VEE_Explosion::boundingBoxOfForce(VEE_BoundingBox *box)
{
  VEE_Vector3 tmp(m_range, m_range, m_range);

  box->low()  = m_location - tmp;
  box->high() = m_location + tmp;
}

VEE_Object * VEE_Explosion::factory()
{
  return new VEE_Explosion;
}
