/* -*- C++ -*- */

/* COPYRIGHT
 *
 * This file is part of the Visual Effects Engine - VEE
 *
 * Read the "VEE-LICENSE" file for the license.
 *
 * Authors & Copyright:   
 *
 * Tommi Ilmonen, Tuukka Heikura, Marko Myllymaa and 
 * Janne Kontkanen 2001-2004
 *
 * Additional copyrights: Tekes 2003-2004
 *
 * firstname.lastname@hut.fi
 *
 */


#ifndef VEE_FORCE_OBJECT_H
#define VEE_FORCE_OBJECT_H

#include <vee_particle2.h>
#include <vee_object_store.h>

class VEE_BoundingBox;

/** Base class for force objects. 

    @author Tommi Ilmonen
*/

class VEE_ForceObject : public VEE_Particle2
{
public:
  /// Default force type is gravity.
  VEE_ForceObject() : m_forceType(VEE_FORCE_GRAVITY), m_moved(false),
		      m_renderAsForce(true),
		      m_lastUpdateFrame(-1)
  { m_updateVelocity = false; m_objectType = VEE_OBJECT_FORCE; }

  VEE_ForceObject(const VEE_Vector3 &location,  const VEE_Vector3 &velocity, 
		  float ageLimit = 1e+10) 
    : VEE_Particle2(location, velocity, ageLimit), 
      m_forceType(VEE_FORCE_GRAVITY), m_moved(false),
		      m_renderAsForce(true),
		      m_lastUpdateFrame(-1) 
  {m_updateVelocity = false; m_objectType = VEE_OBJECT_FORCE; }

  VEE_ForceObject(const VEE_Vector3 &location,
		  float ageLimit = 1e+10) 
    : VEE_Particle2(location, VEE_Vector3ZERO, ageLimit), 
      m_forceType(VEE_FORCE_GRAVITY), m_moved(false),
      m_renderAsForce(true),
      m_lastUpdateFrame(-1) 
  {m_updateVelocity = false; m_objectType = VEE_OBJECT_FORCE; }

  virtual ~VEE_ForceObject();

  virtual bool setValue(const char *, VEE_Value *);
  virtual VEE_Variable getVariable(const char *) const;

  virtual VEE_Vector3 *centralLocation() { return &m_location; }

  virtual void setForceType(VEE_ForceType t) { m_forceType = t; }
  virtual VEE_ForceType getForceType() { return m_forceType; }
  virtual void getForce(VEE_Vector3 *force,
			const VEE_Vector3 &velocity,
			const VEE_Vector3 &location) = 0;

  /** Get the bounding box of this force object. Outside the bounding
      box the force must be zero. This method is needed since the
      scene boxes need to be able to determine where in the scene
      hierarchy the force object should be registered
      (VEE_SceneBox::m_forceObjects). */
  virtual void boundingBoxOfForce(VEE_BoundingBox *box) = 0;

  virtual void render(VEE_Renderer *);

  virtual void update(float dt);

  void linkBox(VEE_SceneBox *);
  void unlinkBox(VEE_SceneBox *);
  void unlinkBoxes();
  void forgetLinks() { m_boxes.clear(); }
  bool isLinked(VEE_SceneBox *b) { return m_boxes.contains(b); }
  void updateSceneLocation(int frame);

  // typedef VEE_PointerStoreList<VEE_SceneBox> boxStore;
  typedef VEE_PointerStoreSet<VEE_SceneBox> boxStore;

  boxStore &boxes() { return m_boxes; }

  bool renderAsForce() const { return m_renderAsForce; }

protected:
  VEE_ForceType m_forceType;

  boxStore    m_boxes;
  // VEE_PointerStoreHash<VEE_SceneBox> m_boxes;
  // VEE_PointerStoreSet<VEE_SceneBox> m_boxes;
  bool m_moved;
  bool m_renderAsForce;
  int m_lastUpdateFrame;
};


void VEE_calculateCubeBounds(VEE_BoundingBox *bb,
			     const VEE_Matrix3 &rotation,
			     const VEE_Vector3 &corner);

#endif
