/* COPYRIGHT
 *
 * This file is part of the Visual Effects Engine - VEE
 *
 * Read the "VEE-LICENSE" file for the license.
 *
 * Authors & Copyright:   
 *
 * Tommi Ilmonen, Tuukka Heikura, Marko Myllymaa and 
 * Janne Kontkanen 2001-2004
 *
 * Additional copyrights: Tekes 2003-2004
 *
 * firstname.lastname@hut.fi
 *
 */

#include <vee_force_plane.h>
#include <vee_math.h>
#include <vee_object.h>
#include <vee_particle.h>
#include <vee_scene.h>
#include <vee_strs.h>
#include <vee_value.h>

using namespace Fluid;

VEE_ForcePlane::VEE_ForcePlane()
  : m_normal(0.0, 1.0, 0.0), 
    m_epsilon(0.1), 
    m_magnitude(1.0), 
    m_range(1.0),
    m_radius(1.0)
{
  m_forceType = VEE_FORCE_GRAVITY;
}

VEE_ForcePlane::~VEE_ForcePlane()
{
}

/** Receive and process a value. 

    Parameters:

    <UL>
    
    <LI> "agelimit", float, range: 0 to inf

    <LI> "epsilon", float, range: 0 to inf

    <LI> "magnitude", float, range: -1 to 1

    <LI> "radius", float, range: 0 to inf

    <LI> "range", float, range: 0 to inf

    <LI> "normal", Vector3

    <LI> "velocity", Vector3

    </UL>

*/
bool VEE_ForcePlane::setValue(const char *id, VEE_Value *msg)
{
  if(VEE_Strs::equals(id, VEE_Strs::epsilon)) {
    float v;

    bool ok = msg->getFloat(&v);

    if(ok)
      m_epsilon = v;
    return ok;
  }
  else if(VEE_Strs::equals(id, VEE_Strs::magnitude)) {
    float v;

    bool ok = msg->getFloat(&v);

    if(ok)
      m_magnitude = v;
    return ok;
  }
  else if(VEE_Strs::equals(id, VEE_Strs::radius)) {
    float v;

    bool ok = msg->getFloat(&v);

    if(ok)
      m_radius = v;
    return ok;
  }
  else if(VEE_Strs::equals(id, VEE_Strs::range)) {
    float v;

    bool ok = msg->getFloat(&v);

    if(ok)
      m_range = v;
    return ok;
  }
  else if(VEE_Strs::equals(id, VEE_Strs::normal)) {
    VEE_Vector3 v;

    bool ok = msg->getVector3(&v);

    if(ok)
      setNormal(v);
    return ok;
  }
  else 
    return VEE_ForceObject::setValue(id, msg);
}

VEE_Variable VEE_ForcePlane::getVariable(const char *id) const
{
  VEE_Variable v;

  if(VEE_Strs::equals(id, VEE_Strs::epsilon)) {
    v = m_epsilon;
  }
  else if(VEE_Strs::equals(id, VEE_Strs::magnitude)) {
    v = m_magnitude;
  }
  else if(VEE_Strs::equals(id, VEE_Strs::radius)) {
    v = m_radius;
  }
  else if(VEE_Strs::equals(id, VEE_Strs::range)) {
    v = m_range;
  }
  else if(VEE_Strs::equals(id, VEE_Strs::normal)) {
    v = m_normal;
  }
  else 
    v = VEE_ForceObject::getVariable(id);

  return v;
}

VEE_Object * VEE_ForcePlane::clone() const
{
  return new VEE_ForcePlane(*this);
}

void VEE_ForcePlane::setNormal(const VEE_Vector3 &n)
{
  m_normal = n;
  m_normal.normalize();
}


void VEE_ForcePlane::getForce(VEE_Vector3 *force,
			      const VEE_Vector3 &velocity,
			      const VEE_Vector3 &location)
{
  (void) velocity;
  VEE_Vector3 p, n;
  float d, l;

  /* Minimum distance between the point and the plane. */
  p = location - m_location;
  d = dot(p, m_normal);
  n = d * m_normal;
  l = (p - n).length();

  if(l > m_radius) {
    force->clear();
  }
  else {
    if(d > m_range) {
      force->clear();
    }
    else {
      // m_normal.normalize();
      *force = m_normal * m_magnitude / (d * d + m_epsilon);
    }
  }
}

void VEE_ForcePlane::boundingBoxOfForce(VEE_BoundingBox *box)
{
  VEE_Vector3 r(m_radius, m_radius, m_radius);

  box->low() = m_location - r;
  box->high() = m_location + r;
}

VEE_Object * VEE_ForcePlane::factory()
{
  return new VEE_ForcePlane();
}
