/* COPYRIGHT
 *
 * This file is part of the Visual Effects Engine - VEE
 *
 * Read the "VEE-LICENSE" file for the license.
 *
 * Authors & Copyright:   
 *
 * Tommi Ilmonen, Tuukka Heikura, Marko Myllymaa and 
 * Janne Kontkanen 2001-2004
 *
 * Additional copyrights: Tekes 2003-2004
 *
 * firstname.lastname@hut.fi
 *
 */

#include <vee_gravitate.h>

#include <vee_error.h>
#include <vee_math.h>
#include <vee_object.h>
#include <vee_particle.h>
#include <vee_scene.h>
#include <vee_strs.h>
#include <vee_value.h>

VEE_Gravitate::VEE_Gravitate()
  : m_epsilon(0.2), 
    m_magnitude(1.0), 
    m_decayPower(2.0),
    m_maxForce(500.0)
{
  m_range = 1.0;
  m_forceType = VEE_FORCE_GRAVITY;  
}

VEE_Gravitate::VEE_Gravitate(VEE_Vector3 location,
			     VEE_Vector3 velocity,
			     float ageLimit,
			     float m,
			     bool updateVelocity)
  : VEE_ForceObject(location, velocity, ageLimit),
    m_epsilon(0.2), 
    m_magnitude(m), 
    m_decayPower(1.0),
    m_maxForce(500.0)
{
  m_updateVelocity = updateVelocity;
  m_range = m / 0.05;
  m_forceType = VEE_FORCE_GRAVITY;
}

/** Receive and process a value. 

    Parameters:

    <UL>
    
    <LI> "decaypower", float, range: 0 to inf

    <LI> "epsilon", float, range: 0 to inf

    <LI> "magnitude", float, range: -inf to inf

    <LI> "range", float, range: 0 to inf

    <LI> "maxforce", float, range: -inf to inf

    </UL>

*/
bool VEE_Gravitate::setValue(const char *id, VEE_Value *msg)
{
  if(VEE_Strs::equals(id, VEE_Strs::decaypower)) {
    float v;

    bool ok = msg->getFloat(&v);

    if(ok)
      m_decayPower = v;
    return ok;
  }
  else if(VEE_Strs::equals(id, VEE_Strs::epsilon)) {
    float v;

    bool ok = msg->getFloat(&v);

    if(ok)
      m_epsilon = v;
    return ok;
  }
  else if(VEE_Strs::equals(id, VEE_Strs::magnitude)) {
    float v;

    bool ok = msg->getFloat(&v);

    if(ok)
      m_magnitude = v;
    return ok;
  }
  else if(VEE_Strs::equals(id, VEE_Strs::range)) {
    float v;

    bool ok = msg->getFloat(&v);

    if(ok)
      m_range = v;
    return ok;
  }
  else if(VEE_Strs::equals(id, "maxforce")) {
    float v;

    bool ok = msg->getFloat(&v);

    if(ok)
      m_maxForce = v;
    return ok;
  }
  else 
    return VEE_ForceObject::setValue(id, msg);
}

VEE_Variable VEE_Gravitate::getVariable(const char *id) const
{
  VEE_Variable v;

  if(VEE_Strs::equals(id, VEE_Strs::decaypower)) {
    v = m_decayPower;
  }
  else if(VEE_Strs::equals(id, VEE_Strs::epsilon)) {
    v = m_epsilon;
  }
  else if(VEE_Strs::equals(id, VEE_Strs::magnitude)) {
    v = m_magnitude;
  }
  else if(VEE_Strs::equals(id, VEE_Strs::range)) {
    v = m_range;
  }
  else if(VEE_Strs::equals(id, "maxforce")) {
    v = m_maxForce;
  }
  else 
    v = VEE_ForceObject::getVariable(id);
  
  return v;
}

VEE_Object * VEE_Gravitate::clone() const
{
  return new VEE_Gravitate(*this);
}

void VEE_Gravitate::getForce(VEE_Vector3 *force,
			     const VEE_Vector3 &,
			     const VEE_Vector3 &location)
{
  // if(m_location != location) {
  VEE_Vector3 diff = m_location - location;

  float d = diff.length();
  
  if(d > m_range)
    force->clear();
  else if(d < m_epsilon)
    force->clear();
  else {
    float f = m_magnitude / powf(d, m_decayPower);

    float fa = VEE_abs(f);
    if(fa > m_maxForce) f = f > 0.0 ? m_maxForce : -m_maxForce;

    diff.normalize(f);
    *force = diff;
  }
}
  
void VEE_Gravitate::boundingBoxOfForce(VEE_BoundingBox *box)
{
  VEE_Vector3 r(m_range, m_range, m_range);

  box->low()  = m_location - r;
  box->high() = m_location + r;
}

VEE_Object * VEE_Gravitate::factory()
{
  return new VEE_Gravitate;
}
