/* COPYRIGHT
 *
 * This file is part of the Visual Effects Engine - VEE
 *
 * Read the "VEE-LICENSE" file for the license.
 *
 * Authors & Copyright:   
 *
 * Tommi Ilmonen, Tuukka Heikura, Marko Myllymaa and 
 * Janne Kontkanen 2001-2004
 *
 * Additional copyrights: Tekes 2003-2004
 *
 * firstname.lastname@hut.fi
 *
 */

#include <vee_tube.h>

#include <vee_bounding_box.h>
#include <vee_error.h>
#include <vee_math.h>
#include <vee_value.h>
#include <vee_strs.h>

VEE_Tube::VEE_Tube()
  : m_height(1.0),
    m_radius(1.0),
    m_effectRange(0.5),
    m_magnitude(1.0)
{
  m_rotation.identity();
  m_rotationInv = m_rotation;
}

VEE_Tube::~VEE_Tube()
{}

/** Receive and process a value. 

    Parameters:

    <UL>

    <LI> "height", float, range: 0 to inf

    <LI> "radius", float, range: 0 to inf

    <LI> "range", float, range: small to inf

    <LI> "magnitude", float, range: -inf to inf

    <LI> "rotation", matrix3, a valid rotation matrix

    </UL>

*/
bool VEE_Tube::setValue(const char *id, VEE_Value *v)
{
  bool ok = false;

  if(strcmp(id, "height") == 0) {
    float tmp;
    ok = v->getFloat(&tmp);
    if(ok)
      m_height = VEE_max(tmp, 0.0f); 
  }
  else if(strcmp(id, VEE_Strs::radius) == 0) {
    float tmp;
    ok = v->getFloat(&tmp);
    if(ok)
      m_radius = VEE_max(tmp, 0.0f); 
  }
  else if(strcmp(id, VEE_Strs::range) == 0) {
    float tmp;
    ok = v->getFloat(&tmp);
    if(ok)
      m_effectRange = VEE_max(tmp, 0.0f); 
  }
  else if(strcmp(id, VEE_Strs::magnitude) == 0) {
    float tmp;
    ok = v->getFloat(&tmp);
    if(ok)
      m_magnitude = tmp;
  }
  else if(strcmp(id, VEE_Strs::rotation) == 0) {
    VEE_Matrix3 tmp;
    ok = v->getMatrix3(&tmp);
    if(ok) {
      m_rotation = tmp;
      m_rotationInv = tmp;
      m_rotationInv.transpose();
    }
  }
  else return VEE_ForceObject::setValue(id, v);

  return ok;
}

VEE_Variable VEE_Tube::getVariable(const char *id) const
{
  VEE_Variable v;

  if(strcmp(id, "height") == 0) {
    v = m_height;
  }
  else if(strcmp(id, VEE_Strs::radius) == 0) {
    v = m_radius;
  }
  else if(strcmp(id, VEE_Strs::range) == 0) {
    v = m_effectRange;
  }
  else if(strcmp(id, VEE_Strs::magnitude) == 0) {
    v = m_magnitude;
  }
  else if(strcmp(id, VEE_Strs::rotation) == 0) {
    v = m_rotation;
  }
  else 
    v = VEE_ForceObject::getVariable(id);

  return v;
}

VEE_Object * VEE_Tube::clone() const
{
  return new VEE_Tube(*this);
}

void VEE_Tube::getForce(VEE_Vector3 *force,
			const VEE_Vector3 &velocity,
			const VEE_Vector3 &location)
{
  VEE_Vector3 relativeLoc = m_rotationInv * (location - m_location);
  
  float horizontal = sqrtf(relativeLoc[0] * relativeLoc[0] +
			   relativeLoc[2] * relativeLoc[2]);

  if(VEE_abs(relativeLoc[1]) >= m_height + m_effectRange ||
     VEE_abs(horizontal - m_radius) > m_effectRange) {
    force->clear();
    return;
  }
  
  VEE_Vector3 relativeVel = m_rotationInv * (velocity - m_location);
  
  relativeLoc.y = 0.0;
  relativeVel.y = 0.0;
  
  VEE_Vector3 directionNormalized = relativeLoc / horizontal;

  float inwardsVelocity = dot(directionNormalized, relativeVel);
  
  float repel;

  if(m_magnitude > 0.0) {
    if(inwardsVelocity < 0.0)
      repel = m_magnitude;
    else 
      repel = m_magnitude * 0.2;
  }
  else {
    if(inwardsVelocity < 0.0)
      repel = m_magnitude * 0.2;
    else 
      repel = m_magnitude;
  }
  
  VEE_Vector3 tmpForce = directionNormalized * repel;

  *force = m_rotation * tmpForce;

  /* VEE_trace("VEE_Tube::getForce # %.3f %.3f %.3f -> %.3f %.3f %.3f",
     location.x, location.y, location.y, force->x, force->y, force->z); */
}

void VEE_Tube::boundingBoxOfForce(VEE_BoundingBox *box)
{
  float maxdistance = m_effectRange + 
    sqrtf(m_height * m_height + m_radius * m_radius);

  VEE_Vector3 tmp;
  tmp.make(maxdistance);
  
  box->low()  = m_location - tmp;
  box->high() = m_location + tmp;
}

VEE_Object * VEE_Tube::factory()
{
  return new VEE_Tube;
}
