/* -*- C++ -*- */

/* COPYRIGHT
 *
 * This file is part of the Visual Effects Engine - VEE
 *
 * Read the "VEE-LICENSE" file for the license.
 *
 * Authors & Copyright:   
 *
 * Tommi Ilmonen, Tuukka Heikura, Marko Myllymaa and 
 * Janne Kontkanen 2001-2004
 *
 * Additional copyrights: Tekes 2003-2004
 *
 * firstname.lastname@hut.fi
 *
 */

#ifndef VEE_VALUE_CURVE_H
#define VEE_VALUE_CURVE_H

#include <vee_interpolators.h>
#include <vee_operator.h>
#include <vee_reference_object.h>


#include <string>

enum VEE_TimeScale {
  VEE_TS_LINEAR,
  VEE_TS_AGE_MATCH,
  VEE_TS_PERIODIC,
  VEE_TS_SIZEOF
};

/** This class outputs a curve at frequent intervals. */

class VEE_ValueCurve : public VEE_Operator
{
public:
  VEE_ValueCurve();
  virtual ~VEE_ValueCurve();

  virtual bool setValue(const char *, VEE_Value *);

  virtual void operate(VEE_Object *, float);
  virtual VEE_Operator * clone() const;

  void setId(const char *id) { m_param = id; }

  void appendPoint(float v, float t)
  { m_curve.appendPoint(v, t); }
  
  void clearCurve() { m_curve.clear(); }

protected:

  float m_interval;
  float m_untilNext;

  std::string m_param;

  VEE_LinearInterpolation2<float> m_curve;
  
};

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////


class VEE_ValueCurveMulti : public VEE_Operator
{
public:
  VEE_ValueCurveMulti();
  virtual ~VEE_ValueCurveMulti();

  virtual bool setValue(const char *, VEE_Value *);

  virtual void operate(VEE_Object *, float);
  virtual VEE_Operator * clone() const;

  void resize(int curves) { m_curves.resize(curves); }

  void setId(const char *name, int n)
  { expand(n); m_curves[n].m_param = name; }

  void appendPoint(float v, float t, int n)
  { expand(n); m_curves[n].m_curve.appendPoint(v, t); }
  
  void clearCurves() { m_curves.clear(); }

protected:

  void expand(uint n) { if(n >= m_curves.size()) m_curves.resize(n+1); }

  float m_interval;
  float m_untilNext;

  class Data {
  public:
    std::string m_param;
    VEE_LinearInterpolation2<float> m_curve;
  };

  std::vector<Data> m_curves;

};

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

/** Sends out a path of values either repeatedly, related to object
    age or at constant rate. */
class VEE_ValueCurve2 : public VEE_Operator
{
public:
  VEE_ValueCurve2();
  virtual ~VEE_ValueCurve2();

  virtual bool setValue(const char *, VEE_Value *);

  virtual void operate(VEE_Object *, float);
  virtual VEE_Operator * clone() const;

  void setId(const char *id) { m_param = id; }

  void appendPoint(float v, float t)
  { (*m_curve).appendPoint(v, t); }
  
  void clearCurve() { (*m_curve).clear(); }

  void setScaling(int scaleMethod) { m_scaling = (VEE_TimeScale) scaleMethod; }

  void scaleValues(float scale) 
  { for(uint i=0; i < (*m_curve).size(); i++) (*m_curve).getPoint(i).m_value *= scale; }
  void scaleTimes(float scale) 
  { for(uint i=0; i < (*m_curve).size(); i++) (*m_curve).getPoint(i).m_time *= scale; }

protected:
  
  VEE_TimeScale m_scaling;

  std::string m_param;

  int m_offset;

  float m_time;

  VEE_ReferenceObject<VEE_LinearInterpolation<float> > m_curve;
  
};


#endif
