
/** This is the ConfigReader library.*/

#ifndef CONFIG_READER_HPP
#define CONFIG_READER_HPP

#include <string>
#include <map>

#include <iostream>

namespace ConfigReader {

  /** A single variant. All variants are stored as text strings. This
      class provides methods to convert the string to the most simple
      variable types including integers, doubles and float vectors. */

  class Variant
  {
  public:

    Variant();
    Variant(const std::string& a);
    Variant(const char *);

    Variant(int);
    Variant(double);
  
    ~Variant();

    operator int () const;
    operator double () const;
    operator const std::string & () const { return m_var; }
  
    double              getDouble(double def = 0.0f) const;
    int                 getInt(int def = 0) const;
    const std::string & getString(const std::string & def) const;
    const std::string & getString() const { return m_var; }

    int                 getFloats(float *, int);
    int                 getDoubles(double *, int);

    void                set(const std::string &s) { m_var = s; }

    void                dump(std::ostream& os) const;

    bool                isEmpty() const 
    { return m_var.size() ? m_var[0] == '\0' : true; }
    
  private:
    std::string m_var;
  };

  /** A template chunk class. These classes can be nested
      easily. Usually it the two default levels are enough. A
      practical example: 

      \code

      ConfigReader::Config config;

      // Read in two configuration files:
      bool ok =  ConfigReader::readConfig(&config, "first-config");
      ok = ok && ConfigReader::readConfig(&config, "second-config");

      // Now retrieve values from the configuration:
      
      ConfigReader::Chunk chunk = config.get("Scene");

      // Now read two variants from the config
      // We can provide default values for variables in case the variable
      // has not been defined in the configuration.

      std::string fileName = chunk.get("filename").getString("default-file");
      float scale = chunk.get("scale").getFloat(1.0);

      \endcode

      <B>Things to keep in mind:</B> All elements must have a unique
      id (unlike Fluid).

  */

  template <class T>
  class ChunkT {
  public:
    typedef typename std::map<std::string, T>::iterator iterator;

    ChunkT() {}
    ~ChunkT() {}
    
    T                  get(const std::string &id)
    {
      T tmp;
      iterator it = m_variants.find(id);
      if(it == m_variants.end())
	return tmp;
      return (*it).second;
    }

    void               set(const std::string& name, const T &v)
    { m_variants[name] = v; }
    // May be specialized at each level.
    void               dump(std::ostream& os);

    void               clear() { m_variants.clear(); }

    unsigned           size() const { return m_variants.size(); }

    bool               isEmpty() const { return m_variants.size() == 0; }

    iterator begin() { return m_variants.begin(); }
    iterator end()   { return m_variants.end(); }

    static T & getType(iterator & it) { return (*it).second; }
    static const std::string & getName(iterator & it) { return (*it).first; }

  private:

    std::map<std::string, T> m_variants;
  };

  

  typedef ChunkT<Variant> Chunk;
  typedef ChunkT<Chunk>   Config;

  bool readConfig(Config *, const char *filename);
  
  // void setLockFunction(void (*function)(bool));

} // namespace

/**
 * Output operator for ConfigReader::Variant
 */

inline std::ostream& operator << 
(std::ostream& os, const ConfigReader::Variant& v)
{
  v.dump(os);
  return os;
}


#endif
