
/* COPYRIGHT
 *
 * This file is part of the Visual Effects Engine - VEE
 *
 * Read the "VEE-LICENSE" file for the license.
 *
 * Authors & Copyright:   
 *
 * Tommi Ilmonen, Tuukka Heikura, Marko Myllymaa and 
 * Janne Kontkanen 2001-2004
 *
 * Additional copyrights: Tekes 2003-2004
 *
 * firstname.lastname@hut.fi
 *
 */

#include <vee_polygon_fire.h>

#include <vee_gravitate.h>
#include <vee_looks_color_table.h>
#include <vee_math.h>
#include <vee_polygon.h>
#include <vee_scene.h>
#include <vee_strs.h>
#include <vee_vortex.h>
#include <vee_value.h>

VEE_PolygonFire::VEE_PolygonFire(float flameHz,
				 float flameAge,
				 float vortexHz,
				 float gustHz)
  : m_flameInterval(1.0f / flameHz),
    m_sinceLastFlame(0.0),
    m_flameAge(flameAge),
    m_vortexInterval(1.0f / vortexHz),
    m_sinceLastVortex(0.0),
    m_gustInterval(1.0f / gustHz),
    m_sinceLastGust(0.0)
{
  m_colorTable.appendColor(VEE_Rgba::Black);
  m_colorTable.appendColor(VEE_Rgba(0.0, 0.0, 0.7, 0.4));

  for(int i=0 ; i < 4 ; i++) {
    // m_colorTable.appendColor(VEE_Rgba::White);
    // m_colorTable.appendColor(VEE_Rgba::Black);
    m_colorTable.appendColor(VEE_Rgba(1.0, 0.0, 0.0, 0.4));
    m_colorTable.appendColor(VEE_Rgba(0.5, 0.5, 0.0, 0.4));
  }
  m_colorTable.appendColor(VEE_Rgba(1.0, 0.0, 0.0, 0.4));
  m_colorTable.appendColor(VEE_Rgba(0.5, 0.3, 0.0, 0.4));
  m_colorTable.appendColor(VEE_Rgba(0.3, 0.25, 0.0, 0.4));
  m_colorTable.appendColor(VEE_Rgba(0.2, 0.15, 0.0, 0.4));

  m_colorTable.appendColor(VEE_Rgba::Black);

  m_updateVelocity = false;
}

VEE_PolygonFire::~VEE_PolygonFire()
{}

bool VEE_PolygonFire::setValue(const char *id, VEE_Value *msg)
{  
  bool ok = false;

  if(VEE_Strs::equals(id, "flamerate")) {
    float v;
    ok = msg->getFloat(&v);
    if(ok)
      m_flameInterval = 1.0 / v;
  }
  else if(VEE_Strs::equals(id, "flameage")) {
    float v;
    ok = msg->getFloat(&v);
    if(ok)
      m_flameAge = v;
  }
  else if(VEE_Strs::equals(id, "vortexrate")) {
    float v;
    ok = msg->getFloat(&v);
    if(ok)
      m_vortexInterval = 1.0 / v;
  }
  else if(VEE_Strs::equals(id, "gustrate")) {
    float v;
    ok = msg->getFloat(&v);
    if(ok)
      m_gustInterval = 1.0 / v;
  }
  else {
    return VEE_Particle::setValue(id, msg);
  }

  return ok;
}

VEE_Object * VEE_PolygonFire::clone() const
{
  return new VEE_PolygonFire(*this);
}

void VEE_PolygonFire::update(float dt)
{
  VEE_Object::update(dt);

  m_sinceLastFlame += dt;

  while(m_sinceLastFlame > m_flameInterval) {
    emitFlame();
    m_sinceLastFlame -= m_flameInterval;
  }

  m_sinceLastVortex += dt;
  
  while(m_sinceLastVortex > m_vortexInterval) {
    emitVortex();
    m_sinceLastVortex -= m_vortexInterval;
  }

  m_sinceLastGust += dt;
  
  while(m_sinceLastGust > m_gustInterval) {
    emitGust();
    m_sinceLastGust -= m_gustInterval;
  }

}

void VEE_PolygonFire::emitFlame()
{
  float diameter = 2.0;

  VEE_Vector3 loc(m_rand.rand11() * diameter,
		  m_rand.rand11() * diameter * 0.2f,
		  m_rand.rand11() * diameter);

  VEE_Vector3 dir(m_rand.rand11(),
		   m_rand.rand11(),
		   m_rand.rand11());

  if(dir[1] < 0.0f) dir[1] *= -1.0;
  dir.normalize();
  dir = dir * 3.0;

  // dir[1] += 4.0;

  VEE_Matrix3 mz, mx, my;

  // float ele = atan2(dir[1], sqrtf(dir[0] * dir[0] + dir[2] * dir[2]));
  float ele = VEE_HALF_PI;
  float azi = atan2(-dir[2], dir[0]);

  mx.rotateX(-VEE_HALF_PI);
  mz.rotateZ(ele);
  my.rotateY(azi);

  VEE_Polygon *poly = VEE_Polygon::createNewQuadMesh(6, 3, 3.5, 1.0);

  poly->move(VEE_Vector3(-1.0, 0.0, -0.5));
  poly->transform(my * mz * mx);

  poly->setLooks(new VEE_LooksColorTable(&m_colorTable));
  poly->setVelocity(dir);
  poly->setAgeLimit(m_flameAge);
  // poly->setStiffness(100.0);
  poly->setStiffness(0.0);
  poly->move(loc + m_location);

  m_sceneBox->addNewObject(poly);
}

void VEE_PolygonFire::emitGust()
{
  float diameter = 2.0;

  VEE_Vector3 loc(m_rand.rand11() * diameter,
		  m_rand.rand11() * diameter * 0.2f,
		  m_rand.rand11() * diameter);

  VEE_Vector3 dir(m_rand.rand11(),
		   m_rand.rand11(),
		   m_rand.rand11());

  if(dir[1] < 0.0f) dir[1] *= -1.0;
  dir.normalize();
  dir = dir * 7.0;

  float magnitude = m_rand.rand11() > 0.0 ? -30.0 : 10.0;

  VEE_Gravitate *grav = new VEE_Gravitate(m_location + loc, dir,
					  m_flameAge, magnitude, false);
  grav->setRange(3.0);

  m_sceneBox->addNewObject(grav);
}

void VEE_PolygonFire::emitVortex()
{  
  bool movingVortex = m_rand.rand01() > 0.1;

  VEE_Vector3 loc, dir;
  float ageLimit;
  float magnitude;

  if(!movingVortex) {

    float diameter = 8.0;

    dir = VEE_Vector3ZERO;

    loc = VEE_Vector3(m_rand.rand11() * diameter,
		       m_rand.rand01() * diameter,
		       m_rand.rand11() * diameter);
    ageLimit = 2;
    magnitude = 3;
  }
  else {

    float diameter = 2.0;

    loc = VEE_Vector3(m_rand.rand11() * diameter,
		      m_rand.rand01() * diameter * 0.2f + diameter * 0.7f,
		      m_rand.rand11() * diameter);

    dir = VEE_Vector3(m_rand.rand11(),
		       m_rand.rand11(),
		       m_rand.rand11());
    ageLimit = m_flameAge;
    magnitude = 2.0;
  }

  dir.normalize();
  dir = dir * 4.0;
  
  VEE_Vortex *vortex = new VEE_Vortex;
  vortex->setLocation(m_location + loc);
  vortex->setAxes(dir);
  vortex->setMagnitude(magnitude);
  vortex->setAgeLimit(ageLimit);
 
  vortex->setAxes(m_rand.rand11(),
		 m_rand.rand11(),
		 m_rand.rand11());

  vortex->setVelocityUpdate(true);
  vortex->setRange(2.0);
  vortex->setVelocityUpdate(movingVortex);
  
  m_sceneBox->addNewObject(vortex);
}
