/* COPYRIGHT
 *
 * This file is part of the Visual Effects Engine - VEE
 *
 * Read the "VEE-LICENSE" file for the license.
 *
 * Authors & Copyright:   
 *
 * Tommi Ilmonen, Tuukka Heikura, Marko Myllymaa and 
 * Janne Kontkanen 2001-2004
 *
 * Additional copyrights: Tekes 2003-2004
 *
 * firstname.lastname@hut.fi
 *
 */

#include <gee_color_around_path.h>

#include <gee_debug.h>
#include <gee_io.h>
#include <gee_macros.h>
#include <gee_world.h>

#include <vee_error.h>

// static bool m_debug = false;

GEE_ColorAroundPath::GEE_ColorAroundPath(GEE_id_t pathId, 
					 bool vertexColors,
					 float radiusNorm,
					 float exaggerateRadius,
					 bool useNapra)
  : GEE_PathToX(pathId),
    m_color(VEE_Rgba::Green),
    m_pressure(1.0f),
    m_vertexColors(vertexColors),
    m_exaggerateRadius(exaggerateRadius),
    m_radiusNorm(radiusNorm),
    m_useNapra(useNapra)
{
  // m_debug = GEE_Debug::debug("GEE_ColorAroundPath");
}

GEE_ColorAroundPath::~GEE_ColorAroundPath()
{}

void GEE_ColorAroundPath::getBoundingBox(VEE_BoundingBox *bb)
{
  GEE_SphereGroup::getBoundingBox(bb);
}
  
bool GEE_ColorAroundPath::recolor(const VEE_Vector3 &location,
				  VEE_Vector4 * color)
{
  bool change = false;

  VEE_Vector4 rgba = * (VEE_Vector4 *) m_color.data();

  const GEE_SphereGroup::Sphere * it  = locationBegin();
  const GEE_SphereGroup::Sphere * end = locationEnd();

  float radius = m_radius;

  for(; it != end; it++) {
    
    float d = (location - it->m_location).length();

    float tmp = it->m_radius;

    if(tmp >= 0.0f)
      radius = tmp;

    float w = (radius - d) / radius * 2;

    if(w <= 0.0)
      continue;
    
    w *= m_pressure;
    
    if(w > 1.0) w = 1.0;
    
    *color = rgba * w + (*color) * (1.0f - w);
    
    change = true;
  }
  return change;
}
 
bool GEE_ColorAroundPath::recolor(const VEE_Vector3 &location,
				  VEE_Rgba * color)
{
  bool change = false;

  VEE_Rgba rgba = m_color;

  const GEE_SphereGroup::Sphere * it  = locationBegin();
  const GEE_SphereGroup::Sphere * end = locationEnd();

  float radius = m_radius;

  for(; it != end; it++) {
    
    float d = (location - it->m_location).length();
    
    float tmp = it->m_radius;

    if(tmp >= 0.0f)
      radius = tmp;

    float w = (radius - d) / radius * 2;
    
    if(w <= 0.0)
      continue;
  
    w *= m_pressure;

    if(w > 1.0) w = 1.0;
    
    *color = rgba * w + (*color) * (1.0f - w);
    change = true;
  }

  return change;
}

void GEE_ColorAroundPath::recolored(GEE_Geometry *g)
{
  addEditedGeometry(g);
}


VEE_Rgba GEE_ColorAroundPath::basicColor()
{
  return m_color;
}

bool GEE_ColorAroundPath::act(GEE_World *w, const GEE_TimeStamp &time)
{
  //  const char * fname = "GEE_ColorAroundPath::act";

  clear();

  if(!GEE_PathToX::act(w, time))
    return false;

  if(locationCount())
    doColor(w);

  return true;
}

static const ushort GEE_ColorId    = GEE_CONST2('C', 'O');
static const ushort GEE_RadiusId   = GEE_CONST2('R', 'A');
static const ushort GEE_PressureId = GEE_CONST2('P', 'R');
static const ushort GEE_VerticesId = GEE_CONST2('V', 'X');
static const ushort GEE_UseNapraId = GEE_CONST2('U', 'N');
static const ushort GEE_ExRadiusId = GEE_CONST2('E', 'R');
static const ushort GEE_RadiusNormId = GEE_CONST2('R', 'N');
static const ushort GEE_EndId      = GEE_CONST2('E', 'N');

bool GEE_ColorAroundPath::read(GEE_Io * io)
{
  std::string str;
  io->readNullTerminatedStdString(str);
  if(str != type())
    return false;
  
  readBasics(io);
  readPathId(io);

  int v = io->readInt32();

  if(v != 1)
    return false;

  ushort id = io->readInt16();

  while(id != GEE_EndId) {
    if(id == GEE_ColorId)
      m_color = io->readRgba();
    else if(id == GEE_VerticesId)
      m_vertexColors = io->readByte();
    else if(id == GEE_PressureId)
      m_pressure = io->readFloat32();
    else if(id == GEE_RadiusId)
      m_radius   = io->readFloat32();
    else if(id == GEE_ExRadiusId)
      m_exaggerateRadius = io->readFloat32();
    else if(id == GEE_RadiusNormId)
      m_radiusNorm = io->readFloat32();
    else if(id == GEE_UseNapraId)
      m_useNapra = io->readByte();
    else {
      VEE_error("GEE_ColorAroundPath::read # Bad id at %d (%c%c)", 
		(int) io->position(), (uchar) (id >> 8), (uchar) id);
      return false;
    }
    id = io->readInt16();
  }

  return true;
}

bool GEE_ColorAroundPath::write(GEE_Io * io)
{
  io->writeNullTerminatedString(type());

  writeBasics(io);
  writePathId(io);
  
  io->writeInt32(1); // version

  io->writeInt16(GEE_ColorId);
  io->writeRgba(m_color);

  io->writeInt16(GEE_VerticesId);
  io->writeByte(m_vertexColors);

  io->writeInt16(GEE_RadiusId);
  io->writeFloat32(m_radius);

  io->writeInt16(GEE_ExRadiusId);
  io->writeFloat32(m_exaggerateRadius);

  io->writeInt16(GEE_RadiusNormId);
  io->writeFloat32(m_radiusNorm);

  io->writeInt16(GEE_UseNapraId);
  io->writeByte(m_useNapra);

  io->writeInt16(GEE_PressureId);
  io->writeFloat32(m_pressure);

  io->writeInt16(GEE_EndId);

  return true;
}

const char * GEE_ColorAroundPath::type() 
{ 
  return "GEE_ColorAroundPath"; 
}

GEE_Operator * GEE_ColorAroundPath::create()
{
  return new GEE_ColorAroundPath();
}

bool GEE_ColorAroundPath::handleNewVertex(GEE_Path3D::iterator &it)
{
  const GEE_PathVertex3D & pv = GEE_Path3D::getVertex(it);

  float radius = m_radius;

  VEE_Vector3 loc = pv.m_location;

  if(m_useNapra) {
    radius = 
      pv.calculateToolRadius(m_radius, m_radiusNorm, m_exaggerateRadius);
    loc = pv.calculateToolCenter();
  }
  addSphere(loc, radius);

  // addSphere(GEE_Path3D::getVertex(it).m_location);
  return true;
}

void GEE_ColorAroundPath::doColor(GEE_World *w)
{
  // GEE_FNAME = "GEE_ColorAroundPath::doColor";

  /* if(m_debug)
   VEE_trace("GEE_ColorAroundPath::doColor # %d", (int) m_locations.size()); */

  GEE_Geometry * g;

  if(m_vertexColors) {
    
    for(GEE_LayerGeometryIterator it(w, m_layer); (g = *it); it.nextVisible())
      g->colorData(this);
  }
  else {
    GEE_Material mat(0.5, VEE_Vector4(m_color.data()));
    mat.setBurn(m_color.theta());

    for(GEE_LayerGeometryIterator it(w, m_layer); (g = *it); it.nextVisible()){
      if(g->intersects(this)) {
	GEE_id_t topId = g->masterGroup();
	if(topId != g->id())
	  g = w->geometry(topId);
	g->setMaterial(mat);
      }
    }
  }
  /* if(m_debug)
     VEE_trace("GEE_ColorAroundPath::doColor # EXIT"); */
}
