/* COPYRIGHT
 *
 * This file is part of the Visual Effects Engine - VEE
 *
 * Read the "VEE-LICENSE" file for the license.
 *
 * Authors & Copyright:   
 *
 * Tommi Ilmonen, Tuukka Heikura, Marko Myllymaa and 
 * Janne Kontkanen 2001-2004
 *
 * Additional copyrights: Tekes 2003-2004
 *
 * firstname.lastname@hut.fi
 *
 */

#include <gee_copy_geometry.h>

#include <gee_engine.h>
#include <gee_geometry.h>
#include <gee_io.h>
#include <gee_macros.h>
#include <gee_world.h>

#include <vee_bounding_box.h>
#include <vee_error.h>

GEE_CopyGeometry::GEE_CopyGeometry(GEE_id_t sourceId,
				   bool sourceIsGeometry)
  : m_sourceId(sourceId),
    m_sourceIsGeometry(sourceIsGeometry)
{
  addIdMember(&m_sourceId);
  m_transform.identity();
}

GEE_CopyGeometry::~GEE_CopyGeometry()
{}

GEE_IdContainer::id_container * GEE_CopyGeometry::getExtraIds()
{
  return & m_ids;
}

bool GEE_CopyGeometry::remapIdMembers(id_remap &remap)
{
  GEE_Operator::remapIdMembers(remap);
  id_remap tmp;

  for(id_remap::iterator it = m_remap.begin(); it != m_remap.end(); it++) {

    id_remap::iterator it2 = remap.find((*it).first);
    id_remap::iterator it3 = remap.find((*it).second);

    if(it2 == remap.end() || it3 == remap.end())
      return false;

    tmp[(*it2).second] = (*it3).second;
  }

  m_remap.swap(tmp);
  
  if(m_sourceModels.geometryCount())
    m_sourceModels.remapIds(remap);

  updateIds();

  return true;
}

bool GEE_CopyGeometry::act(GEE_World *w, const GEE_TimeStamp &)
{
  GEE_FNAME = "GEE_CopyGeometry::act";

  GEE_Geometry * g;
    
  if(m_sourceIsGeometry) {

    g = w->geometry(m_sourceId);
  
    if(!g) {
      VEE_trace("%s # No geometry %d", fname, (int) m_sourceId);
      return false;
    }
    
    GEE_Geometry * g2 = g->copy(w, & m_remap);
    
    g2->setTransformation(m_transform * g2->transformation());

  }
  else {

    GEE_World * source;
    
    if(m_sourceModels.geometryCount())
      source = & m_sourceModels;
    else {
      GEE_Engine * e = w->host();
      
      if(!e) {
	VEE_trace("%s # NULL engine", fname);
	return false;
      }
      
      GEE_Gallery & gal = e->gallery();
      
      source = gal.item(m_sourceId);
      
      if(!source) {
	VEE_trace("%s # No gallery item %d", fname, (int) m_sourceId);
	return false;
      }
    }

    /* VEE_BoundingBox bb;
    bb.clear();
    source->getBoundingBox(&bb);

    VEE_Vector3d center = bb.center();

    VEE_Matrix4d tmp;
    tmp.identity();
    tmp.setColumn3(3, -center);

    VEE_Matrix4d transformation = m_transform * tmp;

    */

    VEE_Matrix4d transformation = m_transform;

    for(GEE_LayerGeometryIterator it(source, -1); (g = *it); ++it) {
      if(g->group() < 0) {
	GEE_Geometry * g2 = g->copy(w, &m_remap);
	g2->setTransformation(transformation * g2->transformation());
      }
    }
    
    for(id_remap::iterator it = m_remap.begin(); it != m_remap.end(); it++) {
      g = w->geometry((*it).second);
      if(g) {
	g->setLayer(m_layer);
	g->setGeneration(0);
      }
    }
  }

  updateIds();

  return true;
}

static const ushort GEE_SourceId = GEE_CONST2('S', 'R');
static const ushort GEE_SouGeoId = GEE_CONST2('S', 'G'); // Source is Geo
static const ushort GEE_RemapId  = GEE_CONST2('R', 'M');
static const ushort GEE_TransId  = GEE_CONST2('T', 'R');
static const ushort GEE_WorldId  = GEE_CONST2('W', 'O');
static const ushort GEE_EndId    = GEE_CONST2('E', 'N');

bool GEE_CopyGeometry::read(GEE_Io *io)
{
  std::string str;
  io->readNullTerminatedStdString(str);
  if(str != type())
    return false;

  readBasics(io);

  int v = io->readInt32();

  if(v != 1)
    return false;

  ushort id = io->readInt16();

  while(id != GEE_EndId) {
    if(id == GEE_SourceId)
      m_sourceId = io->readInt64();
    else if(id == GEE_TransId)
      m_transform = io->readMatrix4d();
    else if(id == GEE_WorldId)
      m_sourceModels.read(io);
    else if(id == GEE_SouGeoId)
      m_sourceIsGeometry = io->readInt16();
    else if(id == GEE_RemapId) {
      GEE_id_t id1 = io->readInt64();
      m_remap[id1] = io->readInt64();
    }
    else {
      VEE_error("GEE_CopyGeometry::read # Bad id at %d (%c%c)", 
		(int) io->position(), (uchar) (id >> 8), (uchar) id);
      return false;
    }
    id = io->readInt16();
  }

  updateIds();

  return true;
}

bool GEE_CopyGeometry::write(GEE_Io *io)
{
  io->writeNullTerminatedString(type());

  writeBasics(io);
  
  io->writeInt32(1); // version

  io->writeInt16(GEE_SourceId);
  io->writeInt64(m_sourceId);

  io->writeInt16(GEE_SouGeoId);
  io->writeInt16(m_sourceIsGeometry);
  
  io->writeInt16(GEE_TransId);
  io->writeMatrix4d(m_transform);

  if(m_sourceModels.geometryCount()) {
    io->writeInt16(GEE_WorldId);
    m_sourceModels.write(io);
  }
  
  for(id_remap::iterator it = m_remap.begin(); it != m_remap.end(); it++) {
    io->writeInt16(GEE_RemapId);
    io->writeInt64((*it).first);
    io->writeInt64((*it).second);
  }
  
  io->writeInt16(GEE_EndId);

  return true;
}

const char * GEE_CopyGeometry::type()
{
  return "GEE_CopyGeometry";
}

GEE_Operator * GEE_CopyGeometry::create()
{
  return new GEE_CopyGeometry();
}

void GEE_CopyGeometry::updateIds()
{
  // Update the ids we have:
  m_ids.clear();
    
  for(id_remap::iterator it = m_remap.begin(); it != m_remap.end(); it++) {
    m_ids.insert((*it).first);
    m_ids.insert((*it).second);
  }

  if(m_sourceModels.geometryCount()) {
    id_remap tmp;
    m_sourceModels.collectIds(tmp);
    
    for(id_remap::iterator it = tmp.begin(); it != tmp.end(); it++)
      m_ids.insert((*it).first);
  }
}
