/* COPYRIGHT
 *
 * This file is part of the Visual Effects Engine - VEE
 *
 * Read the "VEE-LICENSE" file for the license.
 *
 * Authors & Copyright:   
 *
 * Tommi Ilmonen, Tuukka Heikura, Marko Myllymaa and 
 * Janne Kontkanen 2001-2004
 *
 * Additional copyrights: Tekes 2003-2004
 *
 * firstname.lastname@hut.fi
 *
 */

#include <gee_engine.h>

#include <gee_debug.h>
#include <gee_io.h>
#include <gee_macros.h>

#include <vee_error.h>

bool GEE_Engine::m_debug = false;

GEE_Engine::GEE_Engine()
  : m_world(this)
{
  m_actions.setWorld( & m_world);
  m_debug = GEE_Debug::debug("GEE_Engine");
}

GEE_Engine::~GEE_Engine()
{}

void GEE_Engine::addOperator(GEE_Operator *a)
{
  m_actions.addOperator(a);
}

void GEE_Engine::update(GEE_TimeStamp ts)
{
  if(ts < m_prevTime)
    rewind(ts);

  // VEE_trace("GEE_Engine::update # %lf", ts.secondsD());
  m_actions.update(ts);

  m_prevTime = ts;
}

void GEE_Engine::rewind(GEE_TimeStamp ts)
{
  m_actions.stop();
  m_world.clear();
  m_actions.rewind(ts);
  m_prevTime = ts;
}

void GEE_Engine::toBegin()
{
  m_actions.stop();
  m_world.clear();
  m_prevTime = 0;
}

void GEE_Engine::undo(bool update, uint n)
{
  m_actions.stop();
  m_actions.undo(n);

  if(update)
    rewind(m_prevTime);
}

void GEE_Engine::redo(bool update, uint n)
{
  m_actions.stop();
  m_actions.redo(n);

  if(update)
    rewind(m_prevTime);
}

bool GEE_Engine::read(const char *filename, bool merge, bool clean)
{
  VEE_trace("GEE_Engine::read # %s %d", filename, (int) merge);

  GEE_File f;
  if(!f.open(filename, DI_IO_INPUT)) {
    return false;
  }
  
  std::string dirname(filename);
  dirname += "-dir";
  bool ok = false;

  if(!merge) {
    m_world.clear();
    ok = m_actions.read(&f, dirname.c_str());
    ok = ok && m_world.read(&f);
    if(clean)
      m_world.clear();
    ok = ok && m_gallery.read(&f);

    remapThese(m_actions, m_world, m_gallery);
  }
  else {
    m_world.clear();

    GEE_OperatorManager actions;
    GEE_World world;
    GEE_Gallery gallery;
    
    ok = actions.read(&f, dirname.c_str());
    ok = ok && world.read(&f);
    if(clean)
      world.clear();
    bool ok2 = ok && gallery.read(&f);

    remapThese(actions, world, gallery);
    
    if(ok) {
      m_actions.merge(actions);
      m_world.merge(world);
      m_gallery.merge(gallery);
    }

    ok = ok && ok2;
  }

  return ok;
}

/** Writes all data in the engine to a file. */

bool GEE_Engine::write(const char *filename)
{
  // const char * fname = "GEE_Engine::write";

  GEE_File f;
  if(!f.open(filename, DI_IO_OUTPUT)) {
    return false;
  }

  /* std::string dirname(filename);
     dirname += "-dir";
     
     VEE_trace("%s # Creating directory \"%s\"", fname, dirname.c_str());
     
     std::string command("mkdir ");
     command += dirname;
     system(command.c_str()); */
  
  bool ok = m_actions.write(&f, "");
  ok = ok && m_world.write(&f);
  ok = ok && m_gallery.write(&f);
  return ok;
}

/** Reads a file and merges all gallery items and the saves state of
    the world into the gallery. */

bool GEE_Engine::mergeGalleriesFrom(const char *filename)
{
  VEE_trace("GEE_Engine::mergeGalleriesFrom # %s", filename);

  GEE_Engine tmp;
  
  if(!tmp.read(filename, false, false))
    return false;

  m_gallery.takeSnapshot( & tmp.m_world);
  m_gallery.merge(tmp.m_gallery);

  return true;
}

void GEE_Engine::setGarbageBin(VEE_GarbageBin * bin)
{
  m_world.setGarbageBin(bin);
  m_gallery.setGarbageBin(bin);
}

/** Collects all geometry into sorted lists. One list is for solids
    and another for transparencies. */

void GEE_Engine::collectObjects(bool clearOld)
{
  GEE_FNAME = "GEE_Engine::collectObjects";

  if(m_debug)
    VEE_trace(fname);

  if(clearOld)
    m_collector.clear();

  m_world.render(&m_collector);

  if(m_debug)
    VEE_trace("%s # EXIT", fname);
}

/** Clears all member variables. */

void GEE_Engine::clear()
{
  m_world.clear();
  m_actions.clear();
  m_prevTime = 0;
  m_collector.clear();
}

void GEE_Engine::remapThese(GEE_OperatorManager &o, 
			    GEE_World &w, 
			    GEE_Gallery &g)
{
  GEE_IdContainer::id_remap idMap;

  o.collectIds(idMap);
  w.collectIds(idMap);
  g.collectIds(idMap);
  
  // Remap them:
  for(GEE_IdContainer::id_remap::iterator rit = idMap.begin(); 
      rit != idMap.end(); rit++) {
    // Only remap non-negative ids:
    if((*rit).first >= 0)
      (*rit).second = GEE_IdContainer::allocateId();
    else
      (*rit).second = (*rit).first;

    VEE_trace("GEE_Engine::remapThese # %d  ->  %d", 
	      (int) (*rit).first, (int) (*rit).second);
  }
  
  o.remapIds(idMap);
  w.remapIds(idMap);
  g.remapIds(idMap);
}

