/* COPYRIGHT
 *
 * This file is part of the Visual Effects Engine - VEE
 *
 * Read the "VEE-LICENSE" file for the license.
 *
 * Authors & Copyright:   
 *
 * Tommi Ilmonen, Tuukka Heikura, Marko Myllymaa and 
 * Janne Kontkanen 2001-2004
 *
 * Additional copyrights: Tekes 2003-2004
 *
 * firstname.lastname@hut.fi
 *
 */

#include <gee_geometry.h>

#include <gee_bounded.h>
#include <gee_debug.h>
#include <gee_group.h>
#include <gee_io.h>
#include <gee_lines.h>
#include <gee_macros.h>
#include <gee_particle_object.h>
#include <gee_path.h>
#include <gee_triangle_mesh.h>
#include <gee_world.h>

#include <vee_error.h>

#include <typeinfo>

static bool m_debug = false;

typedef std::map<std::string, GEE_Geometry *(*)()> factoryContainer;
static factoryContainer __factories;

GEE_Geometry::GEE_Geometry(VEE_GarbageBin *bin)
  : VEE_Garbage(bin),
    m_world(0),
    m_editorCount(0),
    m_generation(0), 
    m_group(-1),
    m_visible(true), 
    m_selectionState(0)
{
  addIdMember(&m_group);
  m_transformation.identity();
  m_debug = GEE_Debug::debug("GEE_Geometry");
}

GEE_Geometry::~GEE_Geometry()
{}

/** Clone this object, creating an equivalent independent object. */

GEE_Geometry * GEE_Geometry::clone() const
{ 
  VEE_error("GEE_Geometry::clone # unimplemented (in %s)", 
	    typeid(*this).name());
  return 0;
}

/** Copy this object, creating an equivalent independent object into
    the target world. The new object will have a new id. The default
    implementation uses "clone"-call to create a new object and then
    sets the id and world of the object. */

GEE_Geometry * GEE_Geometry::copy(GEE_World *target, id_remap * remap) const
{
  GEE_Geometry * g = clone();
  
  if(!g) 
    return 0;

  id_remap::iterator it = remap->find(m_id);
  
  if(it == remap->end()) {
    g->autoId();
    (*remap)[m_id] = g->id();
  }
  else {
    g->setId((*it).second);
  }

  target->addGeometry(g);

  return g;
}

/** A virtual function that does any necessary operations to erase
    this geometry from the world. In practice this method is only
    interesting in group nodes that may need to erase other stuff. */

void GEE_Geometry::erase()
{}

/** A virtual function that should be overridden to enable us to
    generically delete data within some volume. */

bool GEE_Geometry::eraseData(GEE_Containment *)
{
  return false;
}

bool GEE_Geometry::intersects(GEE_Containment *)
{
  return false;
}

void GEE_Geometry::selectVertices(GEE_Containment *, uint, uint)
{}

void GEE_Geometry::selectFaces(GEE_Containment *, uint, uint)
{}

bool GEE_Geometry::colorData(GEE_ReColor *)
{
  return false;
}

bool GEE_Geometry::moveData(GEE_Mover *)
{
  return false;
}

/** Render this object with the renderer. */

void GEE_Geometry::render(GEE_Renderer *)
{}

/** Get the bounding box of this data set. */

bool GEE_Geometry::getBoundingBox(VEE_BoundingBox *b)
{
  return false;
}

/** Moves the data elements by some offset. */

bool GEE_Geometry::move(const VEE_Vector3 &)
{
  return false;
}

void GEE_Geometry::setMaterial(const GEE_Material &m) 
{ 
  if(m.useVertexColors() != m_material.useVertexColors())
    incrementGeneration();
    
  m_material = m;
}

/** Reads the data from a given file. */

bool GEE_Geometry::read(GEE_Io *)
{
  VEE_error("GEE_Geometry::read # unimplemented (in %s)", 
	    typeid(*this).name());
  return false;
}

/** Reads the data from a given stream. */
bool GEE_Geometry::write(GEE_Io *)
{
  VEE_error("GEE_Geometry::write # unimplemented (in %s)", 
	    typeid(*this).name());
  return false;
}

bool GEE_Geometry::finalizeRead()
{
  return true;
}

bool GEE_Geometry::readBasics(GEE_Io *io)
{
  int v = io->readInt32();

  m_id    = io->readInt64();
  m_layer = io->readInt64();

  io->readNullTerminatedStdString(m_materialName);
  m_material.read(io);
  m_generation = io->readInt64();
  m_visible    = io->readByte();
  m_transformation = io->readMatrix4d();
  m_selectionState = (GEE_SelectionState) io->readInt32();

  /* VEE_trace("GEE_Geometry::readBasics # %d %d %d", 
     (int) m_id, (int) m_layer, (int) m_visible);
     std::cout << "M = " << m_transformation << "\n"; */

  // VEE_trace("GEE_Geometry::readBasics # %d %d %x", v, (int) m_visible, );

  return v == 1;
}

bool GEE_Geometry::writeBasics(GEE_Io *io)
{
  io->writeInt32(1); // version

  io->writeInt64(m_id);
  io->writeInt64(m_layer);

  io->writeNullTerminatedString(m_materialName.c_str());
  m_material.write(io);
  io->writeInt64(m_generation);
  io->writeByte(m_visible);  
  io->writeMatrix4d(m_transformation);
  io->writeInt32(m_selectionState);
  
  //std::cout << "GEE_Geometry::writeBasics # M = " << m_transformation<<"\n";
  
  return true;
}


bool GEE_Geometry::readFile(const char *filename)
{
  GEE_File f;
  if(!f.open(filename, DI_IO_INPUT)) {
    return false;
  }

  return read(&f);
}

bool GEE_Geometry::writeFile(const char *filename)
{
  GEE_File f;
  if(!f.open(filename, DI_IO_OUTPUT)) {
    return false;
  }

  return write(&f);
}


/** Copy all the parameters that are visible to GEE_Geometry from another
    object. Basically an equality operator. */

void GEE_Geometry::copyBasics(const GEE_Geometry &source)
{
  m_id         = source.m_id;
  m_layer      = source.m_layer;

  m_world      = source.m_world;
  m_generation = source.m_generation;
  m_material   = source.m_material;
  m_materialName   = source.m_materialName;
  m_visible        = source.m_visible;
  m_transformation = source.m_transformation;
  m_selectionState = source.m_selectionState;
}

/** Returns the id of the top-level group where this geometry
    belongs. If the geometry is not grouped in any way, then the id of
    this geometry is returned. */

GEE_id_t GEE_Geometry::masterGroup() const
{
  GEE_FNAME = "GEE_Geometry::masterGroup";

  if(!m_world) {
    VEE_error("%s # m_world = NULL", fname);
    return -1;
  }
  
  GEE_id_t i = m_id;
  GEE_id_t g = m_group;
  
  while(g >= 0) {
    const GEE_Geometry * geo = m_world->geometry(g);

    if(!geo) {
      VEE_error("%s # Geometry %d does not exist", fname, (int) g);
      m_world->dump();
      return -1;
    }

    i = geo->id();
    g = geo->group();    
  }
  return i;
}

/** Set the material of this object. */
void GEE_Geometry::setMaterialName(const std::string &mat) 
{ 
  m_materialName = mat; 
}

/** Set the material of this object. */
void GEE_Geometry::setMaterialName(const char *mat)
{
  m_materialName = mat ? mat : ""; 
}

void GEE_Geometry::getWorldTransformation(VEE_Matrix4d *res) const
{
  GEE_FNAME = "GEE_Geometry::getWorldTransformation";

  VEE_Matrix4d tmp = transformation();
  
  GEE_id_t g = m_group;
  
  while(g > 0) {
    const GEE_Geometry * geo = m_world->geometry(g);

    if(!geo) {
      VEE_error("%s # Geometry %d does not exist", fname, (int) g);
      m_world->dump();
      return;
    }

    tmp = geo->transformation() * tmp;

    g = geo->group();    
  }

  *res = tmp;
}

void GEE_Geometry::getWorldTransformation(VEE_Matrix4 *res) const
{
  VEE_Matrix4d tmp;
  getWorldTransformation(&tmp);
  res->copy(tmp.data());
}

void GEE_Geometry::getWorldTransformations
(VEE_Matrix4d *worldTrans, VEE_Matrix4d *invTrans, bool *ok) const
{
  getWorldTransformation(worldTrans);
  *invTrans = worldTrans->inverse(ok);
}

void GEE_Geometry::getWorldTransformations
(VEE_Matrix4 *worldTrans, VEE_Matrix4 *invTrans, bool *ok) const
{
  getWorldTransformation(worldTrans);
  *invTrans = worldTrans->inverse(ok);
}

void GEE_Geometry::centralize()
{
  VEE_BoundingBox bb;

  bb.clear();

  getBoundingBox(&bb);

  VEE_Vector3d center = bb.center();
  m_transformation.addToColumn(3, -center);
}

void GEE_Geometry::setSelectionState(uint s)
{
  m_selectionState = s;
}

GEE_Geometry * GEE_Geometry::readGeometry(GEE_Io * io)
{
  GEE_FNAME = "GEE_Geometry::readGeometry";

  std::string str;
  
  long pos = io->position();

  io->readNullTerminatedStdString(str);

  io->seekPosition(pos);

  GEE_Geometry * g = create(str.c_str());

  if(!g) {
    VEE_error("%s # No geometry type \"%s\"", fname, str.c_str());
    return 0;
  }
  else if(m_debug)
    VEE_trace("%s # Attempting to read \"%s\"",fname, str.c_str());

  if(!g->read(io)) {
    VEE_error("%s # Read failed for \"%s\"", fname, str.c_str());
    delete g;
    return 0;
  }

  return g;
}

/** A factory function that creates Geometry objects as needed. */

GEE_Geometry * GEE_Geometry::create(const char * type)
{
  if(!type)
    return 0;

  static bool once = false;

  if(!once) {
    init();
    once = true;
  }

  factoryContainer::iterator it = __factories.find(type);

  if(it == __factories.end())
    return 0;

  return (*(*it).second)();
}

/** Initializes the data tables used by the #create factory
    function. */

void GEE_Geometry::init()
{
  __factories["GEE_Group"]          = GEE_Group::create;
  __factories["GEE_Lines3D"]        = GEE_Lines3D::create;
  __factories["GEE_TriangleMesh"]   = GEE_TriangleMesh::create;
  __factories["GEE_ParticleObject"] = GEE_ParticleObject::create;
  __factories["PATH_3D"]            = GEE_Path3D::create;
}

bool GEE_Geometry::intersectsBounded(GEE_Bounded *b, 
				     VEE_BoundingBox * myBounds, 
				     const VEE_Matrix4 & myWorldTransform)
{
  VEE_BoundingBox bthat, bthis = myWorldTransform * (*myBounds);
  b->getBoundingBox(& bthat);
  return bthis.intersects(bthat);
}
