/* COPYRIGHT
 *
 * This file is part of the Visual Effects Engine - VEE
 *
 * Read the "VEE-LICENSE" file for the license.
 *
 * Authors & Copyright:   
 *
 * Tommi Ilmonen, Tuukka Heikura, Marko Myllymaa and 
 * Janne Kontkanen 2001-2004
 *
 * Additional copyrights: Tekes 2003-2004
 *
 * firstname.lastname@hut.fi
 *
 */

#include <gee_history_animation.h>

#include <gee_engine.h>
#include <gee_operator.h>

#include <vee_error.h>

GEE_HistoryAnimation::GEE_HistoryAnimation()
  : m_engine(0),
    m_currentGroup(0)
{
  m_iterator = m_groupIds.begin();
}

GEE_HistoryAnimation::~GEE_HistoryAnimation()
{
  stop();
}

void GEE_HistoryAnimation::start(GEE_Engine * e)
{
  if(m_engine)
    stop();

  m_engine = e;

  GEE_OperatorManager::operatorList list;

  m_groupIds.clear();

  e->actions().collectOperators(&list);

  for(GEE_OperatorManager::operatorListIter it = list.begin(); 
      it != list.end(); it++) {
    m_groupIds.insert((*it).ptr()->group());
  }

  VEE_trace(" GEE_HistoryAnimation::start # %u groups", 
	    (uint) m_groupIds.size());
  
  m_iterator = m_groupIds.begin();

  while((*m_iterator) == -1 && m_iterator != m_groupIds.end())
    m_iterator++;

  m_prevTime.getTimeOfDay();

  m_engine->undo(false, 1000000);

  startNewRound(*m_iterator);

  update(0.0f);
}

void GEE_HistoryAnimation::update(float delta)
{
  if(!m_engine)
    return;

  if(delta <= 0.0f) {
    VEE_Time now;
    now.getTimeOfDay();
    delta = m_prevTime.secsTo(now);
    m_prevTime = now;
  }

  if(delta > 0.2)
    delta = 0.2;

  if(m_currentGroup) {
    if(m_currentTime > m_currentGroup->endTime()) {
      endRound();

      if(m_iterator == m_groupIds.end())
	return;

      m_iterator++;

      if(m_iterator == m_groupIds.end())
	return;

      startNewRound(*m_iterator);
    }
  }

  GEE_TimeStamp tmp;
  tmp.setSeconds(delta);

  m_currentTime += tmp;

  m_engine->update(m_currentTime);

}

void GEE_HistoryAnimation::stop()
{
  VEE_trace("GEE_HistoryAnimation::stop");

  if(!m_engine)
    return;

  if(m_currentGroup)
    endRound();

  m_engine->redo(false, 1000000);

  m_engine = 0;
}

bool GEE_HistoryAnimation::isDone()
{
  return m_iterator == m_groupIds.end();
}

void GEE_HistoryAnimation::startNewRound(GEE_id_t groupToInclude)
{

  GEE_OperatorManager * ops = & m_engine->actions();
  GEE_OperatorGroup * gr = ops->getOperatorGroup(groupToInclude);

  assert(gr);

  m_currentGroup = gr;
  m_currentScale = gr->scale();
  m_currentGroup->setScale(1.0);

  GEE_OperatorManager::operatorListIter undoIter = ops->undoIterator();

  uint n = 0;

  for(GEE_OperatorManager::operatorListIter undoIter = ops->undoIterator(); 
      (undoIter != ops->undoEnd()) && 
	((*undoIter).ptr()->group() <= groupToInclude); undoIter++) {
    GEE_Operator * o = (*undoIter).ptr();

    
    VEE_trace("GEE_HistoryAnimation::startNewRound # Including %s %lf %lf",
	      typeid(*o).name(), 
	      o->interval().begin().secondsD(),
	      o->interval().end().secondsD());

    n++;
  }

  if(n)
    m_engine->redo(false, n);

  m_engine->rewind(0);

  m_currentTime = m_currentGroup->offset() + m_currentGroup->firstTime();
  
  if(m_currentTime != 0)
    m_engine->rewind(m_currentTime);

  m_currentGroup->calculateEndTime();

  VEE_trace("GEE_HistoryAnimation::startNewRound # %ld %u %lf %lf", 
	    (long) groupToInclude, n, m_currentTime.secondsD(),
	    m_currentGroup->endTime().secondsD());
}

void GEE_HistoryAnimation::endRound()
{
  m_currentGroup->setScale(m_currentScale);
  m_currentGroup = NULL;
}
