/* COPYRIGHT
 *
 * This file is part of the Visual Effects Engine - VEE
 *
 * Read the "VEE-LICENSE" file for the license.
 *
 * Authors & Copyright:   
 *
 * Tommi Ilmonen, Tuukka Heikura, Marko Myllymaa and 
 * Janne Kontkanen 2001-2004
 *
 * Additional copyrights: Tekes 2003-2004
 *
 * firstname.lastname@hut.fi
 *
 */

#include <gee_id.h>

#include <gee_io.h>
#include <gee_macros.h>

#include <vee_error.h>
#include <vee_ptr.h>

#include <typeinfo>

/* #include <di_mutex.h>

static DI_MutexAuto __GEE_IdLock; */

GEE_IdContainer::GEE_IdContainer()
  : m_id(-1),
    m_layer(-1)
{
  addIdMember(&m_id);
  addIdMember(&m_layer);
}

GEE_IdContainer::GEE_IdContainer(const GEE_IdContainer &that)
  : m_idDistances(that.m_idDistances),
    m_id(that.m_id),
    m_layer(that.m_layer)
{}

GEE_IdContainer::~GEE_IdContainer()
{}

/** This method returns an id that the iterator points to. */

GEE_id_t GEE_IdContainer::getIdMember(id_iterator it)
{
  int offset = *it;

  return * VEE_Ptr::offset((GEE_id_t *) this, offset);
}

/** Adds an id member to be tracked. */

void GEE_IdContainer::addIdMember(GEE_id_t *id)
{
  m_idDistances.insert(VEE_Ptr::distance(id, this));
}

/** Returns all "extra" ids that a child may have. These ids are
    typically something that are dynamically allocated in the child
    object. Thus they cannot be tracked with the normal method. 

    The returned structure should not be allocated for this purpose
    (it shouldbe an ordinary class member that is returned.

    The default implementation returns true.
*/

std::set<GEE_id_t> * GEE_IdContainer::getExtraIds()
{
  return 0;
}

/** This method changes the normal ids based on the remapping
    information. 

    The default implementation remaps the ids that are tracked with
    direct offset pointer method.
*/

bool GEE_IdContainer::remapIdMembers(id_remap &remap)
{
  bool ok = true;

  for(id_iterator it = idBegin(); it != idEnd(); it++) {
    int offset = *it;
    GEE_id_t *oldId = VEE_Ptr::offset((GEE_id_t *) this, offset);

    // Do not remap negative numbers:
    if(*oldId < 0)
      continue;
    
    id_remap::iterator it2 = remap.find(*oldId);
    if(it2 == remap.end()) {
      VEE_error("GEE_IdContainer::remapIdMembers (%s) # No new id for %ld", 
		typeid(*this).name(), (long) oldId);
      ok = false;
    }
    else
      *oldId = (*it2).second;
  }

  id_container * extras = getExtraIds();

  if(!extras)
    return ok;

  id_container tmp;

  for(id_container::iterator it = extras->begin(); it != extras->end(); it++) {
    id_remap::iterator it2 = remap.find(*it);
    if(it2 == remap.end()) {
      VEE_error("GEE_IdContainer::remapIdMembers # No new id for %ld (extra)", 
		(long) *it);
      ok = false;
    }
    else
      tmp.insert((*it2).second);
  }

  extras->swap(tmp);
  
  return ok;
}

void GEE_IdContainer::addIdsToRemap(id_remap &map)
{
  const char * fname = "GEE_IdContainer::addIdsToRemap";
  const char * type = typeid(*this).name();

  for(id_iterator it = idBegin(); it != idEnd(); it++) {
    GEE_id_t id = getIdMember(it);
    map[id] = -1;
    VEE_trace("%s # %s %p uses id %d", fname, type, this, (int) id);
  }

  id_container * extras = getExtraIds();

  if(!extras)
    return;

  for(id_container::iterator it = extras->begin(); it != extras->end(); it++) {
    GEE_id_t id = *it;
    map[id] = -1;
    VEE_trace("%s # %s %p uses extra id %d", fname, type, this, (int) id);
  }
}

GEE_id_t GEE_IdContainer::allocateId()
{
  static GEE_id_t count = 0;

  // __GEE_IdLock.lock();
  GEE_id_t r = ++count;
  // __GEE_IdLock.unlock();

  return r;
}

static const ushort GEE_IdId    = GEE_CONST2('I', 'D');
static const ushort GEE_EndId   = GEE_CONST2('E', 'N');

bool GEE_IdContainer::writeIds(GEE_Io * io, id_container * ids)
{
  for(id_container::iterator it = ids->begin(); it != ids->end(); it++) {
    io->writeInt16(GEE_IdId);
    io->writeInt64((*it));
  }
  io->writeInt16(GEE_EndId);

  return true;
}

bool GEE_IdContainer::readIds(GEE_Io * io, id_container * ids)
{
  ushort id = io->readInt16();

  while(id == GEE_IdId) {
    ids->insert(io->readInt64());
    id = io->readInt16();
  }

  return (id == GEE_EndId);
}
  
