/* -*- C++ -*- */

/* COPYRIGHT
 *
 * This file is part of the Visual Effects Engine - VEE
 *
 * Read the "VEE-LICENSE" file for the license.
 *
 * Authors & Copyright:   
 *
 * Tommi Ilmonen, Tuukka Heikura, Marko Myllymaa and 
 * Janne Kontkanen 2001-2004
 *
 * Additional copyrights: Tekes 2003-2004
 *
 * firstname.lastname@hut.fi
 *
 */

#ifndef GEE_ID_H
#define GEE_ID_H

#include <gee_config.h>

#include <map>
#include <set>

class GEE_Io;

/** This class is used to track the ids that are members of the child
    classes. This enables us to automatically access all the ids that
    a particular object has. This feature is needed when we do the
    id-remapping during file loading. 

    <B>The ids must be relative to an object (normal class members, not
    behind a pointer etc).</B>

    The id-information is stored internally as a list (actually
    std::set) of integers. The integers provide offset values that
    indicate how far an id is located from this-pointer. This may look
    hazardous, but it should be a stable approach.

    @author Tommi Ilmonen
*/

class GEE_IdContainer
{
public:

  typedef std::set<int> id_distances;
  typedef std::set<GEE_id_t> id_container;
  typedef id_distances::iterator id_iterator;

  typedef std::map<GEE_id_t, GEE_id_t> id_remap;

  GEE_IdContainer();
  GEE_IdContainer(const GEE_IdContainer &);
  virtual ~GEE_IdContainer();

  id_iterator idBegin() { return m_idDistances.begin(); }
  id_iterator idEnd()   { return m_idDistances.end(); }

  GEE_id_t getIdMember(id_iterator it);
  void addIdMember(GEE_id_t *);

  virtual id_container * getExtraIds();
  virtual bool remapIdMembers(id_remap &);
  virtual void addIdsToRemap(id_remap &);

  static GEE_id_t allocateId();

  GEE_id_t id() const { return m_id; }
  void setId(GEE_id_t id) { m_id = id; }
  /// Sets the ID with allocateId()
  void autoId() { m_id = allocateId(); }

  GEE_id_t layer() const { return m_layer; }
  void setLayer(GEE_id_t layer) { m_layer = layer; }

  static bool writeIds(GEE_Io *, id_container *);
  static bool readIds(GEE_Io *, id_container *);
  
private:
  id_distances m_idDistances;
protected:
  GEE_id_t m_id;
  GEE_id_t m_layer;
};

#endif
