/* COPYRIGHT
 *
 * This file is part of the Visual Effects Engine - VEE
 *
 * Read the "VEE-LICENSE" file for the license.
 *
 * Authors & Copyright:   
 *
 * Tommi Ilmonen, Tuukka Heikura, Marko Myllymaa and 
 * Janne Kontkanen 2001-2004
 *
 * Additional copyrights: Tekes 2003-2004
 *
 * firstname.lastname@hut.fi
 *
 */

#include <gee_io.h>
#include <gee_time_stamp.h>

#include <di_config.h>

#include <vee_error.h>

#include <vector>

#include <strings.h>

void GEE_Io::readStdString(std::string &str) 
{ 
  char * p = readString(); 
  str = p ? p : ""; 
  free(p); 
}

void GEE_Io::readNullTerminatedStdString(std::string &str)
{
  char * p = readNullTerminatedString(); 
  str = p;
  free(p);
}

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

GEE_File::GEE_File()
  : m_file(0),
    m_littleEndian(true)
{}

GEE_File::~GEE_File()
{}
  
bool GEE_File::open(const char *name, DI_IoMode mode)
{
  close();

  const char * cmode = 0;

  if(mode == DI_IO_INPUT)
    cmode = "rb";
  else if(mode == DI_IO_OUTPUT)
    cmode = "wb";
  else if(mode == DI_IO_INPUT_OUTPUT)
    cmode = "rwb";
  else 
    return false;

  m_file = fopen(name, cmode);

  return m_file != 0;
}

bool GEE_File::close()
{
  if(m_file) {
    fclose(m_file);
    m_file = 0;
    return true;
  }
  return false;
}

long GEE_File::position()
{
  if(!m_file)
    return -1;

  return ftell(m_file);
}

bool GEE_File::seekPosition(long pos)
{
  if(!m_file)
    return -1;

  return fseek(m_file, pos, SEEK_SET) == 0;
}

bool GEE_File::writev(const void *ptr, int bytes)
{
  if(!m_file)
    return false;

  return fwrite(ptr, bytes, 1, m_file) == 1;
}

bool GEE_File::write1(char x)
{
  if(!m_file)
    return false;

  return fwrite(&x, 1, 1, m_file) == 1;
}

bool GEE_File::write2(short x)
{
  if(!m_file)
    return false;

#ifdef WORDS_BIGENDIAN
  if(m_littleEndian) x = DI_byteSwap2(x);
#else
  if(!m_littleEndian) x = DI_byteSwap2(x);
#endif

  return fwrite(&x, 2, 1, m_file) == 1;
}

bool GEE_File::write4(int x)
{
  if(!m_file)
    return false;

#ifdef WORDS_BIGENDIAN
  if(m_littleEndian) x = DI_byteSwap4(x);
#else
  if(!m_littleEndian) x = DI_byteSwap4(x);
#endif

  return fwrite(&x, 4, 1, m_file) == 1;
}

bool GEE_File::write4v(const int *xp, int n)
{
  const int bufSize = 0x1F; // 64

  int buf[bufSize];

#ifdef WORDS_BIGENDIAN
  if(m_littleEndian)
#else
  if(!m_littleEndian) 
#endif
  {
    int i;
    // There is need to convert the data:
    for(i=0; i < n; i++) {
      buf[i & bufSize] = DI_byteSwap4(*xp++);
      if((i & bufSize) == 0 && i != 0)
	fwrite(&buf, 4 * bufSize, 1, m_file);
    }
    i = i & bufSize;
    if(i)
      return fwrite(&buf, 4 * i, 1, m_file) == 1;
  }
  
  return fwrite(xp, 4 * n, 1, m_file) == 1;
}

bool GEE_File::write8(llong x)
{
  if(!m_file)
    return false;

  // int xi

#ifdef WORDS_BIGENDIAN
  if(m_littleEndian) x = DI_byteSwap8(x);
#else
  if(!m_littleEndian) x = DI_byteSwap8(x);
#endif

  return fwrite(&x, 8, 1, m_file) == 1;
}

bool GEE_File::write8v(const llong *xp, int n)
{
#ifdef WORDS_BIGENDIAN
  if(m_littleEndian)
#else
  if(!m_littleEndian) 
#endif
  {
    const uint bufSize = 0x1F; // 64

    llong buf[bufSize];

    uint i;
    // There is need to convert the data:
    for(i = 0; (int) i < n; ) {
      // printf("output %lf\n", * (double *) xp);	     

      buf[i & bufSize] = DI_byteSwap8(*xp++);

      i++;

      if((i & bufSize) == 0) // Flush
	fwrite(&buf, 8 * bufSize, 1, m_file);
    }
    i = i & bufSize;
    if(i)
      return fwrite(&buf, 8 * i, 1, m_file) == 1;

    return true;
  }
  
  return fwrite(xp, 8 * n, 1, m_file) == 1;
}

bool GEE_File::writeString(const char *str)
{
  if(!str)
    return writeInt32(0);

  int l = strlen(str);

  bool ok = writeInt32(l);

  if(!l)
    return ok;

  return ok && writev(str, l + 1);
}

bool GEE_File::writeNullTerminatedString(const char *str)
{
  if(!str)
    str = "";

  int l = strlen(str);

  return writev(str, l + 1);
}

void GEE_File::readv(void *ptr, int n)
{
  if(!m_file)
    return;

  fread(ptr, n, 1, m_file);
}

char GEE_File::read1()
{
  if(!m_file)
    return 0;

  char x;
  fread(&x, 1, 1, m_file);

  return x;
}

short GEE_File::read2()
{  
  if(!m_file)
    return 0;

  short x = 0;
  fread(&x, 2, 1, m_file);

#ifdef WORDS_BIGENDIAN
  if(m_littleEndian) x = DI_byteSwap2(x);
#else
  if(!m_littleEndian) x = DI_byteSwap2(x);
#endif

  return x;
}

int GEE_File::read4()
{  
  if(!m_file)
    return 0;

  int  x = 0;
  fread(&x, 4, 1, m_file);
  
#ifdef WORDS_BIGENDIAN
  if(m_littleEndian) x = DI_byteSwap4(x);
#else
  if(!m_littleEndian) x = DI_byteSwap4(x);
#endif

  return x;
}

void GEE_File::read4v(int *xp, int n)
{
  if(!m_file)
    return;

  /* VEE_trace("GEE_File::read4v # %p %d", xp, n); */
  bzero(xp, 4 * n);
  fread(xp, 4 * n, 1, m_file);
    
#ifdef WORDS_BIGENDIAN
  if(m_littleEndian)
#else
  if(!m_littleEndian) 
#endif
  {
    int * sentinel = xp + n;
    for( ; xp < sentinel; xp++)
      *xp = DI_byteSwap4(*xp);
  }
}

llong GEE_File::read8()
{
  if(!m_file)
    return 0;

  llong  x = 0;

  if(fread(&x, 8, 1, m_file) != 1)
    VEE_error("GEE_File::read8 # read failed");
  
#ifdef WORDS_BIGENDIAN
  if(m_littleEndian) x = DI_byteSwap8(x);
#else
  if(!m_littleEndian) x = DI_byteSwap8(x);
#endif
  
  return x;  
}

void GEE_File::read8v(llong *xp, int n)
{
  if(!m_file)
    return;

  /* VEE_trace("GEE_File::read8v # %p %d", xp, n); */
  bzero(xp, 8 * n);
  fread(xp, 8 * n, 1, m_file);

#ifdef WORDS_BIGENDIAN
  if(m_littleEndian)
#else
  if(!m_littleEndian) 
#endif
  {
    llong * sentinel = xp + n;
    for( ; xp < sentinel; xp++)
      *xp = DI_byteSwap8(*xp);
  }
}

char *GEE_File::readString()
{
  int l = readInt32();

  if(!l)
    return 0;
  
  char * str = (char *) malloc(l + 1);

  if(!str) {
    VEE_error("GEE_File::readString # Failed to allocate string %d bytes llong",
	      l);
    return 0;
  }

  readv(str, l + 1);

  return str;
}

char *GEE_File::readNullTerminatedString()
{
  char tmp;

  std::vector<char> str;

  const uint maxLen = 1000000;

  do {
    tmp = '\0';
    readv(&tmp, 1);
    str.push_back(tmp);
  } while((tmp != '\0') && (str.size() < maxLen));
  
  if(str.size() >= maxLen) {
    VEE_error("GEE_File::readNullTerminatedString # String too long.");
    return strdup("");
  }

  return strdup(&str[0]);
}

void GEE_File::test(bool littleEndian)
{
  VEE_trace("GEE_File::test");
  
  const char * filename = "gee-testfile";

  GEE_File f;

  f.setLittleEndian(littleEndian);

  GEE_TimeStamp ts;
  ts.setSeconds(55.55);

  VEE_Matrix4d m;
  
  m.identity();

  m.setColumn3(3, VEE_Vector3(2, 3, 4));

  // Write some crap:
  f.open(filename, DI_IO_OUTPUT);
  f.writeInt32(37);
  f.writeFloat32(37.7);
  f.writeInt64(ts.stamp());
  f.writeInt64(-3);
  f.writeFloat64(7.89);
  f.writeVector3(VEE_Vector3(111.0, 222.0, 333.3));
  f.writeString("Hello World");
  f.writeMatrix4d(m);
  f.close();

  VEE_trace("GEE_File::test # HALF");

  // Read the same stuff:
  
  if(!f.open(filename, DI_IO_INPUT)) {
    VEE_error("Could not open file");
    return;
  }

  int x = f.readInt32();
  float x2 = f.readFloat32();
  ts = f.readInt64();
  int x4 = f.readInt64();
  double d1 = f.readFloat64();
  VEE_Vector3 v = f.readVector3();
  char * str = f.readString();
  m = f.readMatrix4d();

  VEE_trace("GEE_File::test # ALMOST");

  printf("\nint = %d, float = %f, double = %lf, vector3 = %f %f %f string = \"%s\"\n"
	 "ts = %lf, x4 = %d\n",
	 x, x2, d1, v.x, v.y, v.z, str, ts.secondsD(), x4);
  fflush(0);

  std::cout << "Matrix = " << m << std::endl;

  free(str);
}

