/* -*- C++ -*- */

/* COPYRIGHT
 *
 * This file is part of the Visual Effects Engine - VEE
 *
 * Read the "VEE-LICENSE" file for the license.
 *
 * Authors & Copyright:   
 *
 * Tommi Ilmonen, Tuukka Heikura, Marko Myllymaa and 
 * Janne Kontkanen 2001-2004
 *
 * Additional copyrights: Tekes 2003-2004
 *
 * firstname.lastname@hut.fi
 *
 */

#ifndef GEE_LINES_H
#define GEE_LINES_H

#include <gee_geometry.h>

#include <vee_bounding_box.h>
#include <vee_vector4.h>

#include <map>

class GEE_LineVertex3D
{
public:
  GEE_LineVertex3D(const VEE_Vector3 & location = VEE_Vector3ZERO,
		   const VEE_Vector4 & rgba = VEE_Vector4ONE,
		   bool continues = true) 
    : m_location(location), m_rgba(rgba), 
      m_mixture(0.0), m_continues(continues) {}

  bool read(GEE_Io *);
  bool write(GEE_Io *);

  VEE_Vector3 m_location;
  VEE_Vector4 m_rgba;
  float       m_mixture;
  /// Tells if line continues from this vertex to the next vertex.
  bool        m_continues;
};

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

/** A collection of line segments. This class represents a collection
    of lines (curve) composed of multiple vertices.

    @see GEE_LineVertex3D

    @author Tommi Ilmonen
*/
class GEE_Lines3D : public GEE_Geometry
{
public:

  typedef std::map<int, GEE_LineVertex3D> container;
  typedef container::iterator iterator;

  GEE_Lines3D();
  virtual ~GEE_Lines3D();
  
  static GEE_Geometry * create();
  virtual GEE_Geometry * clone() const;
  virtual bool eraseData(GEE_Containment *);
  virtual bool intersects(GEE_Containment *);
  virtual bool colorData(GEE_ReColor *);
  virtual bool moveData(GEE_Mover *);
  virtual void setMaterial(const GEE_Material &m);
  virtual bool getBoundingBox(VEE_BoundingBox *);
  virtual void render(GEE_Renderer *);

  virtual bool read(GEE_Io *);
  virtual bool write(GEE_Io *);

  void addVertex(int index, const GEE_LineVertex3D &vertex);
  void addVertex(const GEE_LineVertex3D &vertex)
  { addVertex(newIndex(), vertex); }
  
  void addVertex(int index, float x, float y, float z, bool continues = true)
  { 
    addVertex(index, GEE_LineVertex3D(VEE_Vector3(x, y, z), 
				     VEE_Vector4ONE, 
				     continues));
  }

  void addVertex(float x, float y, float z, bool continues = true)
  { addVertex(newIndex(), x, y, z, continues); }

  iterator begin() { return m_vertices.begin(); }
  iterator end()   { return m_vertices.end(); }

  uint vertexCount() const { return m_vertices.size(); }

  static GEE_LineVertex3D & getVertex(iterator & it) { return (*it).second; }
  static VEE_Vector3 & getVertexLocation(iterator & it) 
  { return (*it).second.m_location; }

  float width() const { return m_lineWidth; }
  void setWidth(float w) { m_lineWidth = w; }

  int newIndex()
  {
    if(!m_vertices.size())
      return 0;

    iterator it = m_vertices.end();
    it--;
    return (*it).first + 1;
  }

protected:
  container       m_vertices;

  float           m_lineWidth;
  VEE_BoundingBox m_bbox;
};

#endif
