/* COPYRIGHT
 *
 * This file is part of the Visual Effects Engine - VEE
 *
 * Read the "VEE-LICENSE" file for the license.
 *
 * Authors & Copyright:   
 *
 * Tommi Ilmonen, Tuukka Heikura, Marko Myllymaa and 
 * Janne Kontkanen 2001-2004
 *
 * Additional copyrights: Tekes 2003-2004
 *
 * firstname.lastname@hut.fi
 *
 */

#include <gee_param_writer.h>

#include <gee_io.h>
#include <gee_macros.h>

#include <vee_error.h>
#include <vee_ptr.h>

GEE_ParamWriter::GEE_ParamWriter()
{}

GEE_ParamWriter::~GEE_ParamWriter()
{}

static const ushort GEE_EndId = GEE_CONST2('E', 'N');

bool GEE_ParamWriter::writeParameters(GEE_Io *io)
{
  bool ok = true;

  for(container::iterator it = m_params.begin(); it != m_params.end(); it++) {

    ushort id = (*it).first;
    const Item & i = (*it).second;

    void * ptr = VEE_Ptr::offset(this, i.m_offset);
    
    io->writeInt16(id);

    switch(i.m_type) {
    case FLOAT:
      io->writeFloat32(* (const float *) ptr);
      break;
    case INT:
      io->writeInt32(* (const int *) ptr);
      break;
    case ID:
      io->writeInt64(* (const GEE_id_t *) ptr);
      break;
    case BOOL:
      io->writeByte(* (const bool *) ptr);
      break;
    case MATRIX3:
      io->writeFloat32v((const float *) ptr, 9);
      break;
    case MATRIX4:
      io->writeFloat32v((const float *) ptr, 16);
      break;
    case MATRIX4D:
      io->writeFloat64v((const double *) ptr, 16);
      break;
    case VECTOR3:
      io->writeFloat32v((const float *) ptr, 3);
      break;
    case VECTOR4:
      io->writeFloat32v((const float *) ptr, 4);
      break;
    case STLSTRING:
      io->writeStdString(* ((std::string *) ptr));
      break;
    default:
      VEE_error("GEE_ParamWriter::writeParameters # Bad type = %d",
		(int) i.m_type);
      ok = false;
    }
    
  }

  io->writeInt16(GEE_EndId);
  
  return ok;
}

bool GEE_ParamWriter::readParameters(GEE_Io *io)
{
  bool ok = true;

  ushort id = io->readInt16();

  while(id != GEE_EndId && id != 0) {

    /* VEE_trace("GEE_ParamWriter::readParameters # (%c%c)", 
       (uchar) (id >> 8), (uchar) id); */

    container::iterator it = m_params.find(id);

    if(it == m_params.end()) {
      ok = false;
      id = io->readInt16();
      continue;
    }

    const Item & i = (*it).second;

    void * ptr = VEE_Ptr::offset(this, i.m_offset);

    switch(i.m_type) {
    case FLOAT:
      (* (float *) ptr) = io->readFloat32();
      break;
    case INT:
      (* (int *) ptr) = io->readInt32();
      break;
    case ID:
      (* (GEE_id_t *) ptr) = io->readInt64();
      break;
    case BOOL:
      (* (bool *) ptr) = io->readByte();
      break;
    case MATRIX3:
      io->readFloat32v((float *) ptr, 9);
      break;
    case MATRIX4:
      io->readFloat32v((float *) ptr, 16);
      break;
    case MATRIX4D:
      io->readFloat64v((double *) ptr, 16);
      break;
    case VECTOR3:
      io->readFloat32v((float *) ptr, 3);
      break;
    case VECTOR4:
      io->readFloat32v((float *) ptr, 4);
      break;
    case STLSTRING:
      io->readStdString(* ((std::string *) ptr));
      break;
    default:
      VEE_error("GEE_ParamWriter::readParameters # Bad type = %d",
		(int) i.m_type);
      ok = false;
    }

    id = io->readInt16();
  }
  
  return ok;
}

void GEE_ParamWriter::clear()
{
  m_params.clear();
}

void GEE_ParamWriter::addMember(Type t, ushort id, const void *ptr)
{
  GEE_FNAME = "GEE_ParamWriter::addMember";

  if(id == GEE_EndId) {
    VEE_error("%s # cannot use id \"EN\" (reserved for internal use)");
    return;
  }

  container::iterator it = m_params.find(id);

  if(it != m_params.end()) {
    VEE_error("%s # Member %d (%s%s) exists already", fname,
	      (int) id, (char) (id & 0xFF), (char) ((id >> 8) & 0xFF));
    return;
  }
  
  Item i;
  i.m_type = t;
  i.m_offset = VEE_Ptr::distance(ptr, this);

  m_params[id] = i;
}
