/* COPYRIGHT
 *
 * This file is part of the Visual Effects Engine - VEE
 *
 * Read the "VEE-LICENSE" file for the license.
 *
 * Authors & Copyright:   
 *
 * Tommi Ilmonen, Tuukka Heikura, Marko Myllymaa and 
 * Janne Kontkanen 2001-2004
 *
 * Additional copyrights: Tekes 2003-2004
 *
 * firstname.lastname@hut.fi
 *
 */

#include <gee_particle_object.h>

#include <gee_containment.h>
#include <gee_io.h>
#include <gee_macros.h>
#include <gee_mover.h>
#include <gee_recolor.h>
#include <gee_renderer.h>
#include <gee_world.h>

#include <vee_error.h>
#include <vee_particle.h>

static const int    GEE_ParticleDataId    = GEE_CONST('P', 'R', 'D', 'T');
static const ushort GEE_ParticleLocId     = GEE_CONST2('L', 'O');
static const ushort GEE_ParticleVelId     = GEE_CONST2('V', 'E');
static const ushort GEE_ParticleColorId   = GEE_CONST2('C', 'O');
static const ushort GEE_ParticleTextureId = GEE_CONST2('T', 'X');
static const ushort GEE_ParticleSizeId    = GEE_CONST2('S', 'I');
static const ushort GEE_ParticleRotId     = GEE_CONST2('P', 'R');
static const ushort GEE_RenderClassId     = GEE_CONST2('R', 'C');
static const ushort GEE_EndId             = GEE_CONST2('E', 'N');


bool GEE_ParticleData::read(GEE_Io *io)
{
  if(io->readInt32() != GEE_ParticleDataId)
    return false;

  int v = io->readInt32();

  ushort id = io->readInt16();

  m_texture = -1;

  std::string tmp;

  if(!m_particle.ptr())
    m_particle = new VEE_Particle();

  while(id != GEE_EndId) {
    if(id == GEE_ParticleLocId)
      setLocation(io->readVector3());
    else if(id == GEE_ParticleVelId)
      setVelocity(io->readVector3());
    else if(id == GEE_ParticleColorId)
      setColor(io->readRgba());
    else if(id == GEE_ParticleTextureId) {
      io->readNullTerminatedStdString(tmp);
      setTexture(tmp, GEE_World::textureId(tmp));
    }
    else if(id == GEE_ParticleSizeId)
      setSize(io->readFloat32());
    else if(id == GEE_ParticleRotId)
      setRotation(io->readFloat32());
    else if(id == GEE_RenderClassId) {
      io->readNullTerminatedStdString(tmp);
      setRenderClass(tmp, 
		     GEE_World::renderClass(tmp));
      
    }
    else {
      VEE_error("GEE_ParticleData::read # Bad id at %d (%c%c)", 
		(int) io->position(), (uchar) (id >> 8), (uchar) id);
      return false;
    }

    id = io->readInt16(); 
  }

  if(m_texture < 0)
    VEE_error("GEE_ParticleData::read # No texture \"%s\"", 
	      m_textureName.c_str());

  return v == 1;
}

/** Writes data to a file. Data is written in chunks with each chunk
    having a 2-byte identifier before the data. This is automatically
    extensible. */

bool GEE_ParticleData::write(GEE_Io *io)
{
  io->writeInt32(GEE_ParticleDataId);
  io->writeInt32(1); // Version

  io->writeInt16(GEE_ParticleLocId);
  io->writeVector3(m_location);

  io->writeInt16(GEE_ParticleVelId);
  io->writeVector3(m_velocity);

  io->writeInt16(GEE_ParticleColorId);
  io->writeRgba(m_color);

  io->writeInt16(GEE_ParticleTextureId);
  io->writeNullTerminatedString(m_textureName.c_str());

  io->writeInt16(GEE_ParticleSizeId);
  io->writeFloat32(m_size);

  io->writeInt16(GEE_ParticleRotId);
  io->writeFloat32(m_rotation);
  
  io->writeInt16(GEE_RenderClassId);
  io->writeNullTerminatedString(m_renderClassName.c_str());
  
  io->writeInt16(GEE_EndId);

  return true;
}

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

GEE_ParticleObject::GEE_ParticleObject()
{
  m_bbox.clear();
}

GEE_ParticleObject::~GEE_ParticleObject()
{}

GEE_Geometry * GEE_ParticleObject::create()
{
  return new GEE_ParticleObject();
}

GEE_Geometry * GEE_ParticleObject::clone() const
{
  return new GEE_ParticleObject(*this);
}


bool GEE_ParticleObject::eraseData(GEE_Containment *co)
{
  bool hit = false;

  VEE_Matrix4 m;
  getWorldTransformation(&m);

  if(!intersectsBounded(co, &m_bbox, m))
    return false;

  for(iterator it = begin(); it != end(); ) {
    GEE_ParticleData & pd = getParticleData(it);
    if(co->contains(m * pd.m_location)) {
      iterator tmp = it;
      tmp++;
      m_particles.erase(it);
      it = tmp;
      hit = true;
    }
    else
      it++;
  }

  if(hit) {
    incrementGeneration();
    co->contained(this);
  }

  return hit;
}

bool GEE_ParticleObject::intersects(GEE_Containment *co)
{
  VEE_Matrix4 m;
  getWorldTransformation(&m);

  if(!intersectsBounded(co, &m_bbox, m))
    return false;

  for(iterator it = begin(); it != end(); it++) {
    GEE_ParticleData & pd = getParticleData(it);
    if(co->contains(m * pd.m_location))
      return true;
  }
  return false;
}

bool GEE_ParticleObject::colorData(GEE_ReColor * rc)
{
  bool hit = false;

  VEE_Matrix4 m;
  getWorldTransformation(&m);

  if(!intersectsBounded(rc, &m_bbox, m))
    return false;

  for(iterator it = begin(); it != end(); it++) {
    GEE_ParticleData & pd = getParticleData(it);
    
    bool changed = rc->recolor(m * pd.m_location, & pd.m_color);
    
    if(changed)
      pd.setColor(pd.m_color);
  }

  if(hit) {
    incrementGeneration();
    rc->recolored(this);
  }

  return hit;
}

bool GEE_ParticleObject::moveData(GEE_Mover *m)
{
  bool hit = false;

  VEE_Matrix4 ma, mi;
  bool ok ;
  getWorldTransformations(&ma, & mi, & ok);

  if(!ok) {
    VEE_error("GEE_ParticleObject::moveData # Bad inverse matrix");
    return false;
  }

  if(!intersectsBounded(m, &m_bbox, ma)) {
    // VEE_trace("GEE_ParticleObject::moveData # culled a mover");
    return false;
  }

  for(iterator it = begin(); it != end(); it++) {
    GEE_ParticleData & pd = getParticleData(it);
    VEE_Vector3 tmp = ma * pd.m_location;
    if(m->move(& tmp)) {
      hit = true;
      tmp = mi * tmp;
      pd.setLocation(tmp);
      m_bbox.expand(tmp);
    }
  }

  if(hit) {
    incrementGeneration();
    m->moved(this);
  }

  return hit;
}

void GEE_ParticleObject::render(GEE_Renderer *r)
{
  r->pushTransform(transformation());
  r->renderParticleObject(this);
  r->popTransform();
}

bool GEE_ParticleObject::getBoundingBox(VEE_BoundingBox *b)
{
  *b = m_bbox;
  return true;
}

void GEE_ParticleObject::setMaterial(const GEE_Material &m)
{
  GEE_Geometry::setMaterial(m);
  
  VEE_Rgba rgba;
  rgba.copy(m.ambient().data(), 4);
  rgba.setTheta(m.burn());

  for(iterator it = begin(); it != end(); it++)
    getParticleData(it).setColor(rgba);
}


static const ushort GEE_PdId  = GEE_CONST2('P', 'D');

bool GEE_ParticleObject::read(GEE_Io *io)
{
  std::string name;
  io->readNullTerminatedStdString(name);

  if(name != "GEE_ParticleObject") {
    VEE_error("GEE_ParticleObject::read # Bad type (%s)", name.c_str());
    return false;
  }

  readBasics(io);
  
  int v = io->readInt32();
  if(v != 1)
    return false;
  
  ushort id = io->readInt16();

  GEE_ParticleData pd;
  
  while(id != GEE_EndId) {

    if(id == GEE_PdId) {
      if(pd.read(io))
	addParticle(pd);
    }
    else {
      VEE_error("GEE_ParticleObject::read # Bad id at %d (%c%c)", 
		(int) io->position(), (uchar) (id >> 8), (uchar) id);
      return false;
    }

    id = io->readInt16();
  }

  VEE_trace("GEE_ParticleObject::read # With %d particles", 
	    (int) m_particles.size());

  return true;
}

bool GEE_ParticleObject::write(GEE_Io *io)
{
  io->writeNullTerminatedString("GEE_ParticleObject");

  writeBasics(io);

  io->writeInt32(1); // version

  for(iterator it = m_particles.begin(); it != m_particles.end(); it++) {
    io->writeInt16(GEE_PdId);
    (*it).write(io);
  }
  
  io->writeInt16(GEE_EndId);

  return true;
}

GEE_Geometry * GEE_ParticleObject::removeData(const VEE_Vector3 &loc, float radius)
{
  GEE_ParticleObject * po = new GEE_ParticleObject();

  for(iterator it = begin(); it != end(); it++) {
    GEE_ParticleData & pd = getParticleData(it);
    if((pd.m_location - loc).length() > radius)
      po->addParticle(pd);
  }

  if(!po->particleCount()) {
    delete po;
    return 0;
  }

  return po;
}

void GEE_ParticleObject::addParticle(const GEE_ParticleData &pd)
{
  if(!m_particles.size())
    m_bbox.set(pd.m_location);
  else
    m_bbox.expand(pd.m_location);

  m_particles.push_front(pd);

  /* VEE_Vector3 l = pd.m_location;

  VEE_trace("GEE_ParticleObject::addParticle # %p %p %d [%f %f %f]", 
	    (*m_particles.begin()).m_particle.ptr(), 
	    pd.m_particle.ptr(), (int) m_particles.size(),
	    l.x, l.y, l.z);
  */

}

void GEE_ParticleObject::retexture(GEE_Containment *co, 
				   const std::string & texName, 
				   int texId)
{
  VEE_Matrix4 m;
  getWorldTransformation(&m);

  if(!intersectsBounded(co, &m_bbox, m))
    return;

  bool hit = false;

  for(iterator it = begin(); it != end(); it++) {

    GEE_ParticleData & pd = getParticleData(it);

    if(co->contains(m * pd.m_location)) {
      // VEE_trace("GEE_ParticleObject::retexture");
      pd.setTexture(texName, texId);
    }
  }

  if(hit)
    incrementGeneration();
}
