/* -*- C++ -*- */

/* COPYRIGHT
 *
 * This file is part of the Visual Effects Engine - VEE
 *
 * Read the "VEE-LICENSE" file for the license.
 *
 * Authors & Copyright:   
 *
 * Tommi Ilmonen, Tuukka Heikura, Marko Myllymaa and 
 * Janne Kontkanen 2001-2004
 *
 * Additional copyrights: Tekes 2003-2004
 *
 * firstname.lastname@hut.fi
 *
 */


#include <gee_particle_rotate.h>

#include <gee_io.h>
#include <gee_macros.h>
#include <gee_particle_object.h>
#include <gee_world.h>

#include <vee_error.h>

GEE_ParticleRotate::GEE_ParticleRotate(GEE_id_t pathId)
  : GEE_PathToX(pathId),
    m_pressure(1.0f),
    m_omega(1.0),
    m_radius(1.0f),
    m_radiusBoost(0.0f),
    m_radiusNorm(1.0f),    
    m_useNapra(false)
{
  m_writer.addMemberFloat(GEE_CONST2('P', 'R'), & m_pressure);
  m_writer.addMemberFloat(GEE_CONST2('O', 'M'), & m_omega);
  m_writer.addMemberFloat(GEE_CONST2('R', 'A'), & m_radius);
  m_writer.addMemberFloat(GEE_CONST2('R', 'B'), & m_radiusBoost);
  m_writer.addMemberFloat(GEE_CONST2('R', 'N'), & m_radiusNorm);
  m_writer.addMemberBool (GEE_CONST2('U', 'N'), & m_useNapra);
}

GEE_ParticleRotate::~GEE_ParticleRotate()
{}
  
bool GEE_ParticleRotate::act(GEE_World *w, const GEE_TimeStamp &ts)
{
  m_turns.clear();

  bool ok = GEE_PathToX::act(w, ts);

  if(m_turns.size())
    doTurn(w);

  return ok;
}

bool GEE_ParticleRotate::read(GEE_Io *io)
{
  std::string str;
  io->readNullTerminatedStdString(str);
  if(str != type())
    return false;
  
  readBasics(io);
  readPathId(io);

  int v = io->readInt32();

  if(v != 1)
    return false;

  return m_writer.readParameters(io);
}

bool GEE_ParticleRotate::write(GEE_Io *io)
{
  io->writeNullTerminatedString(type());

  writeBasics(io);
  writePathId(io);

  io->writeInt32(1); // version

  return m_writer.writeParameters(io);
}

const char * GEE_ParticleRotate::type() 
{
  return "GEE_ParticleRotate"; 
}

GEE_Operator * GEE_ParticleRotate::create()
{
  return new GEE_ParticleRotate();
}

void GEE_ParticleRotate::doTurn(GEE_World * w)
{
  if(!m_turns.size())
    return;

  GEE_Geometry * g;

  uint i, n = m_turns.size();

  // calculate bounding box:
  VEE_BoundingBox bbox;
  bbox.set(m_turns[0].m_location, m_turns[0].m_radius);
  for(i = 1; i < n; i++)
    bbox.expand(m_turns[i].m_location, m_turns[1].m_radius);

  // i = 0;
  
  for(GEE_LayerGeometryIterator it(w, m_layer); (g = *it); it.nextVisible()){

    GEE_ParticleObject * po = dynamic_cast<GEE_ParticleObject *> (g);
    
    if(!po)
      continue;
    
    VEE_Matrix4 m;

    po->getWorldTransformation(&m);

    for(GEE_ParticleObject::iterator it = po->begin(); it != po->end(); it++) {

      GEE_ParticleData & pd = GEE_ParticleObject::getParticleData(it);
      
      VEE_Vector3 loc = m * pd.m_location;

      if(!bbox.contains(loc))
	continue;
      
      float rot = pd.m_rotation;
      
      for(uint i = 0; i < n; i++) {
	const Item & item = m_turns[i];
	
	float dist = (item.m_location - loc).length();
	float weight = m_pressure * (dist - item.m_radius) / item.m_radius;
	
	if(weight > 1.0f) weight = 1.0f;

	rot += weight * item.m_turn;
      }

      pd.setRotation(rot);
      // i++;
    }
  }

  // VEE_trace("GEE_ParticleRotate::doTurn # %u particles", i);
}

bool GEE_ParticleRotate::handleNewVertex(GEE_Path3D::iterator & it)
{
  if(!processedPathVertices()) {
    m_last = (*it).second;
    return true;
  }

  const GEE_PathVertex3D & pv = (*it).second;

  VEE_Vector3 v = (pv.m_rotation * VEE_Vector3_UNIT_Z) * m_last.m_rotation;
  
  Item item;

  item.m_turn = -m_omega * atan2f(v.x, v.z);
  
  if(m_useNapra) {
    item.m_radius = 
      pv.calculateToolRadius(m_radius, m_radiusNorm, m_radiusBoost);
    item.m_location = pv.calculateToolCenter();
  }
  else {
    item.m_location = pv.m_location;
    item.m_radius = m_radius;
  }

  m_turns.push_back(item);

  m_last = pv;
  
  return true;
}

