/* COPYRIGHT
 *
 * This file is part of the Visual Effects Engine - VEE
 *
 * Read the "VEE-LICENSE" file for the license.
 *
 * Authors & Copyright:   
 *
 * Tommi Ilmonen, Tuukka Heikura, Marko Myllymaa and 
 * Janne Kontkanen 2001-2004
 *
 * Additional copyrights: Tekes 2003-2004
 *
 * firstname.lastname@hut.fi
 *
 */

#include <gee_path.h>

#include <gee_containment.h>
#include <gee_io.h>
#include <gee_macros.h>
#include <gee_renderer.h>

#include <vee_error.h>
#include <vee_math.h>

static const int GEE_PathVertex3DId = GEE_CONST('P', 'V', 'X', '3');
static const ushort GEE_LocId  = GEE_CONST2('L', 'O');
static const ushort GEE_RotId  = GEE_CONST2('R', 'O');
static const ushort GEE_TimeId = GEE_CONST2('T', 'I');
static const ushort GEE_EndId  = GEE_CONST2('E', 'N');

bool GEE_PathVertex3D::read(GEE_Io *io)
{
  if(io->readInt32() != GEE_PathVertex3DId)
    return false;
  int v = io->readInt32();

  m_time = 0.0f;

  ushort id = io->readInt16();

  while(id != GEE_EndId) {
    if(id == GEE_LocId)
      m_location = io->readVector3();
    else if(id == GEE_RotId)
      m_rotation = io->readMatrix3();
    else if(id == GEE_TimeId)
      m_time = io->readFloat32();
    else if((id & 0xFF) == 't') {
      uint index = id >> 8;
      // VEE_trace("Getting vertex %u", index);
      m_tips[index] = io->readVector3();
    }
    else {
      VEE_error("GEE_PathVertex3D::read # Bad id at %d (%c%c)", 
		(int) io->position(), (uchar) (id >> 8), (uchar) id);
      return false;
    }

    id = io->readInt16();
  }
  /* VEE_trace("GEE_PathVertex3D::read # %d, %f %f %f", v,
     m_location.x, m_location.y, m_location.z); */

  return v == 1;
}

bool GEE_PathVertex3D::write(GEE_Io *io)
{
  io->writeInt32(GEE_PathVertex3DId);
  io->writeInt32(1); // version

  io->writeInt16(GEE_LocId);
  io->writeVector3(m_location);
  io->writeInt16(GEE_RotId);
  io->writeMatrix3(m_rotation);
  io->writeInt16(GEE_TimeId);
  io->writeFloat32(m_time);
  
  for(uint i=0; i < 3; i++) {
    io->writeInt16(GEE_CONST2('t', i));
    io->writeVector3(m_tips[i]);    
  }

  io->writeInt16(GEE_EndId);

  /* VEE_trace("GEE_PathVertex3D::write # 1 %f %f %f",
     m_location.x, m_location.y, m_location.z); */

  return true;
}

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

GEE_Path3D::GEE_Path3D()
  : m_counter(0),
    m_activeNapra(false)
{
  m_bbox.clear();
  m_writer.addMemberBool(GEE_CONST2('N', 'A'), &m_activeNapra);
}

GEE_Path3D::~GEE_Path3D()
{}
  
GEE_Geometry * GEE_Path3D::create()
{
  return new GEE_Path3D();
}

GEE_Geometry * GEE_Path3D::clone() const
{
  return new GEE_Path3D(*this);
}

bool GEE_Path3D::intersects(GEE_Containment *co)
{
  VEE_Matrix4 m;
  getWorldTransformation(&m);

  if(!intersectsBounded(co, &m_bbox, m))
    return false;
  
  for(iterator it = begin(); it != end(); it++) {
    GEE_PathVertex3D & pv = getVertex(it);
    
    if(co->contains(m * pv.m_location))
      return true;
  }
  
  return false;  
}

bool GEE_Path3D::getBoundingBox(VEE_BoundingBox *b)
{
  *b = m_bbox;
  return true;
}

void GEE_Path3D::render(GEE_Renderer *r)
{
  r->pushTransform(transformation());
  r->renderPath3D(this);
  r->popTransform();
}


bool GEE_Path3D::read(const char *filename)
{
  GEE_File f;
  if(!f.open(filename, DI_IO_INPUT)) {
    return false;
  }
  
  return read(&f);
}

bool GEE_Path3D::write(const char *filename)
{
  GEE_File f;
  if(!f.open(filename, DI_IO_OUTPUT)) {
    return false;
  }
  
  return write(&f);
}

void GEE_Path3D::addVertex(const GEE_PathVertex3D &v)
{ 
  if(!m_vertices.size())
    m_bbox.set(v.m_location); 
  else
    m_bbox.expand(v.m_location); 

  m_vertices[m_counter++] = v; 
}

/** Reads the stuff from a stream. Version 1 does not have tagged
    variable values, version 2 introduced them. */

bool GEE_Path3D::read(GEE_Io *io)
{
  // const char * fname = "GEE_Path3D::read";

  std::string str;
  io->readNullTerminatedStdString(str);
  if(str != "PATH_3D")
    return false;

  readBasics(io);
  
  int v = io->readInt32();

  if(v > 2)
    return false;

  if(v == 2) {
    if(!m_writer.readParameters(io))
      return false;
  }

  int vertices = io->readInt32();

  GEE_PathVertex3D pv;

  bool ok = true;

  for(int i=0; i < vertices; i++) {
    int t = io->readInt32();
    if(pv.read(io))
      addVertex(t, pv);
    else
      ok = false;
  }
  
  /* VEE_trace("%s # Read %d vertices (%s)", fname, vertices, 
	    ok ? "SUCCESS" : "FAILURE");

  std::cout << m_bbox.low() << " # " << m_bbox.high() 
  << "\nM = " << transformation() << "\n"; */

  return ok;
}

bool GEE_Path3D::write(GEE_Io *io)
{
  io->writeNullTerminatedString("PATH_3D");
  writeBasics(io);

  io->writeInt32(2); // version
  
  m_writer.writeParameters(io);

  io->writeInt32(m_vertices.size());

  for(iterator it = begin(); it != end(); it++) {
    io->writeInt32((*it).first);
    getVertex(it).write(io);
  }
  
  return true;
}


void GEE_Path3D::create3DSpiral(int rounds, 
				int segmentsPerRound, 
				float secondsPerRound,
				float initialRadius,
				float depth)
{
  int totalPoints = rounds * segmentsPerRound;
  float dt = secondsPerRound / segmentsPerRound;
  float points = totalPoints;

  VEE_Matrix3 m;
  m.rotateAroundAxis(VEE_Vector3(1, 1, 0), 1.4);

  GEE_PathVertex3D pv;

  pv.m_tips[1].make(0.5, 0, 0);
  pv.m_tips[2].make(0.25, 0.5, 0);

  pv.m_rotation = m;
  
  for(int i=0; i < totalPoints; i++) {
    float angle  = i / (float) segmentsPerRound  * VEE_TWO_PI;
    float radius = (1.0 - i / points) * initialRadius;

    VEE_Vector3 location;
    
    pv.m_location.make(sin(angle), cos(angle), 0.0);
    pv.m_location *= radius;
    pv.m_location.z = depth * i / points;

    addVertex(pv);

    pv.m_time += dt;
  }

  
  VEE_trace("GEE_Path3D::create3DSpiral # %d vertices", 
	    (int) m_vertices.size());
}

bool GEE_Path3D::operator == (const GEE_Path3D &that) const
{
  if(materialName() != that.materialName() || 
     generation() != that.generation())
    return false;

  const_iterator i1 = begin();
  const_iterator i2 = that.begin();

  while(i1 != end() && i2 != that.end()) {
    if(!((*i1).second == (*i2).second) ||
	 !((*i1).first == (*i2).first))
      return false;
    
    i1++;
    i2++;
  }

  return true;
}
