/* COPYRIGHT
 *
 * This file is part of the Visual Effects Engine - VEE
 *
 * Read the "VEE-LICENSE" file for the license.
 *
 * Authors & Copyright:   
 *
 * Tommi Ilmonen, Tuukka Heikura, Marko Myllymaa and 
 * Janne Kontkanen 2001-2004
 *
 * Additional copyrights: Tekes 2003-2004
 *
 * firstname.lastname@hut.fi
 *
 */

#include <gee_path_multiply.h>

#include <gee_io.h>
#include <gee_macros.h>
#include <gee_path_to_mesh.h>
#include <gee_world.h>

#include <vee_error.h>


GEE_PathMultiply::GEE_PathMultiply(GEE_id_t pathId)
  : GEE_PathToX(pathId),
    m_scale(1.0),
    m_threshold(0.0),
    m_align(0),
    m_newSegment(true)
{
  m_extrude.clear();

  m_writer.addMemberFloat(GEE_CONST2('S', 'C'), & m_scale);
  m_writer.addMemberFloat(GEE_CONST2('T', 'H'), & m_threshold);
  m_writer.addMemberInt(GEE_CONST2('A', 'L'),   & m_align);
  m_writer.addMemberVector3(GEE_CONST2('E', 'X'),   & m_extrude);
}

GEE_PathMultiply::~GEE_PathMultiply()
{}

GEE_IdContainer::id_container * GEE_PathMultiply::getExtraIds()
{
  return & m_ids;
}

bool GEE_PathMultiply::start(GEE_World *w, const GEE_TimeStamp &ts)
{
  GEE_FNAME = "GEE_PathMultiply::start";

  VEE_trace("%s # %u", fname, m_profile.size());

  if(!GEE_PathToX::start(w, ts))
    return false;

  uint i, n = m_profile.size();
  
  if(n != m_ids.size()) {
    VEE_error("%s # Size mismatch %u != %u", fname, n, (uint) m_ids.size());
    setProfile(m_profile);
  }

  m_paths.resize(n);

  i = 0;

  for(id_container::iterator it = m_ids.begin(); it != m_ids.end(); it++) {

    GEE_Path3D * path = new GEE_Path3D();
    path->setId(*it);

    Item & item = m_paths[i];
    item.m_geo  = path;
    item.m_path = path;
    item.m_vertex = m_profile.getVertex(i).m_location;

    w->addGeometry2(item.m_geo);

    VEE_trace("%s # Created path %p (%u)", fname, path, i);

    i++;
  }
  
  m_newSegment = true;
  m_prevLoc.clear();

  return true;
}

bool GEE_PathMultiply::stop(GEE_World *w, const GEE_TimeStamp &ts)
{
  GEE_PathToX::stop(w, ts);
  
  m_paths.clear();

  return true;
}

bool GEE_PathMultiply::read(GEE_Io *io)
{
  std::string str;
  io->readNullTerminatedStdString(str);
  if(str != type())
    return false;
  
  readBasics(io);
  readPathId(io);

  int v = io->readInt32();

  if(v != 1)
    return false;

  bool ok = m_writer.readParameters(io);

  ok = ok && readIds(io, & m_ids);
  ok = ok && m_profile.read(io);

  return ok;
}
  
bool GEE_PathMultiply::write(GEE_Io *io)
{
  io->writeNullTerminatedString(type());

  writeBasics(io);
  writePathId(io);

  io->writeInt32(1); // version

  m_writer.writeParameters(io);

  writeIds(io, & m_ids);
  m_profile.write(io);

  return true;
}

const char * GEE_PathMultiply::type()
{
  return "GEE_PathMultiply";
}

GEE_Operator * GEE_PathMultiply::create()
{
  return new GEE_PathMultiply();
}

void GEE_PathMultiply::setProfile(const GEE_Profile & profile)
{
  m_profile = profile;

  uint n = m_profile.size();

  if(n != m_ids.size()) {
    m_ids.clear();

    for(uint i = 0; i < n; i++)
      m_ids.insert(allocateId());
  }
}

/** Gets the nth path id. */

GEE_id_t GEE_PathMultiply::pathId(int index)
{
  id_container::iterator it = m_ids.begin(); 

  for(int i = 0; i < index && it != m_ids.end(); i++) 
    it++;
  
  return *it;
}

bool GEE_PathMultiply::handleNewVertex(GEE_Path3D::iterator & it)
{
  GEE_PathVertex3D & tv = GEE_Path3D::getVertex(it);
  
  // VEE_trace("GEE_PathMultiply::handleNewVertex # %f", tv.m_time);

  VEE_Vector3 dir = tv.m_location - m_prevLoc;
  m_dir = dir;

  if(processedPathVertices() == 0) {
    m_prevLoc = tv.m_location;
  }

  VEE_Matrix4 m =  GEE_PathToMesh::createMatrix(tv, m_scale);

  uint i, n = m_paths.size();

  VEE_Vector3 extra = tv.m_rotation * m_extrude;
  
  for(i = 0; i < n; i++) {

    Item & item = m_paths[i];

    VEE_Vector3 loc;

    loc.vector2() = item.m_vertex;
    loc.z = 0.0;
    loc = m * loc + extra;

    GEE_PathVertex3D nv(loc, tv.m_rotation, tv.m_time); // New vertex

    if(m_newSegment)
      item.m_path->addVertex(nv);
    else {
      GEE_Path3D::iterator it = item.m_path->end();
      it--;
      (*it).second = nv;
    }
  }
  
  if((m_dir.length() >= m_threshold)) {
    m_newSegment = true;
    m_prevLoc = tv.m_location;
  }
  else
    m_newSegment = false;

  return true;
}
