/* COPYRIGHT
 *
 * This file is part of the Visual Effects Engine - VEE
 *
 * Read the "VEE-LICENSE" file for the license.
 *
 * Authors & Copyright:   
 *
 * Tommi Ilmonen, Tuukka Heikura, Marko Myllymaa and 
 * Janne Kontkanen 2001-2004
 *
 * Additional copyrights: Tekes 2003-2004
 *
 * firstname.lastname@hut.fi
 *
 */

#include <gee_path_to_x.h>

#include <gee_debug.h>
#include <gee_io.h>
#include <gee_macros.h>
#include <gee_sphere.h>
#include <gee_world.h>

#include <vee_error.h>

static bool m_debug = false;

inline GEE_Path3D * GEE_PathToX::getPath(GEE_World *w, GEE_id_t id)
{
  return dynamic_cast<GEE_Path3D *> (w->geometry(id));
}

GEE_PathToX::GEE_PathToX(GEE_id_t pathId, 
			 GEE_id_t pathId2)
  : m_pathId(pathId),
    m_pathId2(pathId2),
    m_nodeIsUsed(false),
    m_showToolSphere(false),
    m_sphereRadius(1.0)
{
  addIdMember(&m_pathId);
  addIdMember(&m_pathId2);
  m_debug = GEE_Debug::debug("GEE_PathToX");
}

GEE_PathToX::~GEE_PathToX()
{}

bool GEE_PathToX::start(GEE_World *w, const GEE_TimeStamp &)
{
  GEE_FNAME = "GEE_PathToX::start";

  if(m_debug) VEE_trace(fname);

  m_pathIterator1 = -1;
  m_pathIterator2 = -1;

  if(!initializePath(w, m_pathId))
    return false;
    
  if(m_pathId2 >= 0 && !initializePath(w, m_pathId2))
    return false;

  m_nodeIsUsed = false;

  m_pathIterator1 = 0;
  m_pathIterator2 = 0;
  m_processedPathVertices = 0;

  if(m_keepAlive && m_showToolSphere) {
    GEE_Sphere * sphere = new GEE_Sphere();
    sphere->setRadius(m_sphereRadius);
    m_sphereId = w->addGeometry(sphere);
  }
  else
    m_sphereId = -1;

  if(m_debug) VEE_trace("%s # EXIT OK", fname);

  m_edits.clear();

  return true;
}

bool GEE_PathToX::act(GEE_World *w, const GEE_TimeStamp & tworld)
{
  if(m_pathId2 < 0)
    return handlePath(w, tworld);
  else 
    return handleTwoPaths(w, tworld);
}

bool GEE_PathToX::stop(GEE_World *w, const GEE_TimeStamp &)
{
  if(m_debug) VEE_trace("GEE_PathToLines::stop");

  if(m_sphereId >= 0)
    w->eraseGeometry(m_sphereId);

  m_edits.detach(w);

  if(m_keepAlive) {
    
  }

  return true;
}

bool GEE_PathToX::readPathId(GEE_Io * io)
{
  int v = io->readInt32();

  if(v != 1)
    return false;
  
  m_pathId  = io->readInt64();
  m_pathId2 = io->readInt64();
  
  return true;
}

bool GEE_PathToX::writePathId(GEE_Io * io)
{  
  io->writeInt32(1); // version
  io->writeInt64(m_pathId);
  io->writeInt64(m_pathId2);

  return true;
}

bool GEE_PathToX::handleNewVertex(GEE_Path3D::iterator &)
{
  VEE_error("GEE_PathToX::handleNewVertex # MUST BE OVERRIDDEN");
  return false;
}

bool GEE_PathToX::handleNewVertices(GEE_Path3D::iterator &, 
				     GEE_Path3D::iterator &)
{
  VEE_error("GEE_PathToX::handleNewVertices # MUST BE OVERRIDDEN");
  return false;
}


bool GEE_PathToX::initializePath(GEE_World *w, GEE_id_t id)
{  
  GEE_Geometry * d = w->geometry(id);

  if(!dynamic_cast<GEE_Path3D *> (d)) {
    VEE_error("GEE_PathToX::initializePath # No geometry %d (%p)", 
	      (int) id, this);
    return false;
  }

  // path information is hidden:
  d->setVisible(false);
  
  return true;
}

bool GEE_PathToX::handlePath(GEE_World *w, const GEE_TimeStamp & tworld)
{
  GEE_FNAME = "GEE_PathToX::handlePath";

  if(m_debug)
    VEE_trace(fname);

  if(m_pathIterator1 < 0 || !m_manager) {
    VEE_error("%s # m_pathIterator1 = %d, m_manager = %p", 
	      fname, m_pathIterator1, m_manager);
    return false;
  }

  GEE_Path3D * path = getPath(w, m_pathId);

  if(!path) {
    if(m_debug) VEE_trace("%s # No path %d", fname, (int) m_pathId);
    return false;
  }

  path->setVisible(false);

  if(m_pathIterator1 >= path->counter()) {
    if(m_debug) VEE_trace("%s # No new vertices (%d %d).", 
			  fname, (int) m_pathIterator1, 
			  (int) path->vertexCount());
    return true;
  }

  bool ok = true;

  // GEE_TimeStamp time = m_manager->realTimeToOperatorTime(tworld, m_group);
  GEE_TimeStamp time = tworld;
  
  if(m_debug)
    VEE_trace("%s %s # %d %lf %lf", 
	      typeid(*this).name(), fname, (int) m_group, 
	      tworld.secondsD(), time.secondsD());

  double dtime = time.secondsD();

  if(m_debug) VEE_trace(fname);

  GEE_Path3D::iterator it = path->lower_bound(m_pathIterator1);

  VEE_Vector3 loc;
  loc.clear();

  for(; it != path->end(); it++) {

    float vtime = GEE_Path3D::getTime(it);

    if(vtime > dtime && !m_keepAlive)
      break;

    ok = ok && handleNewVertex(it);
    m_pathIterator1 = GEE_Path3D::getIndex(it) + 1;
    m_processedPathVertices++;

    loc = GEE_Path3D::getVertexLocation(it);
  }

  if(m_sphereId >= 0) {
    GEE_Geometry * geo = w->geometry(m_sphereId);
    if(geo) {
      geo->transformation().setColumn3(3, VEE_Vector3d(loc));
    }
  }

  return ok;
}

bool GEE_PathToX::handleTwoPaths(GEE_World *w, const GEE_TimeStamp & tworld)
{
  GEE_FNAME = "GEE_PathToX::handleTwoPaths";
    
  if(m_debug)
    VEE_trace(fname);

  if(m_pathIterator1 < 0 || m_pathIterator2 < 0 || !m_manager) {
    VEE_error("%s # m_pathIterator1 = %d, m_pathIterator2 = %d, m_manager = %p", 
	      fname, m_pathIterator1, m_pathIterator2, m_manager);
    return false;
  }

  GEE_Path3D * path1 = getPath(w, m_pathId);
  GEE_Path3D * path2 = getPath(w, m_pathId2);

  if(!path1 || !path2) {
    if(m_debug) VEE_trace("%s # Paths: %p %p", fname, path1, path2);
    return false;
  }

  if(m_pathIterator1 >= path1->counter() || 
     m_pathIterator2 >= path2->counter()) {
    if(m_debug) VEE_trace("%s # No new vertices.", fname);
    return true;
  }

  bool ok = true;

  // GEE_TimeStamp time = m_manager->realTimeToOperatorTime(tworld, m_group);
  GEE_TimeStamp time = tworld;

  // VEE_trace("%s::act # %d", typeid(*this).name(), (int) m_group);

  double dtime = time.secondsD();

  if(m_debug) VEE_trace(fname);
  
  GEE_Path3D::iterator it1 = path1->lower_bound(m_pathIterator1);
  GEE_Path3D::iterator it2 = path2->lower_bound(m_pathIterator2);

  while(it1 != path1->end() && it2 != path2->end()) {

    float vtime = GEE_Path3D::getTime(it1);

    if(vtime > dtime)
      break;

    int i1 = GEE_Path3D::getIndex(it1);
    int i2 = GEE_Path3D::getIndex(it2);

    if(i1 > i2) {
      it2++;
    }
    else if(i2 > i1) {
      it1++;
    }
    else {
      ok = ok && handleNewVertices(it1, it2);
      
      it1++;
      it2++;
      m_processedPathVertices++;
    }

    m_pathIterator1 = i1 + 1;
    m_pathIterator2 = i2 + 1;

    // VEE_trace("%s # loop: %d %d", fname, i1, i2);
  }

  return ok;  

}

