/* COPYRIGHT
 *
 * This file is part of the Visual Effects Engine - VEE
 *
 * Read the "VEE-LICENSE" file for the license.
 *
 * Authors & Copyright:   
 *
 * Tommi Ilmonen, Tuukka Heikura, Marko Myllymaa and 
 * Janne Kontkanen 2001-2004
 *
 * Additional copyrights: Tekes 2003-2004
 *
 * firstname.lastname@hut.fi
 *
 */

#include <gee_polygon_mesh.h>

#include <gee_renderer.h>

#include <vee_error.h>

GEE_PolygonMesh::GEE_PolygonMesh()
{}

GEE_PolygonMesh::~GEE_PolygonMesh()
{}

GEE_Geometry * GEE_PolygonMesh::clone() const
{
  return new GEE_PolygonMesh(*this);
}

void GEE_PolygonMesh::render(GEE_Renderer *r)
{
  r->pushTransform(transformation());
  r->renderPolygonMesh(this);
  r->popTransform();
}

bool GEE_PolygonMesh::getBoundingBox(VEE_BoundingBox *b)
{
  *b = m_bbox;
  return true;
}

bool GEE_PolygonMesh::move(const VEE_Vector3 & offset)
{
  if(!m_vertices.size())
     return true;

  m_bbox.move(offset);

  for(vertexIterator it = beginVertex(); it != endVertex(); it++)
    (*it).second.m_location += offset;

  return true;
}

void GEE_PolygonMesh::clear()
{
  m_vertices.clear();
  m_polygons.clear();
  m_vertexCounter = 0;
  m_polygonCounter = 0;
}

/** Adds a new vertex point to the mesh and returns the index number
    of the vertex. */
int GEE_PolygonMesh::addVertex(const GEE_PolygonVertex &pv)
{
  if(!m_vertices.size())
    m_bbox.set(pv.m_location);
  else
    m_bbox.expand(pv.m_location);

  int r = m_vertexCounter;
  m_vertices[m_vertexCounter++] = pv;
  return r;  
}

/** Adds a polygon and returns its index. If the vertices of the
    polygon do not exist, then -1 returned and the polygon is not
    added. */

int GEE_PolygonMesh::addPolygon(const GEE_PolygonIndices &poly)
{
  int i, r = m_polygonCounter;

  // Check that vertices exist:
  for(i=0; i < (int) poly.m_indices.size(); i++) {
    if(!getVertex(poly.m_indices[i])) {
      VEE_error("GEE_PolygonMesh::addPolygon # Missing vertex data %d",
		poly.m_indices[i]);
      return -1;
    }
  }
  
  // Vertices do exist, now link:

  for(i=0; i < (int) poly.m_indices.size(); i++) {
    getVertex(poly.m_indices[i])->addPolygon(r);
  }

  m_polygons[r] = poly;

  m_polygonCounter++;
    
  return r;
}

/** Returns a vertex (or NULL if not found). */

GEE_PolygonVertex * GEE_PolygonMesh::getVertex(int id)
{
  vertexIterator it = m_vertices.find(id);
  
  if(it != m_vertices.end()) {
    return & (*it).second;
  }

  return 0;
}

GEE_TriangleMesh * GEE_PolygonMesh::triangleMesh()
{
  GEE_TriangleMesh * tm = new GEE_TriangleMesh();

  tm->copyBasics(*this);

  // Move vertices:
  for(vertexIterator vit = m_vertices.begin(); 
      vit != m_vertices.end(); vit++) {
    int id = (*vit).first;
    GEE_PolygonVertex pv = (*vit).second;
    pv.m_polygons.clear();

    tm->setVertex(id, pv);
  }

  // Split polygons into triangles, the polygon should be a triangle fan

  for(polygonIterator pit = m_polygons.begin(); 
      pit != m_polygons.end(); pit++) {

    GEE_PolygonIndices * pind = & (*pit).second;

    int v0 = pind->m_indices[0];

    for(uint i = 2; i < pind->m_indices.size(); i++) {
      int v1 = pind->m_indices[i - 1];
      int v2 = pind->m_indices[i];

      GEE_TriangleIndices tri(v0, v1, v2);

      tm->addTriangle(tri);      
    }
  }

  VEE_trace
    ("GEE_PolygonMesh::triangleMesh # Converted %u polygons to %u triangles", 
     (uint) m_polygons.size(), (uint) tm->triangleCount());

  return tm;
}
