/* COPYRIGHT
 *
 * This file is part of the Visual Effects Engine - VEE
 *
 * Read the "VEE-LICENSE" file for the license.
 *
 * Authors & Copyright:   
 *
 * Tommi Ilmonen, Tuukka Heikura, Marko Myllymaa and 
 * Janne Kontkanen 2001-2004
 *
 * Additional copyrights: Tekes 2003-2004
 *
 * firstname.lastname@hut.fi
 *
 */

#include <gee_triangle_mesh.h>

#include <gee_containment.h>
#include <gee_io.h>
#include <gee_macros.h>
#include <gee_mover.h>
#include <gee_recolor.h>
#include <gee_renderer.h>

#include <vee_error.h>
#include <vee_math.h>

/** Adds a polygon to this vertex. If the polygon is already present,
    then nothing happens. */

void GEE_PolygonVertex::addPolygon(int p)
{ 
  std::vector<int>::iterator it = 
    std::find(m_polygons.begin(), m_polygons.end(), p);
  
  if(it == m_polygons.end())
    m_polygons.push_back(p);
}

/** Erase a polygon from the list of linked polygons. */

bool GEE_PolygonVertex::removePolygon(int i)
{ 
  std::vector<int>::iterator it = 
    std::find(m_polygons.begin(), m_polygons.end(), i);
  
  if(it != m_polygons.end()) {
    m_polygons.erase(it);
    return true;
  }
  return false;
}

static const int GEE_PolygonVertexId = GEE_CONST('P', 'G', 'V', 'X');

bool GEE_PolygonVertex::read(GEE_Io *io)
{
  if(io->readInt32() != GEE_PolygonVertexId)
    return false;

  int v = io->readInt32();

  m_location = io->readVector3();
  m_normal   = io->readVector3();
  m_uv       = io->readVector2();
  m_rgba     = io->readVector4();

  int polygons = io->readInt32();
  m_polygons.resize(polygons);
  io->read4v(& m_polygons[0], polygons);

  return v == 1;
}

bool GEE_PolygonVertex::write(GEE_Io *io)
{
  io->writeInt32(GEE_PolygonVertexId);
  io->writeInt32(1); // Version  

  io->writeVector3(m_location);
  io->writeVector3(m_normal);
  io->writeVector2(m_uv);
  io->writeVector4(m_rgba);

  io->writeInt32(m_polygons.size());
  io->write4v(& m_polygons[0], m_polygons.size());

  return true;
}

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

static const int GEE_TriangleIndicesId = GEE_CONST('T', 'R', 'I', 'N');

bool GEE_TriangleIndices::read(GEE_Io *io)
{
  if(io->readInt32() != GEE_TriangleIndicesId)
    return false;

  int v = io->readInt32();

  io->read4v(m_indices, 3);
  
  m_continues = io->readByte();
  m_stripIndex = io->readInt32();

  return v == 1;
}

bool GEE_TriangleIndices::write(GEE_Io *io)
{
  io->writeInt32(GEE_TriangleIndicesId);
  io->writeInt32(1); // Version  

  io->write4v(m_indices, 3);
  io->writeByte(m_continues);
  io->writeInt32(m_stripIndex);

  return true;
}

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

GEE_TriangleMesh::GEE_TriangleMesh()
  : m_vertexCounter(0),
    m_triangleCounter(0),
    m_useStrips(false)
{
  m_bbox.clear();
}

GEE_TriangleMesh::~GEE_TriangleMesh()
{}

GEE_Geometry * GEE_TriangleMesh::create()
{
  return new GEE_TriangleMesh();
}

GEE_Geometry * GEE_TriangleMesh::clone() const
{
  return new GEE_TriangleMesh(*this);
}

bool GEE_TriangleMesh::eraseData(GEE_Containment *co)
{
  bool hit = false;

  VEE_Matrix4d m;
  getWorldTransformation(&m);
  VEE_Matrix4 m2(m.data());

  if(!intersectsBounded(co, &m_bbox, m2))
    return false;

  for(vertexIterator it = m_vertices.begin(); it != m_vertices.end(); ) {
    if(co->contains(m2 * (*it).second.m_location)) {
      int id = (*it).first;
      removeVertex(id);
      it = m_vertices.lower_bound(id);
      hit = true;
    }
    else
      it++;
  }
  if(hit) {
    //autoNormalize();
    incrementGeneration();
    co->contained(this);
  }

  return hit;
}

bool GEE_TriangleMesh::intersects(GEE_Containment *co)
{
  VEE_Matrix4d m;
  getWorldTransformation(&m);
  VEE_Matrix4 m2(m.data());

  if(!intersectsBounded(co, &m_bbox, m2))
    return false;

  for(vertexIterator it = m_vertices.begin(); it != m_vertices.end(); it++)
    if(co->contains(m2 * (*it).second.m_location))
      return true;

  return false;
}

void GEE_TriangleMesh::selectVertices(GEE_Containment * co, 
				      uint mask, 
				      uint values)
{
  VEE_Matrix4d m;
  getWorldTransformation(&m);
  VEE_Matrix4 m2(m.data());

  for(vertexIterator it = m_vertices.begin(); it != m_vertices.end(); it++) {

    GEE_PolygonVertex & pv = (*it).second;

    if(co->contains(m2 * pv.m_location)) {
      uchar state = (pv.m_selectionState | ~mask) | 
	(pv.m_selectionState & mask & values);
      pv.m_selectionState = state;
    }
  }
}

void GEE_TriangleMesh::selectFaces(GEE_Containment *, 
				   uint mask, 
				   uint values)
{

}


/** Color the vertices. */

bool GEE_TriangleMesh::colorData(GEE_ReColor *rc)
{
  bool hit = false;
  
  VEE_Matrix4d m;
  getWorldTransformation(&m);
  VEE_Matrix4 m2(m.data());

  if(!intersectsBounded(rc, &m_bbox, m2)) {
    // VEE_trace("GEE_TriangleMesh::colorData # culled a recolor");
    return false;
  }

  for(vertexIterator it = m_vertices.begin(); it != m_vertices.end(); it++) {
    GEE_PolygonVertex & pv = (*it).second;

    VEE_Vector4 color = m_material.useVertexColors() ? 
      pv.m_rgba : m_material.ambient();
    
    bool change = rc->recolor(m2 * pv.m_location, & color);

    if(!change)
      continue;
    
    if(!hit) {
      m_materialName.resize(0);
      if(!m_material.useVertexColors()) {
	m_material.setUseVertexColors(true);
	// Put the ambient color to all the vertices.
	for(vertexIterator it2 = m_vertices.begin(); 
	    it2 != m_vertices.end(); it2++){
	  (*it2).second.m_rgba = m_material.ambient();
	}
	
	// m_material.setColor(VEE_Vector4(rc->basicColor().data()));

	VEE_trace("GEE_TriangleMesh::colorData # Switching to vertex colors %d",
		  (int) m_material.useVertexColors());
      }
      hit = true;
    }

    pv.m_rgba = color;
  }

  if(hit) {
    incrementGeneration();
    rc->recolored(this);
  }

  return hit;
}


bool GEE_TriangleMesh::moveData(GEE_Mover *m)
{
  bool hit = false;

  VEE_Matrix4d ma, mi;
  bool ok ;
  getWorldTransformations(&ma, & mi, & ok);

  if(!ok) {
    VEE_error("GEE_TriangleMesh::moveData # Bad inverse matrix");
    return false;
  }
  
  VEE_Matrix4 ma2(ma.data()), mi2(mi.data());

  if(!intersectsBounded(m, &m_bbox, ma2)) {
    // VEE_trace("GEE_TriangleMesh::moveData # culled a mover");
    return false;
  }

  for(vertexIterator it = m_vertices.begin(); it != m_vertices.end(); it++) {
    GEE_PolygonVertex & pd = (*it).second;
    VEE_Vector3 tmp = ma2 * pd.m_location;
    if(m->move(& tmp)) {
      hit = true;
      tmp = mi2 * tmp;
      pd.m_location = tmp;
      m_bbox.expand(tmp);
    }
  }

  if(hit) {
    incrementGeneration();
    autoNormalize();
    m->moved(this);
  }

  return hit;
}

bool GEE_TriangleMesh::smooth(GEE_Containment *co)
{
  bool hit = false;

  VEE_Matrix4d ma, mi;
  bool ok ;
  getWorldTransformations(&ma, & mi, & ok);

  VEE_Matrix4 ma2(ma.data()), mi2(mi.data());

  if(!intersectsBounded(co, &m_bbox, ma2)) {
    // VEE_trace("GEE_TriangleMesh::smooth # culled a mover");
    return false;
  }

  std::map<int, VEE_Vector3> moves;
  
  for(vertexIterator it = m_vertices.begin(); it != m_vertices.end(); it++) {

    GEE_PolygonVertex & pd = (*it).second;
    VEE_Vector3 tmp = ma2 * pd.m_location;

    if(co->contains(tmp)) {
      hit = true;

      int id = (*it).first;
      
      moves[id] = smoothVertex(id, 0.5f);
    }
  }

  if(hit) {
    incrementGeneration();
    autoNormalize();
    // m->moved(this);
  }

  return hit;

}

void GEE_TriangleMesh::render(GEE_Renderer *r)
{
  r->pushTransform(transformation());
  r->renderTriangleMesh(this);
  r->popTransform();
}

bool GEE_TriangleMesh::getBoundingBox(VEE_BoundingBox *b)
{
  *b = m_bbox;
  return true;
}

bool GEE_TriangleMesh::move(const VEE_Vector3 & offset)
{
  if(!m_vertices.size())
     return true;

  m_bbox.move(offset);

  for(vertexIterator it = beginVertex(); it != endVertex(); it++)
    (*it).second.m_location += offset;

  return true;
}

bool GEE_TriangleMesh::read(GEE_Io *io)
{
  clear();

  std::string id;
  io->readNullTerminatedStdString(id);

  if(id != "GEE_TriangleMesh") {
    VEE_error("GEE_TriangleMesh::read # Bad type (%s)", id.c_str());
    return false;
  }

  int v = io->readInt32();
  if(v > 2)
    return false;

  readBasics(io);

  // Get vertices

  m_vertexCounter = io->readInt32();
  int i, n = io->readInt32();

  for(i=0; i < n; i++) {
    int id = io->readInt32();
    GEE_PolygonVertex v;
    v.read(io);
    setVertex(id, v);
  }

  // Get triangles
  
  m_triangleCounter = io->readInt32();
  n = io->readInt32();
  for(i=0; i < n; i++) {
    int id = io->readInt32();
    GEE_TriangleIndices ind;
    ind.read(io);
    m_triangles[id] = ind;

    if(ind.m_continues && (v == 1))
      m_useStrips = true;
  }

  VEE_trace("GEE_TriangleMesh::read # EXIT (%d tris, %d vs)",
	    (int) m_triangles.size(), (int) m_vertices.size());

  if(v >= 2)
    m_useStrips = io->readByte();
  
  return true;
}

bool GEE_TriangleMesh::write(GEE_Io *io)
{
  io->writeNullTerminatedString("GEE_TriangleMesh");
  io->writeInt32(2); // version
  /* Version 2 introduced m_useStrips, which was missing. */

  writeBasics(io);

  io->writeInt32(m_vertexCounter);
  io->writeInt32(m_vertices.size());

  for(vertexIterator vit = m_vertices.begin(); vit != m_vertices.end(); vit++){
    io->writeInt32((*vit).first);
    (*vit).second.write(io);
  }

  io->writeInt32(m_triangleCounter);
  io->writeInt32(m_triangles.size());
  for(triangleIterator vit = m_triangles.begin(); 
      vit != m_triangles.end(); vit++){

    io->writeInt32((*vit).first);
    (*vit).second.write(io);
  }

  io->writeByte(m_useStrips);

  return true;
}

GEE_Geometry * GEE_TriangleMesh::removeData(const VEE_Vector3 &loc, 
					    float radius)
{
  if(!m_vertices.size())
    return 0;

  GEE_TriangleMesh *mesh = new GEE_TriangleMesh(*this);

  vertexIterator it = mesh->m_vertices.end();
  it--;

  int lowest = (*mesh->m_vertices.begin()).first;

  for(int i = (*it).first; i >= lowest; i--) {

    it = mesh->m_vertices.find(i);
    if(it == mesh->m_vertices.end())
      continue;

    const GEE_PolygonVertex & vertex = (*it).second;

    if((vertex.m_location - loc).length() < radius) {
      mesh->removeVertex(i);
    }
  }

  if(!mesh->m_vertices.size()) {
    delete mesh;
    return 0;
  }
  
  return mesh;
}

void GEE_TriangleMesh::clear()
{
  m_vertexCounter = 0;
  m_triangleCounter = 0;
  m_vertices.clear();
  m_triangles.clear();
}

/** Adds a new vertex point to the mesh and returns the index number
    of the vertex. */

int GEE_TriangleMesh::addVertex(const GEE_PolygonVertex &v)
{
  if(!m_vertices.size())
    m_bbox.set(v.m_location);
  else
    m_bbox.expand(v.m_location);

  int r = m_vertexCounter++;
  m_vertices[r] = v;
  return r;
}

void GEE_TriangleMesh::setVertex(int id, const GEE_PolygonVertex &v)
{
  if(!m_vertices.size())
    m_bbox.set(v.m_location);
  else
    m_bbox.expand(v.m_location);

  m_vertices[id] = v;
}


/** Removes a vertex. Returns true if the operation was
    successful. Automatically erases also the triangles that did
    include this vertex. 

    @author Tommi Ilmonen
*/
bool GEE_TriangleMesh::removeVertex(int id)
{
  vertexIterator it = m_vertices.find(id);

  if(it == m_vertices.end())
    return false;
  
  // Remove the triangles that included this vertex. 
  for(uint i=0; i < (*it).second.m_polygons.size(); i++) {

    int triId = (*it).second.m_polygons[i];
    triangleIterator trit = m_triangles.find(triId);
    
    if(trit == m_triangles.end()) {
      VEE_error("GEE_TriangleMesh::removeVertex # Memory corrupted");
      return false;
    }
    
    // Remove this triangle from the other polygons that it points to;
    for(uint j=0; j < 3; j++) {

      int vertexId = (*trit).second.m_indices[j];
      if(vertexId == id) // Do not break the link to this particular polygon.
	continue;

      vertexIterator vit = m_vertices.find(vertexId);

      if(vit == m_vertices.end())
	continue;

      GEE_PolygonVertex * v = & (*vit).second;
      v->removePolygon(triId);

      // If the vertex is not needed any more delete it:
      if(!v->m_polygons.size())
	m_vertices.erase(vit);
    }

    if(m_useStrips && trit != m_triangles.begin()) {
      triangleIterator trit2 = trit;
      trit2--;
      (*trit2).second.m_continues = false;
    }

    m_triangles.erase(trit);
  }

  m_vertices.erase(it);
  return true;
}

/** Adds a triangle and returns its index. If the vertices of the
    triangle do not exist, then -1 returned and the triangle is not
    added. */

int GEE_TriangleMesh::addTriangle(const GEE_TriangleIndices &tri, int id)
{
  int i, r;

  if(id >= 0)
    r = id;
  else
    r = m_triangleCounter++;

  GEE_PolygonVertex * pv[3];

  // Check that vertices exist:
  for(i=0; i < 3; i++) {
    if(!(pv[i] = getVertex(tri.m_indices[i]))) {
      VEE_error("GEE_TriangleMesh::addTriangle # Missing vertex data %d",
		tri.m_indices[i]);
      return -1;
    }
  }

  if(tri.m_continues)
    m_useStrips = true;
  
  // Vertices do exist, now link:

  for(i=0; i < 3; i++)
    pv[i]->addPolygon(r);

  m_triangles[r] = tri;
    
  return r;
}

/** Gets a triangle with some id. Returns NULL if the triangle is not
    found. */

GEE_TriangleIndices * GEE_TriangleMesh::getTriangle(int id)
{
  triangleIterator it = m_triangles.find(id);

  if(it != m_triangles.end())
    return & (*it).second;

  return 0;
}

/** Recalculate the normals at each vertex (assuming that the model is
    smooth). */

void GEE_TriangleMesh::autoNormalize()
{
  for(vertexIterator it = m_vertices.begin(); it != m_vertices.end(); it++)
    normalizeVertex(it, false);

  incrementGeneration();
}

/** Calculates the vertex normal based on the triangles that are
    connected to the vertex. */

void GEE_TriangleMesh::normalizeVertex(vertexIterator &it, 
				       bool adjustGeneration)
{
  VEE_Vector3 normal;
  normal.clear();
  
  GEE_PolygonVertex & vertex = (*it).second;
  int id = (*it).first;
  
  // We calculate the normal of each separate triangle and sum them up.
  
  for(uint j = 0; j < vertex.m_polygons.size(); j++) {
    normal += triangleNormal(vertex.m_polygons[j], id);
  }
  
  normal.normalize();
  vertex.m_normal = normal;

  /* VEE_trace("GEE_TriangleMesh::normalizeVertex # %d %f %f %f", (*it).first,
     normal.x, normal.y, normal.z); */
  
  if(adjustGeneration) 
    incrementGeneration();
}

void GEE_TriangleMesh::createRectangularGrid(uint xdivision, uint ydivision,
					     float width, float height)
{
  clear();

  setUseStrips(true);

  uint i, j;

  uint xn = xdivision + 1;
  uint yn = ydivision + 1;

  float xdf = xdivision;
  float ydf = ydivision;

  float xs = width  / xdf;
  float ys = height / ydf;

  /** First create the vertices: */

  GEE_PolygonVertex pv;
  pv.m_normal.make(0, 0, -1);
  pv.m_rgba.make(1, 1, 1, 1);

  for(i = 0; i < xn; i++) {
    for(j = 0; j < yn; j++) {

      int index = i * yn + j;
      pv.m_uv.make(i / xdf, j / ydf);
      pv.m_location.make(i * xs, j * ys, 0.0);

      setVertex(index, pv);
    }
  }

  /** Now create triangle strips: */

  GEE_TriangleIndices tri;
  
  for(i = 0; i < xdivision; i++) {

    tri.m_continues = true;

    for(j = 0; j < ydivision; j++) {
      int index1 = i * yn + j;
      int index2 = index1 + 1;
      int index3 = (i + 1) * yn + j;
      int index4 = index3 + 1;

      tri.make(index1, index2, index3);
      addTriangle(tri);

      if(j + 1 < ydivision)
	tri.m_continues = false;

      tri.make(index3, index2, index4);
      addTriangle(tri);
    }
  }  
}

/** Returns the unnormalized triangle normal (calculated from the
    vertex locations). */

VEE_Vector3 GEE_TriangleMesh::triangleNormal(int triId, int vertexId)
{
  const GEE_TriangleIndices * tri = getTriangle(triId);
  
  if(!tri) {
    VEE_error("GEE_TriangleMesh::triangleNormal # Memory error");
    return VEE_Vector3ZERO;
  }
    
  // Now we need to know which corner is this vertex corner.
  
  int offset = 0;
  if(tri->m_indices[1] == vertexId) offset = 1;
  if(tri->m_indices[2] == vertexId) offset = 2;
  
  int w1 = (offset - 1 +  3) % 3;
  int w2 = offset;
  int w3 = (offset + 1) % 3;
  
  VEE_Vector3 v1 = getVertex(tri->m_indices[w1])->m_location;
  VEE_Vector3 v2 = getVertex(tri->m_indices[w2])->m_location;
  VEE_Vector3 v3 = getVertex(tri->m_indices[w3])->m_location;
  
  VEE_Vector3 d1 = v2 - v1;
  VEE_Vector3 d2 = v2 - v3;
  
  d1.normalize();
  d2.normalize();
  
  // Calculate normal (and implicitly weight)
  return cross(d2, d1);
}

float GEE_TriangleMesh::vertexWrinkliness(int vertex)
{
  GEE_PolygonVertex * v = getVertex(vertex);

  if(!v) {
    VEE_error("GEE_TriangleMesh::vertexWrinkliness # Bad argument %d", vertex);
    return 0.0f;
  }

  float wrinkles = 0.0f;

  // First we sum the wrinkliness caused by inconsistent normals.

  const int maxNormals = 100;
  
  int i, polyCount = v->m_polygons.size();
  VEE_Vector3 tmp(0, 0, 0), tmp2;
  VEE_Vector3 normals[maxNormals];
  
  for(i = 0; i <  polyCount; i++) {
    tmp2 = triangleNormal(v->m_polygons[i], vertex);
    normals[i] = tmp2;
    tmp += tmp2;
  }

  tmp.normalize();

  float wsum = 0.0f;
  
  for(i = 0; i <  polyCount; i++) {
    tmp2 = normals[i];
    float len = tmp2.length();

    if(len < 1.0e-6)
      continue;

    tmp2 /= len;
    wrinkles += (tmp2 - tmp).length() * len;
    wsum += len;
  }

  // Normalize
  if(wsum > 1.0e-10)
    wrinkles /= wsum;

  /* Now we check if the vertex is "open" and calculate the
     wrinkliness of its edge angle. */

#if 0
  
  int openTri = -1, openVert = -1;

  GEE_TriangleIndices tris[maxNormals];

  // Collect the triangle indices to a local storage for quick access.
  for(i = 0; i <  polyCount; i++)
    tris[i] = * getTriangle(v->m_polygons[i]);
  
  for(i = 0; i <  polyCount; i++) {
    const GEE_TriangleIndices & tri = tris[i];
    
    for(uint j = 0; j < 3; j++) {
      int vid = tri.m_indices[j];
      
      if(vid == vertex)
	continue;
      
      /* Scan all other triangles to see if another triangle hits this
	 same position: */
      
      bool shared = false;

      for(int k = 0; k <  polyCount && ! shared; k++) {
	if(k == i)
	  continue;
	
	const GEE_TriangleIndices & tri2 = tris[k];
	
	for(uint l = 0; l < 3; l++)
	  if(tri2.m_indices[l] == vid)
	    shared = true;
      }
     
      if(!shared) {
	if(openTri < 0) {
	  openTri = i;
	  openVert = j;
	}
	else {
	  // This is the second open edge:
	  const GEE_PolygonVertex * v1 = 
	    getVertex(tris[openTri].m_indices[openVert]);
	  const GEE_PolygonVertex * v2 = 
	    getVertex(tris[i].m_indices[j]);
	  
	  VEE_Vector3 e1 = v1->m_location - v->m_location;
	  VEE_Vector3 e2 = v->m_location  - v2->m_location;
	  
	  e1.normalize();
	  e2.normalize();

	  float c = dot(e1, e2);
	  c = 1.0f - c;
	  return wrinkles + c * c;
	}	
      }
      
    }
    
  }

#endif

  // VEE_trace("GEE_TriangleMesh::vertexWrinkliness # %d %f", vertex, wrinkles);

  return wrinkles;
}

/** Returns the wrinkliness of this vertex plus the wrinkliness of all
    connected vertices. */
float GEE_TriangleMesh::vertexWrinklinessLinked(int vertex)
{
  GEE_PolygonVertex * v = getVertex(vertex);

  if(!v) {
    VEE_error("GEE_TriangleMesh::vertexWrinklinessLinked # Bad argument %d", 
	      vertex);
    return 0.0f;
  }

  float wrinkles = vertexWrinkliness(vertex);

  int i, polyCount = v->m_polygons.size();
  
  const int maxVertices = 200;
  int hitVertices[maxVertices];
  int foundVertices = 0;

  for(i = 0; i <  polyCount; i++) {

    const GEE_TriangleIndices & tri = * getTriangle(v->m_polygons[i]);

    for(uint j = 0; j < 3; j++) {
      int id = tri.m_indices[j];

      if(id == vertex)
	continue;

      int * begin = hitVertices;
      int * sentinel = begin + foundVertices;

      if(std::find(begin, sentinel, id) != sentinel)
	continue;

      wrinkles += vertexWrinkliness(id);

      hitVertices[foundVertices++] = id;
    }
  }

  return wrinkles;
}

/** The vertex smoother works by using the gradient method to pull the
    vertex towards optimal locations. The vertex is in fact not moved,
    but a new proposal for a location vector is returned. */

VEE_Vector3 GEE_TriangleMesh::smoothVertex(int vertex, float effect)
{
  GEE_PolygonVertex * v = getVertex(vertex);
  
  if(!v) {
    VEE_error("GEE_TriangleMesh::smoothVertex # Bad argument %d", vertex);
    return VEE_Vector3ZERO;
  }
  
  VEE_Vector3 originalLoc = v->m_location;

  const float epsilon = 1.0e-2;
  // const float e2 = epsilon * epsilon;

  float w0 = vertexWrinklinessLinked(vertex);

  VEE_Vector3 wspan[2];

  for(int j = 0; j < 2; j ++) {

    float move = (j * 2 - 1) * epsilon;

    for(uint i = 0; i < 3; i++) {
    
      v->m_location[i] += move;

      wspan[j][i] = vertexWrinklinessLinked(vertex);
      
      // Return the vertex to original location:
      v->m_location[i] = originalLoc[i];
    }
  }

  VEE_Vector3 derivative1 = (VEE_Vector3(w0) - wspan[0]) / epsilon;
  VEE_Vector3 derivative2 = (wspan[1] - w0) / epsilon;
  VEE_Vector3 derivMean = (derivative1 + derivative2) * 0.5f;

  // Approximation of the second order derivative:
  VEE_Vector3 derivSecond = (derivative2 - derivative1) / epsilon;

  VEE_Vector3 zeroDerivLoc;

  for(uint i=0; i < 3; i++) {
    if(VEE_abs(derivSecond[i]) > 1.0e-5)
      zeroDerivLoc[i] = derivMean[i] / derivSecond[i];
    else
      zeroDerivLoc[i] = - derivMean[i];
  }

  VEE_trace("GEE_TriangleMesh::smoothVertex # %f %f %f # \n"
	    "  %f %f %f # %f %f %f ->  %f %f %f (%f)",
	    originalLoc.x, originalLoc.y, originalLoc.z, 
	    derivMean.x, derivMean.y, derivMean.z, 
	    derivSecond.x, derivSecond.y, derivSecond.z, 
	    zeroDerivLoc.x, zeroDerivLoc.y, zeroDerivLoc.z, w0);

  return originalLoc + zeroDerivLoc * effect;
}

/** Moves the vertex in the direction of the normal and returns the
    best place on that axis. */

VEE_Vector3 GEE_TriangleMesh::smoothVertex2(int vertex, 
					    int iterations,
					    float effect)
{
  GEE_PolygonVertex * v = getVertex(vertex);
  
  if(!v) {
    VEE_error("GEE_TriangleMesh::smoothVertex2 # Bad argument %d", vertex);
    return VEE_Vector3ZERO;
  }
  
  normalizeVertex(vertex);

  VEE_Vector3 location = v->m_location;
  VEE_Vector3 normal   = v->m_normal;

  float offset = 0.01;
  
  VEE_Vector3 offsets(-1, 0, 1), wrinkles;

  wrinkles[1] = vertexWrinklinessLinked(vertex);

  v->m_location = location + offset * normal;

  wrinkles[2] = vertexWrinklinessLinked(vertex);

  v->m_location = location - offset * normal;
  
  wrinkles[0] = vertexWrinklinessLinked(vertex);

  for(int i = 0; i < iterations; i++) {
    if(wrinkles[0] > wrinkles[1]) {
      if(wrinkles[2] < wrinkles[1]) {
	// Now we know the center piece is the minumum.

	int edge, insert;
	if(wrinkles[2] > wrinkles[0]) {
	  edge = 0;
	  insert = 0;
	}
	else {
	  edge = 2;
	  insert = 1;
	}

	float newOffset = (offsets[insert] + offsets[insert + 1]) * 0.5f;

	v->m_location = location + newOffset * normal;

	float newWrinkles = vertexWrinklinessLinked(vertex);
	
	if(newWrinkles < wrinkles[1]) {
	  
	}
	else
	  ;
      }
    }
       
    
  }
  
  return location + offset * normal;
}

/** Updates the display list if necessary.
    
    @return True if the list is now up-to-date. False of the list
    could not be generated.

*/

bool GEE_TriangleMesh::genList()
{
  if(editorCount() || m_generation < 0)
    return false;

  if(m_list.m_generation == m_generation)
    return true;

  if(useStrips()) {
    
    m_list.prepareNew();
    
    bool startStrip = true;
    
    GEE_TriMeshListItem item;
    
    m_list.addItem2(GEE_TRIMESH_TRILIST);
    
    for(triangleIterator it = beginTriangle(); it != endTriangle(); it++) {
      
      GEE_TriangleIndices & tri = getTriangle(it);
      
      if(startStrip) {
	for(uint i = 0; i < 3; i++) {
	  GEE_PolygonVertex * v = getVertex(tri.m_indices[i]);
	  listifyVertex(*v);
	}
	startStrip = false;
      }
      else {
	GEE_PolygonVertex * v = getVertex(tri.m_indices[2]);
	listifyVertex(*v);
      }
      
      if(!tri.m_continues) {
	m_list.addItem2(GEE_TRIMESH_END);
	m_list.addItem2(GEE_TRIMESH_TRILIST);
	startStrip = true;
      }      
    }

    m_list.addItem2(GEE_TRIMESH_END);
  }
  else {

    m_list.prepareNew();

    m_list.addItem2(GEE_TRIMESH_TRIS);    
    
    for(triangleIterator it = beginTriangle();  it != endTriangle(); it++) {
      
      GEE_TriangleIndices & tri = getTriangle(it);
      
      for(uint i=0; i < 3; i++) {
	GEE_PolygonVertex * v = getVertex(tri.m_indices[i]);
	listifyVertex(*v);
      }
    }

    m_list.addItem2(GEE_TRIMESH_END);
  }
  
  m_list.m_generation = m_generation;

  return true;
}

void GEE_TriangleMesh::listifyVertex(const GEE_PolygonVertex &v)
{
  GEE_TriMeshListItem item;

  item.m_type = GEE_TRIMESH_COLOR;
  item.m_data = v.m_rgba;
  m_list.addItem(item);
  
  item.m_type = GEE_TRIMESH_NORMAL;
  item.m_data.vector3() = v.m_normal;
  m_list.addItem(item);
  
  item.m_type = GEE_TRIMESH_UV;
  item.m_data.vector2() = v.m_uv;
  m_list.addItem(item);
  
  item.m_type = GEE_TRIMESH_LOCATION;
  item.m_data.vector3() = v.m_location;
  m_list.addItem(item);
}
