/* -*- C++ -*- */

/* COPYRIGHT
 *
 * This file is part of the Visual Effects Engine - VEE
 *
 * Read the "VEE-LICENSE" file for the license.
 *
 * Authors & Copyright:   
 *
 * Tommi Ilmonen, Tuukka Heikura, Marko Myllymaa and 
 * Janne Kontkanen 2001-2004
 *
 * Additional copyrights: Tekes 2003-2004
 *
 * firstname.lastname@hut.fi
 *
 */

#ifndef GEE_GL_MATERIAL_HPP
#define GEE_GL_MATERIAL_HPP

#include <gee_material.h>

#ifdef __APPLE__
# include <OpenGL/gl.h>
#else
# include <GL/gl.h>
#endif

#include <vee_color.h>

#include <map>
#include <string>

class GEE_GlRenderer;


/** Material definition for OpenGL. Includes color, texture and
    shine values. */
class GEE_GlMaterial
{
public:

  GEE_GlMaterial(float shine, 
		 const VEE_Rgba & color,
		 GLuint textureId = (GLuint) -1,
		 int flags = GEE_MAT_LIGHTING | GEE_MAT_SMOOTH | GEE_MAT_TWO_SIDED);
  GEE_GlMaterial();

  virtual ~GEE_GlMaterial();

  void set(float shine, const VEE_Rgba & color, GLuint textureId = (GLuint) -1)
    { m_textureId = textureId; m_shine = shine;  setColor(color); }

  void setShine(float s) { m_shine = s; }
  void setFlags(int f) { m_flags = f; }
  void setTextureId(GLuint i) { m_textureId = i; }
  /// Sets all colors at once
  
  void setColor(const VEE_Rgba & c) { m_ambient = m_diffuse = m_specular = c; }
  void setAmbient(const VEE_Rgba & color)  { m_ambient = color; }
  void setDiffuse(const VEE_Rgba & color)  { m_diffuse = color; }
  void setSpecular(const VEE_Rgba & color) { m_specular = color; }
  
  void applyGl(GEE_GlRenderer *) const;

  GLuint   textureId()    const { return m_textureId; }

  const VEE_Rgba & ambient()  const { return m_ambient; }
  const VEE_Rgba & diffuse()  const { return m_diffuse; }
  const VEE_Rgba & specular() const { return m_specular; }

  float uvScale() const { return m_uvScale; }
  void  setUvScale(float uvScale) { m_uvScale = uvScale; }

  bool useVertexColors() const { return m_flags & GEE_MAT_VERTEX_COLORS; }
  void setUseVertexColors(bool useColors)
  {
    if(useColors) m_flags |= GEE_MAT_VERTEX_COLORS;
    else m_flags &= ~(GEE_MAT_VERTEX_COLORS);
  }

protected:

  GLuint   m_textureId;
  float    m_shine;
  VEE_Rgba     m_ambient, m_diffuse, m_specular;
  float    m_uvScale;

  int      m_flags;
};

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

/** Collection of materials.  */

class GEE_GlMaterialBank
{
public:
  GEE_GlMaterialBank();
  virtual ~GEE_GlMaterialBank();

  void addMaterial(const std::string &, const GEE_GlMaterial &);

  const GEE_GlMaterial & getMaterial(const std::string &);
  
  void addUsualColors();

  void forceMaterial(const std::string & mat) { m_forceMaterial = mat; }
  void unforceMaterial() { m_forceMaterial.clear(); }

  static GEE_GlMaterial & defaultMaterial() { return m_default; }

protected:

  static GEE_GlMaterial m_default;

  std::map<std::string, GEE_GlMaterial> m_materials;

  std::string m_forceMaterial;
};


#endif
