/* COPYRIGHT
 *
 * This file is part of the Visual Effects Engine - VEE
 *
 * Read the "VEE-LICENSE" file for the license.
 *
 * Authors & Copyright:   
 *
 * Tommi Ilmonen, Tuukka Heikura, Marko Myllymaa and 
 * Janne Kontkanen 2001-2004
 *
 * Additional copyrights: Tekes 2003-2004
 *
 * firstname.lastname@hut.fi
 *
 */

#include <gee_gl_renderer.h>

#include <gee_debug.h>
#include <gee_display_lists.h>
#include <gee_gl_material.h>
#include <gee_group.h>
#include <gee_lines.h>
#include <gee_macros.h>
#include <gee_material.h>
#include <gee_path.h>
#include <gee_particle_object.h>
#include <gee_polygon_mesh.h>
#include <gee_profile.h>
#include <gee_sphere.h>
#include <gee_triangle_mesh.h>

#include <vee_error.h>
#include <vee_gl_renderer.h>
#include <vee_gl_texture.h>
#include <vee_object.h>

#include <di_mutex.h>

#include <ConfigReader.hpp>

#ifdef __APPLE__
# include <OpenGL/glu.h>
# include <OpenGL/glext.h>
#else
# include <GL/glu.h>
#endif

#ifdef __OS_LINUX__
	// Fix the broken NVidia headers:
	#define GL_GLEXT_PROTOTYPES 1
	#define GLAPI

	#include <GL/glext.h>
#endif // __OS_LINUX__

bool GEE_GlRenderer::m_debug = false;

GEE_GlRenderer::GEE_GlRenderer()
  : m_displayLists(0),
    m_highlightDisplayLists(0),
    m_materials(0),
    m_currentMaterial(0),
    m_currentLight(0),
    m_veeRenderer(0),
    m_textures(0),
    m_currentTexture((GLuint) -1),
    m_ignoreTransforms(false),
    m_showGroups(true),
    m_renderCount(0),
    m_renderDisplayListCount(0),
    m_generateDisplayListCount(0),
    m_selectionRender(RENDER_USUAL),
    m_quadric(gluNewQuadric())
{
  GEE_GlLight defaultLight;
  GEE_GlLights defls;
  defls.set(GL_LIGHT0, defaultLight);
  
  addLights(defls);

  resetCounters();

  m_debug = GEE_Debug::debug("GEE_GlRenderer");

  m_worldMatrix.identity();
}

GEE_GlRenderer::~GEE_GlRenderer()
{
  gluDeleteQuadric(m_quadric);
  // if(m_debug)
  VEE_trace("GEE_GlRenderer::~GEE_GlRenderer # "
	    "Rendered %ld things (%ld lists)",
	    m_renderCount, m_renderDisplayListCount);
}

void GEE_GlRenderer::renderGroupStructure(GEE_Group *group)
{
  if(!m_showGroups)
    return;

  m_geometryCount++;

  // uint ss = group->selectionState();
  
  VEE_BoundingBox box;

  if(!group->getBoundingBox(&box))
    return;

  if(m_inVee)
    exitVee();

  VEE_Vector3 groupTop = box.topCenter();

  applyGlLighting(false);
  glLineWidth(2);
  
  glColor4f(0, 1, 0, 1);
  glBegin(GL_LINES);

  for(GEE_Group::iterator it = group->begin(); it != group->end(); it++) {

    VEE_RefPtr<GEE_Geometry> * geo = group->getGeometry(it);

    if(!geo)
      continue;

    GEE_Geometry * g = geo->ptr();

    g->getBoundingBox(&box);
    box = VEE_Matrix4(g->transformation().data()) * box;

    glVertex3fv(groupTop.data());
    glVertex3fv(box.center().data());
  }

  glEnd();
}

void GEE_GlRenderer::renderLines3D(GEE_Lines3D *lines)
{
  m_geometryCount++;
  m_vertexCount += lines->vertexCount();

  uint ss = lines->selectionState();

  if(ss == 0) {
    renderLines3DInternal(lines);
  }
  else if(ss | GEE_NEARLY_SELECT_BIT) {
    m_selectionRender = RENDER_EDGES_NS_HIGH1;
    renderLines3DInternal(lines);
    m_selectionRender = RENDER_POINTS_NS_HIGH2;
    renderLines3DInternal(lines);
  }
  else {
    m_selectionRender = RENDER_EDGES_IS_HIGH1;
    renderLines3DInternal(lines);
    m_selectionRender = RENDER_POINTS_IS_HIGH2;
    renderLines3DInternal(lines);
  }

  m_selectionRender = RENDER_USUAL;
}

/** This method renders a path with two alternating colors. You cannot
    adjust the colors (except in code) - sorry. */

void GEE_GlRenderer::renderPath3D(GEE_Path3D *path)
{
  if(m_debug)
    VEE_trace("GEE_GlRenderer::renderPath3D # %p %u", 
	      path, path->vertexCount());

  m_geometryCount++;
  m_vertexCount += path->vertexCount();

  preRender(path);

  applyGlSmooth(false);
  applyGlLighting(false);

  ListMode m = renderDisplayList(path);

  if(m == RENDERED_LIST) {
    return;
  }

  glBegin(GL_LINE_STRIP);

  bool flip = true;

  GLubyte color1[4] = { 0, 255, 0, 255 };
  GLubyte color2[4] = { 0, 0, 255, 255 };

  for(GEE_Path3D::iterator it = path->begin(); it != path->end(); it++) {
    const GEE_PathVertex3D &v = GEE_Path3D::getVertex(it);
    glVertex3fv(v.m_location.data());

    flip = !flip;

    glColor4ubv(flip ? color1 : color2);
  }

  glEnd();

  if(m == GENERATE_LIST)
    glEndList();
}

void GEE_GlRenderer::renderParticleObject(GEE_ParticleObject * po)
{
  m_geometryCount++;

  // VEE_trace("GEE_GlRenderer::renderParticleObject");

  /* 
  for(GEE_ParticleObject::iterator it = po->begin();
      it != po->end(); it++) {
    VEE_Object * obj = GEE_ParticleObject::getParticleData(it).m_particle.ptr();
    if(obj)
      renderVeeObject(obj);
  }
  */
  uint ss = po->selectionState();
  
  if(ss == 0)
    return;

  else if(ss && m_veeRenderer) {
        
    if(ss | GEE_NEARLY_SELECT_BIT)
      m_selectionRender = RENDER_EDGES_NS_HIGH1;
    else 
      m_selectionRender = RENDER_EDGES_IS_HIGH2;

    preRender(po);
    
    glLineWidth(1);

    m_veeRenderer->setRenderOnlyOutline(true);
    glColor4f(1, 1, 1, 1);

    for(GEE_ParticleObject::iterator it = po->begin();
	it != po->end(); it++) {
      VEE_Object * obj = 
	GEE_ParticleObject::getParticleData(it).m_particle.ptr();
      if(obj)
	 renderVeeObject(obj);
    }

    if(m_inVee)
      exitVee();

    m_veeRenderer->setRenderOnlyOutline(false);

    m_selectionRender = RENDER_USUAL;
  }

  // VEE_trace("GEE_GlRenderer::renderParticleObject # EXIT");
}

void GEE_GlRenderer::renderPolygonMesh(GEE_PolygonMesh *mesh)
{
  if(m_debug)
    VEE_trace("GEE_GlRenderer::renderPolygonMesh # %p vs = %u polys = %u", 
	      mesh, mesh->vertexCount(), mesh->polygonCount());

  m_geometryCount++;
  m_vertexCount  += mesh->vertexCount();  
  m_polygonCount += mesh->polygonCount();

  preRender(mesh);
  ListMode m = renderDisplayList(mesh);

  if(m == RENDERED_LIST)
    return;

  bool useVertexColors = m_currentMaterial ? 
    m_currentMaterial->useVertexColors() : false;

  for(GEE_PolygonMesh::polygonIterator it = mesh->beginPolygon(); 
      it != mesh->endPolygon(); it++) {
    
    GEE_PolygonIndices & poly = GEE_PolygonMesh::getPolygon(it);
    
    glBegin(GL_POLYGON);

    for(uint i=0; i < poly.m_indices.size(); i++) {

      GEE_PolygonVertex * v = mesh->getVertex(poly.m_indices[i]);

      if(useVertexColors)
	glColor4fv(v->m_rgba.data());

      glTexCoord2f(v->m_uv.x, v->m_uv.y);
      glNormal3fv(v->m_normal.data());
      glVertex3fv(v->m_location.data());
    }
    glEnd();
  }

  if(m == GENERATE_LIST)
    glEndList();
}

void GEE_GlRenderer::renderProfile(GEE_Profile *profile)
{
  m_geometryCount++;
  m_vertexCount += profile->size();
  

  preRender(profile);
  ListMode m = renderDisplayList(profile);

  if(m == RENDERED_LIST)
    return;

  glBegin(GL_LINE_STRIP);

  bool continues = false;

  for(GEE_Profile::iterator it = profile->begin();
      it != profile->end(); it++) {
    const GEE_ProfileVertex2D & v = (*it);

    glVertex2fv(v.m_location.data());

    if(!v.m_continues) {
      // A hole in the profile
      glEnd();
      glBegin(GL_LINE_STRIP);
      continues = false;
    }
    else
      continues = true;
  }

  if(continues)
    glVertex2fv((*profile->begin()).m_location.data());

  glEnd();

  if(m == GENERATE_LIST)
    glEndList();
}

void GEE_GlRenderer::renderSphere(GEE_Sphere *sphere)
{
  m_geometryCount++;  

  preRender(sphere);

  glLineWidth(2);
  gluQuadricDrawStyle(m_quadric, GLU_LINE);
  glPushMatrix();
  glRotatef(90, 1, 0, 0);
  gluSphere(m_quadric, sphere->radius(), 6, 6);
  glPopMatrix();
}

void GEE_GlRenderer::renderTriangleMesh(GEE_TriangleMesh *mesh)
{
  m_geometryCount++;
  m_vertexCount   += mesh->vertexCount();  
  m_triangleCount += mesh->triangleCount();

  uint ss = mesh->selectionState();
  
  if(ss == 0) {
    renderTriangleMeshInternal(mesh);
  }
  else if(ss | GEE_NEARLY_SELECT_BIT) {
    m_selectionRender = RENDER_FACES_NS_HIGH1;
    renderTriangleMeshInternal(mesh);
    m_selectionRender = RENDER_EDGES_NS_HIGH2;
    renderTriangleMeshInternal(mesh);
    glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
  }
  else {
    m_selectionRender = RENDER_FACES_IS_HIGH1;
    renderTriangleMeshInternal(mesh);
    m_selectionRender = RENDER_EDGES_IS_HIGH2;
    renderTriangleMeshInternal(mesh);
    glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
  }

  m_selectionRender = RENDER_USUAL;
}

void GEE_GlRenderer::renderVeeObject(VEE_Object *o, 
				     const VEE_Matrix4 * m,
				     float scale)
{
  m_particleCount++;
  
  // VEE_Vector3 l = o->location();
  // VEE_trace("GEE_GlRenderer::renderVeeObject # %f %f %f", l.x, l.y, l.z);

  if(m_veeRenderer) {
    if(!m_inVee)
      enterVee();

    if(m)
      m_veeRenderer->setMatrix(*m);
    
    if(scale != 0.0f)
      m_veeRenderer->setParticleSize(scale);
    else
      m_veeRenderer->setParticleSize(1.0f);

    o->render(m_veeRenderer);
  }
  else
    VEE_error("GEE_GlRenderer::renderVeeObject # VEE renderer is NULL");
}

void GEE_GlRenderer::pushTransform(const VEE_Matrix4d &m)
{
  // VEE_trace("GEE_GlRenderer::pushTransform");

  if(m_ignoreTransforms)
    return;

  if(m_inVee)
    exitVee();

  glPushMatrix();

  VEE_Matrix4d tmp(m);
  tmp.transpose();

  glMultMatrixd(tmp.data());

  GLenum e = glGetError();
  if(e) 
    VEE_error("GEE_GlRenderer::pushTransform # GL ERROR 1 %s", 
	      gluErrorString(e));

  /* const VEE_Matrix4 & prev = m_stack[m_stackIndex];

  if(++m_stackIndex >= STACK_MAXIMUM) {
    VEE_error("GEE_GlRenderer::pushTransform # Stack overflow");
    m_stackIndex = STACK_MAXIMUM - 1;
    return;
  }
  VEE_Matrix4 tmp2(m.data());
  m_stack[m_stackIndex] = prev * tmp2; */
}


void GEE_GlRenderer::pushTransform(const VEE_Matrix4 &m)
{
  // VEE_trace("GEE_GlRenderer::pushTransform");

  if(m_ignoreTransforms)
    return;

  if(m_inVee)
    exitVee();

  glPushMatrix();

  VEE_Matrix4 tmp(m);
  tmp.transpose();

  glMultMatrixf(tmp.data());

  GLenum e = glGetError();
  if(e) 
    VEE_error("GEE_GlRenderer::pushTransform # GL ERROR 1 %s", 
	      gluErrorString(e));

  /* const VEE_Matrix4 & prev = m_stack[m_stackIndex];

  if(++m_stackIndex >= STACK_MAXIMUM) {
    VEE_error("GEE_GlRenderer::pushTransform # Stack overflow");
    m_stackIndex = STACK_MAXIMUM - 1;
    return;
  }
  m_stack[m_stackIndex] = prev * m; */

}

void GEE_GlRenderer::popTransform()
{
  if(m_ignoreTransforms)
    return;
  
  if(m_inVee)
    exitVee();

  glPopMatrix();

  /* if(!m_stackIndex) {
    VEE_error("GEE_GlRenderer::popTransform # Stack underflow");
    // m_stackIndex = 0;
    return;
  }

  m_stackIndex--; */
}

void GEE_GlRenderer::enableTransforms(bool enable)
{
  m_ignoreTransforms = !enable;
}

bool GEE_GlRenderer::enableTransforms()
{
  return ! m_ignoreTransforms;
}

void GEE_GlRenderer::addLights(const GEE_GlLights & lights)
{
  m_lights.push_back(lights);
  m_currentLight = & m_lights[0];
}

void GEE_GlRenderer::loadLights(const char * nameBase, int maximum)
{
  GEE_FNAME = "GEE_GlRenderer::loadLights";

  char buf[128];

  for(int i = 0; i < maximum; i++) {
    sprintf(buf, nameBase, i);
    
    ConfigReader::Config config2;
    
    ConfigReader::readConfig(&config2, buf);
    
    GEE_GlLights geeLights;
    geeLights.set(&config2);
    
    VEE_trace("%s # Got %d lights (%d) from \"%s\"", 
	      fname, (int) geeLights.size(), i, buf);
    
    if(geeLights.size()) {
      if(i == 0)
	clearLights();
      addLights(geeLights);
    }
    else
      break;
  }
}

void GEE_GlRenderer::clearLights()
{
  m_lights.clear();
  m_currentLight = 0;
}

/** Initializes the OpenGL state machine for rendering the
    objects. This method needs to be called if the OpenGL state is
    wrong for the rendering. */
void GEE_GlRenderer::initGlState()
{
  glPushAttrib(~0);
  GLenum e = glGetError();
  if(e) 
    VEE_error("GEE_GlRenderer::initGlState # GL ERROR AT BEGIN %s", 
	  gluErrorString(e));

  glDisable(GL_BLEND);
  glDisable(GL_TEXTURE_2D);
  glEnable(GL_LIGHTING);
  glEnable(GL_DEPTH_TEST);
  glDepthFunc(GL_LEQUAL);
  glEnable(GL_CULL_FACE);
  glCullFace(GL_BACK);
  glShadeModel(GL_SMOOTH);
  glDepthMask(GL_TRUE);
  glEnable(GL_NORMALIZE);
  // glDisable(GL_NORMALIZE);
  glLineWidth(2);
  
  glBlendEquation(GL_FUNC_ADD);
  glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

  glGetFloatv(GL_MODELVIEW_MATRIX, m_worldMatrix.data());

  if(m_currentLight)
    m_currentLight->apply(true, true);

  m_currentMaterial = 0;
  m_currentTexture = (GLuint) -1;
  m_blend = false;
  m_lighting = true;
  m_smooth = true;
  m_twoSided = false;
  m_inVee = false;
}

void GEE_GlRenderer::closeGlState()
{
  if(m_inVee && m_veeRenderer)
    exitVee();

  glDisable(GL_COLOR_MATERIAL);
  glDepthMask(GL_TRUE);
  glPopAttrib();
}


/** This method switches OpenGL to a new 2D texture. If the texture is
    the same as the previous texture, then no operation is performed.

    @arg tex The OpenGL texture id. If the texture id equals -1, then
    this method disables textures.
*/
void GEE_GlRenderer::applyGlTexture2D(GLuint tex)
{
  if(tex == m_currentTexture) 
    return;

  //VEE_trace("GEE_GlRenderer::applyGlTexture2D # New texture = %d", (int)tex);
  
  if(tex == (GLuint) -1) {
    glDisable(GL_TEXTURE_2D);
  }
  else {
    if(m_currentTexture == (GLuint) -1)
      glEnable(GL_TEXTURE_2D);
    
    glBindTexture(GL_TEXTURE_2D, tex);
    glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
  }
  m_currentTexture = tex;
}

/** This method enables/disables blending of the OpenGL state
    machine. */
void GEE_GlRenderer::applyGlBlend(bool blend)
{
  if(blend == m_blend) return;

  if(blend) {
    glEnable(GL_BLEND);
    glDepthMask(GL_FALSE);
    // glDisable(GL_DEPTH_TEST);
  }
  else {
    glDisable(GL_BLEND);
    glDepthMask(GL_TRUE);
    // glEnable(GL_DEPTH_TEST);
  }

  m_blend = blend;
}

/** Switch OpenGL lights on/off. */

void GEE_GlRenderer::applyGlLighting(bool lighting)
{
  // VEE_trace("GEE_GlRenderer::applyGlLighting # %d", (int) lighting);

  if(lighting == m_lighting) return;

  if(lighting && m_currentLight) {
    glEnable(GL_LIGHTING);
    m_currentLight->apply(true, false);
  }
  else
    glDisable(GL_LIGHTING);

  m_lighting = lighting;
}

void GEE_GlRenderer::applyGlSmooth(bool smooth)
{
  // VEE_trace("GEE_GlRenderer::applyGlSmooth # %d", (int) smooth);
  
  if(smooth == m_smooth) return;

  if(smooth)
    glShadeModel(GL_SMOOTH);
  else
    glShadeModel(GL_FLAT);

  m_smooth = smooth;
}

void GEE_GlRenderer::applyGlTwoSided(bool twoSided)
{
  if(m_twoSided == twoSided) return;

  if(twoSided) 
    glDisable(GL_CULL_FACE);
  else
    glEnable(GL_CULL_FACE);

  m_twoSided = twoSided;
}

void GEE_GlRenderer::applyMaterial(GEE_Material *mat)
{
  int flags = mat->flags();
  
  /* VEE_trace("GEE_GlRenderer::applyMaterial # %d %x", 
     (int) mat->useVertexColors(), flags); */

  if(flags & GEE_MAT_LIGHTING) {
    glLightModeli(GL_LIGHT_MODEL_TWO_SIDE, 1);
    glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT, mat->ambient().data());
    glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, mat->diffuse().data());
    glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, mat->specular().data());
  }

  applyGlLighting(flags & GEE_MAT_LIGHTING);
  applyGlBlend(flags & GEE_MAT_BLEND);
  applyGlSmooth(flags & GEE_MAT_SMOOTH);
  applyGlTwoSided(flags & GEE_MAT_TWO_SIDED);

  GLuint texId = (GLuint) -1;

  const std::string & tex = mat->texture();

  if(tex.size() && m_textures) {
    VEE_GlTexture2D * tp = m_textures->getTexture(tex);
    if(tp)
      texId = tp->id();
    else 
      VEE_error("GEE_GlRenderer::applyMaterial # Missing texture \"%s\"",
		tex.c_str());

    /* VEE_trace("GEE_GlRenderer::applyMaterial # tex = %s, id = %d",
       tex.c_str(), (int) texId); */
  }

  applyGlTexture2D(texId);
  
  float shine = mat->shine() * 128.0f;
  if(shine > 128.0f) shine = 128.0f;
  glMaterialfv(GL_FRONT_AND_BACK, GL_SHININESS, & shine);

  glColor4fv(mat->ambient().data());

  if(mat->useVertexColors() && m_selectionRender == RENDER_USUAL) {
    glEnable(GL_COLOR_MATERIAL);
    glColorMaterial(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE);
  }
  else
    glDisable(GL_COLOR_MATERIAL);  

  m_currentMaterial = mat;
}

/** Resets the vertex, triangle etc. counters. */

void GEE_GlRenderer::resetCounters()
{
  m_vertexCount = 0;
  m_triangleCount = 0;
  m_polygonCount = 0;
  m_particleCount = 0;
  m_geometryCount = 0;    
}

void GEE_GlRenderer::renderLines3DInternal(GEE_Lines3D *lines)
{  
  if(m_debug)
    VEE_trace("GEE_GlRenderer::renderLines3DInternal # %p vs = %u g = %d", 
	      lines, lines->vertexCount(), lines->generation());

  preRender(lines);

  ListMode m;
  bool isLines;

  applyGlLighting(false);
  
  bool useVertexColors = m_currentMaterial ? 
    m_currentMaterial->useVertexColors() : false;

  if(m_selectionRender != RENDER_USUAL)
    useVertexColors  = false;
    
  if(m_selectionRender < RENDER_POINTS_NS_HIGH2) {
    m = renderDisplayList(lines);
    
    if(m == RENDERED_LIST)
      return;
    
    glLineWidth(lines->width());
    
    glBegin(GL_LINE_STRIP);
    isLines = true;
  }
  else {
    glBegin(GL_POINTS);
    isLines = false;
    m = NO_LIST;
  }

  for(GEE_Lines3D::iterator it = lines->begin(); it != lines->end(); it++) {
    const GEE_LineVertex3D &v = GEE_Lines3D::getVertex(it);

    if(useVertexColors)
      glColor4fv(v.m_rgba.data());

    glVertex3fv(v.m_location.data());

    if(!v.m_continues && isLines) {
      glEnd();
      glBegin(GL_LINE_STRIP);      
    }

    /* const VEE_Vector3 & l = v.m_location;
       VEE_trace("GEE_GlRenderer::renderLines3D # %f %f %f", l.x, l.y, l.z); */
  }

  glEnd();

  if(m == GENERATE_LIST)
    glEndList();

  if(m_debug)
    VEE_trace("GEE_GlRenderer::renderLines3DInternal # "
	      "EXIT (m = %d)", (int) m);
}

static inline void GEE_sendVertexToGl(GEE_PolygonVertex *v, 
				      bool useVertexColors,
				      bool useTextureCoords)
{
  if(!v) {
    VEE_error("GEE_sendVertexToGl # NULL vertex");
    return;
  }

  if(useVertexColors)
    glColor4fv(v->m_rgba.data());
  
  if(useTextureCoords)
    glTexCoord2f(v->m_uv.x, v->m_uv.y);

  glNormal3fv(v->m_normal.data());
  glVertex3fv(v->m_location.data());
}

void GEE_GlRenderer::renderTriangleMeshInternal(GEE_TriangleMesh *mesh)
{
  GEE_FNAME = "GEE_GlRenderer::renderTriangleMesh";

  if(m_debug)
    VEE_trace("%s # %p vs = %u tris = %u", fname,
	      mesh, mesh->vertexCount(), mesh->triangleCount());

  preRender(mesh);
  ListMode m = renderDisplayList(mesh);

  if(m == RENDERED_LIST)
    return;

  bool useVertexColors = m_currentMaterial ? 
    m_currentMaterial->useVertexColors() : false;

  bool useTextureCoords = m_currentMaterial ? 
    (m_currentMaterial->texture().size() != 0) : false;

  if(m_selectionRender != RENDER_USUAL) {
    useVertexColors  = false;
    useTextureCoords = false;
  }

#ifdef __OS_IRIX__
  static DI_MutexAuto genMutex;
  
  bool myLists = (m == NO_LIST);
  
  if(myLists) {
    genMutex.lock();
    myLists = mesh->genList();
    genMutex.unlock();
  }
  if(myLists) {

    // HACK HACK, Use the internal display list format...

    GEE_TriMeshList & list = mesh->list();

    GEE_TriMeshList::iterator end = list.end();

    for(GEE_TriMeshList::iterator it = list.begin(); it != end; it++) {
      const GEE_TriMeshListItem & item = (*it);
      GEE_TriMeshItem type = item.m_type;

      const float * data = item.m_data.data();
      
      switch(type) {
      case GEE_TRIMESH_COLOR:
	if(useVertexColors)
	  glColor4fv(data);
	break;
      case GEE_TRIMESH_UV:
	if(useTextureCoords)
	  glTexCoord2fv(data);
	break;
      case GEE_TRIMESH_NORMAL:
	glNormal3fv(data);
	break;
      case GEE_TRIMESH_LOCATION:
	glVertex3fv(data);
	break;
      case GEE_TRIMESH_TRILIST:
	glBegin(GL_TRIANGLE_STRIP);
	break;
      case GEE_TRIMESH_TRIS:
	glBegin(GL_TRIANGLES);
	break;
      case GEE_TRIMESH_END:
	glEnd();
	break;
      case GEE_TRIMESH_UNKNOWN:
      default:
	VEE_error("%s # Uknown mesh list item %d", fname, (int) type);
	break;
      }      
    }

    return;
  }
#endif

  if(!mesh->useStrips()) {
    glBegin(GL_TRIANGLES);
    
    for(GEE_TriangleMesh::triangleIterator it = mesh->beginTriangle(); 
	it != mesh->endTriangle(); it++) {
      
      GEE_TriangleIndices & tri = GEE_TriangleMesh::getTriangle(it);
      
      for(uint i=0; i < 3; i++) {
	GEE_PolygonVertex * v = mesh->getVertex(tri.m_indices[i]);
	GEE_sendVertexToGl(v, useVertexColors, useTextureCoords);
      }
    }
  }
  else {
    bool startStrip = true;
    
    glBegin(GL_TRIANGLE_STRIP);

    for(GEE_TriangleMesh::triangleIterator it = mesh->beginTriangle(); 
	it != mesh->endTriangle(); it++) {
      
      GEE_TriangleIndices & tri = GEE_TriangleMesh::getTriangle(it);
      
      if(startStrip) {
	for(uint i=0; i < 3; i++) {
	  GEE_PolygonVertex * v = mesh->getVertex(tri.m_indices[i]);
	  GEE_sendVertexToGl(v, useVertexColors, useTextureCoords);
	}
	startStrip = false;
      }
      else {
	GEE_PolygonVertex * v = mesh->getVertex(tri.m_indices[2]);
	GEE_sendVertexToGl(v, useVertexColors, useTextureCoords);
      }
      
      if(!tri.m_continues) {
	glEnd();
	glBegin(GL_TRIANGLE_STRIP);
	startStrip = true;
      }      
    }
  }

  glEnd();

  if(m == GENERATE_LIST)
    glEndList();  
}

void GEE_GlRenderer::preRender(GEE_Geometry *data)
{
  GEE_FNAME = "GEE_GlRenderer::preRender";

  if(m_debug)
    VEE_trace("%s # %p", fname, data);

  if(m_inVee)
    exitVee();

  const std::string & matName = data->materialName();

  if(m_selectionRender == RENDER_USUAL) {

    if(!matName.size()) {
      
      GEE_Material * newMat = & data->material();
      
      /* if(newMat == m_currentMaterial && !closeVee)
	 return; */
      
      applyMaterial(newMat);
    }
    else {
      
      if(!m_materials) {
	if(m_debug)
	  VEE_trace("%s # EXIT (no materials)", fname);
	return;
      }
      const GEE_GlMaterial & mat = m_materials->getMaterial(matName);
      
      GLenum e;
      
      e = glGetError();
      if(e) 
	VEE_error("GlRenderer::preRender # GL ERROR ALREADY \"%s\"", 
		  gluErrorString(e));
      
      /* trace("GlRenderer::preRender # \"%s\" [%.2f %.2f %.2f] %d", 
	 gr->material().c_str(), 
	 mat.color().red(), mat.color().green(), mat.color().blue(),
	 (int) mat.textureId()); */
      
      mat.applyGl(this);
      
      m_currentMaterial = 0;      
    }
  }
  else {
    applyGlTexture2D((GLuint) -1);
    applyGlBlend(false);
    applyGlTwoSided(true);
    applyGlLighting(false);
    m_currentMaterial = 0;
    
    static float nsColor[4] = {1, 1, 1, 1};
    static float isColor[4] = {0, 1, 0, 1};
    
    if(m_selectionRender == RENDER_FACES_NS_HIGH1) {
      glColor4fv(nsColor);
    }
    else if(m_selectionRender == RENDER_FACES_IS_HIGH1) {
      glColor4fv(isColor);
    }
    else if(m_selectionRender == RENDER_EDGES_NS_HIGH2 ||
	    m_selectionRender == RENDER_EDGES_IS_HIGH2) {
      glLineWidth(1);
      glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
      glColor4f(0.2, 0.2, 0.2, 1.0);
    }
    else if(m_selectionRender == RENDER_EDGES_NS_HIGH1) {
      glLineWidth(2);
      glColor4fv(nsColor);
    }
    else if(m_selectionRender == RENDER_EDGES_IS_HIGH1) {
      glLineWidth(2);
      glColor4fv(isColor);
    }
    else if(m_selectionRender == RENDER_POINTS_NS_HIGH2) {
      glPointSize(4);
      glColor4f(1, 0, 0, 1);
    }
    else if(m_selectionRender == RENDER_POINTS_IS_HIGH2) {
      glPointSize(4);
      glColor4f(1, 0, 0, 1);
    }
    else {
      VEE_error("%s # Unsupported selection render mode", fname);
    }
  }

  int e = glGetError();
  if(e) 
    VEE_error("%s # GL ERROR \"%s\" (%s)", fname,
	      gluErrorString(e), matName.c_str());

  if(m_debug)
    VEE_trace("%s # EXIT", fname);
}

/** Checks if there is a diplay list for this graphic. If there is
    no list storage, then no list-related stuff is done and NO_LIST
    is returned. If the list exists, then it is rendered. If the the
    list does not exist, but there is a list storage, then its
    recording is began. */
GEE_GlRenderer::ListMode GEE_GlRenderer::renderDisplayList(GEE_Geometry *g)
{    
#ifdef __OS_IRIX__
  return NO_LIST;
#else
  if(m_debug)
    VEE_trace("GEE_GlRenderer::renderDisplayList # %p %d %d %s %ld", this,
	      (int) g->generation(), (int) g->editorCount(), 
	      typeid(*g).name(), m_renderCount);
  
  m_renderCount++;
  
  /* generation = -1 means that no display lists should be used. */
  

  GEE_GlDisplayLists * lists = (m_selectionRender == RENDER_USUAL) ?
    m_displayLists : m_highlightDisplayLists;

  if(!lists)
    return NO_LIST;
  
  if(g->generation() != -1 && !g->editorCount()) {
      
    GEE_GlDisplayList * l = & lists->getList(g);
      
    if(l->needsUpdate(g->generation())) {

      if(m_debug)
	VEE_trace("GEE_GlRenderer::renderDisplayList # generate for %p (%ld)",
		  g, l->generation());
	
      glNewList(l->listId(), GL_COMPILE_AND_EXECUTE);
      l->setGeneration(g->generation());

      m_generateDisplayListCount++;

      return GENERATE_LIST;
    }
    else {
      glCallList(l->listId());
      m_renderDisplayListCount++;
      return RENDERED_LIST;
    }
  }

  GEE_GlDisplayList * l = lists->getListPtr(g);
  if(l)
    l->setGeneration(-1); // Disable, but do not delete

  return NO_LIST;
#endif 
}

void GEE_GlRenderer::enterVee()
{
  GLenum e = glGetError();
  if(e) 
    VEE_error("GEE_GlRenderer::enterVee # GL ERROR 1 %s", 
	      gluErrorString(e));

  glPushAttrib(~0);
  m_veeRenderer->initRenderState();
  m_veeRenderer->setUseMatrix(true);
  m_inVee = true;

  e = glGetError();
  if(e) 
    VEE_error("GEE_GlRenderer::enterVee # GL ERROR 1.5 %s", 
	      gluErrorString(e));
 
  /* VEE_Matrix4 tmp;
     glGetFloatv(GL_MODELVIEW_MATRIX, tmp.data()); */

  glPushMatrix();
  
  glLoadMatrixf(m_worldMatrix.data());
  
  e = glGetError();
  if(e) 
    VEE_error("GEE_GlRenderer::enterVee # GL ERROR 2 %s", 
	      gluErrorString(e));
 
  // m_worldMatrix = tmp;
}

void GEE_GlRenderer::exitVee()
{
  m_veeRenderer->closeRenderState();

  GLenum e = glGetError();
  if(e) 
    VEE_error("GEE_GlRenderer::exitVee # GL ERROR 1 %s", 
	      gluErrorString(e));


  glPopAttrib();
  m_inVee = false;

  glPopMatrix();

  e = glGetError();
  if(e) 
    VEE_error("GEE_GlRenderer::exitVee # GL ERROR 2 %s", 
	      gluErrorString(e));
 
}
