/* -*- C++ -*- */

/* COPYRIGHT
 *
 * This file is part of the Visual Effects Engine - VEE
 *
 * Read the "VEE-LICENSE" file for the license.
 *
 * Authors & Copyright:   
 *
 * Tommi Ilmonen, Tuukka Heikura, Marko Myllymaa and 
 * Janne Kontkanen 2001-2004
 *
 * Additional copyrights: Tekes 2003-2004
 *
 * firstname.lastname@hut.fi
 *
 */

/** @file vee_cg_sl_renderer.h

    OpenGL Shading Language (GLSL) rendering for project VEE.

*/

#ifndef VEE_GL_SL_RENDERER_H
#define VEE_GL_SL_RENDERER_H

//#ifndef HAVE_GLSL_ARB
//#error "Must have OpenGL SL ARB extensions to compile GLSL code !!"
//#endif

#include <vee_gl_renderer.h>

#include <vee_reference_object.h>

#ifdef __OS_LINUX__
// Fix the broken NVidia headers:
#define GL_GLEXT_PROTOTYPES 1
#define GLAPI
#endif // __OS_LINUX__

#ifdef __APPLE__
# include <OpenGL/glext.h>
#else
# include <GL/glext.h>
#endif

class VEE_GlSlRenderer;

enum VEE_GlSlParam {
  /// Camera location (Vector4)
  VEE_SLP_PARTICLE_SIZE,
  /// Camera location (Vector4)
  VEE_SLP_CAMPOS,
  /// Fragment window size
  VEE_SLP_WINDOW_SIZE,
  /** Fragment depth params (near distance, range inverted and fade
      distance normalized). */
  VEE_SLP_DEPTH_PARAMS,
  /// Age (float)
  VEE_SLP_AGE,
  /// Color (Vector4)
  VEE_SLP_COLOR,
  /// Color additivity (float)
  VEE_SLP_ADDITIVE,
  /// Center (Vector4)
  VEE_SLP_LOCATION,
  /// Velocity (Vector4)
  VEE_SLP_VELOCITY,
  /** Corner + uv (Vector4) */
  VEE_SLP_CORNER_UV,
  /** Main texture sampler (int) */
  VEE_SLP_TEX,
  /** Depth map texture sampler (int) */
  VEE_SLP_DEPTH_MAP,
  
  VEE_SLP_SIZEOF
};

extern const char * VEE_SlParamNames[];

char * VEE_loadFile(const char * fileName);

/** A GLSL shader "driver". Each instance of VEE_GlSlDriver holds GLSL
    vertex and/or fragment shaders plus the linked program. 

    @author Tommi Ilmonen
*/

class VEE_GlSlDriver
{
public:
  VEE_GlSlDriver();
  virtual ~VEE_GlSlDriver();

  virtual bool loadShadersFromFile(const char * fileNameBase);
  virtual void deleteShaders();

  virtual bool bindProgram(VEE_GlSlRenderer *);

  virtual void renderParticle(VEE_Particle *, VEE_GlSlRenderer *);

  static std::string infoLog(GLhandleARB);

protected:

  int m_params[VEE_SLP_SIZEOF];

  GLhandleARB m_vertexShader;
  GLhandleARB m_fragmentShader;
  /// Links fragment and vertex shaders
  GLhandleARB m_program;

  std::string m_programName;
};

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

/** A renderer that uses OpenGL Shading Language (GLSL) shaders in the
    rendering back-end. 

    @author Tommi Ilmonen
*/

class VEE_GlSlRenderer : public VEE_GlRenderer
{
public:
  VEE_GlSlRenderer(VEE_SceneBox *box = 0);
  virtual ~VEE_GlSlRenderer();
  
  virtual void initializeGL();
  virtual void endGlMode();
  virtual void renderParticle(VEE_Particle *p);

  void disableSl();
  void enableSl();

  void loadShaders();

  void setUseShaders(bool use) { m_useShaders = use; }
  bool useShaders() const { return m_useShaders; }

  static bool debug() { return m_debug; }
  
protected:

  std::map<int, VEE_SafePtr<VEE_GlSlDriver> > m_drivers;
  typedef std::map<int, VEE_SafePtr<VEE_GlSlDriver> >::iterator driver_iter;

  VEE_GlSlDriver * m_currentDriver;
  
  int         m_currentShader;
  bool        m_useShaders;

  static bool m_debug;
};

#endif
