/* -*- C++ -*- */

/* COPYRIGHT
 *
 * This file is part of the Visual Effects Engine - VEE
 *
 * Read the "VEE-LICENSE" file for the license.
 *
 * Authors & Copyright:   
 *
 * Tommi Ilmonen, Tuukka Heikura, Marko Myllymaa and 
 * Janne Kontkanen 2001-2004
 *
 * Additional copyrights: Tekes 2003-2004
 *
 * firstname.lastname@hut.fi
 *
 */

#ifndef VEE_IMAGE_H
#define VEE_IMAGE_H

#include <sys/types.h>

class VEE_ImageIO;

class VEE_Image
{
public:

  enum Channel {
    RedChannel,
    GreenChannel,
    BlueChannel,
    AlphaChannel
  };

  VEE_Image();
  ~VEE_Image();

  /// Width of the image.
  uint w() const { return m_width; }
  /// Height of the image
  uint h() const { return m_height; }

  uint bytesPerPixel() const { return m_bytesPerPixel; }

  bool loadJpeg(const char *fileName);
  bool load(const char *fileName, int w = -1, int h = -1, int chans = -1);
  bool save(const char *fileName);

  /// Add alpha channel -> RGBA image
  void addAlpha();

  void alphaFromImageLuminance(const VEE_Image &img);

  double channelAverage(uint chan);

  /** Resizes data buffer. Old values are lost. */
  void resize(uint w, uint h, uint bytesPerPixel);

  void resizeImage(uint w, uint h);
  void resizeImgToPowOfTwo();

  void mirrorXY();

  /** Randomize the data buffer. */
  void randomValues();
  void perlinNoise(uint channel,
		   double att, int octaves, double scale = 1.0,
		   double z = 0.0, double amplitude = 1.0);
  void checkTexture(uint horizontalInterval,
		    uint verticalInterval,
		    const unsigned char * color1 = 0,
		    const unsigned char * color2 = 0);
  void skotchTexture(uint stripeWidth,
		     uint stripeInterval,
		     const unsigned char * color1 = 0,
		     const unsigned char * color2 = 0);
  void stripeTexture(uint stripeWidth,
		     uint stripeInterval,
		     const unsigned char * color1 = 0,
		     const unsigned char * color2 = 0);		    
  
  inline unsigned char *image() { return m_image; }
  inline const unsigned char *image() const { return m_image; }

  unsigned char *pixel(uint x, uint y);
  const unsigned char *pixelConst(uint x, uint y) const;

  unsigned char pixelLuminance(uint x, uint y) const;

  void copyPart(const VEE_Image &img, 
		uint startRow, uint endRow,
		uint startCol, uint endCol,
		uint offsetX = 0, uint offsetY = 0);

  /** Copy part of the image and enlarge it by factor ratio. */
  void copyPartEnlarge(const VEE_Image &img, 
		       uint ratio,
		       uint startRow, uint endRow,
		       uint startCol, uint endCol);

  void copyPartSmaller(const VEE_Image &img, 
		       uint ratio,
		       uint startRow, uint endRow,
		       uint startCol, uint endCol);

  void evenGray(unsigned char v);
  void evenGray(uint channel, unsigned char v);
  void addRandom(uint channel, unsigned char amplitude);
  void channelCopy(uint targetChannel, uint sourceChannel);
  void channelSwap(uint c1, uint c2);
  void channelMultiply(uint target, uint source1, uint source2);
  void cosinePeak(uint channel, float power = 1.0, float radius = 0.5);
  void powerPeak(uint channel, float power = 1.0, float radius = 0.5);
  void boostRgb(int boost);
  void multiply(float mul);
  bool add(const VEE_Image &);

  /// Interpolate a value from adjacent pixels
  void getSubpixel(unsigned char *target, double x, double y) const;
  /// Average value of many pixels.
  void getMultipixel(unsigned char *target, uint x, uint y, uint size) const;

  VEE_Image & operator = (const VEE_Image &);

  static void test();

  static void addIO(VEE_ImageIO *);

protected:

  unsigned char *m_image;
  /// Row size in bytes.
  uint m_rowSize;
  uint m_bytesPerPixel;
  uint m_width;
  uint m_height;

private:
  VEE_Image(const VEE_Image &) {}

};

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////


class VEE_ImageIO
{
public:
  VEE_ImageIO() {}
  virtual ~VEE_ImageIO() {}

  virtual bool read(VEE_Image *, const char *, int, int, int) = 0;
  virtual bool write(VEE_Image *, const char *) = 0;
  
};

#endif
