/* COPYRIGHT
 *
 * This file is part of the Visual Effects Engine - VEE
 *
 * Read the "VEE-LICENSE" file for the license.
 *
 * Authors & Copyright:   
 *
 * Tommi Ilmonen, Tuukka Heikura, Marko Myllymaa and 
 * Janne Kontkanen 2001-2004
 *
 * Additional copyrights: Tekes 2003-2004
 *
 * firstname.lastname@hut.fi
 *
 */

#include <vee_tcb_spline3.h>

#include <vee_error.h>
#include <vee_mgc.h>

WmlVector3 * WmlClear(WmlVector3 *v, int n)
{
  for(int i=0; i < n; i++)
    v[i] = WmlVector3::ZERO;

  return v;
}

template <class T>
T * WmlClear(T *v, T val, int n)
{
  for(int i=0; i < n; i++)
    v[i] = val;

  return v;
}

template <class T>
T * WmlAscending(T *v, T val, int n)
{
  T tmp = 0.0;

  for(int i=0; i < n; i++) {
    v[i] = tmp;
    tmp += val;
  }

  return v;
}

VEE_TCBSpline3::VEE_TCBSpline3(int maxknots,
			       float tension, 
			       float continuity, 
			       float bias)
  : Wml::TCBSpline3<float>(maxknots, 
		    WmlAscending(new float[maxknots+1], 1.0f, maxknots+1),
		    WmlClear(new WmlVector3[maxknots+1], maxknots+1),
		    WmlClear(new float[maxknots+1], tension, maxknots+1),
		    WmlClear(new float[maxknots+1], continuity,maxknots+1),
		    WmlClear(new float[maxknots+1], bias, maxknots+1)),
    m_allocatedKnots(maxknots)
{}

VEE_TCBSpline3::~VEE_TCBSpline3()
{}

void VEE_TCBSpline3::setKnots(int n)
{
  if(n > m_allocatedKnots) {
    VEE_error("VEE_TCBSpline3::setKnots # %d > %d", n, m_allocatedKnots);
    return;
  }
  m_iSegments = n - 1;
}

void VEE_TCBSpline3::setKnotParams
(int n, float tension, float continuity, float bias)
{
  if(n >= m_allocatedKnots) {
    VEE_error("VEE_TCBSpline3::setKnotParams # knot %d out of range (%d)",
	      n, m_allocatedKnots);
    return;
  }
  m_afTension[n]    = tension;
  m_afContinuity[n] = continuity;
  m_afBias[n]       = bias;
}
  
void VEE_TCBSpline3::setKnotTime(int n, float t)
{
  if(n >= m_allocatedKnots) {
    VEE_error("VEE_TCBSpline3::setKnotTime # knot %d out of range (%d)",
	      n, m_allocatedKnots);
    return;
  }
  m_afTime[n] = t;
}

void VEE_TCBSpline3::setPoint(int n, const VEE_Vector3 &loc)
{
  if(n >= m_allocatedKnots) {
    VEE_error("VEE_TCBSpline3::setPoint # knot %d out of range (%d)",
	      n, m_allocatedKnots);
    return;
  }

  m_akPoint[n] = * toWml(&loc);
}

/** Recalculates the weights for the spline. */

void VEE_TCBSpline3::recalculate()
{
  // For now, treat the first point as if it occurred twice.
  ComputePoly(0,0,1,2);

  for (int i = 1; i < m_iSegments-1; i++)
    ComputePoly(i-1,i,i+1,i+2);
  
  // For now, treat the last point as if it occurred twice.
  ComputePoly(m_iSegments-2,m_iSegments-1,m_iSegments,m_iSegments);  
}

