/* COPYRIGHT
 *
 * This file is part of the Simple Object Loader and Renderer - SOLAR
 *
 * Read the file "SOLAR-LICENSE" for the license.
 *
 * Authors & Copyright:
 *
 * Tommi Ilmonen, 2001-2003, except LWO-code by RJ, 2001-2003
 *
 * Tommi's email: firstname.lastname@hut.fi
 *
 */

#include <GraphicTriangleStrips.hpp>
#include <Renderer.hpp>

namespace Solar {

  GraphicTriangleStrips::GraphicTriangleStrips()
  {
    m_boundingBox.clear();
  }

  GraphicTriangleStrips::~GraphicTriangleStrips()
  {}
  
  Graphic::Type GraphicTriangleStrips::type() const
  {
    return TriangleStrips;
  }

  void GraphicTriangleStrips::getBoundingBox(BoundingBox *bbox)
  {
    *bbox = m_boundingBox;
  }
  
  int GraphicTriangleStrips::nofPrimitives() const
  {
    int n=0;
    for(uint i=0; i < m_strips.size(); i++) {
      if(m_strips[i].m_length >= 3)
	n += m_strips[i].m_length - 2;
    }
    return n;
  }

  void GraphicTriangleStrips::render(Renderer *r)
  {
    r->renderTriangleStrips(this);
  }

  void GraphicTriangleStrips::scale(const Fluid::Vector3 &scale)
  {
    Vertex * v = &* m_vertices.begin();
    Vertex * sentinel = &* m_vertices.end();

    while(v < sentinel) {
      v->m_location.scale(scale);
      v++;
    }
      
    m_boundingBox.scale(scale);
  }

  void GraphicTriangleStrips::move(const Fluid::Vector3 &loc)
  {
    Vertex * v = &* m_vertices.begin();
    Vertex * sentinel = &* m_vertices.end();

    while(v < sentinel) {
      v->m_location += loc;
      v++;
    }

    m_boundingBox.move(loc);
  }

  void GraphicTriangleStrips::transform(const Fluid::Matrix3 &transform)
  {  
    Vertex * v = &* m_vertices.begin();
    Vertex * sentinel = &* m_vertices.end();

    while(v < sentinel) {
      v->m_location = transform * v->m_location;
      v++;
    }

    v = &* m_vertices.begin();

    m_boundingBox.clear(v->m_location);
    
    while(v < sentinel) {
      m_boundingBox.expand(v->m_location);
      v++;
    }
  }

  void GraphicTriangleStrips::write(std::ostream *s)
  {
    (*s) << "TRIANGLESTRIP\nVERTICES\n";
    
    Vertex * v = &* m_vertices.begin();
    Vertex * sentinel = &* m_vertices.end();

    uint i = 0;

    while(v < sentinel) {
      (*s) << i++ << ":   " << v->m_location
	   << "  " << v->m_normal << "  " << v->m_uv << "\n";
    }
  }

  void GraphicTriangleStrips::addVertex(const Vertex &v)
  { 
    m_vertices.push_back(v); 

    if(!m_vertices.size())
      m_boundingBox.clear(v.m_location);
    else
      m_boundingBox.expand(v.m_location);
      
  }

  /** Use this method to tell the system that the strip you were
      building is done. */

  void GraphicTriangleStrips::finalizeStrip()
  {
    uint prevEnd;

    if(m_strips.size()) {
      Strip old = m_strips[m_strips.size() - 1];
      prevEnd = old.m_begin + old.m_length;
    }
    else
      prevEnd = 0;

    Strip strip;

    strip.m_begin = prevEnd;
    strip.m_length = m_vertices.size() - prevEnd;

    m_strips.push_back(strip);
  }

  void GraphicTriangleStrips::clear() 
  { 
    m_vertices.clear();
    m_strips.clear();
    m_boundingBox.clear(); 
  }

} // namespace Solar
