
/* COPYRIGHT
 *
 * This file is part of the Simple Object Loader and Renderer - SOLAR
 *
 * Read the file "SOLAR-LICENSE" for the license.
 *
 * Authors & Copyright:
 *
 * Tommi Ilmonen, 2001-2003, except LWO-code by RJ, 2001-2003
 *
 * Tommi's email: firstname.lastname@hut.fi
 *
 */

#include <Model.hpp>

#include <GraphicTriangles.hpp>
#include <GraphicTriangleStrips.hpp>
#include <SolarMath.hpp>
#include <Renderer.hpp>

// #include <Fluid/Math.hpp>

#include <stdlib.h>
#include <stdio.h>

namespace Solar {

  Model::Model()
  {
    clear();
  }

  Model::~Model()
  {
    clear();
  }

  void Model::setGarbageBin(GarbageBin *bin)
  {
    Garbage::setGarbageBin(bin);

    for(uint i=0; i < m_graphics.size(); i++)
      m_graphics[i]->setGarbageBin(bin);
  }

  void Model::render(Renderer *renderer)
  {
    // puts("Model::render");

    renderer->begin(this);

    for(Graphic **iter = &* m_graphics.begin(); 
	iter != &* m_graphics.end(); 
	iter++)
      (*iter)->render(renderer);

    renderer->end();
  }

  void Model::scale(const Fluid::Vector3 &v)
  {
    for(Graphic **gr = &* m_graphics.begin(); gr != &* m_graphics.end(); gr++)
      (*gr)->scale(v);

    m_boundingBox.scale(v);
  }

  void Model::move(const Fluid::Vector3 &v)
  {
    for(Graphic **gr = &* m_graphics.begin(); gr != &* m_graphics.end(); gr++)
      (*gr)->move(v);

    m_boundingBox.move(v);
  }

  void Model::transform(const Fluid::Matrix3 &m)
  {
    for(Graphic **gr = &* m_graphics.begin(); gr != &* m_graphics.end(); gr++)
      (*gr)->transform(m);

    calculateBoundingBox();
  }

  void Model::addGraphic(Graphic *g)
  {
    BoundingBox tmp;
    g->getBoundingBox(&tmp);

    if(!m_graphics.size())
      m_boundingBox = tmp;
    else
      m_boundingBox.expand(tmp);    

    m_graphics.push_back(g);
  }

  void Model::clear()
  {
    for(Graphic **iter = &* m_graphics.begin(); 
	iter != &* m_graphics.end(); 
	iter++)
      delete *iter;

    m_graphics.resize(0);
    m_boundingBox.clear();
    // m_location = Fluid::Vector3::ZERO;
    // m_rotation = Fluid::Matrix3::IDENTITY;
  }

  /** Merges the graphic elements from another model. 

  \b Note: The source model becomes empty. */

  void Model::merge(Model *m)
  {
    for(uint i=0; i < m->m_graphics.size(); i++) {
      addGraphic(m->m_graphics[i]);
    }
    m->m_graphics.resize(0);
  }

  void Model::calculateBoundingBox()
  {
    m_boundingBox.clear();

    if(m_graphics.size()) {
      // m_graphics[0]->recalculateBoundingBox(&m_boundingBox);
      m_graphics[0]->getBoundingBox(&m_boundingBox);
    }

    BoundingBox tmp;

    for(uint i=1; i < m_graphics.size(); i++) {
      // m_graphics[i]->recalculateBoundingBox(&m_boundingBox);
      m_graphics[i]->getBoundingBox(&tmp);
      m_boundingBox.expand(tmp);
    }
  }

  void Model::normalize(float maximumSpan, bool center)
  {
    calculateBoundingBox();
  
    if(center) {
      Fluid::Vector3 center = m_boundingBox.center();
      center *= -1.0f;
      move(center);
    }
  
    Fluid::Vector3 volume = m_boundingBox.high() - m_boundingBox.low();
  
    float maxDim = Solar::Max(Solar::Max(volume[0], volume[1]), volume[2]);
  
    if(maxDim < 1.0e-8)
      return;

    float scaling = maximumSpan / maxDim;
    volume[0] = scaling;
    volume[1] = scaling;
    volume[2] = scaling;

    scale(volume);
  }

  void Model::centralize()
  {
    move(-m_boundingBox.center());
  }

  int Model::nofPrimitives() const
  {
    int n=0;

    for(uint i=0; i < m_graphics.size(); i++)
      n += m_graphics[i]->nofPrimitives();

    return n;
  }

  /** Pyramid creator. */  

  void Model::createPyramid
  (float size, const std::string &baseMaterial, bool shareMaterials )
  {
    uint i;

    char buf[16];

    clear();

    GraphicTriangles::Triangle tri;

    tri.clear();

    Fluid::Vector3 tip(0.0f, size, 0.0f), corners[4], edges[4];
    Fluid::Vector3 normal;

    float d = size * 0.5;
  
    // Construct corrners:
    corners[0][0] = -d;
    corners[0][1] = 0.0;
    corners[0][2] = -d;

    corners[1][0] =  d;
    corners[1][1] = 0.0;
    corners[1][2] = -d;

    corners[2][0] =  d;
    corners[2][1] = 0.0;
    corners[2][2] =  d;

    corners[3][0] = -d;
    corners[3][1] = 0.0;
    corners[3][2] =  d;

    /* Now set the texture coordinates. The coordinates are already set
       to zero. As a result we only need to modify the values that
       should not remain zero. */

    tri.m_vertices[0].m_uv[0] = 0.5;
    tri.m_vertices[0].m_uv[1] = sin(Fluid::Math::PI / 3.0);
    tri.m_vertices[1].m_uv[0] = 1.0;

    for(i=0; i < 4; i++)
      edges[i] = corners[i] - tip;

    GraphicTriangles *triangles;

    for(i = 0; i < 4; i++) {
      sprintf(buf, "Face%d", i);

      triangles = new GraphicTriangles;

      triangles->setMaterial(baseMaterial + (shareMaterials ? "" : buf));

      uint j = (i + 1) & 0x3;

      // Construct edges:
      tri.m_vertices[0].m_vertex = tip;
      tri.m_vertices[2].m_vertex = corners[i];
      tri.m_vertices[1].m_vertex = corners[j];
      normal = cross(edges[j], edges[i]);
      normal.normalize();
      tri.setNormals(normal);
      triangles->addTriangle(tri);
      addGraphic(triangles);
    }

    /* Now create the floor element: */


    triangles = new GraphicTriangles;

    triangles->setMaterial(baseMaterial + (shareMaterials ? "" : "Floor"));

    tri.clear();

    tri.m_vertices[0].m_vertex = corners[0];
    tri.m_vertices[1].m_vertex = corners[1];
    tri.m_vertices[2].m_vertex = corners[3];

    tri.m_vertices[1].m_uv[1] = 1.0;
    tri.m_vertices[2].m_uv[0] = 1.0;

    normal.make(0, -1, 0);
    tri.setNormals(normal);

    triangles->addTriangle(tri);

    tri.m_vertices[0].m_vertex = corners[1];
    tri.m_vertices[1].m_vertex = corners[2];
    tri.m_vertices[2].m_vertex = corners[3];

    tri.m_vertices[0].m_uv = tri.m_vertices[1].m_uv;
    tri.m_vertices[1].m_uv[0] = 1.0;
    tri.m_vertices[1].m_uv[1] = 1.0;
    tri.m_vertices[2].m_uv[0] = 1.0;

    tri.setNormals(normal);
  
    triangles->addTriangle(tri);
  

    addGraphic(triangles);
  }

  /** Constructs a cude model. */
  void Model::createCube
  (float size, const std::string &baseMaterial,bool shareMaterials)
  {
    size *= 0.5;

    Fluid::Vector3 normal, corners[2][4];
    uint i;

    char buf[16];

    clear();

    for(i = 0; i < 4; i++) {
      float x, z;

      if(i == 0 || i == 3)
	x = -size;
      else 
	x = size;

      if(i == 0 || i == 1)
	z = -size;
      else 
	z = size;

      corners[0][i] = Fluid::Vector3(x, size, z);
    }

    for(i = 0; i < 4; i++) {
      corners[1][i] = corners[0][i];
      corners[1][i][1] = -size;
    }

    GraphicTriangles *triangles;
    GraphicTriangles::Triangle tri, textri1, textri2;
  
    textri1.m_vertices[0].m_uv[0] = 0.0;
    textri1.m_vertices[0].m_uv[1] = 0.0;
    textri1.m_vertices[1].m_uv[0] = 0.0;
    textri1.m_vertices[1].m_uv[1] = 1.0;
    textri1.m_vertices[2].m_uv[0] = 1.0;
    textri1.m_vertices[2].m_uv[1] = 1.0;

    textri2.m_vertices[0].m_uv[0] = 0.0;
    textri2.m_vertices[0].m_uv[1] = 0.0;
    textri2.m_vertices[1].m_uv[0] = 1.0;
    textri2.m_vertices[1].m_uv[1] = 1.0;
    textri2.m_vertices[2].m_uv[0] = 1.0;
    textri2.m_vertices[2].m_uv[1] = 0.0;
  
    for(i = 0; i < 4; i++) {
      sprintf(buf, "Face%d", i);

      triangles = new GraphicTriangles;

      triangles->setMaterial(baseMaterial  + (shareMaterials ? "" : buf));

      tri = textri1;

      uint j = (i + 1) & 0x3;

      // First triangle:
      tri = textri1;
      tri.m_vertices[0].m_vertex = corners[0][i];
      tri.m_vertices[1].m_vertex = corners[1][j];
      tri.m_vertices[2].m_vertex = corners[1][i];
      normal = cross((corners[1][j] - corners[1][i]),
		     (corners[0][i] - corners[1][i])) * -1.0;
      normal.normalize();
      tri.setNormals(normal);
      triangles->addTriangle(tri);

      // Second triangle:
      tri = textri2;
      tri.m_vertices[0].m_vertex = corners[0][i];
      tri.m_vertices[1].m_vertex = corners[0][j];
      tri.m_vertices[2].m_vertex = corners[1][j];
      tri.setNormals(normal);
      triangles->addTriangle(tri);

      addGraphic(triangles);
    }

    // Create Ceiling:
    triangles = new GraphicTriangles;
  
    triangles->setMaterial(baseMaterial + (shareMaterials ? "" : "Ceiling"));
    
    tri = textri1;

    normal = Fluid::Vector3(0.0, 1.0, 0.0);

    // First triangle:
    tri = textri1;
    tri.m_vertices[0].m_vertex = corners[0][0];
    tri.m_vertices[1].m_vertex = corners[0][2];
    tri.m_vertices[2].m_vertex = corners[0][1];
    tri.setNormals(normal);
    triangles->addTriangle(tri);
  
    // Second triangle:
    tri = textri2;
    tri.m_vertices[0].m_vertex = corners[0][0];
    tri.m_vertices[1].m_vertex = corners[0][3];
    tri.m_vertices[2].m_vertex = corners[0][2];
    tri.setNormals(normal);
    triangles->addTriangle(tri);
  
    addGraphic(triangles);

    // Create FLOOR:
    triangles = new GraphicTriangles;
  
    triangles->setMaterial(baseMaterial  + (shareMaterials ? "" : "Floor"));
    
    tri = textri1;

    normal = Fluid::Vector3(0.0, -1.0, 0.0);

    // First triangle:
    tri = textri1;
    tri.m_vertices[0].m_vertex = corners[1][0];
    tri.m_vertices[1].m_vertex = corners[1][1];
    tri.m_vertices[2].m_vertex = corners[1][2];
    tri.setNormals(normal);
    triangles->addTriangle(tri);
  
    // Second triangle:
    tri = textri2;
    tri.m_vertices[0].m_vertex = corners[1][0];
    tri.m_vertices[1].m_vertex = corners[1][2];
    tri.m_vertices[2].m_vertex = corners[1][3];
    tri.setNormals(normal);
    triangles->addTriangle(tri);
  
    addGraphic(triangles);

  }

  Fluid::Vector3 __calculateNormal(const unsigned char *map, 
				   unsigned w, unsigned h,
				   int x, int y)
  {
    int xl = x - 1;
    if(xl < 0) xl = 0;

    int xh = x + 1;
    if(xh >= (int) w) xh = w - 1;

    int yl = y - 1;
    if(yl < 0) yl = 0;

    int yh = y + 1;
    if(yh >= (int) h) yh = h - 1;
  
    Fluid::Vector3 vx, vz;

    float xrise = (map[w * y + xh] - map[w * y + xl]) / 255.0;
    float yrise = (map[w * yh + x] - map[w * yl + x]) / 255.0;
  
    vx.make((xh - xl) / (float) (w - 1), xrise, 0.0);
    vz.make(0.0, yrise, (yh - yh) / (float) (h - 1));
  
     Fluid::Vector3 normal = cross(vx, vz);

    normal.normalize();

    return normal;
  }

  GraphicTriangleStrips::Vertex __calculateVertex(const unsigned char *map, 
						  unsigned w, unsigned h,
						  int x, int y)
  {
    GraphicTriangleStrips::Vertex v;

    Fluid::Vector3 location;
    Fluid::Vector3 normal;
    Fluid::Vector2 uv;

    uv.make(x / (float) (w - 1), y / (float) (h-1));

    location[0] = uv[0] - 0.5;
    location[2] = uv[1] - 0.5;
    location[1] = map[y * w + x] / 255.0 - 0.5;
  
    normal = __calculateNormal(map, w, h, x, y);

    v.m_location = location;
    v.m_normal   = normal;
    v.m_uv       = uv;

    return v;
  }

  /** Creates a model from height map data. Each row in the map is
      turned into a triangle strip. This should be efficient enough. */

  void Model::createFromHeightMap(const unsigned char *map, 
				  unsigned w, unsigned h,
				  const char * material)
  {
    // float scale = 1.0 / 255;

    GraphicTriangleStrips * strips = new GraphicTriangleStrips();

    for(unsigned row = 0; row < h - 1; row++) {
    
      for(unsigned col = 0; col < w - 1; col++) {

	unsigned r1, r2;

	// if(col & 0x1) {
	r1 = row;
	r2 = row + 1;
	/* }
	   else {
	   r2 = row;
	   r1 = row + 1;
	   }*/

	strips->addVertex(__calculateVertex(map, w, h, col, r1));
	strips->addVertex(__calculateVertex(map, w, h, col, r2));

	strips->setMaterial(material);
      }
  
      strips->finalizeStrip();    
    }
  
    addGraphic(strips);
  }


} // namespace Solar
