/* COPYRIGHT
 *
 * This file is part of the Simple Object Loader and Renderer - SOLAR
 *
 * Read the file "SOLAR-LICENSE" for the license.
 *
 * Authors & Copyright:
 *
 * Tommi Ilmonen, 2001-2003, except LWO-code by RJ, 2001-2003
 *
 * Tommi's email: firstname.lastname@hut.fi
 *
 */

#include <lwo.hpp>

#include <LwoConvert.hpp>

#include <Error.hpp>
#include <GraphicTriangleMesh.hpp>
#include <Model.hpp>

#include <map>
#include <set>

namespace Solar {

static Graphic * lwoLoadSurface(uint nsurf, LWOReader *lwo) 
{
  const char * fname = "lwoLoadSurface";

  uint i, j;

  std::map<uint, uint> vertexMap;
  std::vector<uint> reverseMap;
  std::set<int>     matIds;

  LWOReader::Surface &surf = lwo->getSurface(nsurf);
  uint triangles = lwo->getTriangleCount();
  uint nsurfacetriangles = 0;

  trace("%s # Preparing to load %d triangles", fname, surf.m_polygons);
  
  // Re-index the relevant vertices:
  for(i = 0; i < triangles; i++) {
    LWOReader::Triangle   &lwoTri = lwo->getTriangle(i);

    matIds.insert(lwoTri.m_surface);

    if((uint) lwoTri.m_surface != nsurf) 
      continue;

    nsurfacetriangles++;

    for(j=0; j < 3; j++) {
      uint lwoVertex = lwoTri.m_v[j];
      
      std::map<uint, uint>::iterator it = vertexMap.find(lwoVertex);

      if(it == vertexMap.end()) {
	// Add the vertex:
	uint tmp = reverseMap.size();
	vertexMap[lwoVertex] = tmp;
	reverseMap.push_back(lwoVertex);
      }
    }
  }
  
  Solar::trace("%s # %d triangles, %d vertices, material =  \"%s\"", 
	       fname, nsurfacetriangles, vertexMap.size(),
	       surf.m_name.c_str());

  if(matIds.size()) {
    std::set<int>::iterator it = matIds.end();
    it--;
    trace("%s # Largest material is %d", fname, *it);
  }

  if(!nsurfacetriangles) 
    return 0;

  // Initialize the triangle mesh:
  GraphicTriangleMesh *trimesh = new GraphicTriangleMesh;
  trimesh->setMaterial(surf.m_name);
  
  // Get relevant vertices in the right order:  
  uint nvertices = reverseMap.size();

  for(i = 0; i < nvertices; i++) {

    LWOReader::Vertex * lwoVertex = & lwo->getVertex(reverseMap[i]);
    GraphicTriangleMesh::VertexData meshVertex;
    
    meshVertex.location() = * ((Fluid::Vector3 *) & lwoVertex->m_position);
    meshVertex.normal() = * ((Fluid::Vector3 *) & lwoVertex->m_normal);
    meshVertex.uv() = * ((Fluid::Vector2 *) & lwoVertex->m_uv);
    
    trimesh->addVertex(meshVertex);
  }
  
  // Get the relevant triangles (in any order)

  for(i = 0; i < triangles; i++) {

    LWOReader::Triangle &lwoTri = lwo->getTriangle(i);
    GraphicTriangleMesh::Triangle meshTri;

    if((uint) lwoTri.m_surface != nsurf) 
      continue;
    
    for(j=0; j < 3; j++) {
      uint meshVertexIndex = vertexMap[lwoTri.m_v[j]];
      meshTri.vertex(j) = meshVertexIndex;
    }
    trimesh->addTriangle(meshTri);
  }

  if(trimesh->nofTriangles())
    return trimesh;
  
  delete trimesh;
  return 0;
}

bool LwoConvert::convert(Model *model, LWOReader *lwo)
{
  model->clear();

  uint surfaces = lwo->getSurfaceCount();

  for(uint i=0; i < surfaces; i++) {
    Graphic *gr = lwoLoadSurface(i, lwo);

    if(gr)
      model->addGraphic(gr);
  }
  
  return model->subObjectCount();
}

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

bool LwoConvert::load(Model *model, const char *filename)
{
  LWOReader lwo(filename);

  bool ok = convert(model, &lwo);

  return  ok;
}

} // namespace Solar
